{@abstract(1D and 2D spline routines )
@author(Nikolai Shokhirev http://www.shokhirev.com/nikolai.html)
@created(2002.02.02)
@lastmod(2006.11.11)
Nikolai V. Shokhirev, 2002-2006
Added splintd -  2006.11.11}
{ 2007.07.07 - removed compiler warnings, extended comments, other cleanup }
unit uSpline;

interface

uses
  uMatTypes, uDynArrays, uDynData;

{
  Pascal version by Nikolai V. Shokhirev.
  Note: Added the estimation of derivatives  (NVS)
  Ref: W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P.
       Numerical Recipes in Fortran  http://www.nr.com/
Input
  x[Lo..Hi],
  y[Lo..Hi] - y[i] = f(x[i])
  yp1 - left boundary condition
        yp1 >  0.99e30 - natural spline: f"(x[Lo]) = 0
        yp1 < -0.99e30 - 3 or 4-point estimation of f'(x[Lo]) - default
        otherwise f'[x[Lo]) = fp1
  ypn - right boundary condition
        yp1 >  0.99e30 - natural spline: f"(x[Hi]) = 0
        yp1 < -0.99e30 - 3 or 4-point estimation of f'(x[Hi]) - default
        otherwise f'[x[Hi]) = fpn
Output
  y2[Lo..Hi] - y2[i] = f"(x[i])             }
procedure spline(const x, y, y2: IFArr1D;  yp1: TFloat = -1.0e30;
                                           ypn: TFloat = -1.0e30);

{
  Pascal version by Nikolai V. Shokhirev.
  Ref: W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P.
       Numerical Recipes in Fortran  http://www.nr.com/
Input
  xa[Lo..Hi],
  ya[Lo..Hi]  - ya[i] = f(xa[i])
  y2a[Lo..Hi] - y2a[i] = f"(xa[i]) - from spline
  x  -  1D point
Output
  y = f(x)                         }
procedure splint(const xa, ya, y2a: IFArr1D; x: TFloat; var y: TFloat);

{
  Nikolai V. Shokhirev.
 
 Input
  xa[Lo..Hi],
  ya[Lo..Hi]  - ya[i] = f(xa[i])
  y2a[Lo..Hi] - y2a[i] = f"(xa[i]) - from spline
  x  -  1D point
Output
  y = f'(x)                         }
procedure splintd(const xa, ya, y2a: IFArr1D; x: TFloat; var y: TFloat);

{
  Pascal version by Nikolai V. Shokhirev.
  Note: Added the estimation of derivatives (accuracy increase)
  Ref: W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P.
       Numerical Recipes in Fortran  http://www.nr.com/
Purpose
  Precomputing of the 2nd-derivative table (over x2).
Note
  This routine is called once before calling splin2.
Input
  x1a[Lo1..Hi1],  - not used, included for consistency with splin2
  x2a[Lo2..Hi2],
  ya[Lo1..Hi1,Lo2..Hi2]  - y[i, j] = f(x1a[i], x2a[j])

Output
  y2a[Lo1..Hi1, Lo2..Hi2] - y2a[i,j] = f"(x1a[i], x2a[j]) over x2 }
procedure splie2(const x1a, x2a: IFArr1D; const ya, y2a: IFArr2D);

{
  Pascal version by Nikolai V. Shokhirev.
  Note: Added the estimation of derivatives (accuracy increase)
  Ref: W.H., Teukolsky S.A., Vetterling W.T., Flannery B.P.
       Numerical Recipes in Fortran  http://www.nr.com/
Purpose
  Calculation of the bicubic spline interpolation at (x1, x2)
Input
  x1a[Lo1..Hi1],
  x2a[1..n],
  ya[Lo1..Hi1,Lo2..Hi2]  - y[i, j] = f(x1a[i], x2a[j])
  y2a[Lo1..Hi1, Lo2..Hi2] - y2a[i,j] = f"(x1a[i], x2a[j]) over x2
  x1, x2 - 2D coordinates
Output
  y = f(x1, x2)     }
procedure splin2(const x1a, x2a: IFArr1D; const ya, y2a: IFArr2D;
                 x1, x2: TFloat; var y: TFloat);

{ Xarr[Index2] < x <= Xarr[Index2+1] }
function Index2(x: TFloat; const Xa: IFArr1D): TInt;

{ Xarr[Index3] is nearest to x and internal (if possible) }
function Index0(x: TFloat; const Xa: IFArr1D): TInt;

{ Pascal version by Nikolai V. Shokhirev.
  Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
       Computer methods for mathematical computations. Prentice-Hall, 1977.
PURPOSE:
  Compute the coefficients b,c,d for a cubic interpolating spline so that the
  interpolated value is given by
    s(x) = y[k] + b[k]*(x-x[k]) + c[k]*(x-x[k])**2 + d[k]*(x-x[k])**3
      when x[k] <= x <= x[k+1]
  The end conditions match the third derivatives of the interpolated curve to
  the third derivatives of the unique polynomials thru the first four and
  last four points.

  Use Seval or Seval3 to evaluate the spline.
INPUT:
   x  -  abscissas of knots
   y  -  ordinates of knots
OUTPUT:
   b  -  linear coeff
   c  -  quadratic coeff.
   d  -  cubic coeff.
  All arrays have the same limits }
procedure FMMSpline(const x, y, b, c, d: IFArr1D);

{ Pascal version by Nikolai V. Shokhirev.
  Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
       Computer methods for mathematical computations. Prentice-Hall, 1977.
PURPOSE:
  Construct the natural spline thru a set of points
NOTES:
  A natural spline has zero second derivative at both endpoints.
INPUT:
  x, y - coordinates of knots
OUTPUT:
   b, c, d  -  linear, quadratic and cubic coefficients          }
procedure NaturalSpline(const x, y, b, c, d: IFArr1D);

{ Pascal version by Nikolai V. Shokhirev.
  Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
       Computer methods for mathematical computations. Prentice-Hall, 1977.
PURPOSE:
  Evaluate the cubic spline function
     Seval=y[i]+b[i]!(u-x[i])+c[i]*(u-x[i])**2+d[i]*(u-x[i])**3
           where  x[i] <= u < x[i+1]
NOTES:
  if u<x[Lo], i = Lo is used;if u>x[Hi], i = Hi is used
INPUT:
  u - abscissa at which the spline is to be evaluated
  x - abscissas of knots
  y - ordinates of knots
  b, c, d - linear, quadratic, cubic coeff
IN-OUT
  i - last used index; should be set to x.Lo1 before the 1st call }
function Seval(const u: TFloat; var i: TInt; const x, y, b, c, d: IFArr1D): TFloat;

{ Pascal version by Nikolai V. Shokhirev.
  Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
       Computer methods for mathematical computations. Prentice-Hall, 1977.
PURPOSE:
  Evaluate the cubic spline function
     Seval=y[i]+b[i]!(u-x[i])+c[i]*(u-x[i])**2+d[i]*(u-x[i])**3
           where  x[i] <= u < x[i+1]

NOTES:
  if u<x[Lo], i=1 is used;if u>x[Hi], i=Dim is used

  u - abscissa at which the spline is to be evaluated
  x - abscissas of knots
  y - ordinates of knots
  b, c, d - linear,quadratic,cubic coeff
  f, fp, fpp, fppp - function, 1st, 2nd, 3rd derivatives
IN-OUT
  i - last used index; should be set to x.Lo1 before the 1st call }
procedure Seval3(const u: TFloat; var i: TInt; const x, y, b, c, d: IFArr1D;
                       var f, fp, fpp, fppp: TFloat);

{
PURPOSE:
  Linear interpolation of 1D data
NOTES:
  Use LinSpline1D for fast lookup of dense data
  Use FMMSpline / Seval or spline / splint for higher accuracy
INPUT:
  x - 1D point ( Dat1D.x1 )
RESULT:
  f(x) - interpolated value ( Dat1D.x2 );              }
function LinSpline1D(x: TFloat; const Dat1D: IFData1D): TFloat; overload;

function LinSpline1D(x: TFloat; const Dat1D: IFSignal1D): TFloat; overload;

function LinSpline1D(x: TFloat; const Dat1D: IFArr1D;
                                      Start, Step: TFloat): TFloat; overload;

{
PURPOSE:
  Linear interpolation of 2D data
NOTES:
  Use LinSpline2D for fast lookup of dense data
  Use splie2 / spline for higher accuracy
INPUT:
  (t1, t2) - 2D point ( Dat2D.x1, Dat2D.x2 )
RESULT:
  f(t1, t2) - interpolated value ( Dat2D.z );              }
function LinSpline2D(t1, t2: TFloat; const Dat2D: IFData2D): TFloat; overload;

function LinSpline2D(t1, t2: TFloat; const Dat2D: IFSignal2D): TFloat; overload;

function LinSpline2D(t1, t2: TFloat; const Dat2D: IFArr2D;
                     Start1, Step1, Start2, Step2: TFloat): TFloat; overload;

{ f = f(t1,t2), d1_ = f'1(t1,t2) , d_2 = f'2(t1,t2) , d12 = f"12(t1,t2) }
procedure LinExpand2D(t1, t2: TFloat; const Dat2D: IFArr2D;
                     Start1, Step1, Start2, Step2: TFloat;
                     var f, d1_, d_2, d12: TFloat);
implementation

uses
  Math, uDynDataUtils;

procedure spline(const x, y, y2: IFArr1D;  yp1: TFloat; ypn: TFloat);
var
  n, Lo, Hi: TInt;
  i, k: TInt;
  p, qn, sig, un: TFloat;
  u: IFArr1D;

  function dyLo: TFloat;
  begin
    if n = 3 then
      result := FD3d1(x[Lo],x[Lo+1],x[Lo+2],y[Lo],y[Lo+1],y[Lo+2],1)
    else
      result := FD4d1(x[Lo],x[Lo+1],x[Lo+2],x[Lo+3],y[Lo],y[Lo+1],y[Lo+2],y[Lo+3],1);
  end;

  function dyHi: TFloat;
  begin
    if n = 3 then
      result := FD3d1(x[Hi-2],x[Hi-1],x[Hi],y[Hi-2],y[Hi-1],y[Hi],3)
    else
      result := FD4d1(x[Hi-3],x[Hi-2],x[Hi-1],x[Hi],y[Hi-3],y[Hi-2],y[Hi-1],y[Hi],4);
  end;

begin
  n := x.Dim1;
  Hi := x.Hi1;
  Lo := x.Lo1;
  Assert( n > 2, 'Spline: N < 3');
  u := TFArr1D.Create(x);

  if (yp1 > 0.99e30) then
  begin
     y2[Lo] := 0.0;
     u[Lo] := 0.0;
  end else
  begin
     y2[Lo] := -0.5;
     if (yp1 < -0.99e30) then yp1 := dyLo;

     u[Lo] := (3.0/(x[Lo+1]-x[Lo]))*((y[Lo+1]-y[Lo])/(x[Lo+1]-x[Lo])-yp1);
  end;
  for i := Lo+1 to Hi-1 do
  begin
    sig := (x[i]-x[i-1])/(x[i+1]-x[i-1]);
    p := sig*y2[i-1]+2.0;
    y2[i] := (sig-1.0)/p;
    u[i] := (6.0*( (y[i+1]-y[i])/(x[i+1]-x[i])
                  -(y[i]-y[i-1])/(x[i]-x[i-1]) )/(x[i+1]-x[i-1])-sig*u[i-1])/p;
//    u[i] := (y[i+1]-y[i])/(x[i+1]-x[i])-(y[i]-y[i-1])/(x[i]-x[i-1]);
//    u[i] := (6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  end;
   if (ypn > 0.99e30) then
   begin
      qn := 0.0;
      un := 0.0
   end else
   begin
      qn := 0.5;
      if (ypn < -0.99e30) then ypn := dyHi;
      un := (3.0/(x[Hi]-x[Hi-1]))*(ypn-(y[Hi]-y[Hi-1])/(x[Hi]-x[Hi-1]));
   end;
   y2[Hi] := (un-qn*u[Hi-1])/(qn*y2[Hi-1]+1.0);
   for k := Hi-1 downto Lo do
   begin
      y2[k] := y2[k]*y2[k+1]+u[k];
   end
end;

procedure splint(const xa, ya, y2a: IFArr1D; x: TFloat; var y: TFloat);
var
  klo, khi, k: TInt;
  h, b, a: TFloat;
begin
   klo := xa.Lo1;
   khi := xa.Hi1;
   while (khi-klo > 1) do
   begin
      k := (khi+klo) div 2;
      if (xa[k] > x) then khi := k else klo := k
   end;
   h := xa[khi]-xa[klo];
   Assert(h<>0.0, 'splint: equal nodes');
   a := (xa[khi]-x)/h;
   b := (x-xa[klo])/h;
   y := a*ya[klo]+b*ya[khi]+ ((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
end;

procedure splintd(const xa, ya, y2a: IFArr1D; x: TFloat; var y: TFloat);
var
  klo, khi, k: TInt;
  h, b, a: TFloat;
begin
   klo := xa.Lo1;
   khi := xa.Hi1;
   while (khi-klo > 1) do
   begin
      k := (khi+klo) div 2;
      if (xa[k] > x) then khi := k else klo := k
   end;
   h := xa[khi]-xa[klo];
   Assert(h<>0.0, 'splint: equal nodes');
   a := (xa[khi]-x)/h;
   b := (x-xa[klo])/h;
   y := (ya[khi]-ya[klo])/h + ((3.0*b*b-1.0)*y2a[khi]-(3.0*a*a-1.0)*y2a[klo])*h/6.0;
end;

procedure splie2(const x1a, x2a: IFArr1D; const ya, y2a: IFArr2D);
var
   k, j: TInt;
   ytmp, y2tmp: IFArr1D;
begin
  ytmp := TFArr1D.Create(ya.Lo2, ya.Hi2);
  y2tmp := TFArr1D.Create(ya.Lo2, ya.Hi2);
  for j := ya.Lo1 to ya.Hi1 do
  begin
    for k := ya.Lo2 to ya.Hi2 do
    begin
      ytmp[k] := ya[j,k];
    end;
    spline(x2a, ytmp, y2tmp); // = spline(x2a, ytmp, y2tmp,-1.0e30,-1.0e30);
    for k := ya.Lo2 to ya.Hi2 do
    begin
      y2a[j,k] := y2tmp[k];
    end;
  end;
end;


procedure splin2(const x1a, x2a: IFArr1D; const ya, y2a: IFArr2D;
                 x1, x2: TFloat; var y: TFloat);
var
//  m, n, nm,
  k, j: TInt;
  ytmp, y2tmp, yytmp, y1tmp: IFArr1D;
  yy: TFloat;
begin
//  m := ya.Dim1;
//  n := ya.Dim2;
//  nm := max(n, m);  // ??
  ytmp  := TFArr1D.Create(ya.Lo2, ya.Hi2);
  y2tmp := TFArr1D.Create(ya.Lo2, ya.Hi2);
  yytmp := TFArr1D.Create(ya.Lo1, ya.Hi1);
  y1tmp := TFArr1D.Create(ya.Lo1, ya.Hi1);

  for j := ya.Lo1 to ya.Hi1 do
  begin
    for k := ya.Lo2 to ya.Hi2 do
    begin
      ytmp[k] := ya[j,k];
      y2tmp[k] := y2a[j,k]
    end;
    splint(x2a,ytmp,y2tmp, x2,yy);
    yytmp[j] := yy;
  end;
  spline(x1a, yytmp, y1tmp); // = spline(x1a, yytmp, y2tmp,-1.0e30,-1.0e30);
  splint(x1a, yytmp, y1tmp, x1 ,y)
end;


{ Xa[i] < x <= Xa[i+1] }
function Index2(x: TFloat; const Xa: IFArr1D): TInt;
var
  klo, khi, k: TInt;
begin
   klo := Xa.Lo1;
   khi := Xa.Hi1;
   while (khi-klo > 1) do
   begin
      k := (khi+klo) div 2;
      if (Xa[k] > x) then khi := k else klo := k
   end;
   result := klo;
end;


function Index0(x: TFloat; const Xa: IFArr1D): TInt;
var
  i: TInt;
begin
  result := Xa.Lo1;
  if Xa.Dim1 < 2 then exit;

  result := Xa.Lo1+1;
  if Xa.Dim1 < 4 then exit;
  if x < Xa[result] then exit;

  i := Index2(x, Xa);
  result := i;
  if i = Xa.Hi1-1 then exit;
  if (abs(Xa[i-1]+Xa[i]+Xa[i+1]-x*3.0) >= abs(Xa[i]+Xa[i+1]+Xa[i+2]-x*3.0)) then
    inc(result);
end;


procedure FMMSpline(const x, y, b, c, d: IFArr1D);
const
  ZERO  = 0.0;
  TWO   = 2.0;
  THREE = 3.0;
var
  k, Dim, Lo, Hi: TInt;
  t: TFloat;
begin
  Dim := x.Dim1;
  Lo := x.Lo1;
  Hi := x.Hi1;

  if (Dim < 3) then
  begin // Straight line - special case for n < 3
    b[Lo] := ZERO;
    if (Dim = 2) then b[Lo] := (y[Lo+1]-y[Lo])/(x[Lo+1]-x[Lo]);
    c[Lo] := ZERO;
    d[Lo] := ZERO;
    if (Dim < 2) then exit;
    b[Lo+1] := b[Lo];
    c[Lo+1] := ZERO;
    d[Lo+1] := ZERO;
    exit;
  end;

//.....Set up tridiagonal system.........................................
//.     b = diagonal, d = offdiagonal, c = right-hand side
  d[Lo]   := x[Lo+1]-x[Lo];
  c[Lo+1] := (y[Lo+1]-y[Lo])/d[Lo];
//  for k  :=  2 to n-1 do
  for k := Lo+1 to Hi-1 do
  begin
    d[k]   := x[k+1]-x[k];
    b[k]   := TWO*(d[k-1]+d[k]);
    c[k+1] := (y[k+1]-y[k])/d[k];
    c[k]   := c[k+1]-c[k];
  end;

//.....End conditions.  third derivatives at x[Lo] and x[Hi] obtained
//.       from divided differences.......................................
  b[Lo] := -d[Lo];
  b[Hi] := -d[Hi-1];
  c[Lo] := ZERO;
  c[Hi] := ZERO;
  if (Dim > 3) then
  begin
    c[Lo] := c[Lo+2]/(x[Lo+3]-x[Lo+1]) - c[Lo+1]/(x[Lo+2]-x[Lo]);
    c[Hi] := c[Hi-1]/(x[Hi]-x[Hi-2]) - c[Hi-2]/(x[Hi-1]-x[Hi-3]);
    c[Lo] := c[Lo]*d[Lo]*d[Lo]/(x[Lo+3]-x[Lo]);
    c[Hi] := -c[Hi]*d[Hi-1]*d[Hi-1]/(x[Hi]-x[Hi-3]);
  end;

  for  k  := Lo+1 to Hi do
  begin // forward elimination
    t := d[k-1]/b[k-1];
    b[k] := b[k]-t*d[k-1];
    c[k] := c[k]-t*c[k-1];
  end;

  c[Hi] := c[Hi]/b[Hi];
// back substitution ( makes c the sigma of text)
  for  k := Hi-1 downto Lo do
  begin
    c[k] := (c[k]-d[k]*c[k+1])/b[k];
  end;

//.....Compute polynomial coefficients...................................
  b[Hi] := (y[Hi]-y[Hi-1])/d[Hi-1]+d[Hi-1]*(c[Hi-1]+c[Hi]+c[Hi]);
  for  k := Lo to Hi-1  do
  begin
    b[k] := (y[k+1]-y[k])/d[k]-d[k]*(c[k+1]+c[k]+c[k]);
    d[k] := (c[k+1]-c[k])/d[k];
    c[k] := THREE*c[k];
  end;
  c[Hi] := THREE*c[Hi];
  d[Hi] := d[Hi-1];

  exit;
end;// of FMMspline ---------------------------------------------------------


procedure NaturalSpline(const x, y, b, c, d: IFArr1D);
const
  ZERO  = 0.0;
  TWO   = 2.0;
  THREE = 3.0;
var
  k, Dim, Lo, Hi: TInt;
begin
  Lo := x.Lo1;
  Hi := x.Hi1;
  Dim := x.Dim1;

  if (Dim < 3) then
  begin // Straight line - special case for Dim < 3
    b[Lo] := ZERO;
    if (Dim = 2) then b[Lo]:=(y[Lo+1]-y[Lo])/(x[Lo+1]-x[Lo]);
    c[Lo] := ZERO;
    d[Lo] := ZERO;
    b[Lo+1] := b[Lo];
    c[Lo+1] := ZERO;
    d[Lo+1] := ZERO;
    exit;
  end;

//  d[1:n-1] := x[2:n]-x[1:n-1]; - Put the h-array of the text into array d
  for k := Lo to Hi-1 do
    d[k] := x[k+1]-x[k];

//.....Set up the upper triangular system in locations 2 thru Dim-1 of
//        arrays b and c. B holds the diagonal and c the right hand side.
  b[Lo+1] := TWO*(d[Lo]+d[Lo+1]);
  c[Lo+1] := (y[Lo+2]-y[Lo+1])/d[Lo+1]-(y[Lo+1]-y[Lo])/d[Lo];
  for k := Lo+2 to Hi-1  do
  begin
    b[k] := TWO*(d[k-1]+d[k])-d[k-1]*d[k-1]/b[k-1];
    c[k] := (y[k+1]-y[k])/d[k]-(y[k]-y[k-1])/d[k-1]-d[k-1]*c[k-1]/b[k-1];
  end;

  c[Hi-1] := c[Hi-1]/b[Hi-1];
// Back substitute to get c-array
  for  k := Hi-2 downto Lo+1  do begin
    c[k] := (c[k]-d[k]*c[k+1])/b[k];
  end;
  c[Lo] := ZERO;
  c[Hi] := ZERO   ;
// c now holds the sigma array of the text

//.....Compute polynomial coefficients ..................................
  b[Hi] := (y[Hi]-y[Hi-1])/d[Hi-1]+d[Hi-1]*(c[Hi-1]+c[Hi]+c[Hi]);
  for  k := Lo to Hi-1 do
  begin
    b[k] := (y[k+1]-y[k])/d[k]-d[k]*(c[k+1]+c[k]+c[k]);
    d[k] := (c[k+1]-c[k])/d[k];
    c[k] := THREE*c[k];
  end;
  c[Hi] := THREE*c[Hi];
  d[Hi] := d[Hi-1];
end;


function Seval(const u: TFloat; var i: TInt; const x, y, b, c, d: IFArr1D): TFloat;
var
  k, Lo, Hi, j: TInt;
  dx: TFloat;
begin
//  Dim := SIZE(x);
  Lo := x.Lo1;
  Hi := x.Hi1;

//.....First check if u is in the same interval found on the last call to Seval
  if (i < Lo) then  i := Lo;
  if (i >= Hi) then  i := Hi;
  if ( (u < x[i]) or (u >= x[i+1]) ) then
  begin
    i := Lo ;
    j := Hi+1;
    // binary search
    repeat
      k := (i+j) div 2;
      if (u < x[k]) then j := k else i := k;
    until (j <= i+1);
  end;
  dx := u-x[i];
// evaluate the spline
  Result := y[i]+dx*(b[i]+dx*(c[i]+dx*d[i]));
end;


procedure Seval3(const u: TFloat; var i: TInt; const x, y, b, c, d: IFArr1D;
                       var f, fp, fpp, fppp: TFloat);
const
  TWO = 2.0;
  THREE = 3.0;
  SIX = 6.0;
var
  k, Lo, Hi,j: TInt;
  dx: TFloat;
begin
//  Dim := SIZE(x);
  Lo := x.Lo1;
  Hi := x.Hi1;

//.....First check if u is in the same interval found on the
//        last call to Seval.............................................
  if (i < Lo) then  i := Lo;
  if (i >= Hi) then  i := Hi;
  if ( (u < x[i]) or (u >= x[i+1]) ) then
  begin
    i := Lo ;
    j := Hi+1;
    // binary search
    repeat
      k := (i+j) div 2;
      if (u < x[k]) then j := k else i := k;
    until (j <= i+1);
  end;
  dx := u-x[i]   ;
// evaluate the spline
  f:=y[i]+dx*(b[i]+dx*(c[i]+dx*d[i]));
  fp:=b[i]+dx*(TWO*c[i] + dx*THREE*d[i]);
  fpp:=TWO*c[i] + dx*SIX*d[i];
  fppp:=SIX*d[i];
end;

function LinSpline1D(x: TFloat; const Dat1D: IFData1D): TFloat;
var
  i: TInt;
  h: TFloat;
begin
  i := Index2(x, Dat1D.x1);
  h := Dat1D.x1[i+1]-Dat1D.x1[i];
  Assert(h<>0.0, 'LinearSpline1D: equal nodes');
  result := Dat1D.x2[i]+(Dat1D.x2[i+1]-Dat1D.x2[i])*(x-Dat1D.x1[i])/h;
end;

function LinSpline1D(x: TFloat; const Dat1D: IFSignal1D): TFloat;
var
  i: TInt;
  h, d: TFloat;
begin
  h := Dat1D.Step1;
  Assert(h<>0.0, 'LinearSpline1D: equal nodes');
  d := (x-Dat1D.Start1)/h;
  i := Dat1D.Lo1+Floor(d);
  if i < Dat1D.Lo1 then i := Dat1D.Lo1;
  if i >= Dat1D.Hi1 then i := Dat1D.Hi1-1;
  d := d - i + Dat1D.Lo1;
  result := Dat1D[i]+(Dat1D[i+1]-Dat1D[i])*d;
end;

function LinSpline1D(x: TFloat; const Dat1D: IFArr1D;
                                      Start, Step: TFloat): TFloat;
var
  i: TInt;
  d: TFloat;
begin
  Assert(Step > 0.0, 'LinearSpline1D: equal nodes');
  d := (x-Start)/Step;
  i := Dat1D.Lo1+Floor(d);
  if i < Dat1D.Lo1 then i := Dat1D.Lo1;
  if i >= Dat1D.Hi1 then i := Dat1D.Hi1-1;
  d := d - i + Dat1D.Lo1;
  result := Dat1D[i]+(Dat1D[i+1]-Dat1D[i])*d;
end;

function LinSpline2D(t1, t2: TFloat; const Dat2D: IFData2D): TFloat;
var
  i1, i2: TInt;
  h1, h2, a, b, c, d: TFloat;
begin
  i1 := Index2(t1, Dat2D.x1);
  i2 := Index2(t2, Dat2D.x2);
  h1 := Dat2D.x1[i1+1]-Dat2D.x1[i1];
  h2 := Dat2D.x2[i2+1]-Dat2D.x2[i2];
  Assert(h1<>0.0, 'LinearSpline2D: equal x1 nodes');
  Assert(h2<>0.0, 'LinearSpline2D: equal x2 nodes');
  a := Dat2D.z[i1, i2];
  b := Dat2D.z[i1+1, i2] - a;
  c := Dat2D.z[i1, i2+1] - a;
  d := Dat2D.z[i1+1,i2+1] - a - b - c;
  result := a + b*(t1-Dat2D.x1[i1])/h1 + c*(t2-Dat2D.x2[i2])/h2 +
            d*(t1-Dat2D.x1[i1])*(t2-Dat2D.x2[i2])/h1/h2;
end;

function LinSpline2D(t1, t2: TFloat; const Dat2D: IFSignal2D): TFloat;
var
  i1, i2: TInt;
  h1, h2, d1, d2, a, b, c, d: TFloat;
begin
  h1 := Dat2D.Step1;
  h2 := Dat2D.Step2;
  Assert(h1<>0.0, 'LinearSpline2D: equal x1 nodes');
  Assert(h2<>0.0, 'LinearSpline2D: equal x2 nodes');

  d1 := (t1-Dat2D.Start1)/h1;
  d2 := (t2-Dat2D.Start2)/h2;
  i1 := Dat2D.Lo1+Floor(d1);
  i2 := Dat2D.Lo2+Floor(d2);

  if i1 < Dat2D.Lo1 then i1 := Dat2D.Lo1;
  if i1 >= Dat2D.Hi1 then i1 := Dat2D.Hi1-1;
  if i2 < Dat2D.Lo2 then i2 := Dat2D.Lo2;
  if i2 >= Dat2D.Hi2 then i2 := Dat2D.Hi2-1;

  d1 := d1 - i1 + Dat2D.Lo1;
  d2 := d2 - i2 + Dat2D.Lo2;
  a := Dat2D[i1, i2];
  b := Dat2D[i1+1, i2] - a;
  c := Dat2D[i1, i2+1] - a;
  d := Dat2D[i1+1,i2+1] - a - b - c;
  result := a + b*d1 + c*d2 + d*d1*d2;
end;

function LinSpline2D(t1, t2: TFloat; const Dat2D: IFArr2D;
                     Start1, Step1, Start2, Step2: TFloat): TFloat;
var
  i1, i2: TInt;
  d1, d2, a, b, c, d: TFloat;
begin
  Assert(Step1 > 0.0, 'LinearSpline2D: equal x1 nodes');
  Assert(Step2 > 0.0, 'LinearSpline2D: equal x2 nodes');

  d1 := (t1-Start1)/Step1;
  d2 := (t2-Start2)/Step2;
//  i1 := Dat2D.Lo1+Floor(d1);
//  i2 := Dat2D.Lo2+Floor(d2);
  i1 := Dat2D.Lo1+round(d1);
  i2 := Dat2D.Lo2+round(d2);

  if i1 < Dat2D.Lo1 then i1 := Dat2D.Lo1;
  if i1 >= Dat2D.Hi1 then i1 := Dat2D.Hi1-1;
  if i2 < Dat2D.Lo2 then i2 := Dat2D.Lo2;
  if i2 >= Dat2D.Hi2 then i2 := Dat2D.Hi2-1;

  d1 := d1 - i1 + Dat2D.Lo1;
  d2 := d2 - i2 + Dat2D.Lo2;
  a := Dat2D[i1, i2];
  b := Dat2D[i1+1, i2] - a;
  c := Dat2D[i1, i2+1] - a;
  d := Dat2D[i1+1,i2+1] - a - b - c;
  result := a + b*d1 + c*d2 + d*d1*d2;
end;

procedure LinExpand2D(t1, t2: TFloat; const Dat2D: IFArr2D;
                     Start1, Step1, Start2, Step2: TFloat;
                     var f, d1_, d_2, d12: TFloat);
var
  i1, i2: TInt;
  d1, d2, a, b, c, d: TFloat;
begin
  Assert(Step1 > 0.0, 'LinearSpline2D: equal x1 nodes');
  Assert(Step2 > 0.0, 'LinearSpline2D: equal x2 nodes');

  d1 := (t1-Start1)/Step1;
  d2 := (t2-Start2)/Step2;
  i1 := Dat2D.Lo1+Floor(d1);
  i2 := Dat2D.Lo1+Floor(d2);

  if i1 < Dat2D.Lo1 then i1 := Dat2D.Lo1;
  if i1 >= Dat2D.Hi1 then i1 := Dat2D.Hi1-1;
  if i2 < Dat2D.Lo2 then i2 := Dat2D.Lo2;
  if i2 >= Dat2D.Hi2 then i2 := Dat2D.Hi2-1;

  d1 := d1 - i1 + Dat2D.Lo1;
  d2 := d2 - i2 + Dat2D.Lo2;
  a := Dat2D[i1, i2];
  b := Dat2D[i1+1, i2] - a;
  c := Dat2D[i1, i2+1] - a;
  d := Dat2D[i1+1,i2+1] - a - b - c;
  f := a + b*d1 + c*d2 + d*d1*d2;
  d1_ := (b+d*d2)/Step1;
  d_2 := (c+d*d1)/Step2;
  d12 := d;
end;

end.
