{*
@abstract(Statistical procedures )
@author(Nikolai V. Shokhirev <nikolai@shokhirev.com> <nikolai@u.arizona.edu>)
@created(2002.02.02)
@lastmod(2006.10.10)
@version 2007.09.30 - TStatRecord, GetStatRecord, enhanced StatMoments
Nikolai V. Shokhirev, 2002-2007
}
unit uStatUtils;

interface

uses
  uMatTypes, uDynArrays, {uComplexType,} uDynData;

type
  TStatRecord = record
    sampleSize: TInt;
    mean: TFloat;
    minValue: TFloat;
    maxValue: TFloat;
    deviation: TFloat;
    variance: TFloat;
    stdDev: TFloat;
    skewness: TFloat;
    excessKurtosis: TFloat;
  end;

{*
  Sample values (not estimators)
  minValue  = min(minV, x[i]);
  maxValue  = max(maxV, x[i]);
  average   = sum(x[i])/n
  deviation = sum(abs(x[i]-average))/n;
  variance  = m2 = sum((x[i]-average)**2)/n   // biased
  stdDev    = sqrt(m2) - standard deviation   // biased
  skewness  = m3/(m2*sqrt(m2));               // biased
  kurtosis  = m4/sqr(m2);                     // biased
}
procedure StatMoments(const x: IFArr1D; var average, variance, stdDev,
                      skewness, kurtosis, deviation, minValue, maxValue: TFloat;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);

{*
  Sample values (not estimators). Direct calculation.
  minValue  = min(minV, x[i]);
  maxValue  = max(maxV, x[i]);
  average   = sum(x[i])/n
  deviation = sum(abs(x[i]-average))/n;
  variance  = m2 = sum((x[i]-average)**2)/n   // biased
  stdDev    = sqrt(m2) - standard deviation   // biased
  skewness  = m3/(m2*sqrt(m2));               // biased
  kurtosis  = m4/sqr(m2);                     // biased
}
procedure StatMoments0(const x: IFArr1D; var average, variance, stdDev,
                       skewness, kurtosis, deviation, minValue, maxValue: TFloat;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);

{*
  Unbiased estimators
  minValue  = min(minV, x[i]);
  maxValue  = max(maxV, x[i]);
  mean      = sum(x[i])/n = average
  deviation = sum(abs(x[i]-average))/n;
  variance  = m2*n/(n-1)
  stdDev    = sqrt(variance) - standard deviation
  biasedSkewness = m3/(m2*sqrt(m2));
  skewness  = sqrt(n*(n-1))*biasedSkewness/(n-2);
  kurtosis = m4/sqr(m2) (biased)
  excessKurtosis := (kurtosis - 3 + 6/(n+1))*(n*n-1)/(n-3)/(n-2)
}
procedure Estimators(const x: IFArr1D; var mean, variance, stdDev,
                     skewness, excessKurtosis, deviation, minValue, maxValue: TFloat;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);

{*
  Unbiased estimators
  minValue  = min(minV, x[i]);
  maxValue  = max(maxV, x[i]);
  mean      = sum(x[i])/n = average
  deviation = sum(abs(x[i]-average))/n;
  variance  = m2*n/(n-1)
  stdDev    = sqrt(variance) - standard deviation
  biasedSkewness = m3/(m2*sqrt(m2));
  skewness  = sqrt(n*(n-1))*biasedSkewness/(n-2);
  kurtosis = m4/sqr(m2) (biased)
  excessKurtosis := (kurtosis - 3 + 6/(n+1))*(n*n-1)/(n-3)/(n-2)
}
function GetStatMoments(const x: IFArr1D;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt): TStatRecord;

{
  xav1 = sum(x1[i])/n
  xav2 = sum(x2[i])/n
  var1 = sum((x1[i]-xav1)**2)/n
  var2 = sum((x2[i]-xav2)**2)/n
  cov - covariance = sum((x1[i]-xav1)*(x2[i]-xav2))/n
  corr := cov/sqrt(var1*var2);                         }
procedure covariance(const x1x2: IFData1D;
                                 var x1av, x2av, var1, var2, cov, corr: TFloat);
                                 
{
Input
  x - data matrix:
  x[i,j] - i-th measurement of j-th parameter
Output
  avg - vector of average parameters
  avg[j] - average of the j-th parameter
  covar - covariance matrix
}
procedure MultiCovar(const x, covar: IFArr2D; const avg: IFArr1D);

{
Input
  x - data matrix:
  x[i,j] - j-th measurement of i-th parameter
Output
  avg - vector of average parameters
  avg[i] - average of the i-th parameter
  d - data with centered parameters: d[i,j] := x[i,j] - avg[i]
  covar - covariance matrix
  covar[i,j] - covariance matrix i >= j
  covar[i,j] - corr[i,j] i < j
}
procedure CovarMatrix(const x, d, covar: IFArr2D; const avg: IFArr1D );


{*
  simple histogram
}
function SimpleHistogram(const Y: IFArr1D; NBins: TInt; var Ymin, Ymax: TFloat;
                         minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt): IFArr1D;

implementation

uses
  SysUtils
  , math
  , uDynArrUtils
  ;

procedure StatMoments(const x: IFArr1D; var average, variance, stdDev,
                      skewness, kurtosis, deviation, minValue, maxValue: TFloat;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);
var
  n, xav2, x1, x2, x3, x4, s1, s2, s3, s4, m2, m3, m4, minV, maxV: TFloat;
  i, Lo, Hi, Dim: TInt;
begin
  SetMinMaxIdx(minIdx, maxIdx, x.Lo1, x.Hi1, Lo, Hi, Dim);
  s1 := 0.0;
  s2 := 0.0;
  s3 := 0.0;
  s4 := 0.0;
  n := Dim;
  minV := MaxFloat;
  maxV := NegMaxFloat;
  for i := Lo to Hi do
  begin
    x1 := x[i];
    minV := min(minV, x1);
    maxV := max(maxV, x1);
    s1 := s1 + x1;
    x2 := sqr(x1);
    s2 := s2 + x2;
    x3 := x1*x2;
    s3 := s3 + x3;
    x4 := sqr(x2);
    s4 := s4 + x4;
  end;

  minValue := minV;
  maxValue := maxV;
  average := s1/n;

  s1 := 0.0;
  for i := Lo to Hi do
    s1 := s1 + abs(x[i]-average);
  deviation := s1/n;

  xav2 := sqr(average);
  m2 := s2/n - xav2;
  if m2 < MachEps then m2 := 0.0;

  variance := m2;
  stdDev := sqrt(m2);

  m3 := s3/n-3.0*average*s2/n+2.0*xav2*average;
  if abs(m3) < MachEps then m3 := 0.0;

  m4 := (s4-4.0*average*s3+6.0*xav2*s2)/n-3.0*sqr(xav2);
  if m4 < MachEps then m4 := 0.0;

  if m2 > 0.0 then
  begin
    kurtosis := m4/sqr(m2);
    skewness := m3/(m2*sqrt(m2));
  end else
  begin
    kurtosis := 0.0;
    skewness := 0.0;
  end;
end;

procedure StatMoments0(const x: IFArr1D; var average, variance, stdDev,
                       skewness, kurtosis, deviation, minValue, maxValue: TFloat;
                       minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);
var
  n, x1, x2, x3, x4, s1, m1, m2, m3, m4, minV, maxV: TFloat;
  i, Lo, Hi, Dim: TInt;
begin
  SetMinMaxIdx(minIdx, maxIdx, x.Lo1, x.Hi1, Lo, Hi, Dim);
  m1 := 0.0;
  n := Dim;
  minV := MaxFloat;
  maxV := NegMaxFloat;
  for i := Lo to Hi do
  begin
    m1 := m1 + x[i];
    minV := min(minV, x[i]);
    maxV := max(maxV, x[i]);
  end;
  minValue := minV;
  maxValue := maxV;

  m1 := m1/n;
  average := m1;

  s1 := 0.0;
  m2 := 0.0;
  m3 := 0.0;
  m4 := 0.0;
  for i := Lo to Hi do
  begin
    x1 := (x[i]-m1);
    s1 := s1 + abs(x1);
    x2 := sqr(x1);
    m2 := m2 + x2;
    x3 := x1*x2;
    m3 := m3 + x3;
    x4 := sqr(x2);
    m4 := m4 + x4;
  end;
  m2 := m2/n;
  if m2 < MachEps then m2 := 0.0;
  variance := m2;
  deviation := s1/n;
  stdDev := sqrt(m2);

  if abs(m3) < MachEps then m3 := 0.0 else m3 := m3/n;
  if m4 < MachEps then m4 := 0.0 else m4 := m4/n;

  if m2 > 0.0 then
  begin
    kurtosis := m4/sqr(m2);
    skewness := m3/(m2*sqrt(m2));
  end else
  begin
    kurtosis := 0.0;
    skewness := 0.0;
  end;
end;

procedure Estimators(const x: IFArr1D; var mean, variance, stdDev,
                     skewness, excessKurtosis, deviation, minValue, maxValue: TFloat;
                     minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt);
var
  n, kurtosis: TFloat;
  Lo, Hi, Dim: TInt;
begin
  StatMoments0(x, mean, variance, deviation, skewness, kurtosis, deviation,
                  minValue, maxValue, minIdx, maxIdx);
  SetMinMaxIdx(minIdx, maxIdx, x.Lo1, x.Hi1, Lo, Hi, Dim);
  n := Dim;
  if Dim > 1 then
    variance := variance*n/(n-1.0)
  else
    variance := 0.0;

  stdDev := sqrt(variance);

  if Dim > 2 then
    skewness := sqrt(n*(n-1.0))*skewness/(n-2.0)
  else
    skewness := 0.0;

  if Dim > 3 then
    excessKurtosis := (kurtosis - 3.0 +6.0/(n+1.0))*(sqr(n)-1.0)/(n-3.0)/(n-2.0)
  else
    excessKurtosis := 0.0;
end;

function GetStatMoments(const x: IFArr1D;
                      minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt): TStatRecord;
var
  sr: TStatRecord;
  Lo, Hi, Dim: TInt;
begin
  Estimators(x, sr.mean, sr.variance, sr.stdDev, sr.skewness, sr.excessKurtosis,
                sr.deviation, sr.minValue, sr.maxValue, minIdx, maxIdx);
  SetMinMaxIdx(minIdx, maxIdx, x.Lo1, x.Hi1, Lo, Hi, Dim);
  sr.sampleSize := Dim;
  result := sr;
end;

procedure covariance(const x1x2: IFData1D;
                                 var x1av, x2av, var1, var2, cov, corr: TFloat);
var
  n, x1, x2, s1, s2, s11, s22, s12: TFloat;
  i: TInt;
begin
  s1 := 0.0;
  s2 := 0.0;
  s11 := 0.0;
  s22 := 0.0;
  s12 := 0.0;
  n := x1x2.Dim1;
  for i := x1x2.Lo1 to x1x2.Hi1 do
  begin
    x1 := x1x2.x1[i];
    x2 := x1x2.x2[i];
    s1 := s1 + x1;
    s2 := s2 + x2;
    s11 := s11 + sqr(x1);
    s22 := s22 + sqr(x2);
    s12 := s12 + x1*x2;
  end;
  x1av := s1/n;
  var1 := s11/n-sqr(x1av);
  x2av := s2/n;
  var2 := s22/n-sqr(x2av);
  cov := (s12 - x1av*s2 - x2av*s1)/n + x1av*x2av;
  corr := cov/sqrt(var1*var2);
end;

{
Input
  x - data matrix:
  x[i,j] - i-th measurement of j-th parameter
Output
  avg - vector of average parameters
  avg[j] - average of the j-th parameter
  covar - covariance matrix
}
procedure MultiCovar(const x, covar: IFArr2D; const avg: IFArr1D);
var
  s, ai, aj, n: TFloat;
  i1, i2, j2: TInt;
begin
  n := x.Dim1;

  if (x.Lo2<>avg.Lo1) or (x.Hi2<>avg.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch+': avg');

  if not IsSquare(covar) or (x.Lo2<>covar.Lo1) or (x.Hi2<>Covar.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch+': covar');

  for i2 := x.Lo2 to x.Hi2 do
  begin
    s := 0.0;
    for i1 := x.Lo1 to x.Hi1 do
      s := s + x[i1,i2];
    avg[i2] := s/n;
  end;

  for i2 := x.Lo2 to x.Hi2 do
  begin
    ai := avg[i2];
    for j2 := x.Lo2 to i2 do
    begin
      aj := avg[j2];
      s := 0.0;
      for i1 := x.Lo1 to x.Hi1 do
      begin
        s := s + (x[i1,i2]-ai)*(x[i1,j2]-aj);
      end;
      covar[i2,j2] := s/(n-1.0);     // j2 <= i2
    end;
  end;

  for i2 := x.Lo2+1 to x.Hi2 do
  begin
    for j2 := x.Lo2 to i2-1 do
    begin
        covar[j2,i2] := covar[i2,j2]/sqrt(covar[i2,i2]*covar[j2,j2]);
    end;
  end;
end;

{
Input
  x - data matrix:
  x[i,j] - j-th measurement of i-th parameter
Output
  avg - vector of average parameters
  d - data with centered parameters: d[i,j] := x[i,j] - avg[i]
  covar - covariance matrix
  covar[i,j] - covariance matrix i >= j
  covar[i,j] - corr[i,j] i < j
}
procedure CovarMatrix(const x, d, covar: IFArr2D; const avg: IFArr1D );
var
  s, n: TFloat;
  i1, i2, j1: TInt;
begin
  n := x.Dim2;

  if not SameLim(x, d) then
    Raise ERangeError.Create(RS_LimMismatch+': d');

  if (x.Lo1<>avg.Lo1) or (x.Hi1<>avg.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch+': avg');

  if not IsSquare(covar) or (x.Lo1<>covar.Lo1) or (x.Hi1<>Covar.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch+': covar');

  for i1 := x.Lo1 to x.Hi1 do
  begin
    s := 0.0;
    for i2 := x.Lo2 to x.Hi2 do
      s := s + x[i1,i2];
    avg[i1] := s/n;
  end;

  for i2 := d.Lo2 to d.Hi2 do
    for i1 := d.Lo1 to d.Hi1 do
      d[i1,i2] := x[i1,i2] - avg[i1];

  for i1 := x.Lo1 to x.Hi1 do
  begin
    for j1 := x.Lo1 to i1 do
    begin
      s := 0.0;
      for i2 := x.Lo2 to x.Hi2 do
      begin
        s := s + d[i1,i2]*d[j1,i2];
      end;
      covar[i1,j1] := s/(n-1.0);
    end;
  end;

  for i1 := x.Lo1+1 to x.Hi1 do
  begin
    for j1 := x.Lo1 to i1-1 do
    begin
        covar[j1,i1] := covar[i1,j1]/sqrt(covar[i1,i1]*covar[j1,j1]);
    end;
  end;
end;

function SimpleHistogram(const Y: IFArr1D; NBins: TInt; var Ymin, Ymax: TFloat;
                         minIdx: TInt = MaxInt; maxIdx: TInt = MaxInt): IFArr1D;
var
  histogram: IFArr1D;
  n, w: TFloat;
  i, Lo, Hi, d: TInt;

  function Idx(x: TFloat): TInt;
  begin
    result := 1 + Trunc((x-Ymin)*n/(Ymax-Ymin));
    if result > NBins then
      result := NBins;
  end;

begin
  SetMinMaxIdx(minIdx, maxIdx, Y.Lo1, Y.Hi1, Lo, Hi, d);
  n := NBins;
  MinMaxVec(Y, Ymin, Ymax, true);
  histogram := TFArr1D.Create(NBins);
  histogram.Fill(0.0);
  w := n/(Ymax - Ymin)/d;
  for i := Lo to Hi do
  begin
    d := Idx(Y[i]);
    histogram[d] := histogram[d] + w;
  end;
  result := histogram;
end;

end.
