{*
 Numerical integration
 Nikolai V. Shokhirev, 2002-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2002.02.02 - created
 @version 2004.12.24 - modified
 @version 2007.08.08 - Added DelphiCodeToDoc comments
 }
unit uIntegration;

interface

uses
  uMatTypes;

{
  Pascal version by Nikolai V. Shokhirev                                   
PURPOSE:
  Estimate the integral of FUN(X) from A to B to a user provided tolerance.
  This is an automatic adaptive routine based on the 8-panel Newton-Cotes rule.
Ref.
  George E. Forsythe, Michael A. Malcolm, and Cleve B. Moler. 
  "Computer Methods for Mathematical Computations". Prentice-Hall, 1977.
INPUT:
  a - lower limit of integration
  b - upper limit of integration
  abserr - absolute error tolerance (>0)
  relerr - relative error tolerance (>0)
OUTPUT:
  result - approx. value of integral
  errest - estimate of actual error
  nofun  - no. of function evaluations
  flag   - reliability indicator. ( = 0 is O.K.)   }
procedure Quanc8(Fun: TFloatFunction1D; const a, b, abserr, relerr: TFloat;
                            var result, errest, flag: TFloat; var nofun: TInt);


implementation

uses
  math;

procedure Quanc8(Fun: TFloatFunction1D; const a, b, abserr, relerr: TFloat;
                            var result, errest, flag: TFloat; var nofun: TInt);
label
  30, 50, 70, 80;

const
  ZERO: TFloat = 0.0;
  HALF = 0.5;
  ONE = 1.0;
  SIXTEEN =16.0;

  LEVMIN = 1;
  LEVOUT = 6;
  NOMAX = 5000;
{
  W0 =  0.27908289241622574955908289241623; // =   3956.0/14175.0;
  W1 =  1.6615167548500881834215167548501;  // =  23552.0/14175.0;
  W2 = -0.26186948853615520282186948853616; // =  -3712.0/14175.0;
  W3 =  2.9618342151675485008818342151675;  // =  41984.0/14175.0;
  W4 = -1.2811287477954144620811287477954;  // = -18160.0/14175.0;
}
var
  area: TFloat;
  cor11: TFloat;
  esterr: TFloat;
  f, x: array[0..16] of TFloat;
  fsave, xsave: array[1..8,1..30] of TFloat;
  i, j, jj: TInt;
  lev, nim: TInt;
  NOFIN: TInt;  //...Trouble if NOFUN reaches NOFIN
  qprev, qnow, qdiff, qleft: TFloat;
  stone, step: TFloat;
  tolerr: TFloat;
  qright: array[1..31] of TFloat;
  LEVMAX: TInt;
  W0, W1, W2, W3, W4: TFloat;


begin
//..... S t a g e   1  (General initialization)..........................
  LEVMAX  :=  30;
  W0 :=   3956.0/14175.0;
  W1 :=  23552.0/14175.0;
  W2 :=  -3712.0/14175.0;
  W3 :=  41984.0/14175.0;
  W4 := -18160.0/14175.0;
//   NOFIN := NOMAX-8*(LEVMAX-LEVOUT+2**(LEVOUT+1));
  NOFIN := NOMAX - 8*(LEVMAX - LEVOUT + Ipower(2, LEVOUT+1));

//...initialize running sums to zero...
  area  := ZERO;
  cor11 := ZERO;
  flag  := ZERO;
  result := ZERO;
  errest := ZERO;
  nofun  := 0;
  if (a = b) then exit;

//..... S T A G E   2   (initialization for first interval)..............
  lev := 0;
  nim := 1;
  x[0] := a;
  x[16] := b;
  qprev := ZERO;
  stone := (b-a)/SIXTEEN;
  x[8] :=  HALF*(x[0]+x[16]);
  x[4] :=  HALF*(x[0]+x[8]);
  x[12] := HALF*(x[8]+x[16]);
  x[2] :=  HALF*(x[0]+x[4]);
  x[6] :=  HALF*(x[4]+x[8]);
  x[10] := HALF*(x[8]+x[12]);
  x[14] := HALF*(x[12]+x[16]);

//  for  j := 0 to 16 step 2 do
//  f[j] := Fun(x[j]);
//  for  j := 0 to 8 do
  for  j := 0 to 8 do
    f[j*2] := Fun(x[j*2]);
  nofun := 9;

//..... S T A G E   3   (central calculation)............................
//...requires qprev, x0, x2, x4,....x[16], f0, f2,f 4,....f16.
//...calculates x1, x3,....x15, f1, f3,....f15, qleft, qright, qnow, qdiff, area

// 30for  j := 1 to 15 step 2 do begin
30: j := 1;
    while j <= 15 do
    begin // keeps coming back here until successful
      x[j] := HALF*(x[j-1] + x[j+1]);
      f[j] := Fun(x[j]);
      inc(j,2);
    end;
    nofun := nofun+8;
    step := (x[16]-x[0])/SIXTEEN;
    qleft := (w0*(f[0]+f[8]) +w1*(f[1]+f[7])+ w2*(f[2]+f[6])+
              w3*(f[3]+f[5]) +w4*f[4])*step;
    qright[lev+1] := (w0*(f[8] +f[16]) +w1*(f[9]+f[15]) +w2*(f[10]+f[14])+
                      w3*(f[11]+f[13]) +w4*f[12])*step;
    qnow := qleft + qright[lev+1];
    qdiff := qnow - qprev;
    area := area + qdiff;

//..... S T A G E   4   (interval convergence test) .....................
    esterr := ABS(qdiff)/1023.0;
    tolerr := MAX(abserr,relerr*ABS(area))*(step/stone);
    if (lev < LEVMIN) then goto 50;

    if (lev >= LEVMAX) then
    begin
      flag := flag+ONE;
      goto 70;
    end;

    if (nofun > NOFIN) then
    begin //..... S T A G E   6  (trouble section)
      nofin := nofin + nofin;
// Number of function values is about to exceed limit
      levmax := levout;
      flag := flag+(b-x[0])/(b-a);
      goto 70;
    end;

    if (esterr <= TOLERR) then goto 70;

//..... S T A G E   5   (no convergence).................................
//...locate next interval...
50: nim := nim + nim;
    lev := lev + 1;

// store right hand elements for future use
//    fsave[1:8,lev] := f[9:16];
//    xsave[1:8,lev] := x[9:16];
    for i := 1 to 8 do
    begin
      fsave[i,lev] := f[i+8];
      xsave[i,lev] := x[i+8];
    end;

    qprev := qleft;
    for  I  :=  1 to 8  do
    begin // assemble left hand elements for immediate use
      f[18-2*i] := f[9-i];
      x[18-2*i] := x[9-i];
    end;
    goto 30;

//..... S T A G E   7   (integral converged) ............................
//...add contributions into running sums...
70: result := result+qnow;
   errest := errest+esterr;
   cor11 := cor11+qdiff/1023.0;

  while odd(nim) do
  begin // 72 loop in text
//    if (nim = 2*(nim/2) then) EXIT   ;
    nim := nim div 2;
    lev := lev-1;
  end;
  nim := nim + 1;

  if (lev <= 0) then goto 80;
// looks like success!
  qprev := qright[lev]   ;
// assemble elements required for next interval
  x[0] := x[16];
  f[0] := f[16];
  for  i  :=  1 to 8  do
  begin
    f[2*i] := fsave[i,lev];
    x[2*i] := xsave[i,lev];
  end;

  goto 30;

//..... S T A G E   8  (Finalize and return).............................
80: result := result+cor11;
  if (errest <> ZERO) then
  begin  // make sure ERREST is not < roundoff level ....
    while (abs(result)+errest) = abs(RESULT) do
    begin
      errest := errest + errest;
    end;
  end;

end;// Quanc8 ---------------------------------------------------



initialization

end.
