{*
 Ordinary differential equations
 Nikolai V. Shokhirev, 2002-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2002.02.02 - created
 @version 2004.12.02 - modified
 @version 2007.08.08 - Added DelphiCodeToDoc comments
 }
{@abstract()
@author(Nikolai Shokhirev <nikolai@u.arizona.edu>
                          http://www.shokhirev.com/nikolai.html)
@created()
@lastmod(2004.12.24)
Nikolai V. Shokhirev, 2001-2004   }
unit uODE;

interface

uses
  uMatTypes, uDynArrays;

type
  TDiffEqProc = procedure(const t: TFloat; const y, yp: IFArr1D);


{
  Pascal version by Nikolai V. Shokhirev                                   
PURPOSE:
  Integrate a system of neqn first order differential equations by the Fehlberg
  fourth-fifth order Runge-Kutta method.
Ref.
  George E. Forsythe, Michael A. Malcolm, and Cleve B. Moler. 
  "Computer Methods for Mathematical Computations". Prentice-Hall, 1977.
ORIGINAL AUTHORS:
    RKF45 - H.A. Watts and L.F. Shampine  (Sandia)
NOTES:
  RKF45 is primarily designed to solve non-stiff and mildly stiff differential
  equations when derivative evaluations are inexpensive.
  Rkf45 should generally not be used when the user is demanding high accuracy.

  The arrays yp, f1, f2, f3, f4, and f5 (of dimension at least neqn) and the
  variables h, savre, savae, nfe, kop, init,jflag,and kflag are used internally
  by the code and appear in the call list to eliminate local retention of
  variables between calls. Accordingly, they should not be altered. Items of
  possible interest are
    yp - derivative of solution vector at t
    h - an appropriate stepsize to be used for the next step
    nfe- counter on the number of derivative function evaluations

ABSTRACT:
  Subroutine Rkf45 integrates a system of neqn first order ordinary differential
  equations of the form

 	  dy[i]/dt = f(t,y[1],y[2],...,y[neqn])

 	where the y[i] are given at t .
  Typically the subroutine is used to integrate from t to tout but it can be
  used as a one-step integrator to advance the solution a Double step in the
  direction of tout. On return, the parameters in the call list are set for
  continuing the integration. The user has only to call Rkf45 again (and perhaps
  define a new value for tout). Rkfs45 calls subroutine Fehl which computes an
  approximate solution over one step.

  Rkf45 uses the Runge-Kutta-Fehlberg(4,5) method described in the reference
  E. Fehlberg:  Low-order Classical Runge-Kutta formulas with Stepsize 
  Control, NASA TR R-315.

  The performance of Rkf45 is illustrated in the reference
  L.F.Shampine,H.A.Watts,S.Davenport: "Solving Non-stiff Ordinary
  Differential Equations - The State of the Art.
   Sandia Laboratories Report SAND75-0182, to appear in SIAM Review. 

  First call to Rkf45:

  The user must provide storage in his calling program for the arrays in
  the call list y, yp, f1, f2, f3, f4, f5 - 1D float arrays with the same
  limits and dimension = neqn, supply procedure F(t,y,yp) and initialize the
  following parameters:
    y - vector of initial conditions
    t - starting point of integration , must be a variable
    tout - output point at which solution is desired.
    t = tout is allowed on the first call only, in which case Rkf45 returns with
      iflag=2 if continuation is possible.
    relerr, abserr - relative and absolute local error tolerances which must be
      non-negative. relerr must be a variable while abserr may be a constant.
      The code should normally not be used with relative error control smaller
      than about i.e-8 to avoid limiting precision difficulties the code
      requires relerr to be larger than an internally computed relative error
      parameter which is machine dependent. In particular, pure absolute error
      is not permitted. If a smaller than allowable value of relerr is
      attempted, rkf45 increases relerr appropriately and returns control to the
      user before  continuing the integration.
    iflag - +1,-1 indicator to initialize the code for each new problem.
      Normal input is +1. the user should set iflag = -1 only when one-step
      integrator control is essential. In this case, Rkf45 attempts to advance
      the solution a Double step in the direction of tout each time it is
      called. Since this mode of operation results in extra computing overhead,
      it should be avoided unless needed.

  Output from Rkf45

  y - solution at t
  t - last point reached in integration.
  iflag =  2 - integration reached tout. Indicates successful return and is the
               normal mode for continuing integration.
       	= -2 - a Double successful step in the direction of tout has been taken.
               Normal mode for continuing integration one step at a time.
 	      =  3 - integration was not completed because relative error tolerance
               was too small. relerr has been increased	appropriately for
               continuing.
 	      =  4 - integration was not completed because more than 3000 derivative
               evaluations were needed. This is approximately 500 steps.
 	      =  5 - integration was not completed because solution vanished making
               a pure relative error test impossible. Must use non-zero abserr
               to continue. Using the one-step integration mode for one step
               is a good way to proceed.
 	      =  6 - integration was not completed because requested accuracy could
               not be achieved using smallest allowable stepsize. User must
               increase the error tolerance before continued integration can
               be attempted.
 	      =  7 - it is likely that Rkf45 is inefficient for solving this problem.
               Too much output is restricting the natural stepsize choice.
               Use the one-step integrator mode.
 	      = 8  - invalid input parameters. This indicator occurs if any of the
               following is satisfied:
                 t = tout and iflag <> +1 or -1
 	               relerr or abserr < 0.
                 iflag = 0 or < -2 or > 8
  h, nfe, yp, f1, f2, f3, f4, f5 - information which is usually of no interest
  to the user but necessary for subsequent calls.
  - yp contain the first derivatives of the solution vector y at t.
  - h contains the stepsize to be attempted on the next step.
  - nfe contains the derivative evaluation counter.

  Subsequent calls to Rkf45

  Subroutine Rkf45 returns with all information needed to continue the
  integration. If the integration reached tout, the user need only define
  a new tout and call Rkf45 again. In the one-step integrator mode (iflag=-2)
  the user must keep in mind that each step taken is in the direction of the
  current tout. Upon reaching tout (indicated by changing iflag to 2),the 
  user must then define a new tout and reset iflag to -2 to continue in the 
  one-step integrator mode. 

  If the integration was not completed but the user still wants to continue 
  (iflag = 3,4 cases), he just calls Rkf45 again. With iflag = 3, the relerr
  parameter has been adjusted appropriately for continuing the integration.
  In the case of iflag=4 the function counter will be reset to 0 and 
  another 3000 function evaluations are allowed. 

  However,in the case iflag = 5, the user must first alter the error criterion
  to use a positive value of abserr before integration can proceed. If he 
  does not, execution is terminated. 

  Also, in the case iflag = 6, it is necessary for the user to reset iflag
  to 2 (or -2 when the one-step integration mode is being used) as well as 
  increasing either abserr,relerr or both before the integration can be 
  continued. If this is not done, execution will be terminated. The 
  occurrence of iflag=6 indicates a trouble spot (solution is changing 
  rapidly,singularity may be present) and it often is inadvisable to continue.

  If (flag = 7 is encountered, the user should use the one-step integration mode
  with the stepsize determined by the code or consider switching to the 
  Adams codes DE/STEP, INTRP. If the user insists upon continuing the
  integration with Rkf45, he must reset iflag to 2 before calling Rkf45 again.
  Otherwise, execution will be terminated.

  If iflag = 8 is obtained, integration can not be continued unless the invalid
  input parameters are corrected. 

  It should be noted that the arrays work,iwork contain information required
  for subsequent integration. Accordingly, work and iwork should not be 
  altered.

  This interfacing routine merely relieves the user of a long calling list 
  via the splitting apart of two working storage arrays.

INPUT
  neqn - number of equations
  y - solution vector at t
  t - independent variable
  tout - output point at which solution is desired
  relerr - relative error tolerance
  abserr -  absolute error tolerance
OUTPUT
  iflag - indicator for status of work
  h - step size
  savre;
  savae;
  nfe - number of function evaluations
  kop;
IN, OUT
  yp - derivatives
  f1, f2, f3, f4, f5
  init;
  jflag;
  kflag;                               }
procedure Rkfs45(F: TDiffEqProc; var t: TFloat; const tout, abserr: TFloat;
                 var relerr: TFloat; var iflag: TInt;
                 const y, yp, f1, f2, f3, f4, f5: IFArr1D;
                 var h, savre, savae: TFloat;
                 var nfe, kop, init, jflag, kflag: TInt);

implementation

uses
  math;

{
  Pascal version by Nikolai V. Shokhirev
Ref.
  George E. Forsythe, Michael A. Malcolm, and Cleve B. Moler.
  "Computer Methods for Mathematical Computations". Prentice-Hall, 1977.
PURPOSE:
  Integrate a system of neqn first-order ordinary differential equations
  of the form dy/dt=F(t,y) [y is a vector], where the initial values of y
  and the initial values of yp, the derivatives are specified at the starting
  point t.  Fehl advances the solution over the fixed step h and returns the
  fifth order (sixth-order locally) solution approximation at t+h in array s.
  neqn: TInt; // number of equations
INPUT
  t - starting point
  h - step size
  y - array of length neqn; function at t
  yp - array of length neqn; derivatives at t
OUTPUT
  f1 - array of length neqn for internal use
  f2 - array of length neqn for internal use
  f3 - array of length neqn for internal use
  f4 - array of length neqn for internal use
  f5 - array of length neqn for internal use
  s  - array of length neqn; the results         }
procedure Fehl(F: TDiffEqProc; const t, h: TFloat;
                               const y, yp, f1, f2, f3, f4, f5, s : IFarr1D);
const
  C1  = 0.25;                // = 0.25;
  C2  = 0.09375;             // = 3.0/32.0;
  C3  = 3.0;                 // = 3.0;
  C4  = 0.375;               // = 3.0/8.0;
  C5  = 4.55166135639508e-4; // = 1.0/2197.0;
  C6  = 1932.0;              // = 1932.0;
  C7  = 7296.0;              // = 7296.0;
  C8  = -7200.0;             // = -7200.0;
  C9  = 0.923076923076923;   // = 12.0/13.0;
  C10 = 2.43664717348928e-4; // = 1.0/4104.0;
  C11 = 8341.0;              // = 8341.0;
  C12 = -845.0;              // = -845.0;
  C13 = 29440.0;             // = 29440.0;
  C14 = -32832.0;            // = -32832.0;
  C15 = 4.87329434697856e-5; // = 1.0/20520.0;
  C16 = -6080.0;             // = -6080.0;
  C17 = 9295.0;              // = 9295.0;
  C18 = -5643.0;             // = -5643.0;
  C19 = 41040.0;             // = 41040.0;
  C20 = -28352.0;            // = -28352.0;
  C21 = 0.5;                 // = 0.5;
  C22 = 1.31267187797402e-7; // = 1.0/7618050.0;
  C23 = 902880.0;            // = 902880.0;
  C24 = 3855735.0;           // = 3855735.0;
  C25 = -1371249.0;          // = -1371249.0;
  C26 = 3953664.0;           // = 3953664.0;
  C27 = 277020.0;            // = 277020.0;

var
//  neqn: TInt; // number of equations
//  t: TFloat;  // starting point
//  h: TFloat;  // step size

  ch: TFloat;
  i, Lo, Hi: TInt;
{
  y: IFarr1D; // array of length neqn; function at t
  yp: IFarr1D; // array of length neqn; derivatives at t
  f1: IFarr1D; // array of length neqn for internal use
  f2: IFarr1D; // array of length neqn for internal use
  f3: IFarr1D; // array of length neqn for internal use
  f4: IFarr1D; // array of length neqn for internal use
  f5: IFarr1D; // array of length neqn for internal use
  s : IFarr1D; // array of length neqn; the results
  C1, C2, C3, C4, C5, C6, C7, C8, C9, C10, C11, C12, C13, C14,
  C15,  C16, C17, C18, C19, C20, C21, C22, C23, C24, C25, C26, C27: TFloat;
  C1  := 0.25;
  C2  := 3.0/32.0;
  C3  := 3.0;
  C4  := 3.0/8.0;
  C5  := 1.0/2197.0;
  C6  := 1932.0;
  C7  := 7296.0;
  C8  := -7200.0;
  C9  := 12.0/13.0;
  C10 := 1.0/4104.0;
  C11 := 8341.0;
  C12 := -845.0;
  C13 := 29440.0;
  C14 := -32832.0;
  C15 := 1.0/20520.0;
  C16 := -6080.0;
  C17 := 9295.0;
  C18 := -5643.0;
  C19 := 41040.0;
  C20 := -28352.0;
  C21 := 0.5;
  C22 := 1.0/7618050.0;
  C23 := 902880.0;
  C24 := 3855735.0;
  C25 := -1371249.0;
  C26 := 3953664.0;
  C27 := 277020.0;   }
begin  //-----------------------------------------------------------------------
  ch := C1*h;

  Lo := y.Lo1;
  Hi := y.Hi1;
  for i := Lo to Hi do
    f5[i] := y[i]+ch*yp[i];
  F(t+ch, f5, f1);

  ch := C2*h;
  for i := Lo to Hi do
    f5[i] := y[i] + ch*(yp[i]+C3*f1[i]);
  F(t+C4*h, f5, f2);

  ch := C5*h;
  for i := Lo to Hi do
    f5[i] := y[i] + ch*(C6*yp[i] + (C7*f2[i]+C8*f1[i]) );
  F(t+C9*h, f5, f3);

  ch := C10*h;
  for i := Lo to Hi do
    f5[i] := y[i] + ch*( (C11*yp[i] + C12*f3[i]) + (C13*f2[i]+C14*f1[i]) );
  F(t+h, f5, f4);

  ch := C15*h;
  for i := Lo to Hi do
    f1[i] := y[i] + ch*( C16*yp[i] + (C17*f3[i]+C18*f4[i]) + (C19*f1[i]+C20*f2[i]) );
  F(t+C21*h, f1, f5);

//.....compute approximate solution at t+h
  ch := C22*h;
  for i := Lo to Hi do
    s[i] := y[i] + ch*( C23*yp[i] + (C24*f3[i]+C25*f4[i]) + (C26*f2[i]+C27*f5[i]) );

end;// of Fehl -----------------------------------------------------


procedure Rkfs45(F: TDiffEqProc; var t: TFloat; const tout, abserr: TFloat;
                 var relerr: TFloat; var iflag: TInt;
                 const y, yp, f1, f2, f3, f4, f5: IFArr1D;
                 var h, savre, savae: TFloat;
                 var nfe, kop, init, jflag, kflag: TInt);
label
  25, 30, 40, 45, 50, 55, 60, 65, 80, 85, 95, 100, 150, 200, 220, 240, 260, 300;
const
  ZERO = 0.0;
  ONE = 1.0;
  REMIN = 1E-12; { - minimum acceptable value of relerr.
                   Attempts to obtain higher accuracy with this routine are
                   usually very expensive and unsuccessful.}
  MAXNFE = 3000; { - the expense is controlled by restricting the number of
                   function evaluations to be approximately MAXNFE.
                   As set, this corresponds to about 500 steps. }
var
  rer, s          : TFloat;
  scale, tol, toln: TFloat;
  ypk, eps, u26: TFloat;
  a, ae, dt, ee, eeoet, esttol, et: TFloat;
  hmin: TFloat;
  k, mflag: TInt;
  Lo, Hi, neqn: TInt;
  hfaild, output: Boolean;
begin
//.....check input parameters............................................
  Lo := y.Lo1;
  Hi := y.Hi1;
  neqn := y.Dim1;
  eps := cMachEps;
  u26 := 26.0*eps;
//  if (neqn < 1) or (relerr < ZERO) or (abserr < ZERO) then
  if (relerr < ZERO) or (abserr < ZERO) then
  begin
    iflag := 8;
    exit;
  end;
  mflag := ABS(iflag);
  if (mflag = 0) or (mflag > 8) then
  begin
    iflag := 8;
    exit;
  end;

  if (mflag = 1) then goto 50;

//.....check continuation posibilities...................................
  if ((t = tout) and (kflag <> 3)) then
  begin
    iflag := 8; // invalid flag handler
    exit;
  end;
  if (mflag <> 2) then goto 25;

//...you get here if iflag=+2 or -2
  if ((kflag = 3) or  (init = 0)) then goto 45;
  if (kflag = 4) then goto 40;
  if ((kflag = 5) and  (abserr = ZERO)) then  goto 30;
  if ((kflag = 6) and  (relerr <= savre)  and  (abserr <= savae)) then goto 30;
  goto 50;

//...you get here if iflag = 3, 4, 5, 6, 7, or 8.....
25: if (iflag = 3) then goto 45;
    if (iflag = 4) then goto 40;
    if ((iflag = 5) and (abserr > ZERO)) then goto 45;

//.....integration cannot be continued since user did not respond to the
//..... instructions pertaining to iflag = 5, 6, 7, or 8
30: exit;
//!! maybe this is too strong

//.....reset function evaluation counter.................................
40: nfe := 0;
    if (mflag = 2) then goto 50;

//.....reset flag value from previous call...............................
45: iflag := jflag;
    if (kflag = 3) then mflag := ABS(iflag);

50: jflag := iflag;  // save for subsequent calls
    kflag := 0;      // set the continuation flag
    savre := relerr; // save for subsequent calls
    savae := abserr; // save for subsequent calls

//.....restrict relative error tolerance to be at least as large as
//       2*eps+remin to avoid limiting precision difficulties arising
//       from impossible accuracy requests.
    rer := eps + eps + remin;
    if (relerr >= rer) then goto 55;

//.....relative error tolerance too small................................
    relerr := rer;
    iflag := 3;
    kflag := 3;
    exit;

55: dt := tout-t;
    if (mflag = 1) then goto 60;
    if (init = 0) then goto 65;
    goto 80;

//.....initialization....................................................
60: init := 0;
    kop := 0;
    a := t;
    F(a,y,yp);
    nfe := 1;
    if (t <> tout) then goto 65;
    iflag := 2;
    exit;

65: init := 1;
    h := ABS(dt);
    toln := ZERO;
    for  k := Lo to Hi do
    begin
      tol := relerr*ABS(y[k])+abserr;
      if (tol <= ZERO) then continue;
      toln := tol;
      ypk := ABS(yp[k]);
      if (ypk*IntPower(h, 5) > tol) then h := Power((tol/ypk), 0.2);
   end;

   if (toln <= ZERO) then h := ZERO;
   h := max(h, u26*MAX(ABS(t), ABS(dt)));
   jflag := ISign2(2,iflag);

//.....set stepsize for integration in the direction from t to tout
80: h := SIGN2(h,dt);

//.....test to see if rkf45 is being severely impacted by too many outputs
   if (ABS(h) >= ABS(dt+dt)) then kop := kop + 1;
   if (kop <> 100) then goto 85;
   kop := 0;
   iflag := 7;
   exit;

85: if (ABS(dt) > u26*ABS(t)) then goto 95;

//.....if too close to output point, extrapolate and return
    for  k := Lo to Hi  do
    begin
      y[k] := y[k]+dt*yp[k];
    end;
    a := tout;
    F(a, y, yp);
    nfe := nfe +1 ;
    goto 300;

//.....initialize output point indicator.................................
95: output := false ;

//.....scale the error tolerances to avoid premature underflow in the
//        error tolerance
    scale := 2.0/relerr;
    ae := scale*abserr;


//.....step by step integration..........................................
100:hfaild :=  false ;
//...set smallest allowable step size...
    hmin := u26*ABS(t);

//.....adjust stepsize if necessary to hit the output point.
//.....look ahead two steps to avoid drastic changes in the stepsize and
//.....thus lessen the impact of output points on the code.
    dt := tout-t;
    if (ABS(dt) >= ABS(h+h)) then goto 200;
    if (ABS(dt) > ABS(h)) then goto 150;


//.....the next successful step will complete the integration to the
//     output point
    output :=  true ;
    h := dt;
    goto 200;

150:h := 0.5*dt;

//.....core integrator for taking a Double step..........................

200:if (nfe <= maxnfe) then goto 220;

//.....too much work...
    iflag := 4;
    kflag := 4;
    exit;

//.....advance an approximate solution over one step of length h
220:Fehl(F,t,h,y,yp,f1,f2,f3,f4,f5,f1);
    nfe := nfe+5;

//.....compute and test allowable tolerances versus local error estimates
//.....  and remove scaling of tolerances. note that relative error is
//.....  is measured with respect to the average of the magnitudes of the
//.....   solution at the beginning and end of the step.
    eeoet := ZERO;
    for  k := Lo to Hi  do
    begin
      et := ABS(y[k])+ABS(f1[k])+ae;
      if (et > ZERO) then goto 240;
//.....Inappropriate error tolerance.....................................
      iflag := 5;
      exit;
240:  ee := ABS( (-2090.0*yp[k]+(21970.0*f3[k]-15048.0*f4[k]))+
                 (22528.0*f2[k]-27360.0*f5[k]) );
      eeoet := MAX(eeoet,ee/et);
    end;
    esttol := ABS(h)*eeoet*scale/752400.0;
    if (esttol <= ONE) then goto 260;

//.....unsuccessful step. reduce the stepsize and start again............
//     (decrease is limited to a factor of 1/10)
    hfaild :=  true ;
    output :=  false ;
    s := 0.1;
    if (esttol < 59049.0) then s := 0.9/Power(esttol,0.2);
    h := s*h;
    if (ABS(h) > hmin) then goto 200;

//.....requested error unobtainable at smallest allowable stepsize.......
    iflag := 6;
    kflag := 6;
    exit;

//.....successful step.   store solution at t+h and evaluate derivatives
//     there..........................
260:t := t+h;
    for  k := Lo to Hi  do
    begin
      y[k] := f1[k];
    end;
    a := t;
    F(a,y,yp);
    nfe := nfe + 1;

// Choose next stepsize.  The increase is limited to a factor of 5.
// If step failure has just occured, next stepsize is not allowed
//  to increase.
    s := 5.0;
    if (esttol > 1.889568E-4) then s := 0.9/Power(esttol,0.2);
    if (hfaild) then s := MIN(s,ONE);
    h := SIGN2(MAX(s*ABS(h),hmin),h);

//.....E N D   O F   C O R E   I N T E G R A T O R   ....................

//.....Should we take another step? .....................................
    if (output) then goto 300;
    if (iflag > 0) then goto 100;


//.....integration successfully completed................................
//...one step mode...
    iflag := -2;
    exit;

//...interval mode...
300:t := tout;
    iflag := 2;
    exit;
end;// of Rkfs45 ---------------------------------------------------------------


initialization

end.
