{*
 PasMatLib - 1D and 2D grid optimization
 Nikolai V. Shokhirev, 2005-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2005.09.04 - created
 @version 2007.07.18 - Added DelphiCodeToDoc comments
 }
unit uArrayOptim;

interface

uses
  uMatTypes, uDynArrays;

{*
  Minimization on a vector.
  result = min(A[i1])  result = A[idx1]
  A.Lo1+margin <= i1 <= A.Hi1-margin;
  @param A vector
  @param margin A.Lo1+margin <= i1 <= A.Hi1-margin;  ;
  @param idx1 output index
  @return min(A[i1]) = A[idx1]
}
function MinArray(const A: IFArr1D; var idx1: TInt; margin: TInt = 0): TFloat; overload;

{*
  Maximization on a matrix.
  result = max(A[i1])  result = A[idx1]
  A.Lo1+margin <= i1 <= A.Hi1-margin;
  @param A vector
  @param margin A.Lo1+margin <= i1 <= A.Hi1-margin;
  @param idx1 output index
  @return min(A[i1]) = A[idx1]
}
function MaxArray(const A: IFArr1D; var idx1: TInt; margin: TInt = 0): TFloat; overload;

{*
  |x1| < 1 - fractional corrections to idx1
  @param A Vector
  @param idx1 input index
  @param return: value - optimal value
  @param return: x1 - fractional corrections to idx1
}
procedure RefineExtremum(const A: IFArr1D; const idx1: TInt; var value, x1: TFloat); overload;

{*
  Minimization on a matrix.
  result = min(A[i1,i2])  result = A[idx1,idx2]
  A.Lo1+margin <= i1 <= A.Hi1-margin;  A.Lo2+margin <= i2 <= A.Hi2-margin;
  @param A matrix
  @param margin A.Lo1+margin <= i1 <= A.Hi1-margin;
                A.Lo2+margin <= i2 <= A.Hi2-margin;
  @param idx1 output index
  @param idx2 output index
  @return min(A[i1,i2]) = A[idx1,idx2]
}
function MinArray(const A: IFArr2D; var idx1, idx2: TInt; margin: TInt = 0): TFloat; overload;

{*
  Maximization on a matrix.
  result = max(A[i1,i2])  result = A[idx1,idx2]
  A.Lo1+margin <= i1 <= A.Hi1-margin; A.Lo2+margin <= i2 <= A.Hi2-margin;
  @param A matrix
  @param margin A.Lo1+margin <= i1 <= A.Hi1-margin;
  A.Lo2+margin <= i2 <= A.Hi2-margin;
  @param idx1 output index
  @param idx2 output index
  @return min(A[i1,i2]) = A[idx1,idx2]
}
function MaxArray(const A: IFArr2D; var idx1, idx2: TInt; margin: TInt = 0): TFloat; overload;

{*
  |x1|, |x2| < 1 - fractional corrections to idx1, idx2
  @param A matrix
  @param idx1 input index
  @param idx2 input index
  @param return: value - optimal value
  @param return: x1 - fractional corrections to idx1
  @param return: x2 - fractional corrections to idx2
}
procedure RefineExtremum(const A: IFArr2D; const idx1, idx2: TInt; var value, x1, x2: TFloat); overload;

implementation

function MinArray(const A: IFArr1D; var idx1: TInt; margin: TInt = 0): TFloat;
var
  i1: TInt;
  Val: TFloat;
begin
  idx1 := A.Lo1+margin;
  Val := MaxFloat;
  for i1 := A.Lo1+margin to A.Hi1-margin do
    if A[i1] < Val then
    begin
     idx1 := i1;
     Val := A[i1];
   end;
  result := Val;
end;

function MaxArray(const A: IFArr1D; var idx1: TInt; margin: TInt = 0): TFloat;
var
  i1: TInt;
  Val: TFloat;
begin
  idx1 := A.Lo1+margin;
  Val := NegMaxFloat;
  for i1 := A.Lo1+margin to A.Hi1-margin do
      if A[i1] > Val then
      begin
       idx1 := i1;
       Val := A[i1];
     end;
  result := Val;
end;

procedure RefineExtremum(const A: IFArr1D; const idx1: TInt; var value, x1: TFloat);
var
  d1, d11, f0, fp, fm: TFloat;
begin
  fp := A[idx1+1];
  f0 := A[idx1  ];
  fm := A[idx1-1];
  d1 := (fp-fm)/2.0;
  d11 := (fp-f0+fm-f0);
  value := f0;
  if (d11 <> 0.0) then
  begin
    x1 := -d1/d11;
    if (abs(x1) < 1.0) then
      value := value - d1*d1/d11/2.0
    else
      x1 := 0.0;
  end else
    x1 := 0.0;
end;

function MinArray(const A: IFArr2D; var idx1, idx2: TInt; margin: TInt = 0): TFloat;
var
  i1, i2: TInt;
  Val: TFloat;
begin
  idx1 := A.Lo1+margin;
  idx2 := A.Lo2+margin;
  Val := MaxFloat;
  for i1 := A.Lo1+margin to A.Hi1-margin do
    for i2 := A.Lo2+margin to A.Hi2-margin do
      if A[i1,i2] < Val then
      begin
       idx1 := i1;
       idx2 := i2;
       Val := A[i1,i2];
     end;
  result := Val;
end;

function MaxArray(const A: IFArr2D; var idx1, idx2: TInt; margin: TInt = 0): TFloat;
var
  i1, i2: TInt;
  Val: TFloat;
begin
  idx1 := A.Lo1+margin;
  idx2 := A.Lo2+margin;
  Val := NegMaxFloat;
  for i1 := A.Lo1+margin to A.Hi1-margin do
    for i2 := A.Lo2+margin to A.Hi2-margin do
      if A[i1,i2] > Val then
      begin
       idx1 := i1;
       idx2 := i2;
       Val := A[i1,i2];
     end;
  result := Val;
end;

procedure RefineExtremum(const A: IFArr2D; const idx1, idx2: TInt; var value, x1, x2: TFloat);
var
  d1, d2, d11, d12, d22, det: TFloat;
  f00, fp0, fm0, f0p, fpp, fmp, f0m, fpm, fmm: TFloat;
begin
  fpp := A[idx1+1,idx2+1];
  f0p := A[idx1  ,idx2+1];
  fmp := A[idx1-1,idx2+1];
  fp0 := A[idx1+1,idx2];
  f00 := A[idx1  ,idx2];
  fm0 := A[idx1-1,idx2];
  fpm := A[idx1+1,idx2-1];
  f0m := A[idx1  ,idx2-1];
  fmm := A[idx1-1,idx2-1];
  d1 := (fp0-fm0)/2.0;
  d2 := (f0p-f0m)/2.0;
  d11 := (fp0-f00+fm0-f00);
  d22 := (f0p-f00+f0m-f00);
  d12 := (fpp-fpm-fmp+fmm)/4.0;
  det := (d22*d11-d12*d12);
  // initial values
  value := f00;
  x1 := 0.0;
  x2 := 0.0;
  // refinement
  if (det <> 0.0) then
  begin
    x1 := -(d22*d1-d2*d12)/det;
    x2 := -(d11*d2-d1*d12)/det;
    if (abs(x1) < 1.0) and (abs(x2) < 1.0) then
    begin
      value := value-(d22*d1*d1-2.0*d1*d12*d2+d11*d2*d2)/det/2.0;
    end else
    begin  // cannot refine
      x1 := 0.0;
      x2 := 0.0;
    end;
  end;
end;

end.
