{*
 PasMatLib - Levenberg-Marquardt fit and optimization, SVD fit
 Nikolai V. Shokhirev, 2003-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2003.06.06 - created
 @version 2004.11.12
 @version 2007.07.18 - Added DelphiCodeToDoc comments
 }
unit uLevenbergMarquardt;

interface

uses
  uMatTypes, uDynArrays, uDynArrUtils;

type
  {*
    Fitting 1D function
    @param x 1D variable
    @param P parameter vector
    @return f(x;P)
  }
  FitFunction = function(x: TFloat; const P: IFArr1D): TFloat;

  {*
    Gradient for fitting 1D function
    @param x 1D variable
    @param P parameter vector
    @param GP = df/dP (return)
  }
  FitGradient = procedure(x: TFloat; const P, GP: IFArr1D);

  {*
    Objective (minimization) function
    @param P parameter vector
    @param GP return: parameter gradient
    @return Objective function value
  }
  OptimFunction = function(const P: IFArr1D): TFloat;

  {*
    Objective function 1st and 2nd derivatives
    @param P parameter vector
    @param GP return: gradient
    @param HP return: Hessian
  }
  OptimDerivatives = procedure(const P, GP: IFArr1D; const HP: IFArr2D);

 {*
  Levenberg-Marquardt fit
  @param F  Fit Function
  @param G  FitGradient
  @param Xe Experimental X
  @param Fe Experimental function values
  @param W  weights
  @param P0 return: optimum parameters
  @param epsP desired accuracy in P
  @param epsG desired accuracy in G
  @param deltaP actual accuracy in P
  @param deltaG actual accuracy in G
  @param err return: error code
  @param Factor lambda increase factor, default = 8
  @param IterMax maximum iterations, default = 999
  }
procedure LMFit(F: FitFunction; G: FitGradient; const Xe, Fe, W: IFArr1D;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; Factor: TFloat = 8.0; IterMax: TInt = 999);

{*
  Fit using SVD
  @param F  Fit Function
  @param G  FitGradient
  @param Xe Experimental X
  @param Fe Experimental function values
  @param W  weights
  @param P0 return optimum parameters
  @param epsP desired accuracy in P
  @param epsG desired accuracy in G
  @param deltaP actual accuracy in P
  @param deltaG actual accuracy in G
  @param err error code
  @param NoiseLevel experimental accuracy, default = MinFloat
  @param IterMax maximum iterations, default = 999
 }
procedure SVDFit(F: FitFunction; G: FitGradient; const Xe, Fe, W: IFArr1D;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; NoiseLevel: TFloat = MinFloat; IterMax: TInt = 999);

 {*
  Levenberg-Marquardt optimization
  @param F  Fit Function
  @param GH G and H calculations
  @param P0 return optimum parameters
  @param epsP desired accuracy in P
  @param epsG desired accuracy in G
  @param deltaP actual accuracy in P
  @param deltaG actual accuracy in G
  @param err error code
  @param Factor lambda increase factor, default = 8
  @param IterMax maximum iterations, default = 999
 }
procedure LMOptim(F: OptimFunction; GH: OptimDerivatives;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; Factor: TFloat = 8.0; IterMax: TInt = 999);

implementation

uses
  uDynArrIO, // debug
  uDynLinAlg, math, uDynObjAlg;

{ Levenberg-Marquardt fit }
procedure LMFit(F: FitFunction; G: FitGradient; const Xe, Fe, W: IFArr1D;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; Factor: TFloat = 8.0; IterMax: TInt = 999);
var
  Ne, Np,Iter: TInt;
  Lambda, Phi, Phi0: TFloat;
  dF: IFArr1D;    // Dim = Ne - Weighted (Ft - Fe)
  P: IFArr1D;     // Dim = Np - Parameters
  GP: IFArr1D;    // Dim = Np - Gradients
  rhs: IFArr1D;   // Dim = Np - dF, JT*dF
  J,              // [Ne x Np] Weighted Jacobian
  HH: IFArr2D;    // [Np x Np] JT*J
  HL: IFArr2D;    // [Np x (Np+1)] = [HH+Lambda*I|rhs]
  First, Done, WasReduced, LambdaFound: boolean;

  procedure CalcH(const P: IFArr1D; var Phi: TFloat);
  var
    m, m1: TInt;
  begin
    m1 := 1;
    Phi := 0.0;
    for m := Xe.Lo1 to Xe.Hi1 do
    begin
      dF[m1] := W[m]*(F(Xe[m],P)-Fe[m]); // Weighted (Ft - Fe)
      Phi := Phi + sqr(dF[m1]);
      G(Xe[m], P, GP);
      ValuesToRow(J, m1, GP); // J.Row[m1].AssignValues(GP); regardless of GP.Base
      CxRow(J, m1, W[m]);     // J.Row[m1].CxSelf(W[m]);
      inc(m1);
    end;
    HH := MtTxMt(J, J);    // HH = JT*J
    rhs := VtxMt(dF, J);   // rhs = JT*dF
    deltaG := rhs.MaxAbs;
  end;

  function FirstLambda: TFloat;
  var
    k: TInt;
    sum: TFloat;
  begin
    First := False;
    sum := 0.0;
    for k := 1 to HH.Dim1 do
      sum := sum + HH[k,k];
    result := sum/HH.Dim1;
  end;

  procedure CalcNewParam(Lambda: TFloat; var err: TInt);
  var
    k: TInt;
  begin
    for k := 1 to Np do
    begin
      ValuesToCol(HL, k, HH.GetSlice(_,k));  // HL.Col[k].AssignValues(HH.Col[k]);
    end;
    MShiftDiag(HL, Lambda, HL, false);  // HL[k,k] := HL[k,k] + Lambda;
    ValuesToCol(HL, Np+1, rhs);  //  HL.Col[Np+1].Values(rhs);
    GaussJordan(HL, false, err);
    P.Assign(HL.GetSlice(_,Np+1)); //  P.AssignValues(HL.Col[Np+1]); - dP
    deltaP := P.MaxAbs;           // dPmax
    for k := 1 to Np do
      P[k] := P0[k] - P[k];
  end;

  function FitError2(const P: IFArr1D): TFloat;
  var
    m: TInt;
    sum: TFloat;
  begin
    sum := 0.0;
    for m := Xe.Lo1 to Xe.Hi1 do
      sum := sum + sqr(W[m]*(F(Xe[m],P)-Fe[m]));
    result := sum;
  end;

begin
  Ne :=  Xe.Dim1; // experiments
  Np :=  P0.Dim1; // parameters
  dF  := TFArr1D.Create(Ne);
  J   := TFArr2D.Create(Ne,Np);
  rhs := TFArr1D.Create(Np);
  P   := TFArr1D.Create(Np);
  GP  := TFArr1D.Create(Np);
  HL  := TFArr2D.Create(Np,Np+1);

  Iter := 0;
  First := true;
  Lambda := 1.0;
  repeat
    CalcH(P0, Phi0);
    WasReduced := false;
    LambdaFound := false;
    repeat
      inc(Iter);
      if First then
        Lambda := FirstLambda;
      CalcNewParam(Lambda, err);
      Phi := FitError2(P);
      if (Phi >= Phi0) or (err <> 0) then
      begin
        LambdaFound := WasReduced;
        Lambda := Lambda*Factor;
      end else
      begin
        Phi0 := Phi;
        P0.Assign(P);
        Lambda := Lambda/Factor;
        WasReduced := true;
      end;
    until LambdaFound;
    Done := (deltaP < epsP) or (deltaG < epsG)
         or (Iter > IterMax) or (err <> 0);
  until Done;
  if (Iter > IterMax) then
    err := 1000+iter;
end;

{SVD fit }
procedure SVDFit(F: FitFunction; G: FitGradient; const Xe, Fe, W: IFArr1D;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; NoiseLevel: TFloat = MinFloat; IterMax: TInt = 999);
const
  Factor = 0.5;
var
  Ne, Np,Iter: TInt;
  Phi, Phi0: TFloat;
  dF: IFArr1D;    // Dim = Ne - Weighted (Ft - Fe)
  P: IFArr1D;     // Dim = Np - Parameters
  dP: IFArr1D;    // Dim = Np
  GP: IFArr1D;    // Dim = Np - Gradients
  gamma: IFArr1D; // Dim = Np - singular values
  UTdF: IFArr1D;  // Dim = Np - expansion coefficients
  U: IFArr2D;     // [Ne x Np] J, U
  V: IFArr2D;     // [Np x Np]
  Done: boolean;

  function FitError2(const P: IFArr1D): TFloat;
  var
    m: TInt;
    sum: TFloat;
  begin
    sum := 0.0;
    for m := Xe.Lo1 to Xe.Hi1 do
      sum := sum + sqr(W[m]*(F(Xe[m],P)-Fe[m]));
    result := sum;
  end;

  procedure CalcJ(const P: IFArr1D);
  var
    m, m1: TInt;
  begin
    m1 := 1;
    for m := Xe.Lo1 to Xe.Hi1 do
    begin
      dF[m1] := W[m]*(F(Xe[m],P)-Fe[m]);
      G(Xe[m], P, GP);
      ValuesToRow(U,m1,GP);  //  U.Row[m1].AssignValues(GP);  regardless of GP.Base
      CxRow(U,m1,W[m]);      // U.Row[m1].CxSelf(W[m]);
      inc(m1);
    end;
    SVD(U, U, V, gamma, err);
    UTdF := VtxMt(dF,U);
  end;

  procedure CalcNewParam(NoiseLevel: TFloat);
  var
    k: TInt;
    Lambda: TFloat;
  begin
    dP.Fill(0.0);
    for k := 1 to Np do
    begin
      if gamma[k] < NoiseLevel then break;
      VAddV(dP,V.GetSlice(_,k),UTdF[k]/gamma[k]); // dP.Add(UTdF[k]/gamma[k],V.Col[k]);
    end;
    deltaP := dP.MaxAbs;    // dPmax
    Lambda := 1.0;
    repeat
      for k := 1 to Np do
        P[k] := P0[k] - Lambda*dP[k];
      Phi := FitError2(P);
      if Phi <= Phi0 then
        break;
      Lambda := Lambda*Factor;
      deltaP := Factor*deltaP;
    until deltaP < epsP;
    Phi0 := Phi;
    P0.Assign(P);
  end;

  procedure CalcDeltaG;
  var
    k: TInt;
  begin
    deltaG := 0.0;
    for k := 1 to Np do
    begin
      deltaG := max(deltaG, MaxAbsCol(V,k)*UTdF[k]*gamma[k] );
//      deltaG := max(deltaG,V.Col[k].MaxAbs*UTdF[k]*gamma[k] );
    end;
  end;

begin
  Ne    :=  Xe.Dim1; // experiments
  Np    :=  P0.Dim1; // parameters
  dF    := TFArr1D.Create(Ne);
  gamma := TFArr1D.Create(Np);
  P     := TFArr1D.Create(Np);
  dP    := TFArr1D.Create(Np);
  GP    := TFArr1D.Create(Np);
  U  := TFArr2D.Create(Ne,Np);
  V  := TFArr2D.Create(Np,Np);

  deltaG := 1.0; // in case it is not used
  Iter := 0;
  Phi0 := FitError2(P0);
  repeat
    CalcJ(P0);
    CalcNewParam(NoiseLevel);
    CalcDeltaG;
    Done := (deltaP < epsP) or (deltaG < epsG)
         or (Iter > IterMax) or (err <> 0);
  until Done;
  if (Iter > IterMax) then
    err := 1000+iter;
end;

{ Levenberg-Marquardt optimization }
procedure LMOptim(F: OptimFunction; GH: OptimDerivatives;
            const P0: IFArr1D; epsP, epsG: TFloat; var deltaP, deltaG: TFloat;
            var err: TInt; Factor: TFloat = 8.0; IterMax: TInt = 999);
var
  Np,Iter: TInt;
  Lambda, Phi, Phi0: TFloat;
  P: IFArr1D;     // Dim = Np - Parameters
  GP: IFArr1D;    // Dim = Np - Gradients
  HH: IFArr2D;     // [Np x Np] Hessian
  HL: IFArr2D;    // [Np x (Np+1)] = [HH+Lambda*I|rhs]
  First, Done, WasReduced, LambdaFound: boolean;

//  s: string; // debug;

  function FirstLambda: TFloat;
  var
    k: TInt;
    sum: TFloat;
  begin
    First := False;
    sum := 0.0;
    for k := 1 to HH.Dim1 do
      sum := sum + HH[k,k];
    result := sum/HH.Dim1;
  end;

  procedure CalcNewParam(Lambda: TFloat; var err: TInt);
  var
    k: TInt;
  begin
    for k := 1 to Np do
    begin
      ValuesToCol(HL,k,HH.GetSlice(_,k)); // HL.Col[k].AssignValues(HH.Col[k]);
    end;
    MShiftDiag(HL, Lambda, HL, false);   // HL[k,k] := HL[k,k] + Lambda;
    ValuesToCol(HL,Np+1,GP);      // HL.Col[Np+1].AssignValues(GP);
    GaussJordan(HL, false, err);
    ColValuesToV(HL,Np+1,P);      // P.AssignValues(HL.Col[Np+1]); - dP
    deltaP := P.MaxAbs;           // dPmax
    for k := 1 to Np do
      P[k] := P0[k] - P[k];
  end;

begin
  Np :=  P0.Dim1; // parameters
  P   := TFArr1D.Create(Np);
  GP  := TFArr1D.Create(Np);
  HH   := TFArr2D.Create(Np,Np);
  HL  := TFArr2D.Create(Np,Np+1);

  Iter := 0;
  First := true;
  Lambda := 1.0;
  repeat
    Phi0 := F(P0);
    GH(P0, GP, HH);
//    MatrixToStr(HH, s);   // debug
//    VectorToStr(GP, s);   // debug
    deltaG := GP.MaxAbs;
    WasReduced := false;
    LambdaFound := false;
    repeat
      inc(Iter);
      if First then
        Lambda := FirstLambda;
      CalcNewParam(Lambda, err);
//    VectorToStr(P, s);   // debug
      Phi := F(P);
      if (Phi >= Phi0) or (err <> 0) then
      begin
        LambdaFound := WasReduced;
        Lambda := Lambda*Factor;
      end else
      begin
        Phi0 := Phi;
        P0.Assign(P);
        Lambda := Lambda/Factor;
        WasReduced := true;
      end;
    until LambdaFound;
    Done := (deltaP < epsP) or (deltaG < epsG) or (Iter > IterMax);
  until Done;
end;

end.
