{*
 Optimization utilities
  Nikolai V. Shokhirev, 2003-2007
  @author Nikolai Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2003.01.06 - created
 @version 2004.12.21
 @version 2007.07.18 - Added DelphiCodeToDoc comments
 }
 unit uOptimUtils;

interface

uses
  uMatTypes;

{*
  Bracketing the minimum of a function within given limits
  @param f  function
  @param xmin  search limit min
  @param xmax  search limit max
  @param x0  start ( xmin < x0 < xmax )
  @param tol  accuracy
  @param step  initial step return - final step
  @param x1 return  bracket1
  @param x2 return  bracket2
  @param ErrCode  =  0 - OK
   = -5 - small interval: |xmax - xmin| < |tol|
   =  5 - xmin > xmax
   = -4 - x0 < xmin
   =  4 - x0 > xmax
   = -3 - flat function: |f(x)-f(x+step)| < dfmin
   =  3 - reserved
   = -2 - x0 = xmin and f(x0) < f(x0+|step|)
   =  2 - x0 = xmax and f(x0) < f(x0-|step|)
   = -1 - minimum at x = xmin
   =  1 - minimum at x = xmax
}
procedure Bracketing(f: TFloatFunction1D; x0, tol: TFloat;
               var step, x1, x2: TFloat; var ErrCode: TInt; xmin, xmax: TFloat);

{*
  Bracketing the minimum of a function in the positive semi-axis
  @param f  function
  @param xmax  search limit ( < MaxFloat )
  @param tol  accuracy
  @param step  initial step, return - final step
  @param x1 return:  bracket1
  @param x2 return:  bracket2
  @param ErrCode =  0 - OK
   = -5 - small interval: |xmax - xmin| < |tol|
   = -3 - flat function: |f(x)-f(x+step)| < dfmin
   = -2 - x0 = xmin and f(x0) < f(x0+|step|)
   =  1 - minimum at x = xmax
  }
procedure Bracketing0(f: TFloatFunction1D; tol: TFloat; var step, x1, x2: TFloat;
                                               var ErrCode: TInt; xmax: TFloat);

{*
  An approximation x to the point where f attains a minimum on
  the interval (ax,bx) is determined.

    The method used is a combination of golden section search and successive
  parabolic interpolation. Convergence is never much slower than that for
  a fibonacci search. If f has a continuous second derivative which is positive
  at the minimum (which is not at ax or bx), then convergence is superlinear,
  and usually of the order of about 1.324....
    The function f is never evaluated at two points closer together than
  eps*abs(fmin) + (tol/3), where eps is approximately the square root of
  the relative machine precision.  if  f  is a unimodal function and the
  computed values of  f  are always unimodal when separated by at least
  eps*abs(x) + (tol/3), then fmin approximates the abcissa of the global minimum
  of f on the interval ax,bx with  an error less than 3*eps*abs(fmin) + tol.
    If  f  is not unimodal, then fmin may approximate a local, but perhaps
  non-global, minimum to the same accuracy.
    This function subprogram is a slightly modified version of the algol 60
  procedure localmin given in richard brent, algorithms for minimization without
  derivatives, prentice - hall, inc. (1973).

Ref:
  George E. Forsythe, Michael A. Malcolm, and Cleve B. Moler.
  "Computer Methods for Mathematical Computations". Prentice-Hall, 1977.

  @param ax  left endpoint of initial interval
  @param bx  right endpoint of initial interval
  @param f   function subprogram which evaluates f(x) for any x in the interval (ax,bx)
  @param tol  desired length of the interval of uncertainty of the final result ( >= 0)
  @param fmin return - abcissa approximating the point where f attains a minimum
  @return x_min
  @author Pascal translation by Nikolai Shokhirev
  @author ORIGINAL AUTHORS: Van Winjngaarden, Decker, Brent, Et Al
}
function FMMfmin(ax, bx: TFloat; f: TFloatFunction1D; tol: TFloat): TFloat;

implementation

uses
  math;

{*
  Message - private unit function
  @param ErrCode Error Code
  @return
    0 - OK
   -5 - small interval: |xmax - xmin| < |tol|
    5 - xmin > xmax
   -4 - x0 < xmin
    4 - x0 > xmax
   -3 - flat function: |f(x)-f(x+step)| < dfmin
    3 - reserved
   -2 - x0 = xmin and f(x0) < f(x0+|step|)
    2 - x0 = xmax and f(x0) < f(x0-|step|)
   -1 - minimum at x = xmin
    1 - minimum at x = xmax
}
function BracketingErrMsg(ErrCode: TInt): string;
begin
 if ErrCode = -5 then result := 'small interval: |xmax - xmin| < |tol|' else
 if ErrCode =  5 then result := 'xmin > xmax' else
 if ErrCode = -4 then result := 'x0 < xmin' else
 if ErrCode =  4 then result := 'x0 > xmax' else
 if ErrCode = -3 then result := 'flat function: |f(x)-f(x+step)| < dfmin' else
// if ErrCode =  3 then result := '' else
 if ErrCode = -2 then result := 'x0 = xmin and f(x0) < f(x0+|step|)' else
 if ErrCode =  2 then result := 'x0 = xmax and f(x0) < f(x0-|step|)' else
 if ErrCode = -1 then result := 'minimum at x = xmin' else
 if ErrCode =  1 then result := 'minimum at x = xmax' else
 if ErrCode =  0 then result := 'OK' else
   result := 'Undefined Err Code' ;
end;

procedure Bracketing(f: TFloatFunction1D; x0, tol: TFloat;
               var step, x1, x2: TFloat; var ErrCode: TInt; xmin, xmax: TFloat);
const
  factor = 2.0;
var
  h, hmin, hmax,
  a, b, c, fa, fb, fc: TFloat;

  function dfIsSmall: boolean;
  var
    dfmin: TFloat;
  begin
    dfmin := max(abs(fb),abs(fc))*MachEps*SafeFactor;
    result := (abs(fb-fc) < dfmin);
    if result then ErrCode := -3;
  end;

  function IsOutOfLim: boolean;
  begin
    if c < xmin then
    begin
      ErrCode := -1;
      c := xmin;
    end else
    if c > xmax then
    begin
      ErrCode := 1;
      c := xmax;
    end;
    result := (ErrCode <> 0);
    fc := f(c);
    if fc > fb then ErrCode := 0;
  end;

  function IsAtLim: boolean;
  begin
    if (h < 0.0) then
    begin
      result :=  ((b - xmin) < tol);
      if result then ErrCode := -1;
    end else
    begin
      result :=  ((xmax - b) < tol);
      if result then ErrCode := 1;
    end;
  end;

begin
  ErrCode := 0;
  x1 := xmin;
  x2 := xmax;
  tol := abs(tol);
  h := step;
  b := x0;

  if xmin > xmax then ErrCode := 5 else
  if (xmax - xmin) < tol then ErrCode := -5 else
  if x0 < xmin then ErrCode := -4 else
  if x0 > xmax then ErrCode := 4;

  if ErrCode <> 0 then exit;

  hmax := (xmax - xmin)/4.0;
  if (abs(h) > hmax) then
    h := sign2(hmax,h); // h := sign1(h)*hmax;

  fb := f(b);
  // check if step is too small
  hmin := max(abs(b),1.0)*SqrtMachEps;
  if abs(h) < hmin then
    h := sign2(max(tol, hmin),h); // h := sign1(h)*max(tol, hmin);

  // point out of Left/Right Limit
  if (b-xmin) <= tol then
  begin // Left Limit
    h := abs(h);
    ErrCode := -2;
  end else
  if (xmax-b) <= tol then
  begin // Right Limit
    h := -abs(h);
    ErrCode := 2;
  end;

  if (ErrCode = 0) then // inside the interval
    if h > 0.0 then h := min(h,xmax-b) else
      h := max(h,xmin-b); // cut step to the nearest limit if necessary

  c := b + h;
  fc := f(c);
  if dfIsSmall then
    exit; // too small change - constant function

  if (fc > fb) and (ErrCode <> 0) then
    exit; //  |ErrCode| = 2

  ErrCode := 0;
  if fc > fb then
  begin
    Swap(c, b);
    Swap(fc, fb);
    h := -h;
  end;
  // fb > fc
  repeat
    if IsAtLim then break;
    a := b;  fa := fb;
    b := c;  fb := fc;
    h := h*factor;
    if abs(h) > hmax then
      h := Sign2(hmax,h);
    c := b + h;
    if IsOutOfLim then break;
    if dfIsSmall then break;
  until fc > fb;

  if h > 0.0 then
  begin
    x1 := a;
    x2 := c;
  end else
  begin
    x1 := c;
    x2 := a;
  end;
  step := h;
end;

{*
  Message - private unit function
  @param ErrCode Error Code
  @return Code description
   =  0 - OK
   = -5 - small interval: |xmax - xmin| < |tol|
   = -3 - flat function: |f(x)-f(x+step)| < dfmin
   = -2 - x0 = xmin and f(x0) < f(x0+|step|)
   =  1 - minimum at x = xmax
}
function Bracketing0ErrMsg(ErrCode: TInt): string;
begin
 if ErrCode = -5 then result := 'small interval: |xmax - xmin| < |tol|' else
 if ErrCode = -3 then result := 'flat function: |f(x)-f(x+step)| < dfmin' else
 if ErrCode = -2 then result := 'x0 = xmin and f(x0) < f(x0+|step|)' else
 if ErrCode =  1 then result := 'minimum at x = xmax' else
 if ErrCode =  0 then result := 'OK' else
   result := 'Undefined Err Code' ;
end;

procedure Bracketing0(f: TFloatFunction1D; tol: TFloat; var step, x1, x2: TFloat;
                                               var ErrCode: TInt; xmax: TFloat);
const
  factor = 2.0;
  x0 = 0.0;
  xmin =0.0; 
var
  h, hmin, hmax, 
  a, b, c, fa, fb, fc: TFloat;

  function dfIsSmall: boolean;
  var
    dfmin: TFloat;
  begin
    dfmin := max(abs(fb),abs(fc))*MachEps*SafeFactor;
    result := (abs(fb-fc) < dfmin);
    if result then ErrCode := -3;
  end;

  function IsOutOfLim: boolean;
  begin
    if c > xmax then
    begin
      ErrCode := 1;
      c := xmax;
    end;
    result := (ErrCode <> 0);
    fc := f(c);
    if fc > fb then ErrCode := 0;
  end;

  function IsAtLim: boolean;
  begin
    result :=  ((xmax - b) < tol);
    if result then ErrCode := 1;
  end;

begin
  ErrCode := 0;
  xmax := abs(xmax); // just in case
  h := step;
  h := abs(h); // just in case ...
  x1 := xmin;
  x2 := xmax;
  tol := abs(tol);
  b := x0;

  if (xmax < tol) then ErrCode := -5;
  if ErrCode <> 0 then exit;

  hmax := xmax/4.0;
  if (h > hmax) then
    h := hmax;

  fb := f(b);
  // check if step is too small
  hmin := max(abs(b),1.0)*SqrtMachEps;
  if h < hmin then
    h := max(tol, hmin);

  c := b + h;
  fc := f(c);
  if dfIsSmall then
    exit; // too small change - constant function

  if (fc > fb) then
  begin
    ErrCode := -2;
    exit;
  end;

  repeat
    if IsAtLim then break;
    a := b;  fa := fb;
    b := c;  fb := fc;
    h := h*factor;
    if h > hmax then
      h := hmax;
    c := b + h;
    if IsOutOfLim then break;
    if dfIsSmall then break;
  until fc > fb;

  x1 := a;
  x2 := c;
  step := h;
end;


function FMMfmin(ax, bx: TFloat; f: TFloatFunction1D; tol: TFloat): TFloat;
const
  IterMax = 9999;
var
  a,b,c,d,e,eps,xm,p,q,r,tol1,tol2,u,v,w: TFloat;
  fu,fv,fw,fx,x: TFloat;
  iter: TInt;
begin

//  c is the squared inverse of the golden ratio

      c := 0.5*(3.0 - sqrt(5.0));

//  eps is approximately the square root of the relative machine
//  precision.

//  eps := 1.00;
//  repeat
//    eps := eps/2.00;
//    tol1 := 1.0 + eps;
//  until tol1  =  1.00;
//  eps := sqrt(eps);
  eps := SqrtMachEps;

//  initialization

  a := ax;
  b := bx;
  v := a + c*(b - a);
  w := v;
  x := v;
  e := 0.0;
  fx := f(x);
  fv := fx;
  fw := fx;

//  main loop starts here
  for iter := 1 to IterMax do
  begin
    xm := 0.5*(a + b);
    tol1 := eps*abs(x) + tol/3.0;
    tol2 := 2.0*tol1;

//  check stopping criterion
    if (abs(x - xm) <= (tol2 - 0.5*(b - a))) then break;

// is golden-section necessary
    if (abs(e) > tol1) then
    begin //  fit parabola
      r := (x - w)*(fx - fv);
      q := (x - v)*(fx - fw);
      p := (x - v)*q - (x - w)*r;
      q := 2.00*(q - r);
      if (q  >  0.0) then p := -p;
      q :=  abs(q);
      r := e;
      e := d;
    end;

    if (abs(e) <= tol1) or (abs(p) >= abs(0.5*q*r)) or
       (p <= q*(a - x)) or (p >= q*(b - x)) then
    begin
//  a golden-section step
      if (x >= xm) then e := a - x;
      if (x < xm) then e := b - x;
      d := c*e;
    end else
    begin
//  a parabolic interpolation step
      d := p/q;
      u := x + d;

//  f must not be evaluated too close to ax or bx
      if ((u - a) < tol2) then d := sign2(tol1, xm - x);
      if ((b - u) < tol2) then d := sign2(tol1, xm - x);
    end;

//  f must not be evaluated too close to x
    if (abs(d) >= tol1) then u := x + d;
    if (abs(d) < tol1) then u := x + sign2(tol1, d);
    fu := f(u);

//  update  a, b, v, w, and x
    if (fu  <=  fx) then
    begin
      if (u >= x) then a := x;
      if (u < x) then b := x;
      v := w;
      fv := fw;
      w := x;
      fw := fx;
      x := u;
      fx := fu;
    end else
    begin
      if (u < x) then a := u;
      if (u >= x) then b := u;
      if (fu <= fw) or (w  =  x) then
      begin
        v := w;
        fv := fw;
        w := u;
        fw := fu;
      end else
      if (fu <= fv) or (v = x) or (v = w) then
      begin
        v := u;
        fv := fu;
      end;
    end;
  end;
  result := x;
end;

end.
