{*
  uzeroin - translation from FORTRAN
  Ref:
  G.E.Forsythe, M.A.Malcolm, C.B.Moler. Computer methods for mathematical computations.
  Prentice-Hall, 1977.
  Nikolai V. Shokhirev, 2003-2007 <nikolai@shokhirev.com> <nikolai@u.arizona.edu>
  @author Nikolai Shokhirev
  @author ORIGINAL AUTHOR: Richard Brent
 @version 2003.01.06 - created
 @version 2004.01.21
 @version 2007.07.18 - Added DelphiCodeToDoc comments
}
unit uzeroin;

interface

uses
  uMatTypes;

{*
  ZEROIN - A zero of the function f(x) is computed in the interval ax, bx .
NOTES:
  It is assumed that  f(ax)  and  f(bx)  have opposite signs without a check.
  zeroin returns a zero x in the given interval [ax, bx] to within a tolerance
  4*macheps*abs(x) + tol, where macheps is the relative machine precision.
   This function subprogram is a slightly modified translation of the algol 60
  procedure zero given in richard brent, algorithms for minimization without
  derivatives, prentice - hall, inc. (1973).

Ref:
  G.E.Forsythe, M.A.Malcolm, C.B.Moler.
  Computer methods for mathematical computations. Prentice-Hall, 1977.

  @param ax   left endpoint of initial interval
  @param bx   right endpoint of initial interval
  @param f   function subprogram which evaluates f(x) for any x in the interval ax, bx
  @param tol  desired length of the interval of uncertainty of the final result ( >= 0)
  @return zeroin abcissa approximating a zero of f in the interval ax, bx

  @author Translation from FORTRAN by Nikolai Shokhirev
  @author ORIGINAL AUTHOR: Richard Brent
  }
function zeroin(ax, bx: TFloat; f: TFloatFunction1D; tol: TFloat): TFloat;

implementation

function zeroin(ax,bx: TFloat; f: TFloatFunction1D; tol: TFloat): TFloat;
const
  IterMax = 9999;
var
  a, b, c, d, e, eps, fa, fb, fc: TFloat;
  tol1, xm, p, q, r, s: TFloat;
  iter: TInt;
begin
  eps := MachEps;

// initialization
  a := ax;
  b := bx;
  fa := f(a);
  fb := f(b);

  c := a;
  fc := fa;
  d := b - a;
  e := d;

  for iter := 1 to IterMax do
  begin
    if (abs(fc) < abs(fb)) then
    begin
      a := b;
      b := c;
      c := a;
      fa := fb;
      fb := fc;
      fc := fa;
    end;

    // convergence test
    tol1 := 2.0*eps*abs(b) + 0.5*tol;
    xm := 0.5*(c - b);
    if (abs(xm) <= tol1) or (fb = 0.0) then break;

    // is bisection necessary
    if (abs(e) >= tol1) or (abs(fa) > abs(fb)) then
    begin // bisection
      d := xm;
      e := d;
    end else
    begin
      if (a = c) then
      begin // inverse quadratic interpolation
        q := fa/fc;
        r := fb/fc;
        s := fb/fa;
        p := s*(2.0*xm*q*(q - r) - (b - a)*(r - 1.0));
        q := (q - 1.0)*(r - 1.0)*(s - 1.0);
      end else
      begin // linear interpolation
        s := fb/fa;
        p := 2.0*xm*s;
        q := 1.0 - s;
      end;

      // adjust signs
      if (p  >  0.0) then q := -q;
      p := abs(p);

      // is interpolation acceptable
      if ((2.0*p) < (3.0*xm*q - abs(tol1*q))) or (p < abs(0.5*e*q)) then
      begin // bisection
        d := xm;
        e := d;
      end else
      begin
        e := d;
        d := p/q;
      end;
    end;

    // complete step
    a := b;
    fa := fb;
    if (abs(d) > tol1) then b := b + d;
    if (abs(d) <= tol1) then b := b + sign2(tol1, xm);
    fb := f(b);

    if ((fb*(fc/abs(fc))) > 0.0) then
    begin
      c := a;
      fc := fa;
      d := b - a;
      e := d;
    end;
  end;
  result := b;
end;

end.
