{@abstract(Derivatives and average routines)
@author(Nikolai Shokhirev <nikolai@shokhirev.com>)
@created(June 6, 2002)
@lastmod(July 16, 2003)
Nikolai V. Shokhirev, 2002-2003 }
uunit uAverage;

interface
uses
  uMatTypes, uArrays;

  { two-point derivative: Ain[i+1]-Ain[i])/h }
  procedure deriv2(const Ain, Aout: IVector; h: TFloat = 1.0);
  { tree-point derivative: Ain[i+1]-Ain[i])/h }
  procedure deriv3(const Ain, Aout: IVector; h: TFloat = 1.0);
  { 5-point average. Ain can be equal to Aout }
  procedure Avg5(const Ain, Aout: IVector);
  { 5-point average derivative. Ain can be equal to Aout }
  procedure Avg5d(const Ain, Aout: IVector; h: TFloat = 1.0);
  { 5-point average the second derivative. Ain can be equal to Aout }
  procedure Avg5d2(const Ain, Aout: IVector; h: TFloat = 1.0);

  procedure Sharpen(const Ain, Aout: IVector);

implementation

uses
  math;

procedure deriv2(const Ain, Aout: IVector; h: TFloat = 1.0);
var
  i, N: TInt;
begin
  N := Ain.Dim;
  for i := 1 to N-1do
    Aout[i] := (Ain[i+1]-Ain[i])/h;
  Aout[N] := Aout[N-1];
end;

procedure deriv3(const Ain, Aout: IVector; h: TFloat = 1.0);
var
  i, N: TInt;
begin
  N := Ain.Dim;
  for i := 2 to N-1do
    Aout[i] := (Ain[i+1]-Ain[i-1])/h/2.0;
  Aout[N] := Aout[N-1];
  Aout[1] := Aout[2];
end;

{ 5-point average. Ain can be equal to Aout <br>
 ( 31,  9, -3, -5,  3)/35     1     <br>
    ^                               <br>
 (  9, 13, 12,  6, -5)/35     2     <br>
        ^                           <br>
 ( -3, 12, 17, 12, -3)/35           <br>
            ^                       <br>
 ( -5,  6, 12, 13,  9)/35    N-1    <br>
                ^                   <br>
 (  3, -5, -3,  9, 31)/35    N      <br>
                    ^      }
procedure Avg5(const Ain, Aout: IVector);
var
  i, N: TInt;
  a3, a2, a1, a0: TFloat;
begin
  N := Ain.Dim;
  a1 := Ain[1];
  a0 := Ain[2];
  Aout[1] := (31.0*a1+9.0*a0-3.0*Ain[3]-5.0*Ain[4]+3.0*Ain[5])/35.0;
  Aout[2] := (9.0*a1+13.0*a0+12.0*Ain[3]+6.0*Ain[4]-5.0*Ain[5])/35.0;

  for i := 3 to N-2 do
  begin
    a2 := a1;  a1 := a0;  a0 := Ain[i];
    Aout[i] := (-3.0*a2+12.0*a1+17.0*a0+12.0*Ain[i+1]-3.0*Ain[i+2])/35.0;
  end;

  a3 := a2;  a2 := a1;  a1 := a0;  a0 := Ain[N-1];
  Aout[N-1] := (-5.0*a3+6.0*a2+12.0*a1+13.0*a0+9.0*Ain[N])/35.0;
  Aout[N]   := (3.0*a3-5.0*a2-3.0*a1+9.0*a0+31.0*Ain[N])/35.0;
end;

{ 5-point average derivative. Ain can be equal to Aout
 (-54, 13, 40, 27,-26)/70     1
    ^
 (-34,  3, 20, 17, -6)/70     2
        ^
 ( -2, -1,  0,  1,  2)/10
            ^
 (  6,-17,-20, -3, 34)/70    N-1
                ^
 ( 26,-27,-40,-13, 54)/70    N
                    ^      }
procedure Avg5d(const Ain, Aout: IVector; h: TFloat = 1.0);
var
  i, N: TInt;
  a3, a2, a1, a0: TFloat;
begin
  N := Ain.Dim;
  a0 := Ain[1];
  Aout[1] := (-54.0*a0+13.0*Ain[2]+40.0*Ain[3]+27.0*Ain[4]-26.0*Ain[5])/70.0/h;
  a1 := a0;
  a0 := Ain[2];
  Aout[2] := (-34.0*a1+3.0*a0+20.0*Ain[3]+17.0*Ain[4]-6.0*Ain[5])/70.0/h;

  for i := 3 to N-2 do
  begin
    a2 := a1;  a1 := a0;  a0 := Ain[i];
    Aout[i] := (-2.0*a2 -a1 + Ain[i+1] +2.0*Ain[i+2])/10.0/h;
  end;

  a3 := a2;  a2 := a1;  a1 := a0;  a0 := Ain[N-1];
  Aout[N-1] := (6.0*a3-17.0*a2-20.0*a1-3.0*a0+34.0*Ain[N])/70.0/h;
  Aout[N]   := (26.0*a3-27.0*a2-40.0*a1-13.0*a0+54.0*Ain[N])/70.0/h;
end;

{ 5-point average the second derivative. Ain can be equal to Aout
 (  2, -1, -2, -1,  2)/7    }
procedure Avg5d2(const Ain, Aout: IVector; h: TFloat = 1.0);
var
  i, N: TInt;
  a2, a1, a0: TFloat;
begin
  N := Ain.Dim;
  a1 := Ain[1];
  a0 := Ain[2];

  for i := 3 to N-2 do
  begin
    a2 := a1;  a1 := a0;  a0 := Ain[i];
    Aout[i] := (2.0*a2 -a1 -2.0*a0 -Ain[i+1] +2.0*Ain[i+2])/7.0/h/h;
  end;

  Aout[1] := Aout[3];
  Aout[2] := Aout[3];
  Aout[N-1] := Aout[N-2];
  Aout[N] := Aout[N-2];
end;

procedure Sharpen(const Ain, Aout: IVector);
var
  i, i1, i2, N: TInt;
  Aextr, Amax, Amin, d2, d2old, d1, d1old: TFloat;
  wasExtremum, done: boolean;
  
  function d(i: TInt): TFloat;
  var
    j: TInt;
  begin
    j := i;
    if j < 3 then j := 3;
    if j > N-2 then j := N-2;
    result := (-2.0*Ain[j-2]-Ain[j-1]+Ain[j+1]+2.0*Ain[j+2]);
  end;
  
  function dd(i: TInt): TFloat;
  var
    j: TInt;
  begin
    j := i;
    if j < 3 then j := 3;
    if j > N-2 then j := N-2;
    result := (2.0*Ain[j-2]-Ain[j-1]-2.0*Ain[j]-Ain[j+1]+2.0*Ain[j+2]);
  end;

begin
  N := Ain.Dim;
  Aextr := Ain[1];
  d1 := d(1);
  d2 := dd(1);
  i2 := 0;

  repeat
    i1 := i2+1;
    i  := i2;
    wasExtremum := false;
    Amax := -MaxFloat;
    Amin := MaxFloat;
    repeat
      inc(i);
      d1old := d1;
      d1 := d(i);
      d2old := d2;
      d2 := dd(i);
      Amax := max(Amax,Ain[i]);
      Amin := min(Amin,Ain[i]);
      if (d1old*d1 < 0.0) then
      begin
        Aextr := Ain[i];
        wasExtremum := true;
      end;
      done := d2old*d2 < 0.0;
    until done or (i=N);
    i2 := i;

    if not wasExtremum then
    begin
      if i=N then
       Aextr := Ain[i]
      else
      begin
        if dd(i-1) > 0.0 then Aextr := Amin else Aextr := Amax;
      end;
    end;
    for i := i1 to i2 do
    Aout[i] := Aextr;
  until i2 >= N;

end;

end.
