{*
 Fourier signal processing library
 Nikolai V. Shokhirev, 2001-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2002.02.02 - created
 @version 2004.12.12 - modified
 @version 2007.08.08 - Added DelphiCodeToDoc comments
 }
unit uFourier;

interface

uses
  uMatTypes, uDynArrays;

type
  TTransformType = (ttDirect, ttInverse);
  TFactorArray = array [1..32] of TInt;

{<pre> if the input signal is exp(i*2*Pi*M*j/N) 
  sampled at j = 0, 1, .. , N-1 then 
  the output array from DFT and CFTG is: 
 idx |  1 |  2 |  3 | .. | N-1|  N | 
  M  |  0 |  1 |  2 | .. | N-2| N-1| 
  M  |+/-N|-N+1|-N+2| .. | -2 | -1 | 
</pre>}

{<pre> returns the corresponding index for a given M
 from the "standard" interval:
    MinFreqIndex <= M <= MaxFreqIndex
    1 <= IndexDFT <= N
    1 <= IndexDFT1 <= N

N=3 M   = -1, 0, 1
    Idx =  3, 1, 2
N=4 M   = -1, 0, 1, 2
    Idx =  4, 1, 2  3
N=5 M   = -2,-1, 0, 1, 2
    Idx =  4, 5, 1, 2, 3              </pre>}
  function IndexDFT(N, M: TInt): TInt;
  function IndexDFT1(N, M: TInt): TInt;

{<pre> returns the corresponding index for a given M
 from the "standard" interval:
    MinFreqIndex <= M <= MaxFreqIndex
    0 <= IndexDFT0 <= N-1

N=3 M   = -1, 0, 1
    Idx =  2, 0, 1
N=4 M   = -1, 0, 1, 2
    Idx =  3, 0, 1  3
N=5 M   = -2,-1, 0, 1, 2
    Idx =  3, 4, 0, 1, 2              </pre>}
  function IndexDFT0(N, M: TInt): TInt;

{<pre> MinFreqIndex = -((N - 1) div 2)
  N   MinFreqIndex  MaxFreqIndex
  1    0            0            0
  2    0            1            0, 1
  3   -1            1           -1, 0, 1
  4   -1            2           -1, 0, 1, 2
  5   -2            2           -2,-1, 0, 1, 2   </pre>}
  function MinFreqIndex(N: TInt): TInt;
{<pre> MaxFreqIndex =  (N div 2) = Nyquist Frequency</pre>}
  function MaxFreqIndex(N: TInt): TInt;

{<pre> ain input frequency array - output from GFTG 
  returns centered frequency array aout 
  aout[1] = MinFreq, . . , aout[N] = MaxFreq </pre>}
  procedure CenterFreq(const ain, aout: IFArr1D);

{<pre> DFT - NOT Fast Fourier Transforms -  for testing pirpose only 
  see R.W.Hamming, Numerical methods for scientists and engineers, 
  McGraw-Hill, N.Y. 1962.</pre>}

{<pre> Forward Discrete Fourier Transform of real data 
  It is used in combinatio with InvDFTR and InterpolDFTR
  x[1..N] of TReal - input signal 
  result[1..(N+2) div 2] of Complex - Fourier transform 
  if x[k] = x(h*(k-1)) then the transform frequencies are 
   0, w, 2*w, 3*w, . .  , where w = 2.0*Pi/(h*N)  </pre>}
  function DFTR(const x: IFArr1D): ICArr1D;
{<pre> Inverse Discrete Fourier Transform to real data 
  It is used in combinatio with DFTR 
  If c = DFTR(x) then x = InvDFTR(c, N) 
  N - Number of real points ( N = x.Dim1 in DFTR) 
  c[1..(N+2) div 2] of Complex - Fourier transform 
  result[1..N] of TReal - real signal  </pre>}
  function InvDFTR(const c: ICArr1D; N: TInt): IFArr1D;
{<pre> Fourier Interpolation of real data
  It is used in combinatio with DFTR 
  N - Number of signal points in DFTR 
  h - step = distance between signal points 
  c[1..(N+2) div 2] of Complex - Fourier transform from DFTR </pre>}
  function InterpolDFTR(t, h: TFloat; const c: ICArr1D; N: TInt): TFloat;

{<pre> Forward Discrete Fourier Transform of complex data
  It is used in combinatio with InvDFT and InterpolDFT: 
  C[1..N] of Complex - input signal 
  result[1..N] of Complex - Fourier transform
  if x[k] = x(h*(k-1)) then the transform frequencies are 
  0, w, 2*w, 3*w, . .  , where w = 2.0*Pi/(h*N)  </pre>}
  function DFT(const x: ICArr1D): ICArr1D;
{<pre> Inverse Discrete Fourier Transform to complex data
  It is used in combinatio with DFT: 
  If c = DFT(z) then z = InvDFT(c) 
  c[1..N] of Complex - Fourier transform 
  result[1..N] of Complex - complex signal  </pre>}
  function InvDFT(const c: ICArr1D): ICArr1D;
{<pre> Fourier Interpolation of complex data  
  It is used in combinatio with DFT: 
  h - step = distance between signal points 
  c[1..N] of Complex - Fourier transform from DFT </pre>}
  function InterpolDFT(t, h: TFloat; const c: ICArr1D): Complex;

{<pre> Forward DFT for even number of signal points N = 2*M
  It is used in combinatio with InvDFTSym and InterpolSym: 
  C[1..N] of Complex - input signal 
  result[1..N] of Complex - Fourier transform 
  if x[k] = x(h*(k-1)) then the transform frequencies are 
  -Pi, -Pi+w, -Pi+2*w, . .  , where w = Pi/(h*M) 
  They map to 1, 2, .. , N indices of the result  </pre>}
  function DFTSym(const x: ICArr1D): ICArr1D;
{<pre> Inverse DFT for even number of signal points N = 2*M
  It is used in combinatio with DFTSym: 
  If c = DFTSym(z) then z = InvDFTSym(c) 
  c[1..N] of Complex - Fourier transform 
  result[1..N] of Complex - complex signal  </pre>}
  function InvDFTSym(const c: ICArr1D): ICArr1D;
{<pre> Fourier Interpolation of complex data with "Symmetric" frequencies 
  It is used in combinatio with DFTSym: 
  h - step = distance between signal points 
  c[1..N] of Complex - Fourier transform from DFT </pre>}
  function InterpolSym(t, h: TFloat; const c: ICArr1D): Complex;

{<pre>John Herbster's implementation of Glassman FFT algorithm. 
  The number of points does not have to be a power of 2. 
  Instead, the transform will take advantage of any factoring 
  that can be done on the number of points.  Which means that if the number 
  of points is prime, the transform will not be "fast" at all. 
  
  Returns a complex Fourier transform of the Cinp array into the Cout array. 
  The Cscr (scratch) array may be same as the input array. The transform will 
  be from the time to the frequency domain if NC is positive of the number   
  of complex points to be converted or from the frequency to the time domain 
  if NC is the negative of the number of complex points.  The only difference 
  between the two is just a divide by the number of points.

  Input:  Cinp[1..N]
  Work:   Cscr[1..N]  Cscr = Cinp - OK
  Output: Cout [1..N]                         </pre>}
  procedure CFTG (const Cinp, Cout, Cscr: ICArr1D; NC: TInt);

{<pre>John Herbster's implementation of Glassman FFT algorithm. 

  TransformType ttDirect  - Fourier transform
    Input
      Cinp - input signal: Arbitrary limits, Dimension = N
      Cout - Initially can have arbitrary limits, Dimension = N
    Work
      Cscr - work array: Arbitrary limits, Dimension = N, Cscr = Cinp is OK.
    Output
      Cout[0..N-1] - zero-based array of the Fourier transform.

    if Cinp[n] = Cinp(tn) where tn = h*n then the transform frequencies are 
     0, w, 2*w, 3*w, . .  (N-1)*w, where w = 2.0*Pi/(h*N)

  TransformType = ttInverse - inverse Fourier transform
    Input
      Cinp[0..N-1] - zero-based array of the Fourier transform.
    Work
      Cscr - work array: Arbitrary limits, Dimension = N, Cscr = Cinp is OK.
    Output
      Cout[Lo1..Hi1] - reconstructed signal: Arbitrary limits, Dimension = N.
                       FFTG keeps the initial limits.     </pre>}
  procedure FFTG (const Cinp, Cout, Cscr: ICArr1D; TransformType: TTransformType);

{<pre>
  Input
    Cinp[0..N-1]
  Output
    Cout[Mmin..Mmax], Mmin = -((N - 1) div 2), Mmax = (N div 2)

  Cout[m] = Cinp[N-m], Mmin <= m <= -1
  Cout[m] = Cinp[m],   0    <= m <= Mmax   </pre>}
  procedure ZeroToCentric(const Cinp, Cout: ICArr1D);

{<pre>
  Input
    Cinp[Mmin..Mmax], Mmin = -((N - 1) div 2), Mmax = (N div 2)
  Output
    Cout[0..N-1]           </pre>}
  procedure CentricToZero(const Cinp, Cout: ICArr1D);

implementation

function IndexDFT(N, M: TInt): TInt;
begin
  if M < 0 then
    result := N + M + 1  // -N < M < 0
  else
    result := M + 1;
end;

function IndexDFT1(N, M: TInt): TInt;
begin
  if M < 0 then
    result := N + M + 1  // -N < M < 0
  else
    result := M + 1;
end;

function IndexDFT0(N, M: TInt): TInt;
begin
  if M < 0 then
    result := N + M  // -N < M < 0
  else
    result := M;
end;

function MinFreqIndex(N: TInt): TInt;
begin
  result := -((N - 1) div 2);
end;

function MaxFreqIndex(N: TInt): TInt;
begin
  result := N div 2;
end;

procedure CenterFreq(const Ain, Aout: IFArr1D);
var
  M, N, k: TInt;
begin
  N := Ain.Dim1;
  k := 0;
  for M := MinFreqIndex(N) to MaxFreqIndex(N) do
  begin
    inc(k);
    Aout[k] := Ain[IndexDFT(N, M)];
  end;
end;

function DFTR(const x: IFArr1D): ICArr1D;
var
  temp: ICArr1D;
  N2, N1, N, k, m: TInt;
  r, i, rN, h, Hk, Hkm: TFloat;
begin
  N := x.Dim1;
  rN := 1.0/N;
  N1 := (N+1) div 2;
  N2 := (N+2) div 2;
  temp := TCArr1D.Create(N2);
  h := 2.0*Pi*rN;
  Hk := 0.0;
  for k := 1 to N2 do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{     Hkm := h*(k-1)*(m-1);}
      r := r+x[m]*Cos(Hkm);
      i := i-x[m]*Sin(Hkm);
      Hkm := Hkm + Hk;
    end;
    if (k>1) and (k<=N1)then
    begin
      r := r+r;
      i := i+i;
    end;
    temp.Re[k] := r*rN;
    temp.Im[k] := i*rN;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFTR(const c: ICArr1D; N: TInt): IFArr1D;
var
  temp: IFArr1D;
  N2, k, m: TInt;
  r, h, Hm, Hkm: TFloat;
begin
  temp := TFArr1D.Create(N);
  N2 := c.Dim1;
  h := 2.0*Pi/N;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    Hkm := 0.0;
//    N2 := (N+2) div 2;
    for k := 1 to N2 do
    begin
{     Hkm := h*(m-1)*(k-1);}
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
    Hkm := Hkm + Hm;
    end;
    temp[m] := r;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolDFTR(t, h: TFloat; const c: ICArr1D; N: TInt): TFloat;
var
  N2, k: TInt;
  r, Hx, Hkx: TFloat;
begin
  N2 := c.Dim1;
  Hx := (2.0*Pi*t/h)/N;
  r := 0.0;
  Hkx := 0.0;
  for k := 1 to N2 do
  begin
{     Hkx := Hx*(k-1) }
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    Hkx := Hkx + Hx;
  end;
  result := r;
end;

function DFT(const x: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  N, k, m: TInt;
  r, i, rN, h, Hk, Hkm: TFloat;
begin
//  c := DFTR(x);
//  xx := InvDFTR(c, N);
  N := x.Dim1;
  rN := 1.0/N;
  temp := TCArr1D.Create(N);
  h := 2.0*Pi*rN;
  Hk := 0.0;
  for k := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{     Hkm := h*(k-1)*(m-1);}
      r := r+x.Re[m]*Cos(Hkm)+x.Im[m]*Sin(Hkm);
      i := i-x.Re[m]*Sin(Hkm)+x.Im[m]*Cos(Hkm);
      Hkm := Hkm + Hk;
    end;
    temp.Re[k] := r*rN;
    temp.Im[k] := i*rN;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFT(const c: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  N, k, m: TInt;
  r, i, rN, h, Hm, Hkm: TFloat;
begin
  N := c.Dim1;
  rN := 1.0/N;
  temp := TCArr1D.Create(N);
  h := 2.0*Pi*rN;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for k := 1 to N do
    begin
{     Hkm := h*(m-1)*(k-1);}
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
      i := i+c.Re[k]*Sin(Hkm)+c.Im[k]*Cos(Hkm);
    Hkm := Hkm + Hm;
    end;
    temp.Re[m] := r;
    temp.Im[m] := i;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolDFT(t, h: TFloat; const c: ICArr1D): Complex;
var
  N, k: TInt;
  r, i, Hx, Hkx: TFloat;
begin
  N := c.Dim1;
  Hx := (2.0*Pi*t/h)/N;
  r := 0.0;
  i := 0.0;
  Hkx := 0.0;
  for k := 1 to N do
  begin
{     Hkx := Hx*(k-1) }
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    i := i+c.Re[k]*Sin(Hkx)+c.Im[k]*Cos(Hkx);
    Hkx := Hkx + Hx;
  end;
  result.Re := r;
  result.Im := i;
end;

function DFTSym(const x: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  MM, N, k, m: TInt;
  r, i, rM, h, Hk, Hkm: TFloat;
begin
  N := x.Dim1;
  if odd(N) then
    exit;
  temp := TCArr1D.Create(N);
  MM := N div 2;
  rM := 1.0/MM;

  h := Pi*rM;
  Hk := -Pi;
  for k := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{      Hkm := h*(k-1-MM)*(m-1);}
      r := r+x.Re[m]*Cos(Hkm)+x.Im[m]*Sin(Hkm);
      i := i-x.Re[m]*Sin(Hkm)+x.Im[m]*Cos(Hkm);
      Hkm := Hkm + Hk;
    end;
    temp.Re[k] := r*rM*0.5;
    temp.Im[k] := i*rM*0.5;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFTSym(const c: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  MM, N, k, m: TInt;
  r, i, rM, h, Hm, Hkm, fk: TFloat;
begin
  N := c.Dim1;
  if odd(N) then
    exit;
  MM := N div 2;
  rM := 1.0/MM;
  temp := TCArr1D.Create(N);
  h := Pi*rM;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
//    Hkm := Hm;
    fk := -MM;
    for k := 1 to N do
    begin
{     Hkm := h*(m-1)*(k-1-MM);}
      Hkm := Hm*fk;
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
      i := i+c.Re[k]*Sin(Hkm)+c.Im[k]*Cos(Hkm);
      fk := fk + 1.0;
    end;
    temp.Re[m] := r;
    temp.Im[m] := i;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolSym(t, h: TFloat; const c: ICArr1D): Complex;
var
  N, k, M: TInt;
  r, i, Hx, Hkx, fk: TFloat;
begin
  result.Re := 0.0;
  result.Im := 0.0;
  N := c.Dim1;
  if odd(N) then
    exit;
  M := N div 2;
  r := 0.0;
  i := 0.0;
  fk := M;
  Hx := (Pi*t/h)/fk;
  for k := N downto 1 do
  begin
    fk := fk - 1.0;
    Hkx := Hx*fk;
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    i := i+c.Re[k]*Sin(Hkx)+c.Im[k]*Cos(Hkx);
  end;
  result.Re := r;
  result.Im := i;
end;

{ John Herbster's implementation of Glassman FFT ---------------------------- }

procedure CScale(const aInp, aOut: ICArr1D; A: TFloat);
var
  i: TInt;
begin
  for i := aInp.Lo1 to aInp.Hi1 do
    aOut[i] := cMul(aInp[i], A);
end;

{ Factor n into its nf primes in array factors[]. }
procedure GetFactors(n: integer; var factors: tFactorArray; var nf: integer);
var
  nr, tf, q: word;
begin
  nr := n;
  q  := nr;
  tf := 2;
  nf := 1;
  while tf < q do
  begin
    q := nr div tf;
    if q*tf=nr then
    begin
      factors[nf]:=tf;
      Inc(nf);
      nr:=q;
    end
    else
      Inc(tf);
  end{while};
  factors[nf]:=nr;
end;

procedure GFTInnerLoop (const aInp, aOut: ICArr1D;  NC, X, S: integer);
var
  WDL1, WD, CTemp, CTemp2: Complex;
  N, L1, L, K, D, I, U, H: word;
  a: TFloat;
begin
  N := abs(NC);
  D := N div S;
  I := 0;
  a := (-(2.0*Pi)*D)/NC;
  WD := cmplx(cos(a), sin(a));
  for L1:=1 to S do
  begin
    If L1 = 1 then
      WDL1 := Cmplx1
    else
      WDL1 := cMul(WDL1,WD);
    for L:=1 to D do
    begin
      K := (L+(L1-1)*D*X) mod N;
      U := K+(X-1)*D;
      CTemp := aInp[U];
      for H := 2 to X do
      begin
        CTemp := cMul(CTemp,WDL1);
        U := U - D;
        CTemp2 := aInp[U];
        CTemp := cAdd(CTemp,CTemp2);
      end;
      Inc(I);
      aOut[i] := CTemp;
    end{L-loop};
  end{L1-loop};
End;

{ Cscr array may be same as the Cinp array.}
procedure CFTG (const Cinp, Cout, Cscr: ICArr1D; NC: TInt);
var
  a: TFloat;
  j, N, x, s, nf: TInt;
  r: TFactorArray;
begin
{ Let N be the number of complex elements in the array.}
  N := abs(NC);
  If N = 0 then
    exit;

  GetFactors(N, r, nf); { Factor N into its primes.}

{ Now do the transformation.}
  s := 1;
  for j:= 1 to nf do
  begin
    x := r[j];
    s := s*x;
    If j=1 then
      GFTInnerLoop (Cinp, Cout, NC, X, S)
    else
      if Odd(j) then
        GFTInnerLoop (Cscr, Cout, NC, X, S)
      else
        GFTInnerLoop (Cout, Cscr, NC, X, S);
  end{j-loop};

{ Finally, check for the required scaling and that data is in out array.}
  If nc < 0 then a := 1.0 else a := 1.0/N;
  If not Odd(nf) then
    CScale({in}Cscr, {out}Cout, a)
  else
    if a <> 1 then
      CScale(Cout, Cout, a);

end;

{ Cscr array may be same as the Cinp array.}
{
  Direct = true - Fourier transform
  Direct = false - inverse Fourier transform
}
procedure FFTG (const Cinp, Cout, Cscr: ICArr1D; TransformType: TTransformType);
var
  a: TFloat;
  j, N, x, s, nf, NC, bInp, bOut, bScr: TInt;
  r: TFactorArray;
//  Restore: IRestore;
begin
{ Let N be the number of complex elements in the array.}
  N := Cinp.Dim1;
  if TransformType = ttDirect then
    NC := N
  else
    NC := -N;

  Assert( (Cout.Dim1>=N) and (Cscr.Dim1>=N), RS_LimMismatch );

//  Restore := TRestore1D.Create([Cinp, Cout, Cscr]);
  bInp := Cinp.Lo1;  Cinp.Lo1 := 1;
  bOut := Cout.Lo1;  Cout.Lo1 := 1;
  bScr := Cscr.Lo1;  Cscr.Lo1 := 1;

  GetFactors(N, r, nf); { Factor N into its primes.}

{ Now do the transformation.}
  s := 1;
  for j:= 1 to nf do
  begin
    x := r[j];
    s := s*x;
    If j=1 then
      GFTInnerLoop (Cinp, Cout, NC, X, S)
    else
      if Odd(j) then
        GFTInnerLoop (Cscr, Cout, NC, X, S)
      else
        GFTInnerLoop (Cout, Cscr, NC, X, S);
  end{j-loop};

{ Finally, check for the required scaling and that data is in out array.}
  If TransformType = ttDirect then
    a := 1.0/N
  else
    a := 1.0;
  If not Odd(nf) then
    CScale({in}Cscr, {out}Cout, a)
  else
    if TransformType = ttDirect then
      CScale(Cout, Cout, a);

  Cscr.Lo1 := bScr;    // Restore
  Cinp.Lo1 := bInp;    // Restore
  if TransformType = ttDirect then
    Cout.Lo1 := 0      // Frequency
  else
    Cout.Lo1 := bOut;  // Signal
end;

procedure ZeroToCentric(const Cinp, Cout: ICArr1D);
var
  c: Complex;
  n, m, i, k, l: TInt;
begin
  n := Cinp.Dim1;
  Assert( Cout.Dim1 = n, RS_LimMismatch );
  m := (n-1) div 2;
  Cinp.Lo1 := 0;
  Cout.Lo1 := 0;
  Cout.Assign(Cinp);
  for i := 1 to m do
  begin
    c := Cout[n-1];
    for k := n-1 downto 1 do
      Cout[k] := Cout[k-1];
    Cout[0] := c;
  end;
  l := MinFreqIndex(n);
  Cout.Lo1 := l;
end;

procedure CentricToZero(const Cinp, Cout: ICArr1D);
var
  c: Complex;
  n, m, i, k, l: TInt;
begin
  n := Cinp.Dim1;
  Assert( Cout.Dim1 = n, RS_LimMismatch );
  m := (n-1) div 2;
  Cinp.Lo1 := 0;
  Cout.Lo1 := 0;
  Cout.Assign(Cinp);
  for i := 1 to m do
  begin
    c := Cout[0];
    for k := 0 to N-2 do
      Cout[k] := Cout[k+1];
    Cout[N-1] := c;
  end;
  l := MinFreqIndex(n);
  Cinp.Lo1 := l;
end;

end.
