{@abstract(Tests for Data and DFT)
@author(Nikolai Shokhirev <nikolai@shokhirev.com> <nikolai@u.arizona.edu>)
@created(January 1, 2003)
@lastmod(June 06, 2004)
Nikolai V. Shokhirev, 2003-2004 }
unit Test_uDFT;

interface

uses
  TestFramework,
  SysUtils,
  uMatTypes,
  uDynArrays,
  uDFT;

type

Check_FArray1D = class(TTestCase)
private
  A: TFArr;
  V1, V2: IFArr1D;
  i, j, d1, l1, h1: integer;
  eps: TFloat;
  n, m: TFloat;
  msg: string;
public
  procedure setUp;  override;
  procedure tearDown; override;
published
  procedure VerifyFill;
  procedure VerifyCopy;
  procedure VerifySwap;
  procedure VerifyGetHiLo;
  procedure VerifyGetDimLo;
  procedure VerifySetGetValue;
  procedure VerifyCreate;
  procedure VerifyCreateGeneric;
  procedure VerifyCreateNatural;
  procedure VerifyCreateCentered;
end;

Check_CArray1D = class(TTestCase)
private
  AR, AI: TFArr;
  V1, V2: ICArr1D;
  i, j, d1, l1, h1: integer;
  eps: TFloat;
  n, m: Complex;
  msg: string;
public
  procedure setUp;  override;
  procedure tearDown; override;
published
  procedure VerifyFill;
  procedure VerifyCopy;
  procedure VerifyConjugate;
  procedure VerifySwap;
  procedure VerifyGetHiLo;
  procedure VerifyGetDimLo;
  procedure VerifySetGetValue;
  procedure VerifyCreate;
  procedure VerifyCreateGeneric;
  procedure VerifyCreateNatural;
  procedure VerifyCreateCentered;
end;

Check_Tdft1D = class(TTestCase)
private
  A: Idft1D;
  F1, F2: IFArr1D;
  C1, C2: ICArr1D;
  N, err: integer;
  x, eps, del: TFloat;
  msg, msg0: string;
public
  procedure setUp;  override;
  procedure tearDown; override;
published
  procedure Verifydft1DCreate;
  procedure VerifyForwardInverse;
  procedure VerifyForwardInvSignal;
  procedure VerifyHarmonicRe;
  procedure VerifyHarmonicIm;
end;


function Suite : ITestSuite;

implementation

uses
  uDynArrUtils, uDynLinAlg, math,
  uDynArrUtilsX,
  uDynArrIO,
  uDynObjAlg;

function Suite : ITestSuite;
begin
  result := TTestSuite.Create('uDFT Tests');

  result.addTest(Check_FArray1D.Suite);
  result.addTest(Check_CArray1D.Suite);
  result.addTest(Check_Tdft1D.Suite);
end;

{ Check_FArray1D }

procedure Check_FArray1D.setUp;
begin
  d1 := 16;
  eps := 1.0e-12;
  Randomize;
end;

procedure Check_FArray1D.tearDown;
begin
end;

{ check: FastInverse, MtxMt }
procedure Check_FArray1D.VerifyFill;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1);
  V2 := TFArray1D.Create(l1,h1);
  V1.Fill(sqrt(2.0));
  Fill(V2,sqrt(2.0));
  Check(SameArr(V1,V2,msg,eps), 'Fail Fill:'+msg);
  V1.Times(0.5);
  CxArr(0.5,V2);
  Check(SameArr(V1,V2,msg,eps), 'Fail Times:'+msg);
end;

{ check: Cholesky, MTtxMt, MtxMtT, MtxVt, LSolve, LTSolve,ChSolve }
procedure Check_FArray1D.VerifyCopy;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1);
  RandArr1D(V1,-9.0,9.0);
  V2 := TFArray1D.Create(l1,h1);
  V2.Assign(V1);
  Check(SameArr(V1,V2,msg,eps), 'Fail VerifyCopy:'+msg);
end;

procedure Check_FArray1D.VerifySwap;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1);
  i := RndNum(l1,h1);
  n := RndFloat(-9.0,9.0);
  V1[i] := n;
  j := RndNumEx(l1,h1,i);
  m := RndFloat(-9.0,9.0);
  V1[j] := m;
  V1.Swap(i,j);
  Check( (V1[i]=m) and (V1[j]=n), 'Fail Swap');
end;

procedure Check_FArray1D.VerifyGetHiLo;
begin
  l1 := RndNum(-5,5); h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1);
  Check( (V1.Hi1=V1.Lo1+d1-1), 'Fail Hi, Lo');
//  i := RndNumEx(-5,5,l1);
//  V1.Lo1 := i;
//  Check( (V1.Hi1=V1.Lo1+d1-1), 'Fail Hi, Lo #2');
end;

procedure Check_FArray1D.VerifyGetDimLo;
begin
  h1 := RndNum(2,7);
  V1 := TFArray1D.Create(h1);
  Check( (h1 = V1.Hi1), 'Fail Hi=Dim');
  Check( (1=V1.Lo1), 'Fail Lo=1');
end;

procedure Check_FArray1D.VerifySetGetValue;
var
  i: integer;
  b: boolean;
  n: TFloat;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1);
  i := RndNum(l1,h1);
  n := RndFloat(-5.5,11.0);
  V1[i] :=n;
  b := abs(V1[i] - n) < eps;
  Check( b, 'Fail SetValue');
end;

procedure Check_FArray1D.VerifyCreate;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1,A, atGeneric);
  RandArr1D(V1,-2.0,2.0);
  V2 := TFArray1D.Create(l1,_,A);
  SetLength(A,0);
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail VerifyCreate:'+msg);
end;

procedure Check_FArray1D.VerifyCreateGeneric;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1,atGeneric);
  RandArr1D(V1,-2.0,2.0);
  V2 := TFArray1D.Create(l1,h1,atGeneric);
  V2.Assign(V1);
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateGeneric:'+msg);
end;

procedure Check_FArray1D.VerifyCreateNatural;
var
  i: TInt;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1,atNatural);
  RandArr1D(V1,-2.0,2.0);
  V2 := TFArray1D.Create(l1,h1,atNatural);
  for i := V1.Lo1 to V1.Hi1 do
   V2[i] := V1[i];
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateNatural:'+msg);
  Check((1=V1.Lo1) and (V1.Hi1=d1),'Fail (0=V1.Lo1) and (V.Hi1=d1):');
end;

procedure Check_FArray1D.VerifyCreateCentered;
var
  i: integer;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1,atCentered);
  RandArr1D(V1,-2.0,2.0);
  V2 := TFArray1D.Create(l1,h1,atCentered);
  for i := V1.Lo1 to V1.Hi1 do
   V2[i] := V1[i];
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateCentered1:'+msg);
  Check( (-((d1-1) div 2)=V1.Lo1) and (V1.Hi1=(d1 div 2)),'Fail (-((d1-1) div 2))=V1.Lo1) and (V.Hi1=(d1 div 2):');

  d1 := d1-1;
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TFArray1D.Create(l1,h1,atCentered);
  RandArr1D(V1,-2.0,2.0);
  V2 := TFArray1D.Create(l1,h1,atCentered);
  for i := V1.Lo1 to V1.Hi1 do
   V2[i] := V1[i];
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateCentered2:'+msg);
  Check( (-((d1-1) div 2)=V1.Lo1) and (V1.Hi1=(d1 div 2)),'Fail (-((d1-1) div 2))=V1.Lo1) and (V.Hi1=(d1 div 2):');
end;

{ Check_CArray1D }

procedure Check_CArray1D.setUp;
begin
  d1 := 16;
  eps := 1.0e-12;
  Randomize;
end;

procedure Check_CArray1D.tearDown;
begin
end;

{ check: FastInverse, MtxMt }
procedure Check_CArray1D.VerifyFill;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1);
  V2 := TCArray1D.Create(l1,h1);
  V1.Fill(cmplx(1.0,1.0));
  Fill(V2,cmplx(1.0,1.0));
  Check(SameArr(V1,V2,msg,eps), 'Fail Fill:'+msg);
  V1.Times(cmplx(1.0,1.0));
  CxArr(cmplx(1.0,1.0),V2);
  Check(SameArr(V1,V2,msg,eps), 'Fail Times:'+msg);
end;

{ check: Cholesky, MTtxMt, MtxMtT, MtxVt, LSolve, LTSolve,ChSolve }
procedure Check_CArray1D.VerifyCopy;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1);
  RandArr1D(V1.Re,-9.0,9.0);
  RandArr1D(V1.Im,-9.0,9.0);
  V2 := TCArray1D.Create(l1,h1);
  V2.Assign(V1);
  Check(SameArr(V1,V2,msg,eps), 'Fail VerifyCopy:'+msg);
end;

procedure Check_CArray1D.VerifySwap;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1);
  i := RndNum(l1,h1);
  n := RndComplex(-9.0,9.0,-9.0,9.0);
  V1[i] := n;
  j := RndNumEx(l1,h1,i);
  m := RndComplex(-9.0,9.0,-9.0,9.0);
  V1[j] := m;
  V1.Swap(i,j);
  Check(SameComplex(V1[i],m,eps) and SameComplex(V1[j],n,eps), 'Fail Swap');
end;

procedure Check_CArray1D.VerifyGetHiLo;
begin
  l1 := RndNum(-5,5); h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1);
  Check( (V1.Hi1=V1.Lo1+d1-1), 'Fail Hi, Lo');
//  i := RndNumEx(-5,5,l1);
//  V1.Lo1 := i;
//  Check( (V1.Hi1=V1.Lo1+d1-1), 'Fail Hi, Lo #2');
end;

procedure Check_CArray1D.VerifyGetDimLo;
begin
  h1 := RndNum(2,7);
  V1 := TCArray1D.Create(h1);
  Check( (h1 = V1.Hi1), 'Fail Hi=Dim');
  Check( (1=V1.Lo1), 'Fail Lo=1');
end;

procedure Check_CArray1D.VerifySetGetValue;
var
  i: integer;
  n: complex;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1);
  i := RndNum(l1,h1);
  n := RndComplex(-5.5,11.0,-5.5,11.0);
  V1[i] := n;
  Check(Cabs(Csub(V1[i],n)) < eps, 'Fail SetValue');
end;

procedure Check_CArray1D.VerifyCreate;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1,AR,AI, atGeneric);
  RandArr1D(V1.Re,-2.0,2.0);
  RandArr1D(V1.Im,-2.0,2.0);
  V2 := TCArray1D.Create(l1,_,AR,AI);
  SetLength(AR,0);
  SetLength(AI,0);
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail VerifyCreate:'+msg);
end;

procedure Check_CArray1D.VerifyCreateGeneric;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1,atGeneric);
  RandArr1D(V1.Re,-2.0,2.0);
  RandArr1D(V1.Im,-2.0,2.0);
  V2 := TCArray1D.Create(l1,h1,atGeneric);
  V2.Assign(V1);
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateGeneric:'+msg);
end;

procedure Check_CArray1D.VerifyCreateNatural;
var
  i: TInt;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1,atNatural);
  RandArr1D(V1.Re,-2.0,2.0);
  RandArr1D(V1.Im,-2.0,2.0);
  V2 := TCArray1D.Create(l1,h1,atNatural);
  for i := V1.Lo1 to V1.Hi1 do
   V2[i] := V1[i];
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateNatural:'+msg);
  Check((1=V1.Lo1) and (V1.Hi1=d1),'Fail (0=V1.Lo1) and (V.Hi1=d1):');
end;

procedure Check_CArray1D.VerifyCreateCentered;
var
  i: integer;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArray1D.Create(l1,h1,atCentered);
  RandArr1D(V1.Re,-2.0,2.0);
  RandArr1D(V1.Im,-2.0,2.0);
  V2 := TCArray1D.Create(l1,h1,atCentered);
  for i := V1.Lo1 to V1.Hi1 do
   V2[i] := V1[i];
  Check(SameArr(V1, V2, msg, 8*MachEps),'Fail CreateCentered:'+msg);
  Check( (-((d1-1) div 2)=V1.Lo1) and (V1.Hi1=(d1 div 2)),'Fail (-((d1-1) div 2))=V1.Lo1) and (V.Hi1=(d1 div 2):');
end;

procedure Check_CArray1D.VerifyConjugate;
begin
  l1 := RndNum(-5,5);  h1 := l1+d1-1;
  V1 := TCArr1D.Create(l1,h1);
  RandArr1D(V1.Re,-9.0,9.0);
  RandArr1D(V1.Im,-9.0,9.0);
  V2 := Conjugate(V1);
  V1.Conjugate;
  Check(SameArr(V1,V2,msg,eps), 'Fail VerifyConjugate:'+msg);
end;

{ Check_Tdft1D }

procedure Check_Tdft1D.setUp;
begin
  inherited;
  eps := 1.0e-12;
  Randomize;
end;

procedure Check_Tdft1D.tearDown;
begin
  inherited;

end;

procedure Check_Tdft1D.Verifydft1DCreate;
var
  i: TInt;
  fi, fN : TFloat;
  MatU, MatV: boolean;
  c: complex;
begin
  msg0 := 'Fail Verifydft1DCreate - ';
  N := RndNum(7,16);
  fN := N;
  A := Tdft1D.Create(N);
  Check(IsEqual(1, A.Data1D.Lo1, msg),msg0+'Data.Lo1:'+ msg);
  Check(IsEqual(N, A.Data1D.Hi1, msg),msg0+'Data.Hi1:'+ msg);
  Check(IsEqual(1, A.dft1D1.Lo1, msg),msg0+'dft1.Lo1:'+ msg);
  Check(IsEqual(N, A.dft1D1.Hi1, msg),msg0+'dft1.Hi1:'+ msg);
  Check(IsEqual(-((N-1) div 2), A.dft1DCenter.Lo1, msg),msg0+'dft1DCenter.Lo1:'+ msg);
  Check(IsEqual((N div 2), A.dft1DCenter.Hi1, msg),msg0+'dft1DCenter.Hi1:'+ msg);

  for i := 1 to N do
  begin
    fi := i-1;
    A.Data1D[i] := cmplx(fi,fN-fi-1.0);
    A.dft1D1[i] := A.Data1D[i];
  end;
  Check(SameArr(A.Data1D, A.dft1D1, msg, eps),msg0+'Data=Generic:'+ msg);

  for i := 1 to N do
    Check(IsEqual(A.dft1D1[i],A.dft1D0[i-1],msg),msg0+'dft1[i] <> dft0[i-1] at i='+IntToStr(i)+':'+msg);

  for i := 1 to N do
    Check(IsEqual(A.Data1D[i],A.Signal1D[i-1],msg),msg0+'dft1[i] <> dft0[i-1] at i='+IntToStr(i)+':'+msg);

  for i := A.dft1DCenter.Lo1 to A.dft1DCenter.Hi1 do
  begin
    fi := i;
    if i < 0 then
      c := cmplx(fN+fi,-fi-1.0)
    else
      c := cmplx(fi,(fN-fi-1.0));
    Check(IsEqual(c,A.dft1DCenter[i],msg),msg0+'Freq at i='+IntToStr(i)+':'+msg);
  end;
end;

procedure Check_Tdft1D.VerifyForwardInverse;
begin
  msg0 := 'Fail VerifyForwardInverse - Data';
  N := RndNum(7,16);
  A := Tdft1D.Create(N);
  C1 := TCArray1D.Create(1,N);
  RandArr1D(A.Data1D,-1.1,1.1);
  C1.Assign(A.Data1D);
  A.ForwardDFT;
  A.Data1D.Fill(Cmplx0);
  A.InverseDFT;
  x := N; x := x*x*x;
  Check(SameArr(A.Data1D, C1, msg, x*MachEps),msg0+msg);
end;

procedure Check_Tdft1D.VerifyForwardInvSignal;
begin
  msg0 := 'Fail VerifyForwardInverse - Signal';
  N := RndNum(7,16);
  A := Tdft1D.Create(N);
  C1 := TCArray1D.Create(0,N-1);
  RandArr1D(A.Data1D,-1.1,1.1);
  C1.Assign(A.Signal1D);
  A.ForwardDFT;
  A.Signal1D.Fill(Cmplx0);
  A.InverseDFT;
  x := N; x := x*x*x;
  Check(SameArr(A.Signal1D, C1, msg, x*MachEps),msg0+msg);
end;

procedure Check_Tdft1D.VerifyHarmonicRe;
var
  M, j, k: TInt;
begin
  msg0 := 'Fail VerifyHarmonicRe M=';
  N := RndNum(7,16);
  A := Tdft1D.Create(N);
  C1 := TCArray1D.Create(1,N,atCentered);
  for M := A.dft1DCenter.Lo1 to A.dft1DCenter.Hi1 do
  begin
    for j := 0 to N-1 do
      A.Signal1D[j] := CexpIm(2.0*Pi*M*j/N);
    A.ForwardDFT;
    for k := A.dft1DCenter.Lo1 to A.dft1DCenter.Hi1 do
      if k=M then
        Check(IsEqual(Cmplx1, A.dft1DCenter[k], msg),msg0+IntToStr(M)+' : '+ msg)
      else
        Check(IsEqual(Cmplx0, A.dft1DCenter[k], msg),msg0+IntToStr(M)+' : '+ msg);
  end;
end;

procedure Check_Tdft1D.VerifyHarmonicIm;
var
  M, j, k: TInt;
begin
  msg0 := 'Fail VerifyHarmonicIm M=';
  N := RndNum(7,16);
  A := Tdft1D.Create(N);
  C1 := TCArray1D.Create(1,N,atCentered);
  for M := A.dft1DCenter.Lo1 to A.dft1DCenter.Hi1 do
  begin
    for j := 0 to N-1 do
      A.Signal1D[j] := cMul(CexpIm(2.0*Pi*M*j/N),CmplxIm1);
    A.ForwardDFT;
    for k := A.dft1DCenter.Lo1 to A.dft1DCenter.Hi1 do
      if k=M then
        Check(IsEqual(CmplxIm1, A.dft1DCenter[k], msg),msg0+IntToStr(M)+' : '+ msg)
      else
        Check(IsEqual(Cmplx0, A.dft1DCenter[k], msg),msg0+IntToStr(M)+' : '+ msg);
  end;
end;

initialization

  RegisterTest('uLinAlg Tests',Suite);

end.
