{*
  Optimization unit tests
Nikolai V. Shokhirev, 2001-2007
@author(Nikolai Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
@created(2005.09.09)
@lastmod(2005.09.09)
}
unit Test_uArrayOptim;

interface

uses
  TestFramework,
  SysUtils,
  uMatTypes,
  uArrayOptim,
  uDynArrays;

type

Check_Extremum1 = class(TTestCase)
private
  Ap, Am: IFArr1D;
  idx1: TInt;
  value, x1, tol: TFloat;
  msg: string;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyMinArray;
   procedure VerifyMaxArray;
   procedure VerifyRefineMin;
   procedure VerifyRefineMax;
end;

Check_Extremum2 = class(TTestCase)
private
  Ap, Am: IFArr2D;
  idx1, idx2: TInt;
  value, x1, x2, tol: TFloat;
  msg: string;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyMinArray;
   procedure VerifyMaxArray;
   procedure VerifyRefineMin;
   procedure VerifyRefineMax;
end;


function Suite : ITestSuite;

implementation


function Suite : ITestSuite;
begin
  result := TTestSuite.Create('uArrayOptim Tests');

  result.addTest(Check_Extremum1.Suite);
  result.addTest(Check_Extremum2.Suite);
end;


{ Check_Extremum1 }

procedure Check_Extremum1.setUp;
var
  i1: TInt;
begin
  inherited;
  tol := 0.1e-9;
  Ap := TFArr1D.Create(9);
  Am := TFArr1D.Create(9);
  for i1 :=1 to 9 do
  begin
    x1 := i1/3.0;
    Am[i1] :=  0.5+sqr(x1-1.25);
    Ap[i1] := -Am[i1];
  end;
end;

procedure Check_Extremum1.tearDown;
begin
  inherited;

end;

procedure Check_Extremum1.VerifyMaxArray;
begin
  value := MaxArray( Ap, idx1);
  msg :=  'faild: MaxArray 1D, idx1';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MaxArray 1D, value';
  Check(abs(value+0.5069444444444) < tol, msg);
end;

procedure Check_Extremum1.VerifyMinArray;
begin
  value := MinArray( Am, idx1);
  msg :=  'faild: MinArray 1D, idx1';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MinArray 1D, value';
  Check(abs(value-0.5069444444444) < tol, msg);
end;

procedure Check_Extremum1.VerifyRefineMax;
begin
  value := MaxArray( Ap, idx1);
  RefineExtremum(Ap, idx1, value, x1);
  msg :=  'faild: RefineMax 1D, x1';
  Check(abs(x1+0.25) < tol, msg);
  msg :=  'faild: RefineMax 1D, value';
  Check(abs(value+0.5) < tol, msg);
end;

procedure Check_Extremum1.VerifyRefineMin;
begin
  value := MinArray( Am, idx1);
  RefineExtremum(Am, idx1, value, x1);
  msg :=  'faild: RefineMin 1D, x1';
  Check(abs(x1+0.25) < tol, msg);
  msg :=  'faild: RefineMin 1D, value';
  Check(abs(value-0.5) < tol, msg);
end;

{ Check_Extremum2 }

procedure Check_Extremum2.setUp;
var
  i1, i2: TInt;
begin
  inherited;
  tol := 0.1e-9;
  Ap := TFArr2D.Create(9,9);
  Am := TFArr2D.Create(9,9);
  for i1 :=1 to 9 do
  begin
    x1 := i1/3.0;
    for i2 :=1 to 9 do
    begin
      x2 := i2/3.0;
      Am[i1,i2] :=  0.5+sqr(x1-1.25)+sqr(x2-1.25);
      Ap[i1,i2] := -Am[i1,i2];
    end;
  end;
end;

procedure Check_Extremum2.tearDown;
begin
  inherited;

end;

procedure Check_Extremum2.VerifyMaxArray;
begin
  value := MaxArray( Ap, idx1, idx2);
  msg :=  'faild: MinArray 2D, idx1';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MinArray 2D, idx2';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MinArray 2D, value';
  Check(abs(value+0.5138888889) < tol, msg);
end;

procedure Check_Extremum2.VerifyMinArray;
begin
  value := MinArray( Am, idx1, idx2);
  msg :=  'faild: MinArray 2D, idx1';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MinArray 2D, idx2';
  CheckEquals(4, idx1, msg);
  msg :=  'faild: MinArray 2D, value';
  Check(abs(value-0.5138888889) < tol, msg);
end;

procedure Check_Extremum2.VerifyRefineMax;
begin
  value := MaxArray( Ap, idx1, idx2);
  RefineExtremum(Ap, idx1, idx2, value, x1, x2);
  msg :=  'faild: RefineMax 2D, x1';
  Check(abs(x1+0.25) < tol, msg);
  msg :=  'faild: RefineMax 2D, x2';
  Check(abs(x2+0.25) < tol, msg);
  msg :=  'faild: RefineMax 2D, value';
  Check(abs(value+0.5) < tol, msg);
end;

procedure Check_Extremum2.VerifyRefineMin;
begin
  value := MinArray( Am, idx1, idx2);
  RefineExtremum(Am, idx1, idx2, value, x1, x2);
  msg :=  'faild: RefineMin 2D, x1';
  Check(abs(x1+0.25) < tol, msg);
  msg :=  'faild: RefineMin 2D, x2';
  Check(abs(x2+0.25) < tol, msg);
  msg :=  'faild: RefineMin 2D, value';
  Check(abs(value-0.5) < tol, msg);
end;

initialization

  RegisterTest('uArrayOptim Tests',Suite);

end.
