{*
  Optimization unit tests
Nikolai V. Shokhirev, 2001-2007
@author(Nikolai Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
@created(2003.06.06)
@lastmod(2004.01.21)
}
unit Test_uOptimization;

interface

uses
  TestFramework,
  SysUtils,
  uMatTypes,
  uOptimization,
  uOptimUtils,
  uzeroin,
  uLevenbergMarquardt,
  uDynArrays;

type

CRACK_TOptim1D = class(TOptim1D);

Check_Bracketing1 = class(TTestCase)
private
  x0, step, xmin, xmax, tol: TFloat;
  x1, x2: TFloat;
  ErrCode: TInt;
  msg: string;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyBracketing;
   procedure VerifyLimErrCodes;
   procedure VerifyErrCodes;
   procedure VerifyBracketingStep0;
   procedure VerifyBracketingBigStep;
   procedure VerifyFmin;
   procedure VerifyZeroin;
end;

Check_Bracketing0 = class(TTestCase)
private
  step, xmax, tol: TFloat;
  x1, x2: TFloat;
  ErrCode: TInt;
  msg: string;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyBracketing;
   procedure VerifyLimErrCodes;
   procedure VerifyErrCodes;
   procedure VerifyBracketingStep0;
   procedure VerifyBracketingBigStep;
end;

Check_TOptim1D = class(TTestCase)
private
  Opt: TOptim1D;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyBracketing;
   procedure VerifyGoldenSection;
   procedure VerifyBernetMinimization;
   procedure VerifyGoldenAndBernet;
end;

Check_TOptimND = class(TTestCase)
private
//  Opt: TOptim1D;
public
   procedure setUp;  override;
   procedure tearDown; override;
published
   procedure VerifyLevenbergMarquardt;
   procedure VerifySVDFit;
   procedure VerifyLMO_RosenBrock3;
end;

function Suite : ITestSuite;

implementation

function fconst(const x: TFloat): TFloat;
begin
  result := 1.0;
end;

function f1(const x: TFloat): TFloat;
begin
  result := -1.0/x;
end;

function f2(const x: TFloat): TFloat;
begin
  result := 1.0/x;
end;
// min at x = 1
function f(const x: TFloat): TFloat;
begin
  result := exp(2.0-x-x)-2.0*exp(1.0-x);
end;
// zero at x = 1
function dfdx(const x: TFloat): TFloat;
begin
  result := -2.0*(exp(2.0-x-x)-exp(1.0-x));
end;

function d2fdx2(const x: TFloat): TFloat;
begin
  result := (4.0*exp(2.0-x-x)-2.0*exp(1.0-x));
end;

function d3fdx3(const x: TFloat): TFloat;
begin
  result := -(8.0*exp(2.0-x-x)-2.0*exp(1.0-x));
end;

function f3p(const x: TFloat):TFloat;
begin
  result := exp(x);
end;

function f3m(const x: TFloat):TFloat;
begin
  result := exp(-x);
end;

function Suite : ITestSuite;
begin
  result := TTestSuite.Create('uOptimization Tests');

  result.addTest(Check_Bracketing1.Suite);
  result.addTest(Check_Bracketing0.Suite);
  result.addTest(Check_TOptim1D.Suite);
  result.addTest(Check_TOptimND.Suite);
end;

{ Check_TOptim1D }

procedure Check_TOptim1D.setUp;
begin
  Opt := TOptim1D.Create;
end;

procedure Check_TOptim1D.tearDown;
begin
  Opt.Free;
end;

procedure Check_TOptim1D.VerifyBracketing;
var
  x1, x2, x3: TFloat;
begin
  Opt.Bracketing(f,0.0,0.321);
  x1 := Opt.Xmin;
  x2 := Opt.OptimalX;
  x3 := Opt.Xmax;
  Check(Opt.Bounded , 'Bracketing not bounded');
  Check((f(x1)>f(x2)) and (f(x3)>f(x2)), 'Forward Bracketing faild');

  Opt.Bracketing(f,10.0,0.321);
  x1 := Opt.Xmin;
  x2 := Opt.OptimalX;
  x3 := Opt.Xmax;
  Check(Opt.Bounded, 'Bracketing not bounded');
  Check((f(x1)>f(x2)) and (f(x3)>f(x2)), 'Backward Bracketing faild');

  Opt.LimitLow := 0.01;
  Opt.Bracketing(f1,10.0,0.321);
  Check(not Opt.Bounded and Opt.OutOfLimits, 'NOT Bounded check');

  Opt.LimitLow := 0.01;
  Opt.LimitHigh := 100.0;
  Opt.Bracketing(f2,1.0,0.321);
  Check(not Opt.Bounded and Opt.OutOfLimits, 'OutOfLimits check');
end;

procedure Check_TOptim1D.VerifyGoldenSection;
var
  x, d, e: TFloat;
begin
  Opt.Xmin := 0.0;
  Opt.Xmax := 3.0;
  Opt.GoldenSection(f, 1.0e-9);
  x := Opt.OptimalX;
  d := 2.0*SqrtMachEps*Sqrt(abs(2.0*f(x)/d2fdx2(x)));
  Check(abs(x-1.0) < d, 'GoldenSection faild');
  e := 1.0e-4;
  Opt.Xmin := 0.0;  // resetting
  Opt.Xmax := 3.0;
  Opt.GoldenSection(f, e);
  x := Opt.OptimalX;
  Check((abs(x-1.0) < e) and (abs(x-1.0) > d), 'GoldenSection eps faild');
end;

procedure Check_TOptim1D.VerifyBernetMinimization;
var
  x, d, e: TFloat;
begin
  Opt.Xmin := 0.0;
  Opt.Xmax := 3.0;
  Opt.BernetMin(f, 1.0e-9);
  x := Opt.OptimalX;
  d := 4.0*SqrtMachEps*Sqrt(abs(f(x)/d2fdx2(x)));
  Check(abs(x-1.0) < d, 'BernetMin faild: dx='+FloatToStr(x-1.0));
  Opt.Xmin := 0.0;
  Opt.Xmax := 2.0;
  e := 1.0e-4;
  Opt.BernetMin(f, e);
  x := Opt.OptimalX;
  Check((abs(x-1.0) < e) and (abs(x-1.0) > d), 'BernetMin eps faild dx='+FloatToStr(x-1.0));
end;

procedure Check_TOptim1D.VerifyGoldenAndBernet;
var
  x, d: TFloat;
begin
  Opt.Xmin := 0.0;
  Opt.Xmax := 3.0;
  Opt.GoldenSection(f, 0.1);
  Opt.BernetMin(f, 1.0e-9);
  x := Opt.OptimalX;
  d := 2.0*SqrtMachEps*Sqrt(abs(2.0*f(x)/d2fdx2(x)));
  Check(abs(x-1.0) < d, 'Golden & Bernet faild dx='+FloatToStr(x-1.0));
end;

{ functions for ND-fit }

function FF(x: TFloat; const P: IFArr1D): TFloat;
begin
  result := P[1]+x*P[2]+x*x*P[3];
end;

procedure G(x: TFloat; const P, GP: IFArr1D);
begin
  GP[1] := 1.0;
  GP[2] := x;
  GP[3] := x*x;
end;

{ Check_TOptimND }

procedure Check_TOptimND.setUp;
begin
  inherited;

end;

procedure Check_TOptimND.tearDown;
begin
  inherited;

end;

procedure Check_TOptimND.VerifyLevenbergMarquardt;
var
  Xe, Fe, W, P0: IFArr1D;
  x, epsP, deltaP, epsG, deltaG: TFloat;
  i, Ne, Np, err: TInt;
begin
  Ne := 4;
  Np := 3;
  Xe   := TFArr1D.Create(Ne);
  Fe   := TFArr1D.Create(Ne);
  W   := TFArr1D.Create(Ne);
  P0   := TFArr1D.Create(Np);
  P0[1] := 6.0;
  P0[2] :=-3.0;
  P0[3] := 2.0;
  for i := 1 to 4 do
  begin
    x := i;
    Xe[i] := x;
    Fe[i] := FF(x,P0);
    W[i] := 1.0;
  end;
  P0[1] := 1.0;
  P0[2] := -1.0;
  P0[3] := 1.0;
  epsP := 0.0001;
  epsG := 0.00001;
  LMFit(FF, G, Xe, Fe, W, P0, epsP, epsG, deltaP, deltaG, err);
  Check(err = 0, 'LevenbergMarquardt err = '+IntToStr(err));
  if err = 0 then
    Check(deltaP < epsP, 'LevenbergMarquardt faild deltaP='+FloatToStr(deltaP));
end;

procedure Check_TOptimND.VerifySVDFit;
var
  Xe, Fe, W, P0: IFArr1D;
  x, epsP, deltaP, epsG, deltaG: TFloat;
  i, Ne, Np, err: TInt;
begin
  Ne := 4;
  Np := 3;
  Xe   := TFArr1D.Create(Ne);
  Fe   := TFArr1D.Create(Ne);
  W   := TFArr1D.Create(Ne);
  P0   := TFArr1D.Create(Np);
  P0[1] := 6.0;
  P0[2] :=-3.0;
  P0[3] := 2.0;
  for i := 1 to 4 do
  begin
    x := i;
    Xe[i] := x;
    Fe[i] := FF(x,P0);
    W[i] := 1.0;
  end;
  P0[1] := 1.0;
  P0[2] := -1.0;
  P0[3] := 1.0;
  epsP := 0.0001;
  epsG := 0.00001;
  SVDFit(FF, G, Xe, Fe, W, P0, epsP, epsG, deltaP, deltaG, err);
  Check(err = 0, 'LevenbergMarquardt err = '+IntToStr(err));
  if err = 0 then
    Check(deltaP < epsP, 'LevenbergMarquardt faild deltaP='+FloatToStr(deltaP));
end;

function RB3F(const P: IFArr1D): TFloat;
begin
  result := sqr(P[1]-1.0)+10.0*sqr(sqr(P[1])-P[2])+10.0*sqr(sqr(P[2])-P[3]);
end;

procedure RB3GH(const P, G: IFArr1D; const H: IFArr2D);
begin
  G[1] :=  2.0*(P[1]-1.0)       +40.0*(sqr(P[1])-P[2])*P[1];
  G[2] := -20.0*(sqr(P[1])-P[2])+40.0*(sqr(P[2])-P[3])*P[2];
  G[3] := -20.0*(sqr(P[2])-P[3]);
  H[1,1] :=  2.0+40.0*(3.0*sqr(P[1])-P[2]);
  H[2,2] := 20.0+40.0*(3.0*sqr(P[2])-P[3]);
  H[3,3] := 20.0;
  H[1,3] := 0.0;
  H[3,1] := 0.0;
  H[1,2] := -40.0*P[1];
  H[2,1] := -40.0*P[1];
  H[3,2] := -40.0*P[2];
  H[2,3] := -40.0*P[2];
end;

procedure Check_TOptimND.VerifyLMO_RosenBrock3;
var
  P0: IFArr1D;
  epsP, deltaP, epsG, deltaG: TFloat;
  Np, err: TInt;
begin
  Np := 3;
  P0   := TFArr1D.Create(Np);
  P0[1] := 0.1;
  P0[2] := 0.1;
  P0[3] := 0.1;
  epsP := 0.0001;
  epsG := 0.00001;
  LMOptim(RB3F, RB3GH, P0, epsP, epsG, deltaP, deltaG, err, 2.0);
  Check(err = 0, 'LMO_RosenBrock3 err = '+IntToStr(err));
  if err = 0 then
  begin
    Check((deltaP < epsP), 'LMO_RosenBrock3 faild deltaP='+FloatToStr(deltaP));
    Check((deltaG < epsG), 'LMO_RosenBrock3 faild deltaG='+FloatToStr(deltaG));
  end;
end;

{ Check_Bracketing1 }

procedure Check_Bracketing1.setUp;
begin
  inherited;

end;

procedure Check_Bracketing1.tearDown;
begin
  inherited;

end;

function f3(const x: TFloat):TFloat;
begin
  result := 1.0/(x+1.0)+1/(1.0-x);
end;

procedure Check_Bracketing1.VerifyLimErrCodes;
begin
{ if ErrCode := -5 then message := 'small interval: |xmax - xmin| < |tol|' else}
  tol  := 0.000001;
  xmin := 0.99;
  xmax := 0.9900001;
  step := 0.01;
  x0 := (xmin+xmax)/2.0;
  Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=-5), 'faild ErrCode: Expected:-5, Found:'+IntToStr(ErrCode));
{ if ErrCode :=  5 then message := 'xmin > xmax' else }
  xmin := 0.99;
  xmax := -0.99;
  Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=5), 'faild ErrCode: Expected:5, Found:'+IntToStr(ErrCode));
{ if ErrCode := -4 then message := 'x0 < xmin' else }
  xmin := -0.99;
  xmax := 0.99;
  x0 := -0.9999;
  Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=-4), 'faild ErrCode: Expected:-4, Found:'+IntToStr(ErrCode));
{ if ErrCode :=  4 then message := 'x0 > xmax' else }
  x0 :=  0.9999;
  Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=4), 'faild ErrCode: Expected:4, Found:'+IntToStr(ErrCode));
end;

procedure Check_Bracketing1.VerifyErrCodes;
begin
{ if ErrCode = -3 then message := 'flat function: |f(x)-f(x+step)| < dfmin' }
  tol  := 0.000001;
  xmin :=-0.99;
  xmax := 0.99;
  step := 0.01;
  x0 := (xmin+xmax)/2.0;
  Bracketing(fconst, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=-3), 'faild ErrCode: Expected:-3, Found:'+IntToStr(ErrCode));
{ if ErrCode = -2 then message := 'x0 = xmin and f(x0) < f(x0+|step|)' }
  x0 := xmin;
  Bracketing(f3p, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=-2), 'faild ErrCode: Expected:-2, Found:'+IntToStr(ErrCode));
{ if ErrCode =  2 then message := 'x0 = xmax and f(x0) < f(x0-|step|)' }
  x0 := xmax;
  Bracketing(f3m, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=2), 'faild ErrCode: Expected:2, Found:'+IntToStr(ErrCode));
{ if ErrCode = -1 then message := 'minimum at x = xmin' }
  x0 := xmax;
  Bracketing(f3p, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=-1), 'faild ErrCode: Expected:-1, Found:'+IntToStr(ErrCode));
{ if ErrCode =  1 then message := 'minimum at x = xmax'; }
  x0 := xmin;
  Bracketing(f3m, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
  Check((ErrCode=1), 'faild ErrCode: Expected:1, Found:'+IntToStr(ErrCode));
end;

procedure Check_Bracketing1.VerifyBracketing;
var
  i : TInt;
begin
  xmin := -0.99;
  xmax :=  0.99;
  step := 0.01;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    x0 := RndFloat(xmin,xmax);
    step := RndFloat(-0.02,0.02);
    Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', x0='+FloatToStr(x0)
             +', i='+IntToStr(i)+', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1<0.0) and (x2>0.0),msg);
  end;
end;

procedure Check_Bracketing1.VerifyBracketingStep0;
var
  i : TInt;
begin
  xmin := -0.99;
  xmax :=  0.99;
  tol := 0.000001;
  step := 0.0;
  for i := 1 to 99 do
  begin
//    step := 0.01;
    x0 := RndFloat(xmin,xmax);
    Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', x0='+FloatToStr(x0)
                        +', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1<0.0) and (x2>0.0),msg);
  end;
end;

procedure Check_Bracketing1.VerifyBracketingBigStep;
var
  i : TInt;
begin
  xmin := -0.99;
  xmax :=  0.99;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    step := RndFloat(xmin,xmax);
    x0 := RndFloat(xmin,xmax);
    Bracketing(f3, x0, tol, step, x1, x2, ErrCode, xmin, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', x0='+FloatToStr(x0)
                        +', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1<0.0) and (x2>0.0),msg);
  end;
end;

procedure Check_Bracketing1.VerifyFmin;
var
  i : TInt;
begin
  xmin := 0.0;
  xmax := 2.0;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    x1 := RndFloat(xmin,0.8);
    x2 := RndFloat(1.2,xmax);
    x0 := FMMfmin(x1, x2, f, tol);
    msg := 'faild VerifyFmin:'+', x0='+FloatToStr(x0)+', i='+IntToStr(i)+
           ', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check(abs(x0-1.0)<tol, msg);
  end;
end;

procedure Check_Bracketing1.VerifyZeroin;
var
  i : TInt;
begin
  xmin := 0.0;
  xmax := 2.0;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    x1 := RndFloat(xmin,0.8);
    x2 := RndFloat(1.2,xmax);
    x0 := zeroin(x1, x2, dfdx, tol);
    msg := 'faild VerifyZeroin:'+', x0='+FloatToStr(x0)+', i='+IntToStr(i)+
           ', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check(abs(x0-1.0)<tol, msg);
  end;
end;

{ Check_Bracketing0 }

procedure Check_Bracketing0.setUp;
begin
  inherited;

end;

procedure Check_Bracketing0.tearDown;
begin
  inherited;

end;
{
function f30(const x: TFloat):TFloat;
begin
  result := x*(x-2.0);
end;
}
procedure Check_Bracketing0.VerifyLimErrCodes;
begin
{ if ErrCode := -5 then message := 'small interval: |xmax - xmin| < |tol|' else}
  tol  := 0.000001;
  xmax := 0.0000005;
  step := 0.01;
  Bracketing0(f, tol, step, x1, x2, ErrCode, xmax);
  Check((ErrCode=-5), 'faild ErrCode: Expected:-5, Found:'+IntToStr(ErrCode));
end;

procedure Check_Bracketing0.VerifyErrCodes;
begin
{ if ErrCode = -3 then message := 'flat function: |f(x)-f(x+step)| < dfmin' }
  tol  := 0.000001;
  xmax := 2.0;
  step := 0.01;
  Bracketing0(fconst, tol, step, x1, x2, ErrCode, xmax);
  Check((ErrCode=-3), 'faild ErrCode: Expected:-3, Found:'+IntToStr(ErrCode));
{ if ErrCode = -2 then message := 'x0 = xmin and f(x0) < f(x0+|step|)' }
  Bracketing0(f3p, tol, step, x1, x2, ErrCode, xmax);
  Check((ErrCode=-2), 'faild ErrCode: Expected:-2, Found:'+IntToStr(ErrCode));
{ if ErrCode =  1 then message := 'minimum at x = xmax'; }
  Bracketing0(f3m, tol, step, x1, x2, ErrCode, xmax);
  Check((ErrCode=1), 'faild ErrCode: Expected:1, Found:'+IntToStr(ErrCode));
end;

procedure Check_Bracketing0.VerifyBracketing;
var
  i : TInt;
begin
  xmax := 2.0;
  step := 0.01;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    step := RndFloat(0.01,0.03);
    Bracketing0(f, tol, step, x1, x2, ErrCode, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', i='+IntToStr(i)+
                                  ', x1='+FloatToStr(x1)+', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1> 0.0) and (x2>x1),msg);
  end;
end;

procedure Check_Bracketing0.VerifyBracketingStep0;
var
  i : TInt;
begin
  xmax := 2.0;
  tol := 0.000001;
  step := 0.0;
  for i := 1 to 99 do
  begin
    step := RndFloat(0.01,0.03);
    Bracketing0(f, tol, step, x1, x2, ErrCode, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', x1='+FloatToStr(x1)+
                                                        ', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1> 0.0) and (x2>x1),msg);
  end;
end;

procedure Check_Bracketing0.VerifyBracketingBigStep;
var
  i : TInt;
begin
  xmax := 2.0;
  tol := 0.000001;
  for i := 1 to 99 do
  begin
    step := RndFloat(0.01,xmax);
    Bracketing0(f, tol, step, x1, x2, ErrCode, xmax);
    msg :=  'faild: ErrCode='+IntToStr(ErrCode)+', x1='+FloatToStr(x1)+
                                                        ', x2='+FloatToStr(x2);
    Check((ErrCode=0) and (x1> 0.0) and (x2>x1),msg);
  end;
end;

initialization

  RegisterTest('uOptimization Tests',Suite);

end.
