{*
 Fourier transform objects TDFT1D, TFArray1D, TCArray1D
 Nikolai V. Shokhirev, 2002-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2002.02.02 - created
 @version 2004.12.12 modified
 @version 2007.08.08 - Added DelphiCodeToDoc comments
 }
unit uDFT;

interface

uses
  uMatTypes, uDynArrays;

type

  TFactorArray = array [1..32] of TInt;

{*
<pre>
 dT = Step1,  T0 = Start1, Signal = Signal1D, Data = Data1D             <br>
 Input signal S[j] is sampled at  t[j] = T0 + dT*j, j = 0, 1, .. , N-1  <br>
 Signal =   (S[ 0], S[ 1], S[ 2], .. , S[N-2], S[N-1])  0-based         <br>
 Data =     (D[ 1], D[ 2], D[ 3], .. , D[N-1], D[ N ])  1-based         <br>
 So that  D[j] = S[j-1]                                                 <br>
 The output array from DFT and CFTG (Transform) is:                     <br>
 DFT1 =   (T1[ 1], T1[ 2], T1[ 3], .. , T1[N-1], T1[ N ])  1-based      <br>
 Zero-based array                                                       <br>
 DFT0 =   (T0[ 0], T0[ 1], T0[ 2], .. , T0[N-2], T0[N-1])  0-based      <br>
 So that DFT1[j] = DFT0[j-1]                                            <br>
 T0[j] corresponds to non-centered frequency                            <br>
                         Q[j] = j*2*Pi/(dT*N)   j = 0, 1, 2, .. , N-1   <br>
 Centered frequency:     F[j] = j*2*Pi/(dT*N)   Mmin <= j <= Mmax       <br>
 where Mmin = -((N-1) div 2), Mmax = (M div 2)                          <br>
 The transform corresponding to the centered frequency                  <br>
              C[m] = T0[ m ]  0 <= m <= Mmax                            <br>
              C[m] = T0[N-m]  Mmin <= m < 0                             <br>
</pre>}
  Idft1D = interface
  ['{36134DFB-49DB-41ED-92A4-757E2B7F31A3}']
    function GetStep1: TFloat;
    procedure SetStep1(const Value: TFloat);
    function GetStart1: TFloat;
    procedure SetStart1(const Value: TFloat);
    function GetDim1: TInt;
    function GetSignal1D: ICArr1D;
    function GetData1D: ICArr1D;
    function Getdft1D0: ICArr1D;
    function Getdft1D1: ICArr1D;
    function GetCenter: ICArr1D;
    procedure ForwardDFT;
    procedure InverseDFT;
    property Step1: TFloat read GetStep1 write SetStep1;
    property Start1: TFloat read GetStart1 write SetStart1;
    property Dim1: TInt read GetDim1;
    property Signal1D: ICArr1D read GetSignal1D;
    property Data1D: ICArr1D read GetData1D;
    property dft1D0: ICArr1D read Getdft1D0;
    property dft1D1: ICArr1D read Getdft1D1;
    property dft1DCenter: ICArr1D read GetCenter;
  end;

  TFArray1D = class(TComment, IFArr1D)
  private
    fHi1: TInt;
    fLo1: TInt;
    fDim1: TInt;
    fat: TArrayType;
    fInternalValue: boolean;
    fValue: TFArr;
  protected
    function GetHi1: TInt;
    function GetLo1: TInt;
    function GetDim1: TInt;
    procedure SetLo1(const Value: TInt);
    function GetValue(i1: TInt): TFloat;// virtual;
    procedure SetValue(i1: TInt; const Value: TFloat);// virtual;
  public
    constructor Create(aLo1, aHi1: TInt; at: TArrayType=atGeneric); overload;
    constructor Create(aHi1: TInt; at: TArrayType=atGeneric); overload;
    constructor Create(aLo1, aHi1: TInt; var A: TFArr; at: TArrayType=atGeneric); overload;
    constructor Create(aHi1: TInt; var A: TFArr; at: TArrayType=atGeneric); overload;
    destructor Destroy; override;
    procedure Fill(C: TFloat);
    procedure Times(C: TFloat);
    procedure Assign(const A: IFArr1D);
    function MaxAbs: TFloat;
    function Norm(Normalize: boolean = false): TFloat;
    function Norm2(): TFloat;
    function Dot(const A: IFArr1D): TFloat;
    procedure Swap(i, j: TInt);
    property Hi1: TInt read GetHi1;
    property Lo1: TInt read GetLo1 write SetLo1;
    property Dim1: TInt read GetDim1;
    property Value[i1: TInt]: TFloat read GetValue write SetValue; default;
  end;

  TCArray1D = class(TComment, ICArr1D)
  private
    fRe: IFArr1D;
    fIm: IFArr1D;
  protected
    procedure SetRe(const Value: IFArr1D);
    procedure SetIm(const Value: IFArr1D);
    function GetRe: IFArr1D;
    function GetIm: IFArr1D;
    function GetHi1: TInt;
    function GetLo1: TInt;
    function GetDim1: TInt;
    procedure SetLo1(const Value: TInt);
    function GetValue(i1: TInt): Complex;
    procedure SetValue(i1: TInt; const Value: Complex);
  public
    constructor Create(aLo1, aHi1: TInt; at: TArrayType=atGeneric); overload;
    constructor Create(aHi1: TInt; at: TArrayType=atGeneric); overload;
    constructor Create(aLo1, aHi1: TInt; var AR, AI: TFArr; at: TArrayType=atGeneric); overload;
    constructor Create(aHi1: TInt; var AR, AI: TFArr; at: TArrayType=atGeneric); overload;
    destructor Destroy; override;
    procedure Conjugate;
    procedure Fill(C: Complex);
    procedure Times(C: Complex);
    procedure Assign(const A: ICArr1D);
    procedure Swap(i, j: TInt);
    function MaxAbs: TFloat;
    property Re: IFArr1D read GetRe write SetRe;
    property Im: IFArr1D read GetIm write SetIm;
    property Hi1: TInt read GetHi1;
    property Lo1: TInt read GetLo1 write SetLo1;
    property Dim1: TInt read GetDim1;
    property Value[i1: TInt]: Complex read GetValue write SetValue; default;
  end;

  Tdft1D = class(TInterfacedObject, Idft1D)
  private
    fStep1: TFloat;
    fStart1: TFloat;
    ADRe: TFArr;
    ADIm: TFArr;
    ATRe: TFArr;
    ATIm: TFArr;
    fCScr: ICArr1D;
    fSignal1D: ICArr1D;
    fData1D: ICArr1D;
    fDFStart1: ICArr1D;
    fDFT1: ICArr1D;
    fCenter: ICArr1D;
//    procedure CFTG (const Cinp, Cout, CScr: ICArr1D; NC: TInt);
  protected
    function GetStep1: TFloat;
    procedure SetStep1(const Value: TFloat);
    function GetStart1: TFloat;
    procedure SetStart1(const Value: TFloat);
    function GetDim1: TInt;
    function GetSignal1D: ICArr1D;
    function GetData1D: ICArr1D;
    function Getdft1D0: ICArr1D;
    function Getdft1D1: ICArr1D;
    function GetCenter: ICArr1D;
  public
    constructor Create(N: TInt);
    destructor Destroy; override;
    procedure ForwardDFT;
    procedure InverseDFT;
    property Signal1D: ICArr1D read GetSignal1D;
    property Data1D: ICArr1D read GetData1D;
    property Step1: TFloat read GetStep1 write SetStep1;
    property Start1: TFloat read GetStart1 write SetStart1;
    property Dim1: TInt read GetDim1;
    property dft1D0: ICArr1D read Getdft1D0;
    property dft1D1: ICArr1D read Getdft1D1;
    property dft1DCenter: ICArr1D read GetCenter;
  end;

implementation

uses
  SysUtils, uDFTUtils;

ResourceString
  RS_IndexError = '%s index = %d : out of range [%d, %d]';
  RS_LimMismatch = 'Lim Mismatch';
  RS_ZeroVector = 'Cannot Normalize: Zero Vector';
  RS_IsReadOnly = '%s is read-only in TFArray1D';

{ TFArray1D }

constructor TFArray1D.Create(aLo1, aHi1: TInt; at: TArrayType=atGeneric);
begin
  inherited Create;
  fInternalValue := true;
  fDim1 := aHi1 - aLo1 + 1;
  SetLength(fValue, fDim1);
  fat := at;
  case fat of
    atGeneric: fLo1 := aLo1;
    atNatural: fLo1 := 1;
    atCentered: fLo1 := -((fDim1-1) div 2);
  else  // Generic
    fLo1 := aLo1;
  end;
  fHi1 := fLo1 + fDim1 - 1;
end;

constructor TFArray1D.Create(aHi1: TInt; at: TArrayType);
begin
  Create(1, aHi1, at);
end;

constructor TFArray1D.Create(aLo1, aHi1: TInt; var A: TFArr; at: TArrayType);
begin
  inherited Create;
  fInternalValue := false;
  fat := at;
  if aHi1 = _ then
  begin
    fDim1 := Length(A);
  end else
  begin
    fDim1 := aHi1 - aLo1 + 1;
    SetLength(A, fDim1);
  end;
  fValue := A;
  case fat of
    atNatural: fLo1 := 1;
    atZeroBased: fLo1 := 0;
    atCentered: fLo1 := -((fDim1-1) div 2);
  else // Generic
    fLo1 := aLo1;
  end;
  fHi1 := fLo1 + fDim1 - 1;
end;

constructor TFArray1D.Create(aHi1: TInt; var A: TFArr; at: TArrayType);
begin
  Create(1, aHi1, A, at);
end;

destructor TFArray1D.Destroy;
begin
  if Assigned(fValue) then
    fValue := nil;
  inherited;
end;

function TFArray1D.GetDim1: TInt;
begin
  result := fHi1 - fLo1 + 1;
end;

function TFArray1D.GetHi1: TInt;
begin
  result := fHi1;
end;

function TFArray1D.GetLo1: TInt;
begin
  result := fLo1;
end;

procedure TFArray1D.SetLo1(const Value: TInt);
begin
  Raise Exception.CreateFMT(RS_IsReadOnly,['Lo1']);
end;

function TFArray1D.GetValue(i1: TInt): TFloat;
begin
  If (i1 < fLo1) or (i1 > fHi1) then
     Raise ERangeError.CreateFmt(RS_IndexError,['',i1, fLo1, fHi1]);
  if fat <> atCentered then
    result := fValue[i1-fLo1]
  else
    if i1 < 0 then
      result := fValue[Dim1 + i1]
    else
      result := fValue[i1];
end;

procedure TFArray1D.SetValue(i1: TInt; const Value: TFloat);
begin
  If (i1 < fLo1) or (i1 > fHi1) then
     Raise ERangeError.CreateFmt(RS_IndexError,['',i1, fLo1, fHi1]);
  if fat <> atCentered then
    fValue[i1-fLo1] := Value
  else
    if i1 < 0 then
      fValue[Dim1 + i1] := Value
    else
      fValue[i1] := Value;
end;

procedure TFArray1D.Assign(const A: IFArr1D);
var
  i1: TInt;
begin
  if (Hi1<>A.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  for i1 := A.Lo1 to A.Hi1 do
      Value[i1] := A[i1];
end;

function TFArray1D.Dot(const A: IFArr1D): TFloat;
var
  i1: TInt;
  sum: TFloat;
begin
  if (Hi1<>A.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  sum := 0.0;
  for i1 := A.Lo1 to A.Hi1 do
    sum := sum + Value[i1]*A[i1];
  result := sum;
end;

procedure TFArray1D.Fill(C: TFloat);
var
  i1: TInt;
begin
  for i1 := Lo1 to Hi1 do
    Value[i1] := C;
end;

function TFArray1D.Norm2: TFloat;
var
  i1: TInt;
  sum: TFloat;
begin
  sum := 0.0;
  for i1 := Lo1 to Hi1 do
    sum := sum + sqr(Value[i1]);
  result := sum;
end;

function TFArray1D.Norm(Normalize: boolean): TFloat;
var
  i1: TInt;
begin
  result := sqrt(Norm2);
  if Normalize then
  begin
    if result > SqrtMinFloat then
      for i1 := Lo1 to Hi1 do
        Value[i1] := Value[i1]/result
    else
      Raise EMathError.Create(RS_ZeroVector);
  end;
end;

procedure TFArray1D.Swap(i, j: TInt);
var
  v: TFloat;
begin
  v := Value[i];
  Value[i] := Value[j];
  Value[j] := v
end;

procedure TFArray1D.Times(C: TFloat);
var
  i1: TInt;
begin
  for i1 := Lo1 to Hi1 do
    Value[i1] := Value[i1]*C;
end;

function TFArray1D.MaxAbs: TFloat;
var
  i1: TInt;
  t: TFloat;
begin
  result := 0.0;
  for i1 := Lo1 to Hi1 do
  begin
    t := abs(Value[i1]);
    if t > result then
     result := t;
  end;
end;

{ TCArray1D }

constructor TCArray1D.Create(aHi1: TInt; at: TArrayType);
begin
  Create(1, aHi1, at);
end;

constructor TCArray1D.Create(aLo1, aHi1: TInt; at: TArrayType);
begin
  inherited Create;
  fRe := TFArray1D.Create(aLo1, aHi1, at);
  fIm := TFArray1D.Create(aLo1, aHi1, at);
end;

constructor TCArray1D.Create(aHi1: TInt; var AR, AI: TFArr; at: TArrayType);
begin
  Create(1, aHi1, AR, AI, at);
end;

constructor TCArray1D.Create(aLo1, aHi1: TInt; var AR, AI: TFArr; at: TArrayType);
begin
  inherited Create;
  fRe := TFArray1D.Create(aLo1, aHi1, AR, at);
  fIm := TFArray1D.Create(aLo1, aHi1, AI, at);
end;

destructor TCArray1D.Destroy;
begin
  if Assigned(fRe) then
    fRe := nil;
  if Assigned(fIm) then
    fIm := nil;
  inherited;
end;

function TCArray1D.GetDim1: TInt;
begin
  result := fRe.Dim1;
end;

function TCArray1D.GetHi1: TInt;
begin
  result := fRe.Hi1;
end;

function TCArray1D.GetLo1: TInt;
begin
  result := fRe.Lo1;
end;

procedure TCArray1D.SetLo1(const Value: TInt);
begin
  Raise Exception.CreateFMT(RS_IsReadOnly,['Lo1']);
end;

procedure TCArray1D.Assign(const A: ICArr1D);
begin
  fRe.Assign(A.Re);
  fIm.Assign(A.Im);
end;

procedure TCArray1D.Fill(C: Complex);
begin
  fRe.Fill(C.Re);
  fIm.Fill(C.Im);
end;

function TCArray1D.GetIm: IFArr1D;
begin
  result := fIm;
end;

function TCArray1D.GetRe: IFArr1D;
begin
  result := fRe;
end;

function TCArray1D.GetValue(i1: TInt): Complex;
begin
  result.Re := fRe[i1];
  result.Im := fIm[i1];
end;

procedure TCArray1D.SetIm(const Value: IFArr1D);
begin
//  fIm := Value;
  Raise Exception.CreateFMT(RS_IsReadOnly,['Im']);
end;

procedure TCArray1D.SetRe(const Value: IFArr1D);
begin
//  fRe := Value;
  Raise Exception.CreateFMT(RS_IsReadOnly,['Re']);
end;

procedure TCArray1D.SetValue(i1: TInt; const Value: Complex);
begin
  fRe[i1] := Value.Re;
  fIm[i1] := Value.Im;
end;

procedure TCArray1D.Swap(i, j: TInt);
begin
  fRe.Swap(i, j);
  fIm.Swap(i, j);
end;

procedure TCArray1D.Times(C: Complex);
var
  i1: TInt;
  x: TFloat;
begin
  for i1 := fRe.Lo1 to fRe.Hi1 do
  begin
    x := fRe[i1]*C.Re - fIm[i1]*C.Im;
    fIm[i1] := fRe[i1]*C.Im + fIm[i1]*C.Re;
    fRe[i1] := x;
  end;
end;

procedure TCArray1D.Conjugate;
begin
  Im.Times(-1.0);
end;

function TCArray1D.MaxAbs: TFloat;
var
  i1: TInt;
  t: TFloat;
begin
  result := 0.0;
  for i1 := Lo1 to Hi1 do
  begin
    t := abs(fRe[i1])+abs(fIm[i1]);
    if t > result then
     result := t;
  end;
end;

{ Tdft1D }

constructor Tdft1D.Create(N: TInt);
begin
  fStep1 := 1.0;
  fCScr := TCArray1D.Create(1,N,atNatural);
  fData1D := TCArray1D.Create(1,N,ADRe,ADIm,atNatural);
  fSignal1D := TCArray1D.Create(0,_,ADRe,ADIm,atZeroBased);
  fDFT1 := TCArray1D.Create(1,N,ATRe,ATIm,atNatural);
  fDFStart1 := TCArray1D.Create(0,_,ATRe,ATIm,atZeroBased);
  fCenter := TCArray1D.Create(1,_,ATRe,ATIm,atCentered);
end;

destructor Tdft1D.Destroy;
begin
  ADRe := nil;
  ADIm := nil;
  ATRe := nil;
  ATIm := nil;
  inherited;
end;

function Tdft1D.GetData1D: ICArr1D;
begin
  result := fData1D;
end;

function Tdft1D.GetCenter: ICArr1D;
begin
  result := fCenter;
end;

function Tdft1D.Getdft1D1: ICArr1D;
begin
  result := fDFT1;
end;

function Tdft1D.GetStep1: TFloat;
begin
  result := fStep1;
end;

procedure Tdft1D.SetStep1(const Value: TFloat);
begin
  fStep1 := Value;
end;

function Tdft1D.GetStart1: TFloat;
begin
  result := fStart1;
end;

procedure Tdft1D.SetStart1(const Value: TFloat);
begin
  fStart1 := Value;
end;

procedure Tdft1D.ForwardDFT;
begin
  uDFTUtils.CFTG(fData1D,fDFT1,fCScr,+Dim1);
end;

procedure Tdft1D.InverseDFT;
begin
  uDFTUtils.CFTG(fDFT1,fData1D,fCScr,-Dim1);
end;
function Tdft1D.GetDim1: TInt;
begin
  result := fData1D.Dim1;
end;

function Tdft1D.Getdft1D0: ICArr1D;
begin
  result := fDFStart1;
end;

function Tdft1D.GetSignal1D: ICArr1D;
begin
  result := fSignal1D;
end;

(*
procedure Tdft1D.CFTG(const Cinp, Cout, CScr: ICArr1D; NC: TInt);
var
  a, fN, rN: TFloat;
  j, N, x, s, nf: TInt;
  r: TFactorArray;

  procedure CScale(const Inp, aOut: ICArr1D; A: TFloat);
  var
    i: TInt;
  begin
    for i := 1 to Inp.Dim1 do
      aOut[i] := cMul(Inp[i], A);
  end;

  { Factor n into its nf primes in array factors[]. }
  procedure GetFactors(n: integer; var factors: tFactorArray; var nf: integer);
  var
    nr, tf, q: TInt;
  begin
    nr := n;
    q  := nr;
    tf := 2;
    nf := 1;
    while tf < q do
    begin
      q := nr div tf;
      if q*tf=nr then
      begin
        factors[nf]:=tf;
        Inc(nf);
        nr:=q;
      end
      else
        Inc(tf);
    end{while};
    factors[nf]:=nr;
  end;

  procedure GFTInnerLoop (const Inp, aOut: ICArr1D;  NC, X, S: integer);
  var
    WDL1, WD, CTemp, CTemp2: Complex;
    L1, L, K, D, I, U, H: TInt;
    a, fD: TFloat;
  begin
    D := N div S;
    fD := D;
    I := 0;
    if NC > 0 then
      a := -2.0*Pi*fD*rN
    else
      a := 2.0*Pi*fD*rN;
    WD := cmplx(cos(a), sin(a));
    for L1:=1 to S do
    begin
      If L1 = 1 then
        WDL1 := Cmplx1
      else
        WDL1 := cMul(WDL1,WD);
      for L:=1 to D do
      begin
        K := (L+(L1-1)*D*X) mod N;
        U := K+(X-1)*D;
        CTemp := Inp[U];
        for H := 2 to X do
        begin
          CTemp := cMul(CTemp,WDL1);
          U := U - D;
          CTemp2 := Inp[U];
          CTemp := cAdd(CTemp,CTemp2);
        end;
        Inc(I);
        aOut[i] := CTemp;
      end{L-loop};
    end{L1-loop};
  end;

begin
{ Let n be the number of complex elements in the array.}
  N := abs(NC);
  If N = 0 then
    exit;
  fN := N;
  rN := 1.0/fN;
  GetFactors(N, r, nf); { Factor n into its primes.}

{ Now do the transformation.}
  s := 1;
  for j:= 1 to nf do
  begin
    x := r[j];
    s := s*x;
    If j=1 then
      GFTInnerLoop (CInp, COut, NC, x, s)
    else
      if Odd(j) then
        GFTInnerLoop (CScr, COut, NC, x, s)
      else
        GFTInnerLoop (COut, CScr, NC, x, s);
  end{j-loop};

{ Finally, check for the required scaling and that data is in out array.}
  If NC < 0 then a := 1.0 else a := rN;
  If not Odd(nf) then
    CScale({in}CScr, {out}COut, a)
  else
    if a <> 1.0 then
      CScale(COut, COut, a);
end;
*)
end.
