{@abstract(Fourier signal processing library )
@author(Nikolai V. Shokhirev <nikolai@shokhirev.com> <nikolai@u.arizona.edu>)
@created(February 02, 2002)
@lastmod(July 07, 2003)
Nikolai V. Shokhirev, 2002-2003 }
unit uDFTUtils;

interface

uses
  uMatTypes, uDynArrays, uDFT;

{<pre> if the input signal is exp(i*2*Pi*M*j/N) <br>
  sampled at j = 0, 1, .. , N-1 then            <br>
  the output array from DFT and CFTG is:        <br>
  I  |  1 |  2 |  3 | .. | N-1|  N |            <br>
  M  |  0 |  1 |  2 | .. | N-2| N-1|            <br>
  M  |+/-N|-N+1|-N+2| .. | -2 | -1 |            <br>
  K  |  0 |  1 |  2 | .. | -2 | -1 |            </pre>}

{<pre>
 returns the corresponding index for a given M from the "standard" interval:<br>
    MinFreqIndex <= M <= MaxFreqIndex </pre>}
//  if M < 0 then  result := N + M + 1  // -N < M < 0
//  else  result := M + 1;
  function IndexDFT(N, M: TInt): TInt;
{<pre> MinFreqIndex = -((N - 1) div 2) </pre>}
  function MinFreqIndex(N: TInt): TInt;
{<pre> MaxFreqIndex =  (N div 2) </pre>}
  function MaxFreqIndex(N: TInt): TInt;

{<pre> Ain input frequency array - output from GFTG >br>
  returns centered frequency array Aout <br>
  Aout[1] = MinFreq, .. , Aout[N] = MaxFreq </pre>}
  procedure CenterFreq(const Ain, Aout: IFArr1D);

{<pre> DFT - Not fast Fourier transforms -  for testing pirpose only <br>
  see R.W.Hamming, Numerical methods for scientists and engineers, <br>
  McGraw-Hill, N.Y. 1962.</pre>}

{<pre> Forward Discrete Fourier Transform of real data <br>
  It is used in combinatio with InvDFTR and InterpolDFTR<br>
  x[1..N] of TReal - input signal <br>
  result[1..(N+2) div 2] of Complex - Fourier transform <br>
  if x[k] = x(h*(k-1)) then the transform frequencies are <br>
   0, w, 2*w, 3*w, . .  , where w = 2.0*Pi/(h*N)  </pre>}
  function DFTR(const x: IFArr1D): ICArr1D;
{<pre> Inverse Discrete Fourier Transform to real data <br>
  It is used in combinatio with DFTR <br>
  If c = DFTR(x) then x = InvDFTR(c, N) <br>
  N - Number of real points ( N = x.Dim1 in DFTR) <br>
  c[1..(N+2) div 2] of Complex - Fourier transform <br>
  result[1..N] of TReal - real signal  </pre>}
  function InvDFTR(const c: ICArr1D; N: TInt): IFArr1D;
{<pre> Fourier Interpolation of real data
  It is used in combinatio with DFTR <br>
  N - Number of signal points in DFTR <br>
  h - step = distance between signal points <br>
  c[1..(N+2) div 2] of Complex - Fourier transform from DFTR </pre>}
  function InterpolDFTR(t, h: TFloat; const c: ICArr1D; N: TInt): TFloat;

{<pre> Forward Discrete Fourier Transform of complex data
  It is used in combinatio with InvDFT and InterpolDFT: <br>
  C[1..N] of Complex - input signal <br>
  result[1..N] of Complex - Fourier transform
  if x[k] = x(h*(k-1)) then the transform frequencies are <br>
  0, w, 2*w, 3*w, . .  , where w = 2.0*Pi/(h*N)  </pre>}
  function DFT(const x: ICArr1D): ICArr1D;
{<pre> Inverse Discrete Fourier Transform to complex data
  It is used in combinatio with DFT: <br>
  If c = DFT(z) then z = InvDFT(c) <br>
  c[1..N] of Complex - Fourier transform <br>
  result[1..N] of Complex - complex signal  </pre>}
  function InvDFT(const c: ICArr1D): ICArr1D;
{<pre> Fourier Interpolation of complex data  <br>
  It is used in combinatio with DFT: <br>
  h - step = distance between signal points <br>
  c[1..N] of Complex - Fourier transform from DFT </pre>}
  function InterpolDFT(t, h: TFloat; const c: ICArr1D): Complex;

{<pre> Forward DFT for even number of signal points N = 2*M
  It is used in combinatio with InvDFTSym and InterpolSym: <br>
  C[1..N] of Complex - input signal <br>
  result[1..N] of Complex - Fourier transform <br>
  if x[k] = x(h*(k-1)) then the transform frequencies are <br>
  -Pi, -Pi+w, -Pi+2*w, . .  , where w = Pi/(h*M) <br>
  They map to 1, 2, .. , N indices of the result  </pre>}
  function DFTSym(const x: ICArr1D): ICArr1D;
{<pre> Inverse DFT for even number of signal points N = 2*M
  It is used in combinatio with DFTSym: <br>
  If c = DFTSym(z) then z = InvDFTSym(c) <br>
  c[1..N] of Complex - Fourier transform <br>
  result[1..N] of Complex - complex signal  </pre>}
  function InvDFTSym(const c: ICArr1D): ICArr1D;
{<pre> Fourier Interpolation of complex data with "Symmetric" frequencies <br>
  It is used in combinatio with DFTSym: <br>
  h - step = distance between signal points <br>
  c[1..N] of Complex - Fourier transform from DFT </pre>}
  function InterpolSym(t, h: TFloat; const c: ICArr1D): Complex;

{<pre> John Herbster's implementation of Glassman FFT algorithm. <br>
  The number of points does not have to be a power of 2. <br>
  Instead, the transform will take advantage of any factoring <br>
  that can be done on the number of points.  Which means that if the number <br>
  of points is prime, the transform will not be "fast" at all. <br>
  <br>
  Returns a complex Fourier transform of the Cinp array into the Cout array. <br>
  The CScr (scratch) array may be same as the input array. The transform will <br>
  be from the time to the frequency domAin if NC is positive of the number <br>
  of complex points to be converted or from the frequency to the time domAin <br>
  if NC is the negative of the number of complex points.  The only difference <br>
  between the two is just a divide by the number of points. </pre>}
  procedure CFTG (const Cinp, Cout, CScr: ICArr1D; NC: TInt);

implementation

function IndexDFT(N, M: TInt): TInt;
begin
  if M < 0 then
    result := N + M + 1  // -N < M < 0
  else
    result := M + 1;
end;

function MinFreqIndex(N: TInt): TInt;
begin
  result := -((N - 1) div 2);
end;

function MaxFreqIndex(N: TInt): TInt;
begin
  result := N div 2;
end;

procedure CenterFreq(const Ain, Aout: IFArr1D);
var
  M, N, k: TInt;
begin
  N := Ain.Dim1;
  k := 0;
  for M := MinFreqIndex(N) to MaxFreqIndex(N) do
  begin
    inc(k);
    Aout[k] := Ain[IndexDFT(N, M)];
  end;
end;

function DFTR(const x: IFArr1D): ICArr1D;
var
  temp: ICArr1D;
  N2, N1, N, k, m: TInt;
  r, i, rN, h, Hk, Hkm: TFloat;
begin
  N := x.Dim1;
  rN := 1.0/N;
  N1 := (N+1) div 2;
  N2 := (N+2) div 2;
  temp := TCArr1D.Create(N2);
  h := 2.0*Pi*rN;
  Hk := 0.0;
  for k := 1 to N2 do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{     Hkm := h*(k-1)*(m-1);}
      r := r+x[m]*Cos(Hkm);
      i := i-x[m]*Sin(Hkm);
      Hkm := Hkm + Hk;
    end;
    if (k>1) and (k<=N1)then
    begin
      r := r+r;
      i := i+i;
    end;
    temp.Re[k] := r*rN;
    temp.Im[k] := i*rN;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFTR(const c: ICArr1D; N: TInt): IFArr1D;
var
  temp: IFArr1D;
  N2, k, m: TInt;
  r, h, Hm, Hkm: TFloat;
begin
  temp := TFArr1D.Create(N);
  N2 := c.Dim1;
  h := 2.0*Pi/N;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    Hkm := 0.0;
//    N2 := (N+2) div 2;
    for k := 1 to N2 do
    begin
{     Hkm := h*(m-1)*(k-1);}
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
    Hkm := Hkm + Hm;
    end;
    temp[m] := r;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolDFTR(t, h: TFloat; const c: ICArr1D; N: TInt): TFloat;
var
  N2, k: TInt;
  r, Hx, Hkx: TFloat;
begin
  N2 := c.Dim1;
  Hx := (2.0*Pi*t/h)/N;
  r := 0.0;
  Hkx := 0.0;
  for k := 1 to N2 do
  begin
{     Hkx := Hx*(k-1) }
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    Hkx := Hkx + Hx;
  end;
  result := r;
end;

function DFT(const x: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  N, k, m: TInt;
  r, i, rN, h, Hk, Hkm: TFloat;
begin
//  c := DFTR(x);
//  xx := InvDFTR(c, N);
  N := x.Dim1;
  rN := 1.0/N;
  temp := TCArr1D.Create(N);
  h := 2.0*Pi*rN;
  Hk := 0.0;
  for k := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{     Hkm := h*(k-1)*(m-1);}
      r := r+x.Re[m]*Cos(Hkm)+x.Im[m]*Sin(Hkm);
      i := i-x.Re[m]*Sin(Hkm)+x.Im[m]*Cos(Hkm);
      Hkm := Hkm + Hk;
    end;
    temp.Re[k] := r*rN;
    temp.Im[k] := i*rN;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFT(const c: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  N, k, m: TInt;
  r, i, rN, h, Hm, Hkm: TFloat;
begin
  N := c.Dim1;
  rN := 1.0/N;
  temp := TCArr1D.Create(N);
  h := 2.0*Pi*rN;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for k := 1 to N do
    begin
{     Hkm := h*(m-1)*(k-1);}
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
      i := i+c.Re[k]*Sin(Hkm)+c.Im[k]*Cos(Hkm);
    Hkm := Hkm + Hm;
    end;
    temp.Re[m] := r;
    temp.Im[m] := i;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolDFT(t, h: TFloat; const c: ICArr1D): Complex;
var
  N, k: TInt;
  r, i, Hx, Hkx: TFloat;
begin
  N := c.Dim1;
  Hx := (2.0*Pi*t/h)/N;
  r := 0.0;
  i := 0.0;
  Hkx := 0.0;
  for k := 1 to N do
  begin
{     Hkx := Hx*(k-1) }
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    i := i+c.Re[k]*Sin(Hkx)+c.Im[k]*Cos(Hkx);
    Hkx := Hkx + Hx;
  end;
  result.Re := r;
  result.Im := i;
end;

function DFTSym(const x: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  MM, N, k, m: TInt;
  r, i, rM, h, Hk, Hkm: TFloat;
begin
  N := x.Dim1;
  if odd(N) then
    exit;
  temp := TCArr1D.Create(N);
  MM := N div 2;
  rM := 1.0/MM;

  h := Pi*rM;
  Hk := -Pi;
  for k := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
    Hkm := 0.0;
    for m := 1 to N do
    begin
{      Hkm := h*(k-1-MM)*(m-1);}
      r := r+x.Re[m]*Cos(Hkm)+x.Im[m]*Sin(Hkm);
      i := i-x.Re[m]*Sin(Hkm)+x.Im[m]*Cos(Hkm);
      Hkm := Hkm + Hk;
    end;
    temp.Re[k] := r*rM*0.5;
    temp.Im[k] := i*rM*0.5;
    Hk := Hk+h;
  end;
  result := temp;
end;

function InvDFTSym(const c: ICArr1D): ICArr1D;
var
  temp: ICArr1D;
  MM, N, k, m: TInt;
  r, i, rM, h, Hm, Hkm, fk: TFloat;
begin
  N := c.Dim1;
  if odd(N) then
    exit;
  MM := N div 2;
  rM := 1.0/MM;
  temp := TCArr1D.Create(N);
  h := Pi*rM;
  Hm := 0.0;
  for m := 1 to N do
  begin
    r := 0.0;
    i := 0.0;
//    Hkm := Hm;
    fk := -MM;
    for k := 1 to N do
    begin
{     Hkm := h*(m-1)*(k-1-MM);}
      Hkm := Hm*fk;
      r := r+c.Re[k]*Cos(Hkm)-c.Im[k]*Sin(Hkm);
      i := i+c.Re[k]*Sin(Hkm)+c.Im[k]*Cos(Hkm);
      fk := fk + 1.0;
    end;
    temp.Re[m] := r;
    temp.Im[m] := i;
    Hm := Hm+h;
  end;
  result := temp;
end;

function InterpolSym(t, h: TFloat; const c: ICArr1D): Complex;
var
  N, k, M: TInt;
  r, i, Hx, Hkx, fk: TFloat;
begin
  result.Re := 0.0;
  result.Im := 0.0;
  N := c.Dim1;
  if odd(N) then
    exit;
  M := N div 2;
  r := 0.0;
  i := 0.0;
  fk := M;
  Hx := (Pi*t/h)/fk;
  for k := N downto 1 do
  begin
    fk := fk - 1.0;
    Hkx := Hx*fk;
    r := r+c.Re[k]*Cos(Hkx)-c.Im[k]*Sin(Hkx);
    i := i+c.Re[k]*Sin(Hkx)+c.Im[k]*Cos(Hkx);
  end;
  result.Re := r;
  result.Im := i;
end;

{ John Herbster's implementation of Glassman FFT ---------------------------- }

{ CScr array may be same as the Cinp array.}
procedure CFTG (const Cinp, Cout, CScr: ICArr1D; NC: TInt);
var
  a, fN, rN: TFloat;
  j, N, x, s, nf: TInt;
  r: TFactorArray;

  procedure CScale(const Inp, Aout: ICArr1D; A: TFloat);
  var
    i: TInt;
  begin
    for i := 1 to Inp.Dim1 do
      Aout[i] := cMul(Inp[i], A);
  end;

  { Factor n into its nf primes in array factors[]. }
  procedure GetFactors(n: integer; var factors: tFactorArray; var nf: integer);
  var
    nr, tf, q: TInt;
  begin
    nr := n;
    q  := nr;
    tf := 2;
    nf := 1;
    while tf < q do
    begin
      q := nr div tf;
      if q*tf=nr then
      begin
        factors[nf]:=tf;
        Inc(nf);
        nr:=q;
      end
      else
        Inc(tf);
    end{while};
    factors[nf]:=nr;
  end;

  procedure GFTInnerLoop (const Inp, Aout: ICArr1D;  NC, X, S: integer);
  var
    WDL1, WD, CTemp, CTemp2: Complex;
//    N,
    L1, L, K, D, I, U, H: TInt;
//    fN, rN,
    a, fD: TFloat;
  begin
//    N := abs(NC);
//    fN := N;
//    rN := 1.0/fN;
    D := N div S;
    fD := D;
    I := 0;
    if NC > 0 then
      a := -2.0*Pi*fD*rN
    else
      a := 2.0*Pi*fD*rN;
    WD := cmplx(cos(a), sin(a));
    for L1:=1 to S do
    begin
      If L1 = 1 then
        WDL1 := Cmplx1
      else
        WDL1 := cMul(WDL1,WD);
      for L:=1 to D do
      begin
        K := (L+(L1-1)*D*X) mod N;
        U := K+(X-1)*D;
        CTemp := Inp[U];
        for H := 2 to X do
        begin
          CTemp := cMul(CTemp,WDL1);
          U := U - D;
          CTemp2 := Inp[U];
          CTemp := cAdd(CTemp,CTemp2);
        end;
        Inc(I);
        Aout[i] := CTemp;
      end{L-loop};
    end{L1-loop};
  end;

begin
{ Let n be the number of complex elements in the array.}
  N := abs(NC);
  If N = 0 then
    exit;
  fN := N;
  rN := 1.0/fN;
  GetFactors(N, r, nf); { Factor n into its primes.}

{ Now do the transformation.}
  s := 1;
  for j:= 1 to nf do
  begin
    x := r[j];
    s := s*x;
    If j=1 then
      GFTInnerLoop (CInp, COut, NC, x, s)
    else
      if Odd(j) then
        GFTInnerLoop (CScr, COut, NC, x, s)
      else
        GFTInnerLoop (COut, CScr, NC, x, s);
  end{j-loop};

{ Finally, check for the required scaling and that data is in out array.}
  If NC < 0 then a := 1.0 else a := rN;
  If not Odd(nf) then
    CScale({in}CScr, {out}COut, a)
  else
    if a <> 1.0 then
      CScale(COut, COut, a);

End;

end.
