{@abstract(Data and signal processing utilities )
@author(Nikolai V. Shokhirev <nikolai@shokhirev.com> <nikolai@u.arizona.edu>)
@created(2002.02.02)
@lastmod(2004.12.26)
Nikolai V. Shokhirev, 2002-2004 }
{ 2007.07.07 - removed compiler warnings, extended comments, other cleanup }
unit uDynDataUtils;

interface

uses
  uMatTypes, uDynArrays, uDynData;

type
  TPoint2 = 1..2;
  TPoint3 = 1..3;
  TPoint4 = 1..4;
  TPoint5 = 1..5;

{ n-th derivative of (Base**Exponent)     }
function dnPower(const Base: TFloat; const Exponent, n: Integer): TFloat;

{ df/dx at xn ; *** does not depend on n *** }
function FD2d1(x1,x2,y1,y2: TFloat; n: TPoint2): TFloat;
{ df/dx at xn }
function FD3d1(x1,x2,x3,y1,y2,y3: TFloat; n: TPoint3): TFloat;
{ d2f/dx2 at xn  ; *** does not depend on n *** }
function FD3d2(x1,x2,x3,y1,y2,y3: TFloat; n: TPoint3): TFloat;
{ df/dx at xn}
function FD4d1(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;
{ d2f/dx2 at xn }
function FD4d2(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;
{ d3f/dx3 at xn ; *** does not depend on n *** }
function FD4d3(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;

{ 5-point 2nd degree average.
INPUT:
  y1, y2, y3, y4, y5 - function values at equi-distant points
OUTPUT:
  averaged (smoothed) value at the point n
  c1  c2  c3  c4  c5        n
( 31,  9, -3, -5,  3)/35    1
  ^
(  9, 13, 12,  6, -5)/35    2
      ^
( -3, 12, 17, 12, -3)/35    3
          ^
( -5,  6, 12, 13,  9)/35    4
              ^
(  3, -5, -3,  9, 31)/35    5
                  ^           }
function Avg5(const y1, y2, y3, y4, y5: TFloat; n: TPoint5): TFloat;

{ 5-point 2nd degree smoothed derivative.
INPUT:
  y1, y2, y3, y4, y5 - function values at equi-distant points
  h - step
OUTPUT:
  1st derivative value at the point n
  c1  c2  c3  c4  c5        n
(-54, 13, 40, 27,-26)/70    1
  ^
(-34,  3, 20, 17, -6)/70    2
      ^
( -2, -1,  0,  1,  2)/10    3
          ^
(  6,-17,-20, -3, 34)/70    4
              ^
( 26,-27,-40,-13, 54)/70    5
                  ^              }
function Avg5d1(const y1, y2, y3, y4, y5, h: TFloat; n: TPoint5): TFloat;

{ 5-point 2nd degree smoothed 2nd derivative.
INPUT:
  y1, y2, y3, y4, y5 - function values at equi-distant points
  h - step
OUTPUT:
  2nd derivative value (does not depend on n)
  c1  c2  c3  c4  c5
(  2, -1, -2, -1,  2)/7    }
function Avg5d2(const y1, y2, y3, y4, y5, h: TFloat; n: TPoint5): TFloat;

{ 2-point forward 1st derivative: (Ain[i+1]-Ain[i])/h. Ain can be equal to D1Aout. }
procedure deriv1f(const Ain: IFSignal1D; const Aout: IFArr1D);

{ 2-point backward 1st derivative: (Ain[i]-Ain[i-1])/h. Ain can be equal to D1Aout. }
procedure deriv1b(const Ain: IFSignal1D; const Aout: IFArr1D);

{ 3-point centered 1st derivative: (Ain[i+1]-Ain[i-1])/(2*h). Ain can be equal to D1Aout. }
procedure deriv1c(const Ain: IFSignal1D; const Aout: IFArr1D);

{ 3-point 2nd derivative: {Ain[i+1]-2*Ain[i]+Ain[i-1])/(h*h). Ain can be equal to D1Aout. }
procedure deriv2c(const Ain: IFSignal1D; const Aout: IFArr1D);

{ 5-point average. Ain can be equal to Aout
( 31,  9, -3, -5,  3)/35     1
  ^                                        
(  9, 13, 12,  6, -5)/35     2              
      ^                                    
( -3, 12, 17, 12, -3)/35   3...N-2
          ^                                
( -5,  6, 12, 13,  9)/35    N-1             
              ^                            
(  3, -5, -3,  9, 31)/35     N
                  ^           }
procedure Avg5Arr(const Ain, Aout: IFArr1D);

{ 5-point average derivative. Ain can be equal to Aout
(-54, 13, 40, 27,-26)/70     1
  ^                                                  
(-34,  3, 20, 17, -6)/70     2                        
       ^
( -2, -1,  0,  1,  2)/10   3...N-2
           ^
(  6,-17,-20, -3, 34)/70    N-1                       
               ^
( 26,-27,-40,-13, 54)/70     N
                  ^              }
procedure Avg5Arrd1(const Ain: IFSignal1D; const Aout: IFArr1D);

{ 5-point average the second derivative. Ain can be equal to Aout
 (  2, -1, -2, -1,  2)/7    }
procedure Avg5Arrd2(const Ain: IFSignal1D; const Aout: IFArr1D);

{ replaces a smooth function with a step-wise one }
procedure Sharpen(const Ain, Aout: IFArr1D);

{3-point derivative. Ain can be equal to Aout }
procedure DerivData1D(const Ain, Aout: IFData1D);

{3-point 2-nd derivative. Ain can be equal to Aout }
procedure Deriv2Data1D(const Ain, Aout: IFData1D);

{
PURPOSE:
  Generation of the coefficients for smoothing over M = 2*n+1 points using
  the least squares cubic (degree = 3) or quadratic (degreet = 2) fit
INPUT:
  n - smoothing half-interval (M = 2*n+1)
  order = 0 - function (Savitzky-Golay smoothing filter)
        = 1 - 1st derivative, Avg should be divided by h, space step
        = 2 - 2nd derivative, Avg should be divided by sqr(h)
OUTPUT:
  c - the matrix of smoothing coefficients
  c[0,m] - symmetric formula, used for the center of the smoothing interval
  -m, ... ,-2,-1, 0, 1, 2, ... ,m
                  ^
  c[-k,m] - offset k to the left, used for the first n points
  -m, ... ,-2,-1, 0, 1, 2, ... ,m
            ^
  c[+k,m] - offset k to the right, used for the last n points
  -m, ... ,-2,-1, 0, 1, 2, ... ,m
                     ^                         }
function GenSmoothCoeff(const n, order: TInt; degree: TInt = 3):IFArr2D;

{
PURPOSE:
  Calculation of smoothed values
NOTE:
  This function calculates the smoothed values of a function (Savitzky-Golay
  smoothing) or its derivatives depending on the generated coefficients
INPUT:
  idx - the center of the smoothing interval
  offset < 0 - offset to the left
         > 0 - offset to the right
  c -  coefficients generated in GenSmoothCoeff
OUTPUT:
  <f[(idx-offset)*h]>       if order = 0 (Savitzky-Golay smoothing filter)
  <h*f'[(idx-offset)*h]>    if order = 1
  <h*h*f"[(idx-offset)*h]>  if order = 2                          } 
function Avg(const y: IFArr1D; const idx, offset: TInt; const c:IFArr2D): TFloat;

implementation

uses
  math;

{ n-th derivative of (Base**Exponent)     }
function dnPower(const Base: TFloat; const Exponent, n: Integer): TFloat;
var
  f, t: TFloat;
  i: TInt;
begin
  Assert( n>=0, 'Negative derivative');
  f := 1.0;
  t := Exponent;
  for i := 1 to n do
  begin
    f := f*t;
    t := t - 1.0;
  end;
  result := f*IntPower(Base,Exponent-n);
end;

{ df/dx at xn ; *** does not depend on n *** }
function FD2d1(x1,x2,y1,y2: TFloat; n: TPoint2): TFloat;
begin
  result := (y2-y1)/(x2-x1);
end;

{ df/dx at xn }
function FD3d1(x1,x2,x3,y1,y2,y3: TFloat; n: TPoint3): TFloat;
var
  h21, h31, h32: TFloat;
begin
  h21 := x2-x1;
  h31 := x3-x1;
  h32 := x3-x2;
  case n of
  1: result := -y1*(1.0/h21+1.0/h31) +y2*h31/h21/h32 -y3*h21/h31/h32;
  2: result :=  y2*(1.0/h21-1.0/h32) +y3*h21/h31/h32 -y1*h32/h31/h21;
  3: result :=  y3*(1.0/h31+1.0/h32) +y1*h32/h21/h31 -y2*h31/h21/h32;
  else
    assert(false, 'Incorrect index in FD3d1');
    result := 0.0;
  end;
end;

{ d2f/dx2 at xn  ; *** does not depend on n *** }
function FD3d2(x1,x2,x3,y1,y2,y3: TFloat; n: TPoint3): TFloat;
var
  h21, h31, h32: TFloat;
begin
  h21 := x2-x1;
  h31 := x3-x1;
  h32 := x3-x2;
  result := 2.0*(y1/h21/h31 -y2/h21/h32 +y3/h31/h32);
end;

{ df/dx at xn}
function FD4d1(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;
var
  h21, h31, h41, h32, h42, h43: TFloat;
begin
  h21 := x2-x1;
  h31 := x3-x1;
  h41 := x4-x1;
  h32 := x3-x2;
  h42 := x4-x2;
  h43 := x4-x3;
  case n of
  1: result := -y1*(1.0/h21+1.0/h31+1.0/h41) +y2*h31*h41/h21/h32/h42
                     -y3*h21*h41/h31/h32/h43 +y4*h21*h31/h41/h42/h43;
  2: result :=  y2*(1.0/h21-1.0/h32-1.0/h42) -y1*h32*h42/h21/h31/h41
                     -y4*h21*h32/h41/h42/h43 +y3*h21*h42/h31/h32/h43;
  3: result :=  y3*(1.0/h31+1.0/h32-1.0/h43) +y4*h31*h32/h41/h42/h43
                     +y1*h32*h43/h21/h31/h41 -y2*h31*h43/h21/h32/h42;
  4: result :=  y4*(1.0/h41+1.0/h42+1.0/h43) -y3*h41*h42/h31/h32/h43
                     +y2*h41*h43/h21/h32/h42 -y1*h42*h43/h21/h31/h41;
  else
    assert(false, 'Incorrect index in FD4d1');
    result := 0.0;
  end;
end;

{ d2f/dx2 at xn }
function FD4d2(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;
var
  h21, h31, h41, h32, h42, h43: TFloat;
begin
  h21 := x2-x1;
  h31 := x3-x1;
  h41 := x4-x1;
  h32 := x3-x2;
  h42 := x4-x2;
  h43 := x4-x3;
  case n of
  1: result := 2.0*( y1*(1.0/h21/h31+1.0/h41/h21+1.0/h31/h41)
                    -y2*( h31+h41)/h21/h32/h42
                    +y3*( h21+h41)/h31/h32/h43 -y4*( h21+h31)/h41/h42/h43);
  2: result := 2.0*( y2*(-1.0/h21/h32-1.0/h21/h42+1.0/h32/h42)
                    +y1*( h32+h42)/h21/h31/h41
                    +y4*( h21-h32)/h41/h42/h43 +y3*(-h21+h42)/h31/h32/h43);
  3: result := 2.0*( y3*( 1.0/h31/h32-1.0/h31/h43-1.0/h32/h43)
                    +y4*( h31+h32)/h41/h42/h43
                    +y1*(-h32+h43)/h21/h31/h41 +y2*( h31-h43)/h21/h32/h42);
  4: result := 2.0*( y4*( 1.0/h41/h42+1.0/h41/h43+1.0/h42/h43)
                    -y3*( h41+h42)/h31/h32/h43
                    +y2*( h41+h43)/h21/h32/h42 -y1*( h42+h43)/h21/h31/h41);
  else
    assert(false, 'Incorrect index in FD4d2');
    result := 0.0;
  end;
end;

{ d3f/dx3 at xn ; *** does not depend on n *** }
function FD4d3(x1,x2,x3,x4,y1,y2,y3,y4: TFloat; n: TPoint4): TFloat;
var
  h21, h31, h41, h32, h42, h43: TFloat;
begin
  h21 := x2-x1;
  h31 := x3-x1;
  h41 := x4-x1;
  h32 := x3-x2;
  h42 := x4-x2;
  h43 := x4-x3;
  result := 6.0*(-y1/h21/h31/h41 +y2/h21/h32/h42 -y3/h31/h32/h43 +y4/h41/h42/h43);
end;

{ 5-point 2nd degree average.}
function Avg5(const y1,y2,y3,y4,y5: TFloat; n: TPoint5): TFloat;
begin
  case n of
  1: result := ( 31.0*y1 +9.0*y2 -3.0*y3 -5.0*y4 +3.0*y5)/35.0;
  2: result := (  9.0*y1+13.0*y2+12.0*y3 +6.0*y4 -5.0*y5)/35.0;
  3: result := ( -3.0*y1+12.0*y2+17.0*y3+12.0*y4 -3.0*y5)/35.0;
  4: result := ( -5.0*y1 +6.0*y2+12.0*y3+13.0*y4 +9.0*y5)/35.0;
  5: result := (  3.0*y1 -5.0*y2 -3.0*y3 +9.0*y4+31.0*y5)/35.0;
  else
    assert(false, 'Incorrect index in Avg5');
    result := 0.0;
  end;
end;

{ 5-point 2nd degree smoothed derivative.}
function Avg5d1(const y1, y2, y3, y4, y5, h: TFloat; n: TPoint5): TFloat;
begin
  case n of
  1: result := (-54.0*y1+13.0*y2+40.0*y3+27.0*y4-26.0*y5)/70.0/h;
  2: result := (-34.0*y1 +3.0*y2+20.0*y3+17.0*y4 -6.0*y5)/70.0/h;
  3: result := ( -2.0*y1     -y2            + y4 +2.0*y5)/10.0/h;
  4: result := (  6.0*y1-17.0*y2-20.0*y3 -3.0*y4+34.0*y5)/70.0/h;
  5: result := ( 26.0*y1-27.0*y2-40.0*y3-13.0*y4+54.0*y5)/70.0/h;
  else
    assert(false, 'Incorrect index in Avg5d1');
    result := 0.0;
  end;
end;

{ 5-point 2nd degree smoothed 2nd derivative.}
function Avg5d2(const y1, y2, y3, y4, y5, h: TFloat; n: TPoint5): TFloat;
begin
  result := (2.0*y1 -y2 -2.0*y3 -y4 +2.0*y5)/7.0/h/h;
end;

{ 2-point forward 1st derivative: (Ain[i+1]-Ain[i])/h. Ain can be equal to D1Aout. }
procedure deriv1f(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  for i := Lo to Hi-1 do
    Aout[i] := (Ain[i+1]-Ain[i])/Ain.Step1;
  Aout[Hi] := Aout[Hi-1];
end;

{ 2-point backward 1st derivative: (Ain[i]-Ain[i-1])/h. Ain can be equal to D1Aout. }
procedure deriv1b(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  for i := Hi downto Lo+1 do
    Aout[i] := (Ain[i]-Ain[i-1])/Ain.Step1;
  Aout[Lo] := Aout[Lo+1];
end;

{ 3-point centered 1st derivative: (Ain[i+1]-Ain[i-1])/(2*h). Ain can be equal to D1Aout. }
procedure deriv1c(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
  dp, dm, h, yp, y0, ym: TFloat;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  h := Ain.Step1;
  yp := Ain[Lo+1];
  y0 := Ain[Lo];
  dp := 0.5*(3.0*Ain[Hi]-4.0*Ain[Hi-1]+Ain[Hi-2])/h;
  dm := 0.5*(-Ain[Lo+2]+4.0*Ain[Lo+1]-3.0*Ain[Lo])/h;
  for i := Lo+1 to Hi-1do
  begin
    ym := y0;
    y0 := yp;
    yp := Ain[i+1];
    Aout[i] := 0.5*(yp - ym)/h;
  end;
  Aout[Hi] := dp;
  Aout[Lo] := dm;
end;

{ 3-point 2nd derivative: {Ain[i+1]-2*Ain[i]+Ain[i-1])/(h*h). Ain can be equal to D1Aout. }
procedure deriv2c(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
  h, yp, y0, ym: TFloat;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  h := Ain.Step1;
  yp := Ain[Lo+1];
  y0 := Ain[Lo];
  for i := Lo+1 to Hi-1 do
  begin
    ym := y0;
    y0 := yp;
    yp := Ain[i+1];
    Aout[i] := (yp - y0 + ym - y0)/h/h;
  end;
  Aout[Hi] := Aout[Hi-1];
  Aout[Lo] := Aout[Lo+1];
end;

{ 5-point average. Ain can be equal to Aout }
procedure Avg5Arr(const Ain, Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
  am2, am1, a0, ap1, ap2: TFloat;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;

  am2 := 0.0;
  am1 := Ain[Lo];
  a0 := Ain[Lo+1];
  ap1 := Ain[Lo+2];
  ap2 := Ain[Lo+3];

  Aout[Lo] := (31.0*am1+9.0*a0-3.0*ap1-5.0*ap2+3.0*Ain[Lo+4])/35.0;
  Aout[Lo+1] := (9.0*am1+13.0*a0+12.0*ap1+6.0*ap2-5.0*Ain[Lo+4])/35.0;

  for i := Lo+2 to Hi-2 do
  begin
    am2 := am1;  am1 := a0;  a0 := ap1; ap1 := ap2; ap2 := Ain[i+2];
    Aout[i] := (-3.0*am2+12.0*am1+17.0*a0+12.0*ap1-3.0*ap2)/35.0;
  end;

  Aout[Hi-1] := (-5.0*am2+6.0*am1+12.0*a0+13.0*ap1+9.0*ap2)/35.0;
  Aout[Hi]   := (3.0*am2-5.0*am1-3.0*a0+9.0*ap1+31.0*ap2)/35.0;
end;

{ 5-point average derivative. Ain can be equal to Aout }
procedure Avg5Arrd1(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
  am2, am1, a0, ap1, ap2, h: TFloat;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  h := Ain.Step1;

  am2 := 0.0;
  am1 := Ain[Lo];
  a0 := Ain[Lo+1];
  ap1 := Ain[Lo+2];
  ap2 := Ain[Lo+3];

  Aout[Lo] := (-54.0*am1+13.0*a0+40.0*ap1+27.0*ap2-26.0*Ain[Lo+4])/70.0/h;
  Aout[Lo+1] := (-34.0*am1+3.0*a0+20.0*ap1+17.0*ap2-6.0*Ain[Lo+4])/70.0/h;

  for i := Lo+2 to Hi-2 do
  begin
    am2 := am1;  am1 := a0;  a0 := ap1; ap1 := ap2; ap2 := Ain[i+2];
    Aout[i] := (-2.0*am2 -am1 + ap1 +2.0*ap2)/10.0/h;
  end;

  Aout[Hi-1] := (6.0*am2-17.0*am1-20.0*a0-3.0*ap1+34.0*ap2)/70.0/h;
  Aout[Hi]   := (26.0*am2-27.0*am1-40.0*a0-13.0*ap1+54.0*ap2)/70.0/h;
end;

{ 5-point average 2nd derivative. Ain can be equal to Aout }
procedure Avg5Arrd2(const Ain: IFSignal1D; const Aout: IFArr1D);
var
  i, Lo, Hi: TInt;
  am2, am1, a0, ap1, ap2, h: TFloat;
begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  h := Ain.Step1;
  am1 := Ain[Lo];
  a0 := Ain[Lo+1];
  ap1 := Ain[Lo+2];
  ap2 := Ain[Lo+3];

  for i := Lo+2 to Hi-2 do
  begin
    am2 := am1;  am1 := a0;  a0 := ap1; ap1 := ap2; ap2 := Ain[i+2];
    Aout[i] := (2.0*am2 - am1 - 2.0*a0 - ap1 + 2.0*ap2)/7.0/h/h;
  end;

  Aout[Lo] := Aout[Lo+2];
  Aout[Lo+1] := Aout[Lo+2];
  Aout[Hi-1] := Aout[Hi-2];
  Aout[Hi] := Aout[Hi-2];
end;

{ replaces a smooth function with a step-wise one }
procedure Sharpen(const Ain, Aout: IFArr1D);
var
  i, Lo, Hi, i1, i2: TInt;
  Aextr, Amax, Amin, d2, d2old, d1, d1old: TFloat;
  wasExtremum, done: boolean;
  
  function d(i: TInt): TFloat;
  var
    j: TInt;
  begin
    j := i;
    if j < Lo+2 then j := Lo+2;
    if j > Hi-2 then j := Hi-2;
    result := (-2.0*Ain[j-2]-Ain[j-1]+Ain[j+1]+2.0*Ain[j+2]);
  end;
  
  function dd(i: TInt): TFloat;
  var
    j: TInt;
  begin
    j := i;
    if j < Lo+2 then j := Lo+2;
    if j > Hi-2 then j := Hi-2;
    result := (2.0*Ain[j-2]-Ain[j-1]-2.0*Ain[j]-Ain[j+1]+2.0*Ain[j+2]);
  end;

begin
  Lo := Ain.Lo1;
  Hi := Ain.Hi1;
  Aextr := Ain[Lo];
  d1 := d(Lo);
  d2 := dd(Lo);
  i2 := Lo-1;

  repeat
    i1 := i2+1;
    i  := i2;
    wasExtremum := false;
    Amax := -MaxFloat;
    Amin := MaxFloat;
    repeat
      inc(i);
      d1old := d1;
      d1 := d(i);
      d2old := d2;
      d2 := dd(i);
      Amax := max(Amax,Ain[i]);
      Amin := min(Amin,Ain[i]);
      if (d1old*d1 < 0.0) then
      begin
        Aextr := Ain[i];
        wasExtremum := true;
      end;
      done := d2old*d2 < 0.0;
    until done or (i=Hi);
    i2 := i;

    if not wasExtremum then
    begin
      if i=Hi then
       Aextr := Ain[i]
      else
      begin
        if dd(i-1) > 0.0 then Aextr := Amin else Aextr := Amax;
      end;
    end;
    for i := i1 to i2 do
    Aout[i] := Aextr;
  until i2 >= Hi;
end;

{3-point derivative }
procedure DerivData1D(const Ain, Aout: IFData1D);
var
  i: TInt;
  hm, hp, hh, y0, ym, yp: TFloat;
begin
  Aout.x1.Assign(Ain.x1);
  i := Ain.Lo1;
  hp := Ain.x1[i+1]-Ain.x1[i];
  yp := Ain.x2[i+1];
  y0 := Ain.x2[i];
  for i := Ain.Lo1+1 to Ain.Hi1-1 do
  begin
    hm := hp;
    hp := Ain.x1[i+1]-Ain.x1[i];
    hh := hm+hp;
    ym := y0;
    y0 := yp;
    yp := Ain.x2[i+1];
    Aout.x2[i] := yp*(1.0/hp-1.0/hh)+y0*(1.0/hm-1.0/hp)-ym*(1.0/hm-1.0/hh);
    if i=(Ain.Lo1+1) then
      Aout.x2[i-1] := -yp*(1.0/hp-1.0/hh)+y0*(1.0/hm+1.0/hp)-ym*(1.0/hm+1.0/hh);
    if i=(Ain.Hi1-1) then
      Aout.x2[i+1] := yp*(1.0/hp+1.0/hh)-y0*(1.0/hm+1.0/hp)+ym*(1.0/hm-1.0/hh);
  end;
end;

{3-point 2-nd derivative }
procedure Deriv2Data1D(const Ain, Aout: IFData1D);
var
  i: TInt;
  hm, hp, hh, y0, ym, yp: TFloat;
begin
  Aout.x1.Assign(Ain.x1);
  i := Ain.Lo1;
  hp := Ain.x1[i+1]-Ain.x1[i];
  yp := Ain.x2[i+1];
  y0 := Ain.x2[i];
  for i := Ain.Lo1+1 to Ain.Hi1-1 do
  begin
    hm := hp;
    hp := Ain.x1[i+1]-Ain.x1[i];
    hh := hm+hp;
    ym := y0;
    y0 := yp;
    yp := Ain.x2[i+1];
    Aout.x2[i] := 2.0*(yp/hp/hh - y0/hm/hp + ym/hm/hh);
  end;
  Aout.x2[Ain.Lo1] :=Aout.x2[Ain.Lo1+1];
  Aout.x2[Ain.Hi1] :=Aout.x2[Ain.Hi1-1];
end;

procedure DerivData2D(const Ain, Aout: IFData1D);
var
  i: TInt;
  hm, hp, hh, y0, ym, yp: TFloat;
begin
  Aout.x1.Assign(Ain.x1);
  i := Ain.Lo1;
  hp := Ain.x1[i+1]-Ain.x1[i];
  yp := Ain.x2[i+1];
  y0 := Ain.x2[i];
  for i := Ain.Lo1+1 to Ain.Hi1-1 do
  begin
    hm := hp;
    hp := Ain.x1[i+1]-Ain.x1[i];
    hh := hm+hp;
    ym := y0;
    y0 := yp;
    yp := Ain.x2[i+1];
    Aout.x2[i] := yp*(1.0/hp-1.0/hh)+y0*(1.0/hm-1.0/hp)-ym*(1.0/hm-1.0/hh);
    if i=(Ain.Lo1+1) then
      Aout.x2[i-1] := -yp*(1.0/hp-1.0/hh)+y0*(1.0/hm+1.0/hp)-ym*(1.0/hm+1.0/hh);
    if i=(Ain.Hi1-1) then
      Aout.x2[i+1] := yp*(1.0/hp+1.0/hh)-y0*(1.0/hm+1.0/hp)+ym*(1.0/hm-1.0/hh);
  end;
end;

function GenSmoothCoeff(const n, order: TInt; degree: TInt):IFArr2D;
var
  c: IFArr2D;
  f0, f1, f2, f3: IFArr1D;
  s0, s2, s4, s6, t, t2, t3 : TFloat;
  r33, r31, r11, r22, r20, r00: TFloat;
  k, m: TInt;
begin
  Assert(n > 0,'Invalid half-interval: n < 1');
  Assert(degree <= 2*n,'Interval too small: degree > 2*n');
  Assert((2 <= degree) and (degree <= 3),'Currently degree = 2 or 3');

  c := TFArr2D.Create(-n,n,-n,n);
  f0 := TFArr1D.Create(-n,n);
  f1 := TFArr1D.Create(-n,n);
  f2 := TFArr1D.Create(-n,n);
  f3 := TFArr1D.Create(-n,n);

  s0 := 0.0;  s2 := 0.0;  s4 := 0.0;  s6 := 0.0;
  for k := -n to n do
  begin
    t := k;
    s0 := s0 + 1.0;
    f0[k] := 1.0;
    f1[k] := t;
    t2 := sqr(t);
    s2 := s2 + t2;
    f2[k] := t2;
    t3 := t*t2;
    f3[k] := t3;
    s4 := s4 + sqr(t2);
    s6 := s6 + sqr(t3);
  end;
  if degree = 2 then
  begin
    r33 := 0.0;
    r11 := 1.0/s2;
    r31 := 0.0;
  end else
  begin
    t := s6*s2-s4*s4;
    r33 := s2/t;
    r11 := s6/t;
    r31 := -s4/t;
  end;
  t := s4*s0-s2*s2;
  r22 := s0/t;
  r00 := s4/t;
  r20 := -s2/t;

  for k := -n to n do
  begin
    t := k;
    t2 := sqr(t);
    for m := -n to n do
      case order of
      0: c[k, m] := f0[m]*(r00+r20*t2) + f1[m]*(r11+r31*t2)*t
                   +f2[m]*(r20+r22*t2) + f3[m]*(r31+r33*t2)*t;
      1: c[k, m] := f0[m]*2.0*r20*t + f1[m]*(r11+3.0*r31*t2)
                   +f2[m]*2.0*r22*t + f3[m]*(r31+3.0*r33*t2);
      2: c[k, m] := 2.0*(f0[m]*r20 + f1[m]*3.0*r31*t
                        +f2[m]*r22 + f3[m]*3.0*r33*t);
      end;
  end;
  result := c;
end;

function Avg(const y: IFArr1D; const idx, offset: TInt; const c:IFArr2D): TFloat;
var
  s: TFloat;
  i, k, n: TInt;
begin
  n := c.Hi2;
  Assert((n > 0) and (c.Lo2 = -n),'Invalid coeff matrix');
  s := 0.0;
  i := idx - n;
  for k := -n to n do
  begin
    s := s + y[i]*c[offset,k];
    inc(i);
  end;
  result := s;
end;

end.
