{*
 Linear algebra routines - Based on EBK&NVS Library for Turbo/Object Pascal
   by Nikolai V. Shokhirev and Eugene B. Krissinel
 Nikolai V. Shokhirev, 2002-2007
 @author Nikolai Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @author Eugene B. Krissinel <keb@ebi.ac.uk> http://www.fh.huji.ac.il/~krissinel/)
 @version 2002.02.02 - created
 @version 2004.12.22 - extended
 @version 2007.07.07 - added Slice as FArr2D property, extended comments, other cleanup
 @version 2007.07.11 - TriDiagSLE was generalised for an arbitrary [Lo,..,Hi] arrays
 @version 2007.07.15 - Added DelphiCodeToDoc comments
 @version 2010.01.06 - Added TriDiagMxV
 @version 2010.01.09 - Added overloaded TriDiagSLE
}
unit uDynLinAlg;

interface

uses
  uMatTypes, uDynArrays;

{*
  d, e -> M
  author Nikolai V. Shokhirev, 2004
 | d1  e2  0   0  ...              |          
 | 0   d2  e3  0  ...              |          
 |                ...              |          
 |                ...  0   dN-1 eN |          
 |                ...  0   0    dN |          
                                              
 M is square dim x dim  e.Dim1 = d.Dim1 = dim 
}
procedure BiDiagMt(const d, e: IFArr1D; const M: IFArr2D);

 {*
 a, b, c -> M
 author Nikolai V. Shokhirev, 2004
 | b1  c1  0   0  ...               |      
 | a2  b2  c2  0  ...               |      
 |                ...               |      
 |                ... aN-1 bN-1 cN-1|      
 |                ...  0   aN   bN  |      
                                           
 M is square dim x dim                     
 a.Dim1 = b.Dim1 = c.Dim1 = dim            
}
procedure TriDiagMt(const a, b, c: IFArr1D; const M: IFArr2D);

{*
PURPOSE: Solution of linear system, a*x = b - translation from FORTRAN
Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
  Computer methods for mathematical computations. Prentice-Hall, 1977.
*** Do not use if FMMDecomp has detected singularity. ***
input:                                                                     
  a[1..n,1..n] - triangularized matrix obtained from FMMDecomp.            
  b[1..n] - right hand side vector.                                        
  ipvt[1..n] - pivot vector obtained from FMMDecomp .                      
output:                                                                    
  b[1..n] - solution vector, x .                                        
Pascal version by Nikolai V. Shokhirev, 2004
}
procedure FMMSolve(const a: IFArr2D; const b: IFArr1D; const ipvt: IIArr1D);

{*
PURPOSE: Decomposes a matrix by Gaussian elimination and estimates the condition
  of the matrix. - translation from FORTRAN                                
Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.
  Computer methods for mathematical computations. Prentice-Hall, 1977.

*** Use FMMSolve to compute solutions to linear systems. ***
input:                                                                     
  a[1..n,1..n] - matrix to be triangularized.                              
output:                                                                    
  a - contains an upper triangular matrix u and a permuted version of a lower
      triangular matrix i-l so that (permutation matrix)*a = l*u .         
  cond - an estimate of the condition of a .                               
      For the linear system a*x = b, changes in a and b may cause changes cond
      times as large in x. If cond+1.0 = cond , a is singular to working   
      precision. cond is set to 1.0e+32 if exact singularity is detected.  
  ipvt - the pivot vector.                                                 
      ipvt[k] = the index of the k-th pivot row                            
      ipvt[n] = (-1)**(number of interchanges)                             
  work space..                                                             
     the vector work must be declared and included in the call. Its input  
     contents are ignored. Its output contents are usually unimportant.    
   The determinant of a can be obtained on output by                       
     det(a) = ipvt[n] * a[1,1] * a[2,2] * ... * a[n,n].                 
  Pascal version by Nikolai V. Shokhirev, 2004
}
procedure FMMDecomp(const a: IFArr2D; var cond: TFloat; const ipvt: IIArr1D);

{*
  Solution of linear system  a*x = b
  The procedure uses pascal translation of FMMDecomp and FMMSolve
  from G.E.Forsythe, M.A.Malcolm, C.B.Moler.                                
  Computer methods for mathematical computations. Prentice-Hall, 1977.      
input:                                                                      
  a[Lo1..Hi1,Lo2..Hi2] - n x n matrix                                       
  b[Lo1..Hi1] - right hand side n-vector                                    
output:                                                                     
  b[Lo2..Hi2] - solution vector, x (n vector, the program changes its limits)
algorithm:                                                                  
  FMMDecomp(a, cond, ipvt);                                                 
  FMMSolve(a, b, ipvt);                                                    
 Author: Nikolai V. Shokhirev, 2004
}
procedure DecompSolve(const a: IFArr2D; const b: IFArr1D; var cond: TFloat);

{*
  Householder (tridiagonal) reduction of a real, symmetric matrix a.
  a[Lo..Hi,Lo..Hi]                                                        
  if (vectors = true) then                                                
  a is replaced by the orthogonal transformation matrix Q                 
  d[Lo..Hi] returns the diaginal elements of the tridiagonal matrix       
  e[Lo..Hi] returns the off-giagonal elements with e[1] = 0.              
  Epsilon is elimination threshold                                      
 Pascal version by N. V. Shokhirev and E. B. Krissinel, 1991
}
procedure tred2(const a: IFArr2D; const d, e: IFArr1D;
                       vectors: boolean = true; Epsilon: TFloat = SqrtMinFloat);

{*
  QL algorithm with implicit shifts for a real tridiaginal symmetric matrix.
On input, d[Lo..Hi] is the vector of diagonal elements,                    
  e[Lo..Hi] - subdiagonal elements, e[Lo] is arbitrary.                    
On input z is identity matrix for initially tridiagonal input matrix       
      or z is output from tred2 for general real symmetric matrix          
On output, d contains eigenvalues                                          
  The k-th column z[i,k] returns the normalized eigenvectors,              
  corresponding to d[k]                                                    
  IterMax is the maximal number of transformations                         
  if (Iter <= Itermax) then err := 0 else err := Itermax                
 Pascal version by Nikolai V. Shokhirev and Eugene B. Krissinel
}
procedure tqli(const d, e: IFArr1D; const z: IFArr2D; var err: TInt;
                                  vectors: boolean = true; IterMax: TInt = 999);

{*
  Simple sort algorithm of the vector d[Lo..Hi] and corresponding columns
  of matrix v[Lo1..Hi1,Lo..Hi]                                             
  if (ascending = true) then d[Lo] <= d[Lo+1] <=, ... <= d[Hi]             
  the columns are sorted only if (vectors = true)                       
 Author: Nikolai V. Shokhirev, 1994
}
procedure SortVal0(const d: IFArr1D; const v: IFArr2D;
                            vectors: boolean = true; ascending: boolean = true);

{*
  Quick sort, Functionally the same as SortVal0
 Author: Nikolai V. Shokhirev, 2002
}
procedure SortVal(const d: IFArr1D; const v: IFArr2D;
                            vectors: boolean = true; ascending: boolean = true);

{*
  Diagonalization of the real symmetric matrix a[Lo..Hi,Lo..Hi] .
  d[Lo..Hi] returns eigenvalues                                            
  if (vectors = true) then a returns eigenvectors by columns               
  the k-th eigenvector a[i,k] corresponds to the eigenvalue d[k]           
  if (sort = true) then values (and vectors) are sorted                    
  if (ascending = true) then d[Lo] <= d[Lo+1] <=, . . . <= d[Hi]           
  Epsilon is tridiagonalization elimination threshold                      
  IterMax is the maximal number of diagonalization transformations         
  err = 0 if diagonalization is successful,                                
  err = IterMax if the number of iterations reached IterMax                
Algorithm:                                                                 
          tred2(a, d, e, vectors, Epsilon);                                
          tqli(d, e, a, err, vectors, IterMax);                            
          SortVal(d, a, vectors, ascending);                             
 Author: Nikolai V. Shokhirev, 2002
}
procedure diag(const a: IFArr2D; const d: IFArr1D; sort: boolean = true;
                     vectors: boolean = true; ascending: boolean = true;
                     Epsilon: TFloat = SqrtMinFloat; IterMax: TInt = 999);

{*
  Diagonalization of real symmetric matrix by JACOBI  method
Input:                                                                     
  A[Lo..Hi,Lo..Hi] - the matrix to be diagonalized;                        
  A[i,j] for i < j is destroyed                                            
Output:                                                                    
  T[Lo..Hi,Lo..Hi] - the matrix of eigenvectors by columns: Tn[i] = T[i,n] 
  Eigen[Lo..Hi] - the vector of eigenvalues:                               
                             Eigen[Lo] <= Eigen[Lo+1] <= ... <= Eigen[Hi]  
  err = 0  O.K.                                                            
      = IterMax if convergence has not beeb reached after IterMax iterations
 Author: Nikolai V. Shokhirev, 1994
}
procedure Jacobi(const A, T: IFArr2D; const Eigen: IFArr1D;
                                       var err: integer; first: boolean = true);

{*
  Jacoby diagonalization of Hermitian matrix
Input:                                                                     
  A + i*B - the matrix to be diagonalized ([Lo..Hi,Lo..Hi])                
Output:                                                                    
  U + i*V - the matrix of eigenvectors by columns ([Lo..Hi,Lo..Hi])        
  E[Lo..Hi] - the vector of eigenvalues: E[Lo] <= E[Lo+1] <= ... <= E[Hi]  
  err = 0  O.K.                                                            
      = IterMax if convergence has not beeb reached after IterMax iterations
      < 0 for incorrect limits                                             
  convergence criteria:                                                    
                        Norm2(OffD(A+iB))/2 < sqr(RelEps)*N*Norm2(A+iB)/2  
                    OR  NoRotation OR Iter >= IterMax                   
 author: Nikolai V. Shokhirev, 1994
}
procedure CJacobi (const A, B, U, V: IFArr2D; const E: IFArr1D;
                  var err: TInt; first: boolean = true; order: boolean = true);

{*
  1-Based variant of Jacobi diagonalization of Hermitian matrix
Input:                                                                     
  N - dimension of the matrices                                            
  A + i*V - the matrix to be diagonalized                                  
Output:                                                                    
  U + i*V - the matrix of eigenvectors by columns                          
  E - the vector of eigenvalues: E[1] <= E[2] <= ... <= E[N]               
  err = 0  O.K.                                                            
      = IterMax if convergence has not beeb reached after IterMax iterations
      < 0 incorrect dimensions                                          
  author: Nikolai V. Shokhirev, 1994
}
procedure JacobiC (const A, B, U, V: IFArr2D; const E: IFArr1D;
                  var err: TInt; first: boolean = true; order: boolean = true);

{*
  This subroutine reduces a complex hermitian matrix to a real symmetric
  tridiagonal matrix using unitary similarity transformations.             
On input:                                                                  
  ar[1..n,1..n] and ai[1..n,1..n] contain the real and imaginary parts,    
    respectively, of the complex hermitian input matrix.                   
    Only the lower triangle of the matrix need be supplied (a[i,j], i>=j). 
On output:                                                                 
  ar[1..n,1..n] and ai[1..n,1..n] contain information about the unitary    
    transformations used in the reduction in their full lower triangles.   
    Their strict upper triangles and the diagonal of ar are unaltered.     
  d[1..n] contains the diagonal elements of the the tridiagonal matrix.    
  e[1..n] contains the subdiagonal elements of the tridiagonal matrix in   
    last n-1 its positions.  e[1] is set to zero.                          
  e2[1..n]  contains the squares of the corresponding elements of e.       
    e2 may coincide with e if the squares are not needed.                  
  tau[1..2,1..n] contains further information about the transformations.   
                                                                           
  calls pythag for dsqrt(a*a + b*b) .                                      
  (see the original EISPACK comment for htridi in implementation)       
 pascal version by Nikolai V. Shokhirev 2002
}
procedure htridi(const ar, ai, tau: IFArr2D; const d, e, e2: IFArr1D);

{*
  This subroutine forms the eigenvectors of a complex Hermitian matrix
  by back transforming those of the corresponding real symmetric tridiagonal 
  matrixdetermined by  htridi.                                             
On input:                                                                  
  ar[1..n,1..n] and ai[1..n,1..n] contain information about the unitary    
    transformations used in the reduction by  htridi  in their full lower  
    triangles except for the diagonal of ar.                               
  tau[1..2,1..n] contains further information about the transformations.   
  m  is the number of eigenvectors to be back transformed.                 
  zr contains the eigenvectors to be back transformed in its first         
    m columns (from tqli or tql2).                                         
On output:                                                                 
  zr[1..n,1..n] and zi[1..n,1..n]  contain the real and imaginary parts,   
    respectively, of the transformed eigenvectors in their first m columns.
                                                                           
  Note that the last component of each returned vector is real             
  and that vector euclidean norms are preserved.                           
  ( see the original EISPACK comment for htribk in implementation ).    
 pascal variant by Nikolai V. Shokhirev 2002
}
procedure htribk(const ar,ai, tau: IFArr2D; m: TInt; const zr,zi: IFArr2D);

{*
  Simple sort algorithm of the first m components of vector d[Lo..Hi]
  and corresponding columns of matrices vr and vi ([Lo1..Hi1,Lo..Hi]).   
  if (ascending = true) then d[Lo] <= d[Lo+1] <=, . . . <= d[Hi]         
  the columns are sorted only if (vectors = true)                        
 Author: Nikolai V. Shokhirev, 2002
}
procedure SortValC(const d: IFArr1D; const vr, vi: IFArr2D; m: TInt;
                           vectors: boolean = true; ascending: boolean = true);

{*
  Diagonalization of Hermitian matrix a = ar + i*ai.
  d returns eigenvalues                                                   
  if (m > 0) then z = zr + i*zi returns m eigenvectors by columns         
  if (sort = true) then values (and vectors) are sorted                   
  if (ascending = true) then d[1] <= d[2] <=, . . . <= d[n]               
  err = 0 if diagonalization is successful,                               
  err = IterMax if the number of iterations reached IterMax               
  err < 0 for incorrect limits (see implementation)                       
  Algorithm:                                                              
            htridi(ar, ai, tau, d, e, e2);                                
            tqli(d, e, zr, err);                                          
            htribk(ar, ai, tau, m, zr,zi);                                
            SortValC(d, a, m, true, ascending);                           
  Lim checks:                                                             
    if (ar.Dim1<>ar.Dim2) then exit;                                      
    n := ar.Dim1;                                                         
    if zr.Dim1 < n then exit;                                             
    if not SameLim(ar,ai) or not SameLim(zr,zi) then exit;                
    if (d.Dim1 < n) then exit;                                            
    mm := min(ar.Dim2,zr.Dim2);                                           
    if m > 0 then                                                         
      if m > mm then m := mm;                                           
 author: Nikolai V. Shokhirev 2002
}
procedure diagc(const ar, ai, zr, zi: IFArr2D; const d: IFArr1D; m: TInt;
                var err: TInt; sort: boolean = true; ascending: boolean = true);

{*
Fast Inversion of the matrix A by the Gauss-Jordan elimination algorithm
Input:                                                                   
     A[Lo..Hi,Lo..Hi] - the matrix to be inversed.                       
Output:                                                                  
     A   -   the inversed matrix                                         
     Det - the determinant of A                                         
authors: Eugene B. Krissinel and Nikolai V. Shokhirev, 1991
}
procedure FastInverse(const A: IFArr2D; var Det: TFloat);

{*
  Singular Value Decomposition - translation from FORTRAN
  Ref: G.E.Forsythe, M.A.Malcolm, C.B.Moler.                               
       Computer methods for mathematical computations. Prentice-Hall, 1977.
                                                                           
  2004.12.21 - moved to uFMM                                               
                                                                           
procedure FMMSVD(NA,M,N: integer; const A,U,V: IFArr2D; const W,RV1: IFArr1D;
                                       MatU,MatV: boolean; var RetCode: integer);
pascal version by Eugene B. Krissinel, 1991
}

{*
  ascending SVD sorting by Eugene B. Krissinel, 1991
  W[1..N], U[1..M,1,..N], V[1..M,1,..N]                                    
  pascal version by Eugene B. Krissinel, 1991
 }
procedure OrderSVD(M, N: integer; const U, V: IFArr2D; const W: IFArr1D;
                                                           MatU, MatV: boolean);

{*
  Descending singular value sorting with sorting of U and V
  Based on: Sebastian Boung, Sorting Algorithms in Pascal                 
  http://privat.schlund.de/b/bossung/prog/delphi/sorting.html            
 Author: Nikolai V. Shokhirev, 2002
}
procedure SortSVD(const W: IFArr1D; const U, V: IFArr2D; MatU, MatV: boolean);

{*
  used in SVD, based om  FMMSVD
Ref:
  G.E.Forsythe, M.A.Malcolm, C.B.Moler.
  Computer methods for mathematical computations. Prentice-Hall, 1977.
ORIGINAL AUTHORS:
  SVD - Golub and Reinsch

*** All arrays are 1-based ***                                    
  M := max(A.Dim1,A.Dim2);                                        
  N := min(A.Dim1,A.Dim2);                                        
  if FullMat then mn := M else mn := N;                           
  if (U.Dim1 < M) or (U.Dim2 < mn) or ( V.Dim1 < N) or            
     (V.Dim2 < N) or (W.Dim1 < mn) then RetCode := -1;            
  Author: Nikolai V. Shokhirev, 2003
}
procedure SVD_(const A, U, V: IFArr2D; const W: IFArr1D; var RetCode: integer;
         MatU, MatV: boolean; FullMat: boolean = false; ItnLimit: integer = 30);

{*
  A is a rectangular D1xD2 matrix: A[L1..H1,L2..H2]
  if FullMat = true then                                                   
    U is a square D1xD1 matrix: U[L1..H1,1..D1],                           
    V is a square D2xD2 matrix: V[L2..H2,1..D2] and                        
         UT*A*V = W                                                        
    where W is a rectangular D1xD2 matrix: W[1..D1,1..D2]                  
    and only W[i,i] <> 0 i = 1.. N, N = min(D1,D2)                         
                                                                           
  if FullMat = false then                                                  
    U is a D1xN matrix: U[L1..H1,1..N], N = min(D1,D2)                     
    V is a D2xN matrix: V[L2..H2,1..N] and                                 
         UT*A*V = W                                                        
    where W is a square NxN diagonal matrix: W[1..N,1..N]                  
    the initial matrix A can be expressed as the following                 
    product: A = U*W*VT (Singular Value Decomposition)                     
N1 > N2                                                                    
  Full:   A[N1,N2] = U[N1,N1] * W[N1,N2] * VT[N2,N2]  V[N2,N2]  mn = N1    
  Small:  A[N1,N2] = U[N1,N2] * W[N2,N2] * VT[N2,N2]  V[N2,N2]  mn = N2    
N1 < N2                                                                    
  Full:   A[N1,N2] = U[N1,N1] * W[N1,N2] * VT[N2,N2]  V[N2,N2]  mn = N2    
  Small:  A[N1,N2] = U[N1,N1] * W[N1,N1] * VT[N1,N2]  V[N2,N1]  mn = N1

 Author: Nikolai V. Shokhirev, 2002.
}
procedure SVD(const A, U, V: IFArr2D; const W: IFArr1D; var RetCode: integer;
                    MatU: boolean = true; MatV: boolean = true;
                    FullMat: boolean = false; ItnLimit: integer = 30);

{*
Bi-diagonal SLE

|d1 e2 0  ...          | | x1 | | b1 |            
|0  d2 e3 ...          | | x2 | | b2 |            
|         ...          |*| .. |=| .. |            
|         ...  dN-1 eN | |xN-1| |bN-1|            
|         ...   0   dN | | xN | | bN |            
                                                  
 a[1..n], b[1..n], c[1..n], d[1..n], x[1..n]      
                                                  
*** can be used in combination with householder ***

Author: Nikolai V. Shokhirev, 2004.
}
procedure BiDiagSLE(d, e, b, x: IFArr1D);

{*
Tri-diagonal SLE

|b1 c1 0  0  ...               | | x1 | | d1 |
|a2 b2 c2 0  ...               | | x2 | | d2 |
| 0 a3 b3 c3 ...               | | x3 | | d3 |
|            ...               |*| .. |=| .. |
|            ... aN-1 bN-1 cN-1| |xN-1| |dN-1|
|            ...  0   aN   bN  | | xN | | dN |

 a[1..n], b[1..n], c[1..n], d[1..n], x[1..n]

 M(a,b,c)*x = d  => x = M^(-1)*d

Generalised for an arbitrary [Lo,..,Hi] 2007.07.11

   Author: Nikolai V. Shokhirev, 2002.
}
// perfotmamce optimization
procedure TriDiagSLE(const a, b, c, d, x, temp: IFArr1D); overload;

procedure TriDiagSLE(const a, b, c, d, x: IFArr1D); overload;

{*
Tri-diagonal Matrix-Vector multiplication

 result = M(a,b,c)*x

   Author: Nikolai V. Shokhirev, 2010.
}
function TriDiagMxV(const a, b, c, x: IFArr1D): IFArr1D;

{*
       Perturbated Cholesky Decomposition        -
 author: Eugene B. Krissinel, 1991
}
procedure CholDecomp(N: integer; const HDiag: IFArr1D; MaxOff: TFloat;
                      const L: IFArr2D; var MaxAdd: TFloat);

{*
       Cholesky Decomposition A = L*LT
Input                                              
  L = A - positive-definite symmetric matrix       
  the upper triangle of L[i,j], j > i is not used  
Output                                             
  lower triangle L[i,j], i <= j contains L         
  upper triangle L[i,j], j > i is set to 0

 author: Eugene B. Krissinel, 1991
}
procedure Cholesky(const L: IFArr2D);

{*
 Solution  of  L*Y  =  B;
 L is a lower-triangle matrix: L[i,j], i >= j

  author: Eugene B. Krissinel, 1991
}
procedure LSolve(const L: IFArr2D; const B,Y: IFArr1D);

{*
 Solution  of LT*X  =  Y L[j,i], j >= i
 LT = Transpose(L) is a upper-triangle matrix   
 L is a lower-triangle matrix: L[i,j], i >= j

 author: Eugene B. Krissinel, 1991
}
procedure LTSolve(const L: IFArr2D; const Y,X: IFArr1D);

{*
 Solution of L*LT*X = B by the Cholesky method
       L*Y = B   // LSolve                         
       LT*X = Y  // LTSolve                        
 author: Eugene B. Krissinel, 1991
}
procedure ChSolve (const L: IFArr2D; const B, X: IFArr1D);

{*
  Cholesky decomposition of the band matrix H[1..p+1,1..N].
  The decomposition will be written into  L[1..p+1,1..N].  -
author: Eugene B. Krissinel, 1991
}
procedure CholBandDec (p: integer; const H,L: IFArr2D; var err: TInt);

{*
 Cholesky L - Solution  of L*Y  =  B (for given  B)
  for the band matrix L[1..p+1,1..N].               -
 author: Eugene B. Krissinel, 1991
}
procedure LBandSolve(p: integer; const L: IFArr2D; const B,Y: IFArr1D;
                     var err: TInt);

{*
  Cholesky LT - Solution  of LT*X  =  Y  (for given  Y)
  for the  band matrix L[1..p+1,1..N].                  -
  author: Eugene B. Krissinel, 1991
}
procedure LTBandSolve(p: integer; const L: IFArr2D; const Y,X: IFArr1D;
                      var err: TInt);

{*
 Solution of the equation L*LT*S = G by the Cholesky method
  for the band matrix L[1..p+1,1..N].                       -
 author: Eugene B. Krissinel, 1991
}
procedure ChBandSolve(p: integer; const L: IFArr2D; const G,S: IFArr1D;
                      var err: TInt);

{*
  QR-decomposition of the n x n matrix M: M = Q*R
  Q - orthogonal m x n matrix so that QT*Q = 1 (n x n)                     
  R - upper triangle matrix: R[i,j] = 0 , i > j                            
Input                                                                      
  M[Lo..Hi,Lo..Hi] - matrix to be decomposed                               
  N[Lo..Hi], D[Lo..Hi]                                                     
Oitput                                                                     
  D[k] - the first non-zero component of the k-th transformation vector uk 
  M[j,k], j =k=1,..,Hi - non-zero components of the k-th transform vector  
  M[j,k], j <= k - non-zero part of the upper-right R matrix               
  N[k] = <uk|uk>/2                                                         
Algorithm                                                                  
  The program performs the Householder's transformations:                  
                                                                           
  M(k+1) = Q(k)*M(k), M(1) = M                                             
  Q =Q(n-1)* .. *Q(2)*Q(1)                                                 
                                                                           
  Q(k) = 1 - 2*|uk><uk|/<uk|uk> = 1 - |uk><uk|/(<uk|uk>/2)                 
                                                                           
  define v as                                                              
                                                                           
  v[j] = 0,         j = 1,..,k-1                                           
  v[j] = M(k)[j,k], j = k,..,n                                             
                                                                           
  uk[j] = v[j], j <> k                                                     
  uk[k] = v[k]+sign(v[k])*|v|, |v| = sqrt(<v|v>)                           
                                                                           
  Q(k) does not change the columns 1,..,k-1;                               
       makes M[j,k] = 0, j = k+1,..,m;                                     
       changes the columns k+1,..,n                                     
 authors: Eugene B. Krissinel and Nikolai V. Shokhirev, 1991
}
procedure QRDecomp(const M: IFArr2D; const N, D: IFArr1D; tol: TFloat = MinFloat);

{*
  solution of the equation Q*R*X = B
Input                                                             
  M, D and N are the output arrays from QRDecomp                  
  B - r.h.s. vector                                               
Otput                                                             
  the computed X is returned in B                                 
  M, D and N are not changed                                     
authors: Eugene B. Krissinel and Nikolai V. Shokhirev, 1991
}
procedure QRSolve(const M: IFArr2D; const N,D, B: IFArr1D);

{*
  solution of a set of linear equations by the Gauss Jordan elimination with partial pivoting
  a[Lo1..Hi1,Lo2..Hi2] = |m|b| - n x M matrix                           
  m - N x N coefficient matrix, b - N x K r.h.s. matrix                 
  m is destroyed, b returns K solutions                                 
  if the r.h.s = Identity matrix then b returns the inverse matrix of M 
  DoBalance = true - balancing so that MaxAbs(row) = 1                  
                     and check for degeneracy                           
           - usually not necessary to use this option                  
 author: Nikolai V. Shokhirev, 2002
}
procedure GaussJordan(const a: IFArr2D; DoBalance: boolean; var err: TInt);

{*
  Direct use of GaussJordan for solving m*x =b
  m[Lo1..Hi1,Lo2..Hi2], x[Lo2..Hi2], b[Lo1..Hi1], x[Lo2..Hi2],          
  x = b is OK, in this case the limits of x will be reset to [Lo2..Hi2] 
 author: Nikolai V. Shokhirev, 2002
}
procedure GJSolve(const m: IFArr2D; DoBalance: boolean; const x, b: IFArr1D; var err: TInt);

{*
  solution of a set of linear equations a*x = b
    by the Gauss Jordan elimination with partial pivoting                
  a[Lo1..Hi1,Lo2..Hi2] is a M x N matrix of the                          
    system of M linear equations of N variables                          
    if M > N then the "best" N equations are uaed                        
  b[Lo1..Hi1] is a M-vector of the r.h.s.                                
  x is a N-vector of the solution, Lims are set to [Lo2..Hi2]            
  b and x can be the same                                                
  a and b are destroyed

 authors Eugene B. Krissinel and Nikolai V. Shokhirev, 1991
}
procedure FastSolve(const a: IFArr2D; const b, x: IFArr1D);

{*
  solution of a set of linear equations by the Gauss Jordan elimination with partial pivoting
  a is a M x N matrix of the system of M linear equations of N variables 
  Use is a boolean N-vector: if Use[i] = false then the i-th variable is 
    excluded from a solution and added to the r.h.s.                     
  if M > Nvar then the "best" Nvar equations are uaed;                   
  Here Nvar is the number of selected variables (Use[i] = true)          
  b is a M-vector of the r.h.s., x is a N-vector of the solution         
  a is NOT destroyed, but b IS destroyed

Author: Nikolai V. Shokhirev, 1994
}
procedure MaskSolve(const m: IFArr2D; const x, b: IFArr1D; const Use: IBArr1D);

{*
 Gram-Schmidt Orthonormalization of vectors
input:                                                                  
  U - matrix of column-vectors:                                         
      U[Lo1..Hi1,Lo2..Hi2] = |Col[Lo2],Col[Lo2+1],...,Col[Hi2]|         
  m - number of required vectors                                        
output:                                                                 
  U - the first m columns contain orthogonal and normalized vectors     
  m - actual number of independent vectors                            
Author: Nikolai V. Shokhirev, 2001
}
procedure GramSchmidt(const U: IFArr2D; var m: TInt);

{*
  Householder reduction (m >= n) extracted from xsvd
    A   =  U  *   J   *  VT      UT  *  A  *   V   =  J         
  m x n  m x n  n x n  n x n   n x m  m x n  n x n  n x n       
                                                                
Full (square) U:                                                
    A   =  U  *   J   *  VT      UT  *  A  *   V   =  J         
  m x n  m x m  m x n  n x n   m x m  m x n  n x n  m x n       
                                                                
    |q1 e2            |                                         
    |   q2 e2         |                                         
    |      q2 e2  .   |                                         
J = |          .  . en|                                         
    |             . qn|                                         
    |-----------------|                                         
    |        0        |

Author Nikolai V. Shokhirev, 2003
}
procedure householder(const a: IFArr2D; const q, e: IFArr1D; const u, v: IFArr2D;
                      eps, tol: TFloat; fullmatrix: boolean = false);

{*
  Singular Value Decomposition  (m >= n)
  J.H.Wilkinson, C.Reinsch, Handbook for Automatic Computation    
    A   =  U  *   Q   *  VT      UT  *  A  *   V   =  Q           
  m x n  m x n  n x n  n x n   n x m  m x n  n x n  n x n         
                                                                  
Full (square) U:                                                  
    A   =  U  *   Q   *  VT      UT  *  A  *   V   =  Q           
  m x n  m x m  m x n  n x n   m x m  m x n  n x n  m x n

Author: Nikolai V. Shokhirev, 2003
}
procedure xsvd(eps, tol: TFloat; const a: IFArr2D; const q: IFArr1D;
               const u, v: IFArr2D; withu: boolean = true;
               withv: boolean = true; fullmat: boolean = false);

implementation

uses
  Math, SysUtils,
  uDynArrUtils;

ResourceString
  /// 'Matrix is not square';
  RS_NotSquare = 'Matrix is not square';
  /// 'Lim Mismatch';
  RS_LimMismatch = 'Lim Mismatch';
  /// 'Lim Mismatch';
  RS_IterMaxError = 'No convergion after %d iterations';

procedure BiDiagMt(const d, e: IFArr1D; const M: IFArr2D);
var
  i, j: TInt;
begin
  if not IsSquare(M) then
    Raise ENonSquareMatrix.Create('in BiDiagMt');
  if not SameLim(d,e) then
    Raise ELimMismatch.Create('in BiDiagMt');
  M.Fill(0.0);
  j := d.Lo1;
  for i := M.Lo1 to M.Hi1 do
  begin
    M[i,i] := d[j];
    if i < M.Hi1 then
      M[i,i+1] := e[j+1];
    inc(j)
  end;
end;

 {
 | b1  c1  0   0  ...               |      
 | a2  b2  c2  0  ...               |      
 |                ...               |      
 |                ... aN-1 bN-1 cN-1|      
 |                ...  0   aN   bN  |        }
procedure TriDiagMt(const a, b, c: IFArr1D; const M: IFArr2D);
var
  i,j: TInt;
begin
  if not IsSquare(M) then
    Raise ENonSquareMatrix.Create('in TriDiagMt');
  if not SameLim(a,b) or not SameLim(b,c) then
    Raise ELimMismatch.Create('in TriDiagMt');
  M.Fill(0.0);
  j := b.Lo1;
  for i := M.Lo1 to M.Hi1 do
  begin
    if i > M.Lo1 then
      M[i,i-1] := a[j];
    M[i,i] := b[j];
    if i < M.Hi1 then
      M[i,i+1] := c[j];
    inc(j)
  end;
end;

procedure FMMSolve(const a: IFArr2D; const b: IFArr1D; const ipvt: IIArr1D);
var
  n, kb, km1, nm1, kp1, i, k, m: TInt;
  t: TFloat;
begin
  // no Lim checks
  n := a.Dim1;
  if (n > 1) then
  begin
    // forward elimination
    nm1 := n-1;
    for k := 1 to nm1 do
    begin
      kp1 := k+1;
      m := ipvt[k];
      t := b[m];
      b[m] := b[k];
      b[k] := t;
      for i := kp1 to n do
        b[i] := b[i] + a[i,k]*t;
    end;
    // back substitution
    for kb := 1 to nm1 do
    begin
       km1 := n-kb;
       k := km1+1;
       b[k] := b[k]/a[k,k];
       t := -b[k];
       for i := 1 to km1 do
         b[i] := b[i] + a[i,k]*t;
    end;
  end;
  b[1] := b[1]/a[1,1];
end;

procedure FMMDecomp(const a: IFArr2D; var cond: TFloat; const ipvt: IIArr1D);
var
  work: IFArr1D;
  ek, t, anorm, ynorm, znorm: TFloat;
  n, nm1, i, j, k, kp1, kb, km1, m: TInt;
begin
  // no Lim checks
  n := a.Dim1;
  ipvt[n] := 1;
  if (n = 1) then
  begin// 1-by-1
    if (a[1,1] <> 0.0) then cond := 1.0 else cond := 1.0e+32;
    exit;
  end;

  work := TFArr1D.Create(n); // work space..
  //     compute 1-norm of a
  anorm := 0.0;
  for j := 1 to n do
  begin
    t := 0.0;
    for i := 1 to n do
      t := t + abs(a[i,j]);
    if (t  >  anorm) then anorm := t;
  end;

  //     gaussian elimination with partial pivoting
  nm1 := n - 1;
  for k := 1 to nm1 do
  begin
    kp1:= k+1;

    // find pivot
    m := k;
    for i := kp1 to n do
    begin
      if (abs(a[i,k]) > abs(a[m,k])) then m := i;
    end;
    ipvt[k] := m;
    if (m <> k) then ipvt[n] := -ipvt[n];
    t := a[m,k];
    a[m,k] := a[k,k];
    a[k,k] := t;
//     if (t  =  0.0) goto 35;
    // skip step if pivot is zero
    if (t <> 0.0) then
    begin
      // compute multipliers
      for i := kp1 to n do
        a[i,k] := -a[i,k]/t;
        // interchange and eliminate by columns
       for j := kp1 to n do
       begin
         t := a[m,j];
         a[m,j] := a[k,j];
         a[k,j] := t;
         if (t <> 0.0) then
           for i := kp1 to n do
              a[i,j] := a[i,j] + a[i,k]*t;
      end;
    end;
  end;
{ cond = (1-norm of a)*(an estimate of 1-norm of a-inverse)
  Estimate obtained by one step of inverse iteration for the small singular
  vector. This involves solving two systems of equations,
    (a-transpose)*y = e  and  a*z = y
  where e is a vector of +1 or -1 chosen to cause growth in y.
  estimate = (1-norm of z)/(1-norm of y) }

//     solve (a-transpose)*y = e
  for k := 1 to  n do
  begin
    t := 0.0;
    if (k > 1) then
    begin
      km1 := k-1;
      for i := 1 to  km1 do
        t := t + a[i,k]*work[i];
    end;
    ek := 1.0;
    if (t < 0.0) then ek := -1.0;
    if (a[k,k] = 0.0) then
    begin
      cond := 1.0e+32;
      exit;
    end;
    work[k] := -(ek + t)/a[k,k];
  end;

  for kb := 1 to nm1 do
  begin
    k := n - kb;
    t := 0.0;
    kp1 := k+1;
    for i := kp1 to  n do
      t := t + a[i,k]*work[i];
    work[k] := t + work[k];
    m := ipvt[k];
    if (m <> k) then
    begin
       t := work[m];
       work[m] := work[k];
       work[k] := t;
    end;
  end;

  ynorm := 0.0;
  for i := 1 to  n do
    ynorm := ynorm + abs(work[i]);

//     solve a*z = y
  FMMSolve(a, work, ipvt);

  znorm := 0.0;
  for i := 1 to  n do
    znorm := znorm + abs(work[i]);

  // estimate condition
  cond := anorm*znorm/ynorm;
  if (cond < 1.0) then cond := 1.0;

end;

procedure DecompSolve(const a: IFArr2D; const b: IFArr1D; var cond: TFloat);
var
  ipvt: IIArr1D;
  Lo1, Lo2: TInt;
begin
  if a.Dim2 <> a.Dim1 then
    Raise ENonSquareMatrix.Create('In DecompSolve');
  if (a.Lo1 <> b.Lo1) or (a.Hi1 <> b.Hi1) then
    Raise ELimMismatch.Create('In DecompSolve');
  Lo1 := a.Lo1;
  Lo2 := a.Lo2;
  a.Lo1 := 1;
  a.Lo2 := 1;
  b.Lo1 := 1;
  ipvt := TIArr1D.Create(a.Dim1);
  FMMDecomp(a, cond, ipvt);
  if cond < 1.0e32 then
  begin
    FMMSolve(a, b, ipvt);
    b.Lo1 := Lo2;
  end else
    raise ESingularMatrix.Create('In DecompSolve');
//    b.Lo1 := Lo1;
  a.Lo1 := Lo1;
  a.Lo2 := Lo2;
end;

procedure tred2(const a: IFArr2D; const d, e: IFArr1D;
                       vectors: boolean = true; Epsilon: TFloat = SqrtMinFloat);
var
  Lo, Hi, i, j, k, l: integer;
  scale, hh, h, g, f: TFloat;
begin { of tred2 }
// No Lim checks - All checks are made in diag
  Lo := a.Lo1;
  Hi := a.Hi1;
  l := 0; // remove warning

  if Hi > Lo then
  begin
    for i := Hi downto Lo+1 do
    begin
      l:= i - 1;
      h:= 0.0;
      scale := 0.0;
      if l > Lo then
      begin
        for k := Lo to l do
        begin
          scale := scale + abs(a[i,k]);
        end;{ k }
        if scale < Epsilon then{ Skip transformation }
        begin
          e[i] := a[i,l];
        end else{ transformation }
        begin{ scale > Epsilon }

          for k := Lo to l do
          begin
            a[i,k] := a[i,k]/scale;
            h:= h + sqr(a[i,k])
          end;{ k }

          f:=  a[i,l];
          g:= -sign2(sqrt(h), f);
          e[i]:=  scale*g;
          h:=  h - f*g;
          a[i,l] :=  f - g;
          f:=  0.0;
          for j := Lo to l do
          begin
            if vectors then
            begin
              a[j,i] := a[i,j]/h;
            end;
            g := 0.0;
            for k := Lo to j do
            begin
              g := g + a[j,k]*a[i,k];
            end;{ k }
            if l > j then
            begin
              for k := j + 1 to l do
              begin
                g := g + a[k,j]*a[i,k];
              end; { k }
            end; { l > j }
            e[j] := g/h;
            f := f + e[j]*a[i,j];
          end;{ j }

          hh := f/(h + h);
          for j := Lo to l do
          begin
            f := a[i,j];
            g := e[j] -hh*f;
            e[j] := g;
            for k := Lo to j do
            begin
              a[j,k] := a[j,k] - f*e[k] -g*a[i,k];
            end;{ k }
          end;{ j }

        end; { scale > Epsilon, transformation }
      end { l > Lo }
      else
      begin
        e[i] := a[i,l];
      end; { l = Lo }
      d[i] := h;
    end;{ i }
  end;{ Hi > Lo }

  d[Lo] := 0.0;
  e[Lo] := 0.0;
  for i := Lo to Hi do
  begin
    if vectors then
    begin
      l := i - 1;
      { if d[i] <> 0.0 }
      if abs(d[i]) > MinFloat then
// ??      if i > 0 then
      begin{ This block skipped when i = 1 }
        for j := Lo to l do
        begin
          g := 0.0;
          for k := Lo to l do
          begin
            g := g + a[i,k]*a[k,j];
          end;{ k }
          for k := Lo to l do
          begin
            a[k,j] := a[k,j] -g*a[k,i];
          end;{ k }
        end;{ j }
      end;{ if d[i] <> 0.0 }
    end; { if vectors }

    d[i] := a[i,i];

    if vectors then
    begin
      a[i,i] := 1.0;
      if l >= Lo then
      begin
        for j := Lo to l do
        begin
          a[i,j] := 0.0;
          a[j,i] := 0.0;
        end;{ j }
      end;{ l > 0 }
    end;{ if vectors }
  end;{ i }
end;{ of tred2 }

procedure tqli(const d, e: IFArr1D; const z: IFArr2D; var err: TInt;
                                  vectors: boolean = true; IterMax: TInt = 999);
label
  1, 2;
var
  i,  k, l, m, iter, Lo, Hi: integer;
  s, r, p, g, f, dd, c, b: TFloat;
begin { of tqli }
// All checks are made in diag
  Lo := z.Lo1;
  Hi := z.Hi1;
  err := 0;
  if Hi > Lo then
  begin
    for i := Lo+1 to Hi do
    begin
      e[i-1] := e[i];
    end;{ i }
    e[Hi] := 0.0;

    for l := Lo to Hi do
    begin
      iter := 0;
1:  for m := l to Hi - 1 do
      begin
        dd := abs(d[m]) + abs(d[m+1]);
        if (abs(e[m]) + dd = dd) then goto 2;
      end; { m }
      m := Hi;
2:  if m <> l then
      begin
        if iter = itermax then
        begin
          err := iter;
          exit;
        end;
        iter := iter + 1;
        g := (d[l+1] - d[l])/(2.0*e[l]); { Form shift }
        r := sqrt(sqr(g) + 1.0);
        g := d[m] - d[l] + e[l]/(g + sign2(r,g));
        s := 1.0;
        c := 1.0;
        p := 0.0;
        for i := m - 1 downto l do
        begin
          f := s*e[i];
          b := c*e[i];
          if abs(f) >= abs(g) then
          begin
            c := g/f;
            r := sqrt(sqr(c) + 1.0);
            e[i+1] := f*r;
            s := 1.0/r;
            c := c*s;
          end else
          begin{ abs(f) < abs(g) }
            s := f/g;
            r := sqrt(sqr(s) + 1.0);
            e[i+1] := g*r;
            c := 1.0/r;
            s := c*s;
          end;

          g:= d[i+1] - p;
          r:= (d[i] - g)*s + 2.0*c*b;
          p:= s*r;
          d[i+1] := g + p;
          g:= c*r - b;
          if vectors then
          begin
            for k := Lo to Hi do
            begin
              f := z[k,i+1];
              z[k,i+1] := s*z[k,i] + c*f;
              z[k,i]:= c*z[k,i] - s*f;
            end;{ k }
          end; { if vectors }

        end;{ i }

        d[l] := d[l] - p;
        e[l] := g;
        e[m] := 0.0;
        goto 1;

      end; {  m <> l }
    end;{ l }
  end; { Hi > Lo }
end;{ of tqli }

{ Functionally the same as SortVal - simple sort }
procedure SortVal0(const d: IFArr1D; const v: IFArr2D;
                            vectors: boolean = true; ascending: boolean = true);
var
  i, j, k: TInt;
  t: TFloat;
begin
// All checks are made in diag
  for i := v.Lo2 to (v.Hi2-1) do
    for j := i+1 to v.Hi2 do
      if     ascending and (d[j] < d[i]) or
         not ascending and (d[j] > d[i]) then
      begin
        t := d[i];
        d[i] := d[j];
        d[j] := t;
        if vectors then
          for k := v.Lo1 to v.Hi1 do
          begin
            t := v[k,i];
            v[k,i] := v[k,j];
            v[k,j] := t;
          end;
      end;
end;{ of SortVal0 }

{ Functionally the same as SortVal0 - Quick sort }
procedure SortVal(const d: IFArr1D; const v: IFArr2D;
                            vectors: boolean = true; ascending: boolean = true);
begin
  if vectors then
    QSortF2D(d, v,ascending, _Col)
  else
    QSortF1D(d, ascending);
end;{ of SortVal }

{ author: Nikolai V. Shokhirev 2002 Diagonalization of real symmetric matrix }
procedure diag(const a: IFArr2D; const d: IFArr1D; sort: boolean = true;
               vectors: boolean = true; ascending: boolean = true;
               Epsilon: TFloat = SqrtMinFloat; IterMax: TInt = 999);
var
  e: IFArr1D;
  err, Lo, Hi: TInt;
begin
  if not IsSquare(a) then
    Raise ERangeError.Create(RS_NotSquare);
  if (a.Lo1 <> d.Lo1) or (a.Hi1 <> d.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  Lo := a.Lo1;
  Hi := a.Hi1;
  if Lo = Hi then
  begin
    d[Lo] := a[Lo,Lo];
    a[Lo,Lo] := 1.0;
  end else
  begin
    e := TFArr1D.Create(Lo, Hi);
    tred2(a, d, e, vectors, Epsilon);
    tqli(d, e, a, err, vectors, IterMax);
    if err <> 0 then
       Raise ERangeError.CreateFMT(RS_IterMaxError, [IterMax]);
    if sort then
    SortVal(d, a, vectors, ascending);
  end;
end;

{ Eps1  - is used at  SNA  and  CSA  calculations
  AbsEps  - is the level of the elimination of the non-diagonal matrix elements
  RelEps  - convergence criterion: if OffDsq < RelEps*Norm then goto END; }
procedure Jacobi(const A, T: IFArr2D; const Eigen: IFArr1D;
                                       var err: integer; first: boolean = true);
const
  Eps1 = 1.0e-12;
  AbsEps  = SqrtMinFloat;
  RelEps = 1.0e-10;
  ItMax = 999;
var
  NoRotation: boolean;
  Lo, Hi, i,   j,   k,   Iter: integer;
  omegaMin, // 0.5*n*sigma*sqr(RelEps)
  omega,    // Norm2(OffD(A))/2
  tau,      // elimination threshold
  sigma,    // Norm2(A)
  n,        // Dim of A
  SPQ,    CSA,    SNA,    Q,   P,
  HoldIK, HoldKI: TFloat;
  Aik: IFArr1D;
begin
  if not IsSquare(A) then
    Raise ERangeError.Create(RS_NotSquare);
  if not SameLim(A,T) or (Eigen.Lo1<>A.Lo1) or (Eigen.Hi1<>A.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);

  Lo := A.Lo1;
  Hi := A.Hi1;
  Aik := TFArr1D.Create(Lo, Hi);
  n := A.Dim1;

  if first then
    for i:= Lo to Hi  do
    begin
      for j := Lo to Hi  do
        T[i,j] := 0.0;
      T[i,i] := 1.0;
    end;

  for i:= Lo to Hi  do
    Eigen[i] := A[i,i];

  if Hi = Lo then exit;

  omega := 0.0;
  sigma := sqr(A[Lo,Lo]);
  for i := Lo+1 to Hi  do
  begin
    sigma := sigma + sqr(A[i,i]);
    for j := Lo to i-1  do
      omega := omega+sqr(A[i,j]);
  end;
  sigma := sigma +2.0*omega;
  omegaMin := 0.5*n*sigma*sqr(RelEps);

  if omega > sqr(n*AbsEps) then
  begin // Not diagonal
    Iter := 1;
    n := n*(n-1.0)/2.0;
    tau := sqrt(omega/n);
    repeat
      NoRotation := true;
      for i := Lo to Hi - 1  do
        for j:=i + 1 to Hi  do
        begin
          Q := abs(A[i,i]-A[j,j]);
          if abs(A[i,j]) > tau*AbsEps then
          begin // rotation
            omega := omega - sqr(A[i,j]);
            tau := sqrt(abs(omega/n));
            NoRotation := false;
            if (Q > Eps1) then
            begin
              P := 2.0*A[i,j]*sign1(A[i,i]-A[j,j]);
              SPQ := sqroot2(P, Q);  // sqrt(sqr(P)+sqr(Q))
              CSA := sqrt((1.0+Q/SPQ)/2.0);
              SNA := P/(SPQ*CSA*2.0);
            end else
            begin
              CSA := sqrt(0.5);
              SNA := CSA;
            end;
            for k := Lo to Hi  do
            begin
              HoldKI := T[k,i];
              T[k,i] := HoldKI*CSA + T[k,j]*SNA;
              T[k,j] := HoldKI*SNA - T[k,j]*CSA;
            end;

            for k := i to Hi  do
              if k <= j then
              begin
                Aik[k] := A[i,k];
                A[i,k] := CSA*Aik[k] + SNA*A[k,j];
                if k = j then
                begin
                  A[j,k] := SNA*Aik[k] - CSA*A[j,k];
                  Aik[j] := SNA*Aik[i] - CSA*Aik[j];
                end;
              end else
              begin
                HoldIK := A[i,k];
                A[i,k] := CSA*HoldIK + SNA*A[j,k];
                A[j,k] := SNA*HoldIK - CSA*A[j,k];
              end;

            for k := Lo to j  do
              if k > i then
                A[k,j] := SNA*Aik[k] - CSA*A[k,j]
              else
              begin
                HoldKI := A[k,i];
                A[k,i] := CSA*HoldKI + SNA*A[k,j];
                A[k,j] := SNA*HoldKI - CSA*A[k,j];
              end;
          end; // rotation
        end; { of i, j - loop }

      for i := Lo to Hi do
        Eigen[i] := A[i,i];

      inc(Iter);
//    until (Iter >= ItMax) or (omega < omegaMin);
    until (Iter >= ItMax) or (omega < omegaMin) or NoRotation ;

    if Iter > ItMax then
      err := ItMax;
  end;

{ sorting of eigenvalues and eigenvectors in increasing order }
  for i := Lo to Hi  do
  begin
    k := i;
    for j := i to Hi  do
      if Eigen[j]<Eigen[k] then  k := j;
    if k<>i then
    begin
      P := Eigen[k];
      Eigen[k] := Eigen[i];
      Eigen[i] := P;
      for j := Lo to Hi  do
      begin
        P := T[j,k];
        T[j,k] := T[j,i];
        T[j,i] := P;
      end;
    end;
  end;
end; { Of the procedure Jacobi }


{ author: Nikolai V. Shokhirev 1994
  convergence criteria:
                           Norm2(OffD(A+iB))/2 < sqr(RelEps)*N*Norm2(A+iB)/2
                       OR  NoRotation OR Iter >= IterMax                     }
procedure CJacobi (const A, B, U, V: IFArr2D; const E: IFArr1D;
                  var err: TInt; first: boolean = true; order: boolean = true);
const
  ItMax = 999;
  Eps1 = 1.0e-12;        // is used at  SNA  and  CSA  calculations
  AbsEps  = SqrtMinFloat;// level of elimination of the Off-diagonal matrix elements
  RelEps = 1.0e-10;
var
  l, m, i, j, k, Iter, Lo, Hi: TInt;
  NoRotation: boolean;
  omegaMin, // 0.5*n*sigma*sqr(RelEps)
  omega,    // Norm2(OffD(A+iB))/2
  tau2,     // elimination threshold
  sigma,    // Norm2(A+iB)
  rn,       // Dim of A
  Hii,     Hjj,     Aji,     Bji,
  Hii_Hjj, HijxHji, C,       CC,
  SS,      S_R,     S_I,     T2,
  Bki,     Aki,     Bik,     Aik,
  Uki,     Vki,     Ujk,     Vjk,
  signum,  Ek,      Nd,      Noff: TFloat;

begin{of CJacobi }

  if (not IsSquare(A)) or (not IsSquare(B)) or
     (not IsSquare(U)) or (not IsSquare(V)) then
    Raise ERangeError.Create(RS_NotSquare);
  if (not SameLim(A,B) ) or (not SameLim(U,V) ) or (not SameLim(U,A) ) or
     (A.Lo1<>E.Lo1) or (A.Hi1<>E.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  Lo := A.Lo1;
  Hi := A.Hi1;

  if Hi = Lo then
    begin
      U[Lo,Lo] := 1.0;
      V[Lo,Lo] := 0.0;
      E[Lo]:= A[Lo,Lo];
      exit;
    end;

  if first then
    for i:= Lo to Hi  do
    begin
      for j := Lo to Hi  do
      begin
        U[i,j] := 0.0;
        V[i,j] := 0.0;
      end;
      U[i,i] := 1.0;
      V[i,i] := 0.0;
      E[i]:= A[i,i];
    end;

  if Lo = Hi then exit;

  rn:= (Lo - Hi +1);
  omega := 0.0;
  sigma := sqr(A[Lo,Lo]);
  for i := Lo+1 to Hi  do
  begin
    sigma := sigma + sqr(A[i,i]);
      for j := Lo to i-1  do
        omega := omega + sqr(A[i,j]) + sqr(B[i,j]);
  end;
  sigma := sigma + 2.0*omega;
  omegaMin := 0.5*rn*sigma*sqr(RelEps);

  if omega > sqr(rn*AbsEps) then
  begin
    Iter := 1;
    rn := rn*(rn-1.0)/2.0;
    tau2 := omega/rn;
    repeat {of iteration loop}
      NoRotation := true;
      for i := Lo to Hi - 1  do
        for j:= i + 1 to Hi  do
        begin
          Aji:= A[j,i];
          Bji:= B[j,i];
  {
          HijxHji := 0.0;
          if abs(Aji) > EpsAbs then HijxHji := HijxHji + sqr(Aji);
          if abs(Bji) > EpsAbs then HijxHji := HijxHji + sqr(Bji);
  }
          HijxHji := sqr(Aji) + sqr(Bji);

          if HijxHji > tau2 then
          begin
            omega := omega - HijxHji;
            tau2 := abs(omega/rn);
            NoRotation := false;
            Hii:= A[i,i];
            Hjj:= A[j,j];
            Hii_Hjj := Hii - Hjj;

            if abs(Hii_Hjj) > Eps1 then
            begin
              T2:= 4.0*HijxHji/sqr(Hii_Hjj);
              S_R := 2.0*Aji/(Hii_Hjj)/sqrt(2.0*(1.0 + T2 + sqrt(1.0 + T2)));
              S_I := 2.0*Bji/(Hii_Hjj)/sqrt(2.0*(1.0 + T2 + sqrt(1.0 + T2)));
              SS:= sqr(S_R) + sqr(S_I);
              CC:= 1.0 - SS;
              C:= sqrt(abs(CC));
            end else{ degenerate case }
            begin
              CC:= 0.5;
              SS:= CC;
              C:= sqrt(CC);
              if Hii_Hjj < 0.0 then signum := -1.0 else signum :=  1.0;
              S_R := signum*Aji/sqrt(2.0*HijxHji);
              S_I := signum*Bji/sqrt(2.0*HijxHji);
            end;
            for k := Lo to Hi  do
            begin
              Uki := U[k,i];
              Vki := V[k,i];
              U[k,i] := Uki*C      + U[k,j]*S_R - V[k,j]*S_I;
              V[k,i] := Vki*C      + U[k,j]*S_I + V[k,j]*S_R;
              U[k,j] := U[k,j]*C - Uki*S_R      - Vki*S_I;
              V[k,j] := V[k,j]*C + Uki*S_I      - Vki*S_R;
            end;
            for k := Lo to Hi  do
            if (k <> i) and (k <> j) then
            begin
              Aki := A[k,i];
              Bki := B[k,i];
              A[k,i] := Aki*C      + A[k,j]*S_R - B[k,j]*S_I;
              B[k,i] := Bki*C      + A[k,j]*S_I + B[k,j]*S_R;
              A[k,j] := A[k,j]*C - Aki*S_R      - Bki*S_I;
              B[k,j] := B[k,j]*C + Aki*S_I      - Bki*S_R;
              A[i,k] :=  A[k,i];
              B[i,k] := -B[k,i];
              A[j,k] :=  A[k,j];
              B[j,k] := -B[k,j];
            end;
  {          for k := Lo to Hi  do
            if (k <> i) and (k <> j) then
            begin
              Aik := A[i,k];
              Bik := B[i,k];
              A[i,k] := Aik*C      + A[j,k]*S_R + B[j,k]*S_I;
              B[i,k] := Bik*C      - A[j,k]*S_I + B[j,k]*S_R;
              A[j,k] := A[j,k]*C - Aik*S_R      + Bik*S_I;
              B[j,k] := B[j,k]*C - Aik*S_I      - Bik*S_R;
            end;
  }
            A[i,i] := Hii*CC + Hjj*SS + 2.0*C*(S_R*Aji + S_I*Bji);
            A[j,j] := Hjj*CC + Hii*SS - 2.0*C*(S_R*Aji + S_I*Bji);
            A[j,i] := 0.0;
            B[j,i] := 0.0;
            A[i,j] := 0.0;
            B[i,j] := 0.0;
          end;
        end; { of i, j - loop }
      for i := Lo to Hi  do
      begin
        E[i] := A[i,i];
      end;
    inc(Iter);
  until (Iter >= ItMax) or (omega < omegaMin) or NoRotation ;

    if Iter >= ItMax then err := ItMax;
  end;

{ sorting of eigenvalues and eigenvectors in increasing order }
  if order then
    SortValC(E, U, V, E.Dim1);
end; { Of the procedure CJacobi }

procedure JacobiC (const A, B, U, V: IFArr2D; const E: IFArr1D;
                  var err: TInt; first: boolean = true; order: boolean = true);
const
  ItMax = 999;
  Eps1 = 1.0e-12;        // is used at  SNA  and  CSA  calculations
  AbsEps  = SqrtMinFloat;// level of elimination of the Off-diagonal matrix elements
  RelEps = 1.0e-10;
var
  l,       m,
  i,       j,       k,       Iter: integer;
  NoRotation: boolean;
  omegaMin, // 0.5*n*sigma*sqr(RelEps)
  omega,    // Norm2(OffD(A+iB))/2
  tau2,     // elimination threshold
  sigma,    // Norm2(A+iB)
  rn,       // Dim of A
  Hii,     Hjj,     Aij,     Aji,
  Bij,     Bji,     Hii_Hjj, HijxHji,
  C,       CC,
  SS,      S_R,     S_I,     T2,
  Bki,     Aki,     Bik,     Aik,
  Uki,     Vki,     Ujk,     Vjk,
  signum,  Ek: TFloat;
  Lo, Hi: TInt;

begin{of JacobiC }

  if (not IsSquare(A)) or (not IsSquare(B)) or
     (not IsSquare(U)) or (not IsSquare(V)) then
    Raise ERangeError.Create(RS_NotSquare);
  if (not SameLim(A,B)) or (not SameLim(U,V)) or (not SameLim(U,A))  or
     (A.Lo1<>E.Lo1) or (A.Hi1<>E.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);

  Lo := a.Lo1;
  Hi := a.Hi1;

  if first then
    for i:= Lo to Hi  do
    begin
      for j := Lo to Hi  do
      begin
        U[i,j] := 0.0;
        V[i,j] := 0.0;
      end;
      U[i,i] := 1.0;
//      V[i,i] := 0.0;
      E[i]:= A[i,i];
    end;

  if Hi = Lo then exit;

  rn := (Hi-Lo+1);
  omega := 0.0;
  sigma := sqr(A[Lo,Lo]);
  for i := Lo+1 to Hi  do
  begin
    sigma := sigma + sqr(A[i,i]);
      for j := Lo to i-1  do
        omega := omega + sqr(A[i,j]) + sqr(B[i,j]);
  end;
  sigma := sigma +2.0*omega;
  omegaMin := 0.5*rn*sigma*sqr(RelEps);

  if omega < sqr(rn*AbsEps) then
    exit;  // already diagonal

  Iter := 1;
  rn := rn*(rn-1.0)/2.0;
  tau2 := omega/rn;
  repeat {of iteration loop}
    NoRotation := true;
    for i := Lo to Hi - 1  do
      for j:=i + 1 to Hi  do
      begin
        Aji:= A[j,i];
        Bji:= B[j,i];
        HijxHji := sqr(Aji) + sqr(Bji);
        if HijxHji > tau2 then
        begin
          omega := omega - HijxHji;
          tau2 := abs(omega/rn);
          NoRotation := false;
          Hii:= A[i,i];
          Hjj:= A[j,j];
          Hii_Hjj := A[i,i]-A[j,j];
          Aij:= A[i,j];
          Bij:= B[i,j];
          if abs(Hii_Hjj) > Eps1 then
          begin
            T2:= 4.0*HijxHji/sqr(Hii_Hjj);
            S_R := 2.0*Aji/(Hii_Hjj)/sqrt(2.0*(1.0 + T2 + sqrt(1.0 + T2)));
            S_I := 2.0*Bji/(Hii_Hjj)/sqrt(2.0*(1.0 + T2 + sqrt(1.0 + T2)));
            SS:= sqr(S_R) + sqr(S_I);
            CC:= 1.0 - SS;
            C:= sqrt(abs(CC));
          end else{ degenerate case }
          begin
            CC:= 0.5;
            SS:= CC;
            C:= sqrt(CC);
            if Hii_Hjj < 0.0 then signum := -1.0 else signum :=  1.0;
            S_R := signum*Aji/sqrt(2.0*HijxHji);
            S_I := signum*Bji/sqrt(2.0*HijxHji);
          end;
          for k := Lo to Hi  do
          begin
            Uki := U[k,i];
            Vki := V[k,i];
            U[k,i] := Uki*C      + U[k,j]*S_R - V[k,j]*S_I;
            V[k,i] := Vki*C      + U[k,j]*S_I + V[k,j]*S_R;
            U[k,j] := U[k,j]*C - Uki*S_R      - Vki*S_I;
            V[k,j] := V[k,j]*C + Uki*S_I      - Vki*S_R;
          end;
          for k := Lo to Hi  do
          if (k <> i) and (k <> j) then
          begin
            Aki := A[k,i];
            Bki := B[k,i];
            A[k,i] := Aki*C      + A[k,j]*S_R - B[k,j]*S_I;
            B[k,i] := Bki*C      + A[k,j]*S_I + B[k,j]*S_R;
            A[k,j] := A[k,j]*C - Aki*S_R      - Bki*S_I;
            B[k,j] := B[k,j]*C + Aki*S_I      - Bki*S_R;
          end;
          for k := Lo to Hi  do
          if (k <> i) and (k <> j) then
          begin
            Aik := A[i,k];
            Bik := B[i,k];
            A[i,k] := Aik*C      + A[j,k]*S_R + B[j,k]*S_I;
            B[i,k] := Bik*C      - A[j,k]*S_I + B[j,k]*S_R;
            A[j,k] := A[j,k]*C - Aik*S_R      + Bik*S_I;
            B[j,k] := B[j,k]*C - Aik*S_I      - Bik*S_R;
          end;
          A[i,i] := Hii*CC + Hjj*SS + 2.0*C*(S_R*Aji + S_I*Bji);
          A[j,j] := Hjj*CC + Hii*SS - 2.0*C*(S_R*Aji + S_I*Bji);
{             _ _
 SS -> S*S or S*S
          A[j,i] := Aji*CC - Aij*SS + C*S_R*(Hjj - Hii);
          B[j,i] := Bji*CC - Bij*SS + C*S_I*(Hjj - Hii);
          A[i,j] := Aij*CC - Aji*SS + C*S_R*(Hjj - Hii);
          B[i,j] := Bij*CC - Bji*SS - C*S_I*(Hjj - Hii);
}
          A[j,i] := 0.0;
          B[j,i] := 0.0;
          A[i,j] := 0.0;
          B[i,j] := 0.0;
        end;
      end; { of i, j - loop }
      for i := Lo to Hi  do
      begin
        E[i] := A[i,i];
      end;
      inc(Iter);
  until (Iter >= ItMax) or (omega < omegaMin) or NoRotation ;

  if Iter >= ItMax then err := ItMax;

{ sorting of eigenvalues and eigenvectors in increasing order }
  if order then
    SortValC(E, U, V, E.Dim1);
end; { Of the procedure JacobiC }

{ Pascal version by Nikolai V. Shokhirev 2002

  Original EISPACK comment for htridi:
  This subroutine is a translation of a complex analogue of the algol procedure
  tred1, num. math. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
  handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).

  this subroutine reduces a complex hermitian matrix to a real symmetric
  tridiagonal matrix using unitary similarity transformations.

on input
  nm must be set to the row dimension of two-dimensional array parameters as
  declared in the calling program dimension statement.

  n is the order of the matrix.

  ar and ai contain the real and imaginary parts, respectively, of the complex
            hermitian input matrix.
            only the lower triangle of the matrix need be supplied.

on output
  ar and ai contain information about the unitary transformations used in the
            reduction in their full lower triangles.
            Their strict upper triangles and the diagonal of ar are unaltered.

  d contains the diagonal elements of the the tridiagonal matrix.

  e contains the subdiagonal elements of the tridiagonal matrix in its last n-1
    positions.  e(1) is set to zero.

  e2 contains the squares of the corresponding elements of e. e2 may coincide
     with e if the squares are not needed.

  tau contains further information about the transformations.

  calls pythag for  dsqrt(a*a + b*b) .

   questions and comments should be directed to burton s. garbow,
   mathematics and computer science div, argonne national laboratory

   this version dated august 1983. }
procedure htridi(const ar, ai, tau: IFArr2D; const d, e, e2: IFArr1D);
var
  i,j,k,l,n,ii,jp1: TInt;
  f,g,h,fi,gi,hh,si,scale: TFloat;
  Lo, Hi: TInt;
  //dimensions: ar[nm,n],ai[nm,n],d[n],e[n],e2[n],tau[2,n];  // nm = m
label
  160, 270, 170;

begin
  Lo := ar.Lo1;
  Hi := ar.Hi1;

  n := (Hi-Lo+1);
{  All checks are in diagc
  err := -4;
  if not IsSquareDim(ar) then exit;
  err := -3;
  if (d.Dim < n) or (e.Dim < n) or (e2.Dim < n) then exit;
  err := FitMt(ar,ai); if err <> 0 then exit;
}
  tau[1,n] := 1.0;
  tau[2,n] := 0.0;

  for i := 1 to  n do
    d[i] := ar[i,i];
  // for i=n step -1 until 1 do
  for ii := 1 to  n do
  begin
    i := n + 1 - ii;
    l := i - 1;
    h := 0.0;
    scale := 0.0;
    if (l < 1) then
    begin
      e[i] := 0.0;
      e2[i] := 0.0;
    end else
      for k := 1 to l do  // scale row (algol tol then not needed)
        scale := scale + abs(ar[i,k]) + abs(ai[i,k]);

    if (scale = 0.0) then
    begin
      if l >= 1 then
      begin
        tau[1,l] := 1.0;
        tau[2,l] := 0.0;
      end;
    end else
    begin
    for k := 1 to  l do
      begin
         ar[i,k] := ar[i,k] / scale;
         ai[i,k] := ai[i,k] / scale;
         h := h + ar[i,k]*ar[i,k] + ai[i,k]*ai[i,k];
      end;

      e2[i] := scale*scale*h;
      g := sqrt(h);
      e[i] := scale*g;
      f := pythag(ar[i,l],ai[i,l]);
      // form next diagonal element of matrix t
      if (f <> 0.0) then
      begin
        tau[1,l] := (ai[i,l]*tau[2,i] - ar[i,l]*tau[1,i]) / f;
        si := (ar[i,l]*tau[2,i] + ai[i,l]*tau[1,i]) / f;
        h := h + f*g;
        g := 1.0 + g / f;
        ar[i,l] := g*ar[i,l];
        ai[i,l] := g*ai[i,l];
      end else
        goto 160;
      if (l  =  1) then goto 270;
      goto 170;
  160:tau[1,l] := -tau[1,i];
      si := tau[2,i];
      ar[i,l] := g;
  170:f := 0.0;

      for j := 1 to  l do
      begin
        g := 0.0;
        gi := 0.0;
        // form element of a*u
        for k := 1 to  j do
        begin
           g := g + ar[j,k]*ar[i,k] + ai[j,k]*ai[i,k];
           gi := gi - ar[j,k]*ai[i,k] + ai[j,k]*ar[i,k];
        end;

        jp1 := j + 1;
        if (l >= jp1) then
        for k := jp1 to  l do
        begin
           g := g + ar[k,j]*ar[i,k] - ai[k,j]*ai[i,k];
           gi := gi - ar[k,j]*ai[i,k] - ai[k,j]*ar[i,k];
        end;
        // form element of p
        e[j] := g / h;
        tau[2,j] := gi / h;
        f := f + e[j]*ar[i,j] - tau[2,j]*ai[i,j];
      end;

      hh := f / (h + h);
      // form reduced a
      for j := 1 to  l do
      begin
        f := ar[i,j];
        g := e[j] - hh*f;
        e[j] := g;
        fi := -ai[i,j];
        gi := tau[2,j] - hh*fi;
        tau[2,j] := -gi;

        for k := 1 to  j do
        begin
           ar[j,k] := ar[j,k] - f*e[k] - g*ar[i,k] + fi*tau[2,k] + gi*ai[i,k];
           ai[j,k] := ai[j,k] - f*tau[2,k] - g*ai[i,k] - fi*e[k] - gi*ar[i,k];
        end;
      end;

  270:for k := 1 to  l do
      begin
         ar[i,k] := scale*ar[i,k];
         ai[i,k] := scale*ai[i,k];
      end;

      tau[2,l] := -si;
    end;
    hh := d[i];
    d[i] := ar[i,i];
    ar[i,i] := hh;
    ai[i,i] := scale*sqrt(h);
  end;
end;

{ Pascal version by Nikolai V. Shokhirev 2002

  Original EISPACK comment for htribk:
  This subroutine is a translation of a complex analogue of the algol procedure
  trbak1, num. math. 11, 181-195(1968) by Martin, Reinsch, and Wilkinson.
  Handbook for auto. comp., vol.ii-linear algebra, 212-226(1971).

  This subroutine forms the eigenvectors of a complex Hermitian matrix by back
  transforming those of the corresponding real symmetric tridiagonal matrix
  determined by  htridi.

on input
  nm must be set to the row dimension of two-dimensional array parameters as
     declared in the calling program dimension statement.

  n is the order of the matrix.

  ar and ai contain information about the unitary transformations used in the
            reduction by  htridi  in their full lower triangles except for the
            diagonal of ar.

  tau contains further information about the transformations.

  m is the number of eigenvectors to be back transformed.

  zr contains the eigenvectors to be back transformed in its first m columns.

on output
  zr and zi contain the real and imaginary parts, respectively, of the
            transformed eigenvectors in their first m columns.

     note that the last component of each returned vector
     is real and that vector euclidean norms are preserved.

     questions and comments should be directed to burton s. garbow,
     mathematics and computer science div, argonne national laboratory

     this version dated august 1983. }
procedure htribk(const ar, ai, tau: IFArr2D; m: TInt; const zr, zi: IFArr2D);
var
  i, j, k, l, n: TInt;
  h, s, si: TFloat;
//  Lo, Hi: TInt;
//  dimensions: ar[nm,n],ai[nm,n],tau[2,n],zr[nm,m],zi[nm,m];  nm = n
begin
  n := ar.Dim1;

{ All checks are in diagc }

  if (m = 0) then exit;
{ transform the eigenvectors of the real symmetric tridiagonal matrix to those
  of the hermitian tridiagonal matrix. }
  for k := 1 to  n do
  begin
    for j := 1 to  m do
    begin
      zi[k,j] := -zr[k,j]*tau[2,k];
      zr[k,j] :=  zr[k,j]*tau[1,k];
    end;
  end;
  if (n = 1) then exit;

//  recover and apply the householder matrices
  for i := 2 to  n do
  begin
    l := i - 1;
    h := ai[i,i];
    if (h <> 0.0) then
    begin
      for j := 1 to  m do
      begin
        s := 0.0;
        si := 0.0;

        for k := 1 to  l do
        begin
          s  := s  + ar[i,k]*zr[k,j] - ai[i,k]*zi[k,j];
          si := si + ar[i,k]*zi[k,j] + ai[i,k]*zr[k,j];
        end;
        //  double divisions avoid possible underflow
        s := (s / h) / h;
        si := (si / h) / h;

        for k := 1 to  l do
        begin
          zr[k,j] := zr[k,j] - s*ar[i,k] - si*ai[i,k];
          zi[k,j] := zi[k,j] - si*ar[i,k] + s*ai[i,k];
        end;
      end;
    end;
  end;
end;

procedure SortValC(const d: IFArr1D; const vr, vi: IFArr2D; m: TInt;
                  vectors: boolean = true; ascending: boolean = true);
var
  i, j, k, i1, j1: TInt;
  t: TFloat;
begin
  i := vr.Lo2;
  for i1 := 1 to (m-1) do
  begin
    j := i + 1;
    for j1 := i1+1 to m do
    begin
      if     ascending and (d[j] < d[i]) or
         not ascending and (d[j] > d[i]) then
      begin
        t := d[i];
        d[i] := d[j];
        d[j] := t;
        if vectors then
          for k := vr.Lo1 to vr.Hi1 do
          begin
            t := vr[k,i];
            vr[k,i] := vr[k,j];
            vr[k,j] := t;
            t := vi[k,i];
            vi[k,i] := vi[k,j];
            vi[k,j] := t;
          end;
      end;
      inc(j);
    end;
    inc(i);
  end;
end;{ of SortValC }

{ author: Nikolai V. Shokhirev 2002 Diagonalization of Hermitian matrix }
procedure diagc(const ar, ai, zr, zi: IFArr2D; const d: IFArr1D; m: TInt;
                var err: TInt; sort: boolean = true; ascending: boolean = true);
var
  e, e2: IFArr1D;
  tau: IFArr2D;
  n, i, mm : TInt;
  R1, R2 : IRestore;
begin
  if (ar.Dim1<>ar.Dim2) then exit;
  n := ar.Dim1;

  if zr.Dim1 < n then exit;
  if not SameLim(ar,ai) or not SameLim(zr,zi) then exit;
  if (d.Dim1 < n) then exit;
  mm := min(ar.Dim2,zr.Dim2);
  if m > 0 then
    if m > mm then m := mm;

  e := TFArr1D.Create(n);
  e2 := TFArr1D.Create(n);
  tau := TFArr2D.Create(2,n);

  R1 := TRestore1D.Create([d]);
  R2 := TRestore2D.Create([ar, ai, zr, zi]);
  for i := 1 to mm do
    zr[i,i] := 1.0;
  htridi(ar, ai, tau, d, e, e2);
  tqli(d, e, zr, err);
  if err <> 0 then
    exit;
  if m > 0 then
    htribk(ar, ai, tau, m, zr,zi);
  if sort and ( m > 0) then
    SortValC(d, zr, zi, m, true, ascending);
end;

procedure FastInverse(const A: IFArr2D; var Det: TFloat);
const
  Eps = 1.0e-16;
var
  Hi, Lo, i,  j,  k,  i0: integer;
  A0,  B: TFloat;
  J0: IIArr1D;

begin
  if not IsSquare(A) then
    raise ENonSquareMatrix.Create('in FastInverse');
  Lo := A.Lo1;
  Hi := A.Hi1;
  J0 := TIArr1D.Create(Lo,Hi);
  if  Lo = Hi then
  begin
    if abs(A[Lo,Lo]) < MinFloat then
      raise ESingularMatrix.Create('in FastInverse 1 x 1');
    A[Lo,Lo] := 1.0/A[Lo,Lo];
    Det:= A[Lo,Lo];
    Exit;
  end;

  Det := 1.0;
  for i:= Lo to Hi  do
    begin
      { 1. Finding of the pivot element A0 i0  is the pivot row index }
      A0 := 0.0;
      i0 := 0;
      for j := i to Hi do
        if abs(A[j,i]) > A0 then
        begin
          A0 := abs(A[j,i]);
          i0 := j;
        end;
      if A0 < Eps then
        raise ESingularMatrix.CreateFmt('in FastInverse i = %d',[i]);

      { 2. Swapping the rows }
      J0[i] := i0;
      B := 1.0/A[i0,i];
//      A.SwapRows(i, i0);
      A.Swap(i, i0, _Row);
      for j := Lo to Hi  do
        A[i,j] := A[i,j]*B;
      A[i,i] := B;

      { 3. Subtracting the rows }
      for j := Lo to Hi  do
        if i<>j then
        begin
          B:= A[j,i];
          A[j,i] := 0.0;
          for k:=Lo to Hi  do
            A[j,k] := A[j,k] - B*A[i,k];
        end;
      Det := Det/A[i0,i];
    end;

    { 4. Back Swapping the columns }
    for i:= Hi downto Lo  do
    begin
      j := J0[i];
      if j<>i then
      begin
        Det := -Det;
        A.Swap(i, j, _Col);
      end;
    end;
end;{   Of the procedure FastInverse  }

procedure SVD(const A, U, V: IFArr2D; const W: IFArr1D; var RetCode: integer;
                MatU: boolean = true; MatV: boolean = true;
                FullMat: boolean = false; ItnLimit: integer = 30);
var
  ALo1, ALo2: TInt;
begin
  ALo1 := A.Lo1;  ALo2 := A.Lo2;
  A.Lo1 := 1;   A.Lo2 := 1;
  U.Lo1 := 1;   U.Lo2 := 1;
  V.Lo1 := 1;   V.Lo2 := 1;
  W.Lo1 := 1;

  if (A.Dim1 > A.Dim2) then
    SVD_(A, U, V, W, RetCode, MatU, MatV, FullMat, ItnLimit)
  else
    SVD_(A, V, U, W, RetCode, MatU, MatV, FullMat, ItnLimit);

  SortSVD(W, U, V, MatU, MatV);
  
  A.Lo1 := ALo1;   A.Lo2 := ALo2;
  U.Lo1 := ALo1; { U.Lo2 := 1; }
  V.Lo1 := ALo2; { V.Lo2 := 1; }
{ W.Lo1  := 1; }
end;

procedure SVD_(const A, U, V: IFArr2D; const W: IFArr1D; var RetCode: integer;
         MatU, MatV: boolean; FullMat: boolean = false; ItnLimit: integer = 30);
var
  i, j, k, l, i1, k1, l1, its, mn, ExitKey: integer;
  C, G, F, X, S, H, Y, Z, Scale, ANorm, GG: TFloat;
  NA, M, N: integer;
  RV1: IFArr1D;
begin
  RetCode := 0;

  M := max(A.Dim1,A.Dim2);
  N := min(A.Dim1,A.Dim2);
  if FullMat then mn := M else mn := N;
  if (U.Dim1 < M) or (U.Dim2 < mn) or ( V.Dim1 < N) or (V.Dim2 < N) or
     (W.Dim1 < mn) then
  begin
    RetCode := -1;
    exit;
  end;
  RV1 := TFArr1D.Create(N);

  if (A.Dim1 >= A.Dim2) then
    for i:=1 to M  do
      for j:=1 to N  do
        U[i,j] := A[i,j]
  else
    for i:=1 to M  do
      for j:=1 to N  do
        U[i,j] := A[j,i];

  G :=  0.0;
  Scale :=  0.0;
  ANorm :=  0.0;

  for i := 1 to N  do
  begin
    l := i+1;
    RV1[i] := Scale*G;
    G:= 0.0;
    S:= 0.0;
    Scale:= 0.0;
    if i <= M then
    begin
      for k:=1 to M  do
        Scale := Scale+abs(U[k,i]);
      if Scale <> 0.0 then
      begin
        for k := i to M  do
        begin
          U[k,i] := U[k,i]/Scale;
          S := S+sqr(U[k,i]);
        end;
        F := U[i,i];
        G := -Sign2(sqrt(S),F);
        H := F*G - S;
        U[i,i] := F - G;
        if i<>N then
          for j := l to N  do
          begin
            S := 0.0;
            for k := i to M  do
              S := S + U[k,i]*U[k,j];
            F := S/H;
            for k := i to M  do
              U[k,j] := U[k,j] + F*U[k,i];
          end;// j
          for k:=i to M  do
            U[k,i] := Scale*U[k,i];
        end;// if i<>N
      end;// if i<=M

      W[i] := Scale*G;
      G := 0.0;
      S := 0.0;
      Scale := 0.0;

      if (i <= M) and (i <> N) then
      begin
        for k:=l to N  do
          Scale := Scale + abs(U[i,k]);
        if Scale <> 0.0 then
        begin
          for k:=l to N  do
          begin
            U[i,k] := U[i,k]/Scale;
            S:= S+sqr(U[i,k]);
          end;
          F := U[i,l];
          G := -Sign2(sqrt(S),F);
          H := F*G - S;
          U[i,l] := F - G;
          for k := l to N do
            RV1[k] := U[i,k]/H;
          if i <> M then
            for j := l to M do
            begin
              S := 0.0;
              for k := l to N do
                S := S+U[j,k]*U[i,k];
              for k := l to N do
                U[j,k] := U[j,k] + S*RV1[k];
            end;
          for k := l to N do
            U[i,k] := Scale*U[i,k];
        end;
    end;
    ANorm := Max(ANorm, abs(W[i]) + abs(RV1[i]));
  end; { of the loop  on  i }

  { Accumulation of the right-hand transformations  }
  if MatV then
    for i := N downto 1 do
    begin
      if i <> N then
      begin
        if G <> 0.0 then
        begin
          for j:=l to N do
            V[j,i] := (U[i,j]/U[i,l]) / G;
          for j := l to N do
          begin
            S := 0.0;
            for k:=l to N do
              S := S+U[i,k]*V[k,j];
            for k := l to N do
              V[k,j] := V[k,j]+S*V[k,i];
          end;
        end;
        for j:=l to N do
        begin
          V[i,j] := 0.0;
          V[j,i] := 0.0;
        end;
      end;

      V[i,i] := 1.0;
      G:= RV1[i];
      l:= i;

    end; // i

  {   Accumulation of the left-hand transformations   }
  if  MatU then
  begin
    for i := mn downto 1 do
    begin
      l := i+1;
      G := W[i];
      if i <> N then
        for j := l to N do
          U[i,j] := 0.0;
      if G <> 0.0 then
      begin
        if i <> mn then
//          for j:=l to N do
          for j := l to mn do
          begin
            S := 0.0;
            for k:=l to M do
              S := S+U[k,i]*U[k,j];
            F := (S/U[i,i])/G;
            for k := i to M do
              U[k,j] := U[k,j] + F*U[k,i];
          end;
        for j := i to M do
          U[j,i] := U[j,i]/G;
      end else
        for j:=i to M do
          U[j,i] := 0.0;
      U[i,i] := U[i,i]+1.0;
    end;
  end;

  { Diagonalization of the two-diagonal form. }
  for k := N downto 1 do
  begin
    k1 := k-1;
    its := 0;

    repeat
      ExitKey := 0;
      l := k + 1;
      while (ExitKey = 0) and (l > 1) do
      begin
        dec(l);
        l1 := l - 1;
        if  abs(RV1[l]) + ANorm = ANorm  then   ExitKey := 1
        else  if  l1 > 0 then
          if abs(W[l1]) + ANorm = ANorm  then   ExitKey := 2;
      end;

      if ExitKey <> 1 then
      begin
        C := 0.0;
        S := 1.0;
        ExitKey := 0;
        i := l;
        while (ExitKey = 0) and (i <= k) do
        begin
          F :=  S*RV1[i];
          RV1[i]:=  C*RV1[i];
          if abs(F) + ANorm = ANorm then
            ExitKey := 1
          else
          begin
            G := W[i];
            H := sqroot2(F,G);
            W[i] := H;
            C := G/H;
            S := -F/H;
            if  MatU then
              for j:=1 to M do
              begin
                Y := U[j,l1];
                Z := U[j,i];
                U[j,l1] :=  Y*C+Z*S;
                U[j,i]  := -Y*S+Z*C
              end;
            inc(i);
          end;
        end;
      end;

      { Convergence Checking }
      Z := W[k];
      if l <> k then
        begin
          if its = Itnlimit then
          begin
            RetCode := k;
            Exit;
          end;
          inc(its);
          X := W[l];
          Y := W[k1];
          G := RV1[k1];
          H := RV1[k];
          F := ((Y-Z)*(Y+Z) + (G-H)*(G+H)) / (2.0*H*Y);
          if abs(F)<=1.0 then   GG := Sign2(sqrt(F*F+1.0),F)
                           else   GG := F*sqrt(1.0+1.0/F/F);
          F:= ((X-Z)*(X+Z) + H*(Y/(F+GG)-H)) / X;

     { Next  QR - Transformation }
          C := 1.0;
          S := 1.0;
          for i1 := l to k1 do
          begin
            i := i1+1;
            G := RV1[i];
            Y := W[i];
            H := S*G;
            G := C*G;
            Z := sqroot2(F,H);
            RV1[i1] := Z;
            C := F/Z;
            S := H/Z;
            F := X*C+G*S;
            G := -X*S+G*C;
            H := Y*S;
            Y := Y*C;
            if MatV then
              for j:=1 to N do
              begin
                X := V[j,i1];
                Z := V[j,i];
                V[j,i1] :=  X*C+Z*S;
                V[j,i]  := -X*S+Z*C;
              end;

            Z := sqroot2(F,H);
            W[i1] := Z;
            if Z <> 0.0 then
            begin
              C := F/Z;
              S := H/Z;
            end;
            F := C*G+S*Y;
            X := -S*G+C*Y;
            if MatU then
              for J:=1 to M do
              begin
                Y := U[j,i1];
                Z := U[j,i];
                U[j,i1] :=  Y*C+Z*S;
                U[j,i]  := -Y*S+Z*C;
              end;
          end;

          RV1[l] := 0.0;
          RV1[k] := F;
          W[k]   := X;
        end

        else if Z < 0.0 then
        begin
          W[k] := -Z;
          if  MatV then
            for j:=1 to N do
              V[j,k] := -V[j,k];
        end;
      until  l=k;
  end; { of kk - loop }
end;

{
  Descending singular value sorting with sorting of U and V
  Based on: Sebastian Boung, Sorting Algorithms in Pascal
  http://privat.schlund.de/b/bossung/prog/delphi/sorting.html }
procedure SortSVD(const W: IFArr1D; const U, V: IFArr2D; MatU, MatV: boolean);

  procedure Quicksort(l, r: TInt);
  var
    i, j: TInt;
    x: TFloat;
  begin
    x := W[(l + r) div 2]; { Wert des Teilungselements }
    i := l; j := r;
    repeat
      while W[i] > x do
        inc(i);
      while x > W[j] do
        dec(j);
      if i <= j then
      begin
        W.Swap(i,j);
        if MatV then
//          V.SwapColumns(i,j);
          V.Swap(i,j,_Col);
        if MatU then
//          U.SwapColumns(i,j);
          U.Swap(i,j,_Col);
        inc(i);
        dec(j); { sonst kreuzen die sich nie }
      end;
    until i > j; { bis die Zeiger sich kreuzen }
    if l < j then Quicksort(l, j);
    if i < r then Quicksort(i, r);
  end;

begin
  Quicksort(W.Lo1, W.Hi1);
end;


procedure OrderSVD(M, N: integer; const U, V: IFArr2D; const W: IFArr1D;
                                                           MatU, MatV: boolean);
var
  i,k,j: integer;
  P: TFloat;
begin
  {  External loop of the re-ordering  }
  for i:=1 to N-1 do
    begin
      k := i;
      P := W[i];
      { Internal loop: finding of the index of the largest singular value  }
      for j:=i+1 to N do
        if  W[j]>P then
        begin
          k := j;
          P := W[j];
        end;

      if  k<>i then
        begin { swapping  of the singular value  }
          W[k] := W[i];
          W[i] := P;
          for j:=1 to M do
            begin { swapping of the U's columns }
             if  MatU then
              begin
                P:= U[j,i];
                U[j,i] := U[j,k];
                U[j,k] := P;
              end;
              { swapping of the V's columns }
              if  MatV then
              begin
                P:= V[j,i];
                V[j,i] := V[j,k];
                V[j,k] := P;
              end;
          end;
        end;
    end;
end;

procedure BiDiagSLE(d, e, b, x: IFArr1D);
var
  i, n: TInt;
begin
  n := d.Hi1;
  if d[n] = 0.0 then
    raise ESingularMatrix.CreateFmt('in BiDiagSLE at i = ',[n]);
  x[n] := b[n]/d[n];
  for i := n-1 downto 1 do
  begin
    if d[n] = 0.0 then
      raise ESingularMatrix.CreateFmt('in BiDiagSLE at i = ',[i]);
    x[i] := (b[i]-e[i+1]*x[i+1])/d[i];
  end;
end;

procedure TriDiagSLE(const a, b, c, d, x, temp: IFArr1D);
var
  t: TFloat;
  i, Lo: TInt;
begin
  // It is assumed that all arrays have the same Lim
  Lo := a.Lo1;
  // Decomposition and forward substitution
  if b[Lo] = 0.0 then
    raise ESingularMatrix.CreateFmt('in TriDiagSLE at i = ',[Lo]);
  t := b[Lo];
  x[Lo] := d[Lo]/t;
  for i := Lo+1 to a.Hi1 do
  begin
    temp[i] := c[i-1]/t;
    t := b[i] - a[i]*temp[i];
    if t = 0.0 then
      raise ESingularMatrix.CreateFmt('in TriDiagSLE at i = ',[i]);
    x[i] := (d[i]-a[i]*x[i-1])/t;
  end;
  // Backsubstitution
  for i := a.Hi1-1 downto Lo do
  begin
     x[i] := x[i] - temp[i+1]*x[i+1];
  end;
end;

procedure TriDiagSLE(const a, b, c, d, x: IFArr1D);
var
  temp: IFArr1D;
begin
  // It is assumed that all arrays have the same Lim
  temp := TFArr1D.Create(a);
  TriDiagSLE(a, b, c, d, x, temp);
end;

function TriDiagMxV(const a, b, c, x: IFArr1D): IFArr1D;
var
  i: TInt;
begin
  // It is assumed that all arrays have the same Lim
  result := TFArr1D.Create(a);
  i := a.Lo1;
  result[i] := b[i]*x[i] + c[i]*x[i+1];
  for i := a.Lo1+1 to a.Hi1-1 do
  begin
    result[i] := a[i]*x[i-1] + b[i]*x[i] + c[i]*x[i+1];
  end;
  i := a.Hi1;
  result[i] := a[i]*x[i-1] + b[i]*x[i];
end;


procedure CholDecomp(N: integer; const HDiag: IFArr1D; MaxOff: TFloat;
                        const L: IFArr2D; var MaxAdd: TFloat);
var
  i,j,k: integer;
  MinL, MinL2, S, MinLjj: TFloat;
  MaxOffl: TFloat;
begin
  MaxOffl := MaxOff;
  MinL:= sqrt(cSqrtMachEps)*MaxOffl;
  if MaxOffl=0.0 then
    begin
      for i:=1 to N do
        if abs(HDiag[i])>MaxOffl then
                    MaxOffl := abs(HDiag[i]);
      MaxOffl := sqrt(MaxOffl);
    end;

  MinL2:= sqrt(MachEps)*MaxOffl;
  MaxAdd := 0.0;
  for j:=1 to N do
    begin
      S := 0.0;
      if j>1 then
        for i:=1 to j-1 do
          S := S+sqr(L[j,i]);
      L[j,j] := HDiag[j]-S;
      MinLjj := 0.0;
      if j<N then
        for i:=j+1 to N do
          begin
            S := 0.0;
            if j>1 then
              for k:=1 to j-1 do
                S := S+L[i,k]*L[j,k];
            L[i,j] := L[j,i]-S;
            if abs(L[i,j])>MinLjj then
                   MinLjj := abs(L[i,j]);
          end;
      if MinLjj/MaxOffl>MinL then
                   MinLjj := MinLjj/MaxOffl
            else   MinLjj := MinL;
      if L[j,j]>sqr(MinLjj) then
                 L[j,j] := sqrt(L[j,j])
      else
      begin
        if MinL2>MinLjj then  MinLjj := MinL2;
        if sqr(MinLjj)-L[j,j]>MaxAdd then
          MaxAdd := sqr(MinLjj)-L[j,j];
        L[j,j] := MinLjj;
      end;

      if j<N then
        for i:=j+1 to N do
          L[i,j] := L[i,j]/L[j,j];
    end;
end;

procedure Cholesky(const L: IFArr2D);
var
  d: IFArr1D;
  Lo, Hi, i, j, k: TInt;
  sum: TFloat;
begin
  if not IsSquare(L) then
    raise ENonSquareMatrix.Create('Cholesky: Non-Square Matrix');

  Lo := L.Lo1;
  Hi := L.Hi1;
  d := TFArr1D.Create(L.Dim1);
  d.Lo1 := Lo;

  for i := Lo to Hi do
  begin
    for j := i to Hi do
    begin
//      sum := L[i,j]; does not use Upper triangle
      sum := L[j,i];
      for k := i-1 downto Lo do
        sum := sum - L[j,k]*L[i,k];
      if i = j then
      begin
        if sum < MinFloat then
          raise ESingularMatrix.CreateFmt('Cholesky: i = %d',[i]);
        d[i] := sqrt(sum);
        L[i,i] := d[i];
      end else
        L[j,i] := sum/d[i];
    end;
  end;
  
  for i := Lo to Hi-1 do
    for j := i+1 to Hi do
      L[i,j] := 0.0;
end;

procedure LSolve(const L: IFArr2D; const B, Y: IFArr1D);
var
  Lo, Hi, i, j: integer;
begin
  Lo := L.Lo1;
  Hi := L.Hi1;
  for i:=Lo to Hi do
    begin
      Y[i] := B[i];
      for j:=i-1 downto Lo do
        Y[i] := Y[i]-L[i,j]*Y[j];
      Y[i] := Y[i]/L[i,i];
    end;
end;

procedure LTSolve(const L: IFArr2D; const Y,X: IFArr1D);
var
  Lo, Hi, i, j: integer;
begin
  Lo := L.Lo1;
  Hi := L.Hi1;
  for i := Hi downto Lo do
    begin
      X[i] := Y[i];
      for j:=i+1 to Hi do
        X[i] := X[i]-L[j,i]*X[j];
      X[i] := X[i]/L[i,i];
    end;
end;

{ Solution of the equation L*LT*X = B by the  Cholesky  method }
procedure ChSolve (const L: IFArr2D; const B, X: IFArr1D);
begin
  if not IsSquare(L) then
    raise ENonSquareMatrix.Create('in ChSolve');
  if not SameLim(B,X) or (L.Lo1 <> B.Lo1) or (L.Hi1 <> B.Hi1) then
    Raise ELimMismatch.Create('in ChSolve');

  LSolve  (L, B, X);

  LTSolve (L, X, X);
end;

procedure CholBandDec (p: integer; const H,L: IFArr2D; var err: TInt);
var
  Lo, Hi, i,j,k,m: integer;
  p1,i1,j1: integer;
  S: TFloat;
begin
  if not IsSquare(L) then
  begin
    err := -1;
    exit;
  end else
    err := 0;

  Lo := L.Lo1;
  Hi := L.Hi1;

  p1 := p+1;
  for j := Lo to Hi do
    begin
      S:= 0.0;
      j1 := p+1-j;
      if j > Lo then
        for i:=Max(j-p,Lo) to j-1 do
          S := S+sqr(L[i+j1,i]);
      L[p1,j] := sqrt(H[p1,j]-S);
      if j < Hi then
        for i:=j+1 to Min(j+p, Hi) do
          begin
            S:= 0.0;
            i1 := p+1-i;
            m:= Max(i-p,Lo);
            if j>m then
              for k:=m to j-1 do
                S := S+L[k+i1,k]*L[k+j1,k];

            if L[p1,j] < MinFloat then
            begin
              err := 1;
              exit;
            end;
            L[j+i1,j] := (H[j+i1,j]-S)/L[p1,j];
          end;
    end;
end;

procedure LBandSolve (p: integer; const L: IFArr2D; const B,Y: IFArr1D; var err: TInt);
var
  Lo, Hi, i,j,p1,i1: integer;
  S: TFloat;
begin
  if not IsSquare(L) then
  begin
    err := -1;
    exit;
  end else
    err := 0;

  Lo := L.Lo1;
  Hi := L.Hi1;

  p1 := p+1;
  Y[Lo] := B[Lo]/L[p1,Lo];
  if  Hi > Lo then
    for i:= Lo+1 to Hi do
      begin
        S:= B[i];
        i1 := p+1-i;
        for j:=Max(i-p,Lo) to i-1 do
          S := S - L[j+i1,j]*Y[j];
        Y[i] := S/L[p1,i];
      end;
end;

procedure LTBandSolve(p: integer; const L: IFArr2D; const Y,X: IFArr1D; var err: TInt);
var
  Lo, Hi, i,j,p1,i1: integer;
begin
  if not IsSquare(L) then
  begin
    err := -1;
    exit;
  end else
    err := 0;

  Lo := L.Lo1;
  Hi := L.Hi1;

  p1 := p+1;
  X[Hi] := Y[Hi]/L[p1,Hi];
  if  Hi > Lo then
    for i:= Hi-1 downto Lo do
      begin
        X[i] := Y[i];
        i1:= p+i+1;
        for j:=i+1 to Min(i+p,Hi) do
          X[i] := X[i]-L[i1-j,i]*X[j];
        X[i] := X[i]/L[p1,i];
      end;
end;

procedure ChBandSolve (p: integer;  const L: IFArr2D; const G,S: IFArr1D; var err: TInt);
begin
  LBandSolve(p,L,G,S, err);
  if err <> 0 then
    exit;
  LTBandSolve(p,L,S,S, err);
end;

procedure QRDecomp(const M: IFArr2D; const N, D: IFArr1D; tol: TFloat = MinFloat);
var
  Lo, Hi, i,j,k: integer;
  Etha,Norm,Tau: TFloat;
begin
  if not IsSquare(M) then
    raise ENonSquareMatrix.Create('in QRDecomp');

  Lo := M.Lo1;
  Hi := M.Hi1;

  for k := Lo to Hi-1 do
  begin
    Etha := 0.0;
    for i:=k to Hi do
      if abs(M[i,k]) > Etha then
        Etha := abs(M[i,k]);
    if Etha < tol then
      begin
        {  degeneracy is found  }
        N[k] := 0.0;
        D[k] := 0.0;
        raise ESingularMatrix.Create('in QRDecomp');
      end else
      begin
        Norm := 0.0;
        for i:=k to Hi do
          begin
            M[i,k] := M[i,k]/Etha;
            Norm := Norm + sqr(M[i,k]);
          end;
        Norm := sqrt(Norm);
        if M[k,k] < 0.0 then  Norm := -Norm;
        M[k,k] := M[k,k] + Norm;
        N[k] := Norm*M[k,k]; // 0.5*norm2(u)
        D[k] := -Etha*Norm;  // u[k], first non-zero element
        for j:=k+1 to Hi do
          begin
            Tau := 0.0;
            for i:=k to Hi do
              Tau := Tau + M[i,k]*M[i,j];
            Tau := Tau/N[k];
            for i:=k to Hi do
              M[i,j] := M[i,j] - Tau*M[i,k];
          end;
      end;
  end;

  if abs(M[Hi,Hi]) < MinFloat then
    raise ESingularMatrix.Create('in QRDecomp');
  D[Hi] := M[Hi,Hi];

end;

procedure QRSolve (const M: IFArr2D; const N, D, B: IFArr1D);
var
  Hi, Lo, i,j: integer;
  Tau: TFloat;
begin
{ No checks for Lim }
  Lo := M.Lo1;
  Hi := M.Hi1;

  { 1. B := QT*B  } //  QSolve(M, B);
  for j:= Lo to Hi-1 do
  begin
    {  B := Qj*B  }
    Tau := 0.0;
    for i := j to Hi do
      Tau := Tau + M[i,j]*B[i];
    Tau := Tau/N[j];
    for i:= j to Hi do
      B[i] := B[i] - Tau*M[i,j];
  end;

  { 2. B := R**(-1)*B  } //  RSolve(M, D, B);
  B[Hi] := B[Hi]/D[Hi];
  for i := Hi-1 downto Lo do
  begin
    for j:=i+1 to Hi do
      B[i] := B[i]-M[i,j]*B[j];
    B[i] := B[i]/D[i];
  end;
end;

{ Direct use of GaussJordan for solving m*x = b }
procedure GJSolve(const m: IFArr2D; DoBalance: boolean; const x, b: IFArr1D; var err: TInt);
var
  a: IFArr2D;
  i: TInt;
begin
  err := -3;
  if not SameLim(x,b) then exit;
  err := -2;
  if (b.Lo1 <> m.Lo1) or (m.Hi1<>b.Hi1) then exit;
  err := 0;
  a := TFArr2D.Create(m.Lo1,m.Hi1,m.Lo2,m.Hi2+1);
  for i := a.Lo2 to a.Hi2-1 do
    CopyCC(a, i, m, i);
  a.SetSlice(_,a.Hi2,b); //  CopyArr1D(a, b, a.Hi2, _Col);
  GaussJordan(a, DoBalance, err);
  x.Lo1 := a.Lo1;
  x.Assign(a.GetSlice(_,a.Hi2)); // x.Assign(GetArr1D(a, a.Hi2, _Col));
  x.Lo1 := a.Lo2;
end;

procedure GaussJordan(const a: IFArr2D; DoBalance: boolean; var err: TInt);
var
 r, c, rr, rank: TInt;
  val: TFloat;
begin
  err := -1;
  if a.Dim1 >= a.Dim2 then exit;  // nothing to solve

  err := 0;
  rank := 1;
  if DoBalance then
    for r := a.Lo1 to a.Hi1 do
    begin // search only the matrix itself (not r.h.s.)
      c := MaxAbsIndexRow(a, val, r, a.Lo2, a.Lo2+a.Dim1-1);
      if abs(val) < MinFloat then
      begin // degeneracy
        err := rank;
        exit;
      end;
      inc(rank);
      // the whole row, including r.h.s.
      CxRow(a, r, 1.0/val);
    end;

  rank := 1;
  c := a.Lo2;
  for r := a.Lo1 to a.Hi1 do
  begin
    rr := MaxAbsIndexCol(a, val, c, r, a.Hi1);
    if abs(val) < MinFloat then
    begin // degeneracy
      err := rank;
      exit;
    end;
    if rr <> r then
      a.Swap(r,rr,_Row);
    CxRow(a,r,1.0/val,c,a.Hi2); // val = a[r,c]
    for rr := a.Lo1 to a.Hi1 do
      if rr <> r then
        RowAddRow(a,rr,r,-a[rr,c],c,a.Hi2);
    inc(c);
    inc(rank);
  end;
end;

procedure FastSolve(const a: IFArr2D; const b, x: IFArr1D);
var
  r, c, rank, rr: TInt;
  val: TFloat;
begin
  if (a.Lo1 <> b.Lo1) or (a.Hi1 > b.Hi1) or(a.Dim1 < a.Dim2) then
    raise ELimMismatch.Create('in FastSolve');
  rank := 1;
  c := a.Lo2;
  for r := a.Lo1 to a.Hi1 do
  begin
    rr := MaxAbsIndexCol(a, val, c, r, a.Hi1);
    if abs(val) < MinFloat then
      raise ESingularMatrix.CreateFmt('in FastSolve, rank = %d',[rank]);

    if rr <> r then
    begin
      a.Swap(r,rr,_Row);
      b.Swap(r,rr);
    end;
    CxRow(a,r,1.0/val);
    b[r] := b[r]/val;  // val = a[r,c]
    for rr := a.Lo1 to a.Hi1 do
      if rr <> r then
      begin
        val := -a[rr,c];
        RowAddRow(a, rr, r, val);
        b[rr] := b[rr] + val*b[r];
      end;
    if c = a.Hi2 then
      break;
    inc(c);
    inc(rank);
  end;

  x.Lo1 := a.Lo2;
  r := b.Lo1;
  { if x <> b then }
  for c := x.Lo1 to x.Hi1 do
  begin
    x[c] := b[r];
    inc(r);
  end;
end;

procedure MaskSolve(const m: IFArr2D; const x, b: IFArr1D; const Use: IBArr1D);
var
  r, c, c1, Nvar: TInt;
  a: IFArr2D;
begin
  Nvar := 0;
  for c := m.Lo2 to m.Hi2 do
    if Use[c] then inc(Nvar);

  if m.Dim2 < Nvar then
    raise ELimMismatch.Create('in MaskSolve: too fiew equations');

  c1 := 1;
  a := TFArr2D.Create(m.Dim1, Nvar);
  a.Lo1 := b.Lo1;

  for c := m.Lo2 to m.Hi2 do
    if not Use[c] then
    begin
      for r := m.Lo1 to m.Hi1 do
        b[r] := b[r] - m[r,c]*x[c];
    end else
    begin
      for r := a.Lo1 to a.Hi1 do
        a[r,c1] := m[r,c];
      inc(c1);
    end;

  FastSolve(a, b, b);

  c1 := b.Lo1;
  for c := x.Lo1 to x.Hi1 do
    if Use[c] then
    begin
      x[c] := b[c1];
      inc(c1);
    end;
end;

// Gram-Schmidt Orthonormalization of columns
procedure GramSchmidt(const U: IFArr2D; var m: TInt);
var
  imax, j, i, err: TInt;
  nmax, n: TFloat;
begin
  for i := U.Lo2 to U.Lo2+m-1 do
  begin
    nmax := 0.0;
    imax := i;
    for j := i to U.Hi2 do
    begin
      n := Norm2Col(U,j);
      if n > nmax then
      begin
        nmax := n;
        imax := j;
      end;
    end;

    if imax <> i then
      U.Swap(imax, i, _Col);

    NormalizeCol(U, i, err);
    if err <> 0 then
    begin
      m := i-1;
      break;
    end;
    for j := i+1 to m do
    begin
      nmax := ColxCol(U, i, j);
      ColAddCol(U, j, i, -nmax)
    end;
  end;
end;

procedure householder(const a: IFArr2D; const q, e: IFArr1D; const u, v: IFArr2D;
                      eps, tol: TFloat; fullmatrix: boolean = false);
label
  1, 2, 3, 4;
var
  i, j, k, l, n, m, nm: TInt;
  f, g, h, s, x, y: TFloat;
//  e: TFArr1D;
begin
//  u.Assign(a, false);
  AssignMatrix(u, a, false);
  // Householder (bidiagonal) reduction
  x := 0.0;
  g := 0.0;
  m := a.Dim1;
  n := a.Dim2;

  for i := 1 to n do
  begin
    e[i] := g;
    l := i + 1;
    s := 0.0;
    for j := i to m do
      s := s + sqr(u[j,i]);
    if s < tol then g := 0.0 else
    begin
      f := u[i,i];
      g := -Sign2(sqrt(s), f); // if f < 0 then g := sqrt(s) else g := -sqrt(s)
      h := f*g - s;
      u[i,i] := f - g;
      for j := l to n do
      begin
        s := 0.0;
        for k := i to m do
          s := s + u[k,i]*u[k,j];
        f := s/h;
        for k := i to m do
          u[k,j] := u[k,j] + f*u[k,i];
      end;// j
    end;// s
    q[i] := g;

    s := 0.0;
    for j := l to n do
      s := s + sqr(u[i,j]);
    if s < tol then g := 0.0 else
    begin
      f := u[i,i+1];
      g := -Sign2(sqrt(s), f); // if f < 0 then g := sqrt(s) else g := -sqrt(s)
      h := f*g - s;
      u[i,i+1] := f-g;
      for j := l to n do
        e[j] := u[i,j]/h;
      for j := l to m do
      begin
        s := 0.0;
        for k := l to n do
          s := s + u[j,k]*u[i,k];
        for k := l to n do
          u[j,k] := u[j,k] + s*e[k];
      end;// j
    end;// s
    y := abs(q[i]) + abs(e[i]);
    x := max(x,y);

  end;// i
    for j := 1 to n do
      for i := 1 to n do
        v[j,i] := u[i,j];

      for i := 1 to n do
        v[i,i] := q[i];

  // right transformations matrix formation
  for i := n downto 1 do
  begin
    if g <> 0.0 then
    begin
      h := u[i,i+1]*g;
      for j := l to n do
        v[j,i] := u[i,j]/h;
      for j := l to n do
      begin
        s := 0.0;
        for k := l to n do
          s := s + u[i,k]*v[k,j];
        for k := l to n do
          v[k,j] := v[k,j] + s*v[k,i];
      end; //j
    end;// g
    for j := l to n do
    begin
      v[i,j] := 0.0;
      v[j,i] := 0.0;
    end;
    v[i,i] := 1.0;
    g := e[i];
    l := i;
  end;// i

  // left transformations matrix formation
  if fullmatrix then
  begin
    nm := m;
    for i := n+1 to m do
    begin
      for j := n+1 to m do
        u[i,j] := 0.0;
      u[i,i] := 1.0;
    end;
  end else
    nm := n;

  for i := n downto 1 do
  begin
    l := i + 1;
    g := q[i];
    for j := l to nm do
      u[i,j] := 0.0;
    if g <> 0.0 then
    begin
      h := u[i,i]*g;
      for j := l to nm do
      begin
        s := 0.0;
        for k := l to m do
          s := s + u[k,i]*u[k,j];
        f := s/h;
        for k := i to m do
          u[k,j] := u[k,j] + u[k,i]*f;
      end;// j
      for j := i to m do
        u[j,i] := u[j,i]/g;
    end else
      for j := 1 to m do
        u[j,i] := 0.0;
    // g
    u[i,i] := u[i,i] + 1.0;
  end;// i
end;

procedure xsvd(eps, tol: TFloat; const a: IFArr2D; const q: IFArr1D;
               const u, v: IFArr2D; withu: boolean = true;
               withv: boolean = true; fullmat: boolean = false);
label
  1, 2, 3, 4;
var
  i, j, k, l, l1, n, m, nm: TInt;
  c, f, g, h, s, x, y, z, t: TFloat;
  e: TFArr1D;
  r1, r2: IRestore;
begin
  m := a.Dim1;
  n := a.Dim2;
  if m < n then
    raise ELimMismatch.Create('In xsvd: Dim1 < Dim2)');
  r2 := TRestore2D.Create([a,u,v]);
  r1 := TRestore1D.Create([q]);
//  u.Assign(a, false);
  AssignMatrix(u, a, false);
  // Householder (bidiagonal) reduction
  x := 0.0;
  g := 0.0;
  e := TFArr1D.Create(n);
  for i := 1 to n do
  begin
    e[i] := g;
    l := i + 1;
    s := 0.0;
    for j := i to m do
      s := s + sqr(u[j,i]);
//    s := U.Norm2(_,i); { = Norm2Col(U, i) }
    if s < tol then g := 0.0 else
    begin
      f := u[i,i];
      g := -Sign2(sqrt(s), f); // if f < 0 then g := sqrt(s) else g := -sqrt(s)
      h := f*g - s;
      u[i,i] := f - g;
      for j := l to n do
      begin
        s := 0.0;
        for k := i to m do
          s := s + u[k,i]*u[k,j];
        f := s/h;
        for k := i to m do
          u[k,j] := u[k,j] + f*u[k,i];
      end;// j
    end;// s
    q[i] := g;
    s := 0.0;
    for j := l to n do
      s := s + sqr(u[i,j]);
    if s < tol then g := 0.0 else
    begin
      f := u[i,i+1];
      g := -Sign2(sqrt(s), f); // if f < 0 then g := sqrt(s) else g := -sqrt(s)
      h := f*g - s;
      u[i,i+1] := f-g;
      for j := l to n do
        e[j] := u[i,j]/h;
      for j := l to m do
      begin
        s := 0.0;
        for k := l to n do
          s := s + u[j,k]*u[i,k];
        for k := l to n do
          u[j,k] := u[j,k] + s*e[k];
      end;// j
    end;// s
    y := abs(q[i]) + abs(e[i]);
    x := max(x,y);
  end;// i

  // right transformations matrix formation
  if withv then
    for i := n downto 1 do
    begin
      if g <> 0.0 then
      begin
        h := u[i,i+1]*g;
        for j := l to n do
          v[j,i] := u[i,j]/h;
        for j := l to n do
        begin
          s := 0.0;
          for k := l to n do
            s := s + u[i,k]*v[k,j];
          for k := l to n do
            v[k,j] := v[k,j] + s*v[k,i];
        end; //j
      end;// g
      for j := l to n do
      begin
        v[i,j] := 0.0;
        v[j,i] := 0.0;
      end;
      v[i,i] := 1.0;
      g := e[i];
      l := i;
    end;// i

  // left transformations matrix formation
  if withu then
  begin
    if fullmat then
    begin
      nm := m;
      for i := n+1 to m do
      begin
        for j := n+1 to m do
          u[i,j] := 0.0;
        u[i,i] := 1.0;
      end;
    end else
      nm := n;

    for i := n downto 1 do
    begin
      l := i + 1;
      g := q[i];
      for j := l to nm do
        u[i,j] := 0.0;
      if g <> 0.0 then
      begin
        h := u[i,i]*g;
        for j := l to nm do
        begin
          s := 0.0;
          for k := l to m do
            s := s + u[k,i]*u[k,j];
          f := s/h;
          for k := i to m do
            u[k,j] := u[k,j] + u[k,i]*f;
        end;// j
        for j := i to m do
          u[j,i] := u[j,i]/g;
      end else
        for j := 1 to m do
          u[j,i] := 0.0;
      // g
      u[i,i] := u[i,i] + 1.0;
    end;// i
  end;

  // bidiagonal matrix diagonalization
  eps := eps*x;
  for k := n downto 1 do
  begin
  1: // test f splitting
     for l := k downto 1 do
     begin
//       l := ll;
       if abs(e[l]) <= eps then
         goto 3; // 3: test f convergence
       if l > 1 then
         if abs(q[l-1]) <= eps then
           goto 2; // 2: cancellation
     end;// l
  2: // cancellation
    c := 0.0;
    s := 1.0;
    l1 := l - 1;
  if l > 1 then
    for i := l to k do
    begin
      f := s*e[i];
      e[i] := c*e[i];
      if abs(f) <= eps then
        goto 3; // 3: test f convergence
      g := q[i];
      h := sqroot2(f,g); // safe sqrt(f*f+g*g)
      q[i] := h;
      c := g/h;
      s := -f/h;
      if withu then
        for j := 1 to m do
        begin
          y := u[j,l1];
          z := u[j,i];
          u[j,l1] := y*c +z*s;
          u[j,i] := -y*s +z*c;
        end;// j
    end;// i
  3: // test f convergence
    z := q[k];
    if l = k then
      goto 4;  // 4: convergence
    // QR transform shift formation
    x := q[l];
    y := q[k-1];
    g := e[k-1];
    h := e[k];
    f := ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
    g := sqroot2(f,1.0);
    if f < 0.0 then t := f-g else t := f+g;
    f := ((x-z)*(x+z)+h*(y/t-h))/x;
    // next QR step
    c := 1.0;
    s := 1.0;
    for i := l+1 to k do
    begin
      g := e[i];
      y := q[i];
      h := s*g;
      g := c*g;
      z := sqroot2(f,h);
      e[i-1] := z;
      c := f/z;
      s := h/z;
      f := x*c+g*s;
      g := -x*s+g*c;
      h := y*s;
      y := y*c;
      if withv then
        for j := 1 to n do
        begin
          x := v[j,i-1];
          z := v[j,i];
          v[j,i-1] := x*c+z*s;
          v[j,i] := -x*s+z*c;
        end;// j
      z := sqroot2(f,h);
      q[i-1] := z;
      c := f/z;
      s := h/z;
      f := c*g+s*y;
      x := -s*g+c*y;
      if withu then
        for j := 1 to m do
        begin
          y := u[j,i-1];
          z := u[j,i];
          u[j,i-1] := y*c+z*s;
          u[j,i] := -y*s+z*c;
        end;// j
    end;//i
    e[l] := 0.0;
    e[k] := f;
    q[k] := x;
    goto 1; // 1: test f splitting
  4: // convergence
    if z < 0.0 then
    begin
      q[k] := -z;
      if withv then
        for j := 1 to n do
          v[j,k] := -v[j,k];
    end;
  end;// k

end;

end.
