{*
 Object algebra routines
 Nikolai V. Shokhirev, 2002-2007
 @author Nikolai Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2002.06.06 - created
 @version 2004.11.11 - expanded
 @version 2005.05.28 agged IdentityMt
 @version 2007.07.15 - Added DelphiCodeToDoc comments
}
unit uDynObjAlg;

interface

uses
  uMatTypes, uDynArrays, uDynLinAlg;

/// Identity Matrix generator 
function IdentityMt(const Dim: TInt; Lo: TInt = 1): IFArr2D;

/// Matrix transposition 
function Transpose(const M: IFArr2D): IFArr2D; overload;
/// Matrix transposition 
function Transpose(const M: ICArr2D): ICArr2D; overload;

/// SumVt = c1*V1 + c2*V2 + . . . 
function SumVt(cc: array of TFloat; const VV: array of IFArr1D): IFArr1D;
/// SumMt = c1*M1 + c2*M2 + . . . 
function SumMt(cc: array of TFloat; const MM: array of IFArr2D): IFArr2D;

/// M := M + d 
function MtShiftDiag(const M: IFArr2D; const d: TFloat): IFArr2D;
/// M := M + D 
function MtAddDiag(const M: IFArr2D; const D: IFArr1D): IFArr2D;

/// Dot (scalar) product of two vectors VtxVt = VL*VR 
function VtxVt(const VL, VR: IFArr1D): TFloat;
/// vector-matrix product: VtxMt = V*M = MT*V 
function VtxMt(const VL: IFArr1D; const MR: IFArr2D): IFArr1D;
/// matrix-vector product: MtxVt = M*V = V*MT 
function MtxVt(const ML: IFArr2D; const VR: IFArr1D): IFArr1D;

/// matrix-matrix product: MtxMt = ML*MR 
function MtxMt(const ML, MR: IFArr2D): IFArr2D;
/// matrix-matrix product: MtTxMt = Transpose(ML)*MR  
function MtTxMt(const ML, MR: IFArr2D): IFArr2D;
/// matrix-matrix product: MtxMtT = ML*Transpose(MR) 
function MtxMtT(const ML, MR: IFArr2D): IFArr2D;
/// matrix-matrix product: MtTxMtT = Transpose(ML)*Transpose(MR) 
function MtTxMtT(const ML, MR: IFArr2D): IFArr2D;

/// matrix-matrix product: DxMt = DiagonalMatrix(DL)*MR 
function DxMt(const DL: IFArr1D; const MR: IFArr2D): IFArr2D;
/// matrix-matrix product: MtxD = ML*DiagonalMatrix(DR) 
function MtxD(const ML: IFArr2D; const DR: IFArr1D): IFArr2D;

/// matrix-matrix product: MtxDxMtT = M*DiagonalMatrix(D)*Transpose(M) 
function MtxDxMtT(const M: IFArr2D; const D: IFArr1D): IFArr2D;
/// matrix-matrix product: MtxDxMtT = Transpose(M)*DiagonalMatrix(D)*M 
function MtTxDxMt(const M: IFArr2D; const D: IFArr1D): IFArr2D;
/// matrix-matrix product: Mt1xDxMt2T = M*DiagonalMatrix(D)*Transpose(M) 
function Mt1xDxMt2T(const M1: IFArr2D; const D: IFArr1D; const M2: IFArr2D): IFArr2D;
/// matrix-matrix product: Mt1TxDxMt2 = Transpose(M)*DiagonalMatrix(D)*M 
function Mt1TxDxMt2(const M1: IFArr2D; const D: IFArr1D; const M2: IFArr2D): IFArr2D;

/// Complex Algebra routines:  

/// complex Conjugate 
function Conjugate(const A: ICArr1D): ICArr1D; overload;
/// complex Conjugate 
function Conjugate(const A: ICArr2D): ICArr2D; overload;
/// complex Hermitian Conjugate = Transpose (Conjugate(A)) 
function HConjugate(const A: ICArr2D): ICArr2D;

/// Dot (scalar) product of two vectors CVtxCVt=VL*VR 
function CVtxCVt(const VL, VR: ICArr1D): Complex;
/// vector-Vector product: CVtHxVRt = HConjugate(VL)*VR 
function CVtHxCVt(const VL, VR: ICArr1D): Complex;

/// vector-matrix product: CVtxCMt = V*M = MT*V  
function CVtxCMt(const VL: ICArr1D; const MR: ICArr2D): ICArr1D;
/// matrix-vector product: CMtxCVt = M*V = V*MT 
function CMtxCVt(const ML: ICArr2D; const VR: ICArr1D): ICArr1D;

/// matrix-matrix product: CMtxCMt = ML*MR 
function CMtxCMt(const ML, MR: ICArr2D): ICArr2D;

/// matrix-matrix product: DxCMt = DiagonalMatrix(DL)*MR 
function DxCMt(const DL: IFArr1D; const MR: ICArr2D): ICArr2D;
/// matrix-matrix product: CMtxD = ML*DiagonalMatrix(DR) 
function CMtxD(const ML: ICArr2D; const DR: IFArr1D): ICArr2D;

/// Diagonalization of the real symmetric matrix a. 
function DiagMt(const a: IFArr2D; var code: TInt;
  sort: boolean = true; vectors: boolean = true; Ascending: boolean = true;
  Epsilon : TFloat = SqrtMinFloat; IterMax: TInt = 999): IEigenSys;

/// result := MtxDxMtT(ES.Vectors,ES.Values)  
function EigenSysToMt(const ES: IEigenSys): IFArr2D;

{*
  SVD for an arbitrary matrix a:                               
  A[aLo1..aHi1, aLo2..aHi2],  m = A.Dim1, n = A.Dim2, nm = min(n,m)                              
  fullmat = false (default)                                          
    A   =  U    *   Q    *  VT        UT  *  A  *   V    =   Q       
  m x n  m x nm  nm x nm  nm x n   nm x m  m x n  n x nm  nm x nm    
  U[aLo1..aHi1, 1..nm], V[aLo2..aHi2, 1..nm2], Q[1..nm]                                                                    
  fullmat = true                                                     
    A   =  U  *   Q   *  VT      UT  *  A  *   V   =  Q              
  m x n  m x m  m x n  n x n   m x m  m x n  n x n  m x n            
  U[aLo1..aHi1, 1..m], V[aLo2..aHi2, 1..n]                           
  Q[1, nm] - first nm values                                       
}
function SVDMt(const a: IFArr2D; fullmat: boolean = false;
               eps : TFloat = cMachEps; tol : TFloat = MinFloat): ISVDSys;

/// Inverse to SVDMt: A = U * Q * VT  
function SVDSysToMt(const SVDS: ISVDSys): IFArr2D;

{*
 author: Nikolai V. Shokhirev 2002      
  Diagonalization of complex Hermitian matrix 
}
function DiagHMt(const a: ICArr2D; var code: TInt; sort: boolean = true;
                vectors: boolean = true; Ascending: boolean = true): IHEigenSys;

/// result := MtxDxMtH(HES.Vectors,HES.Values)  
function HEigenSysToMt(const HES: IHEigenSys): ICArr2D;

{*
 author: Nikolai V. Shokhirev 2004                            
  Diagonalization of complex Hermitian matrix                       
 Moore-Penrose Matrix Inverse (Pseudoinverse)                       
Input:                                                              
  a - m x n arbitrary rectangular matrix; mn = min(m,n)             
  SVD: a = Sum(i=1,nm)[wi*|ui><vi|]                                 
Pseudoinverse:                                                      
  a~ = Sum(j=1,k)[1/(wi+s)*|ui><vi|]  wi > e;  e, s >= 0;  k <= mn  
                                                                    
  Moore-Penrose Inverse takes place when  e = s = 0.                
  Properties: a*(a~*a)=a  (a*a~)*a = a                              
  For non-singular matrix a~ = a^(-1)  (or a**(-1) )                
                                                                    
  The parameters e, s > 0;  k < mn provide regularization
                                           (stabilization)          
}
function PseudoinverseMt(const a: IFArr2D; k: TInt = _; e : TFloat = MinFloat;
                                                    s : TFloat = 0.0): IFArr2D;
implementation

uses
  SysUtils, Math, uDynArrUtils;

ResourceString
  /// 'Lim Mismatch';
  RS_LimMismatch = 'Lim Mismatch';
  /// 'Matrix is not square';
  RS_NotSquare = 'Matrix is not square';

function IdentityMt(const Dim: TInt; Lo: TInt): IFArr2D;
var
  t: IFArr2D;
  i: TInt;
begin
  t := TFArr2D.Create(Lo,Dim+Lo-1, Lo,Dim+Lo-1);
  for i := t.Lo1 to t.Hi1 do t[i,i] := 1.0;
  result := t;
end;

function Transpose(const M: IFArr2D): IFArr2D;
var
  t: IFArr2D;
begin
  t := TFArr2D.Create(M.Lo2,M.Hi2, M.Lo1,M.Hi1);
  MtT(M, t);
  result := t;
end;

{ SumVt = c1*V1 + c2*V2 }
function SumVt(cc: array of TFloat; const VV: array of IFArr1D): IFArr1D;
var
  i1, j, M: TInt;
  t: IFArr1D;
  s: TFloat;
begin
  result := nil;
  M := Length(cc);
  if M<>Length(VV) then
    Raise ERangeError.Create(RS_LimMismatch);

  for j := 1 to M-1 do
    if not SameLim(VV[0], VV[j]) then
      Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr1D.Create(VV[0]);
  for i1 := t.Lo1 to t.Hi1 do
  begin
    s := 0.0;
    for j := 0 to M-1 do
      s := s + cc[j]*VV[j].Value[i1];
    t[i1] := s;
  end;
  result := t;
end;

{ SumMt = c1*M1 + c2*M2  }
function SumMt(cc: array of TFloat; const MM: array of IFArr2D): IFArr2D;
var
  i1, i2, j, M: TInt;
  t: IFArr2D;
  s: TFloat;
begin
  result := nil;
  M := Length(cc);
  if M<>Length(MM) then
    Raise ERangeError.Create(RS_LimMismatch);

  for j := 1 to M-1 do
    if not SameLim(MM[0], MM[j]) then
      Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr2D.Create(MM[0]);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      s := 0.0;
      for j := 0 to M-1 do
        s := s + cc[j]*MM[j].Value[i1,i2];
    t[i1,i2] := s;
    end;
  result := t;
end;

function MtShiftDiag(const M: IFArr2D; const d: TFloat): IFArr2D;
var
  i: TInt;
  t: IFArr2D;
begin
  if not IsSquare(M) then
    Raise ERangeError.Create(RS_NotSquare);
  t := TFArr2D.Create(M,true);
  for i := t.Lo1 to t.Hi1 do
    t[i,i] := t[i,i] + d;
  result := t;
end;

function MtAddDiag(const M: IFArr2D; const D: IFArr1D): IFArr2D;
var
  i: TInt;
  t: IFArr2D;
begin
  if not IsSquare(M) or (M.Lo1<>D.Lo1) or ( M.Hi1<>D.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TFArr2D.Create(M,true);
  for i := t.Lo1 to t.Hi1 do
    t[i,i] := t[i,i] + D[i];
  result := t;
end;

{ Dot (scalar) product of two vectors }
function VtxVt(const VL, VR: IFArr1D): TFloat;
var
  i1: TInt;
begin
  result := 0.0;
  Assert(SameLim(VL, VR), RS_LimMismatch);
  for i1 := VL.Lo1 to VL.Hi1 do
    result := result + VL[i1]*VR[i1];
end;

{ VtxMt = V*M = MT*V }
function VtxMt(const VL: IFArr1D; const MR: IFArr2D): IFArr1D;
var
  i1,i2: TInt;
  s: TFloat;
  t: IFArr1D;
begin
  Assert( (MR.Lo1=VL.Lo1) and ( MR.Hi1=VL.Hi1), RS_LimMismatch);

  t := TFArr1D.Create(MR.Lo2,MR.Hi2);
  for i2 := MR.Lo2 to MR.Hi2 do
  begin
    s := 0.0;
    for i1 := MR.Lo1 to MR.Hi1 do
      s := s + VL[i1]*MR[i1,i2];
    t[i2] := s;
  end;
  result := t;
end;

{ MtxVt = M*V = V*MT  }
function MtxVt(const ML: IFArr2D; const VR: IFArr1D): IFArr1D;
var
  i1,i2: TInt;
  s: TFloat;
  t: IFArr1D;
begin
  if (ML.Lo2<>VR.Lo1) or (ML.Hi2<>VR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TFArr1D.Create(ML.Lo1,ML.Hi1);
  for i1 := ML.Lo1 to ML.Hi1 do
  begin
    s := 0.0;
    for i2 := ML.Lo2 to ML.Hi2 do
      s := s + ML[i1,i2]*VR[i2];
    t[i1] := s;
  end;
  result := t;
end;

{ MtxMt = ML*MR }
function MtxMt(const ML, MR: IFArr2D): IFArr2D;
var
  i1, i2, j: TInt;
  s: TFloat;
  t: IFArr2D;
begin
  if (ML.Lo2<>MR.Lo1) or (ML.Hi2<>MR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TFArr2D.Create(ML.Lo1,ML.Hi1,MR.Lo2,MR.Hi2);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      s := 0.0;
      for j := ML.Lo2 to ML.Hi2 do
        s := s + ML[i1,j]*MR[j,i2];
      t[i1,i2] := s;
    end;
  result := t;
end;

{ MtTxMt = Transpose(ML)*MR }
function MtTxMt(const ML, MR: IFArr2D): IFArr2D;
var
  i1, i2, j: TInt;
  s: TFloat;
  t: IFArr2D;
begin
  Assert( (ML.Lo1=MR.Lo1) and (ML.Hi1=MR.Hi1), RS_LimMismatch);

  t := TFArr2D.Create(ML.Lo2,ML.Hi2,MR.Lo2,MR.Hi2);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      s := 0.0;
      for j := ML.Lo1 to ML.Hi1 do
        s := s + ML[j,i1]*MR[j,i2];
      t[i1,i2] := s;
    end;
  result := t;
end;

{ MtxMtT = ML*Transpose(MR) }
function MtxMtT(const ML, MR: IFArr2D): IFArr2D;
var
  i1, i2, j: TInt;
  s: TFloat;
  t: IFArr2D;
begin
  if (ML.Lo2<>MR.Lo2) or (ML.Hi2<>MR.Hi2) then
    Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr2D.Create(ML.Lo1,ML.Hi1,MR.Lo1,MR.Hi1);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      s := 0.0;
      for j := ML.Lo2 to ML.Hi2 do
        s := s + ML[i1,j]*MR[i2,j];
      t[i1,i2] := s;
    end;
  result := t;
end;

{ MtTxMtT = Transpose(ML)*Transpose(MR) }
function MtTxMtT(const ML, MR: IFArr2D): IFArr2D;
var
  i1, i2, j: TInt;
  s: TFloat;
  t: IFArr2D;
begin
  if (ML.Lo1<>MR.Lo2) or (ML.Hi1<>MR.Hi2) then
    Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr2D.Create(ML.Lo2,ML.Hi2,MR.Lo1,MR.Hi1);
  t.Lo1 := ML.Lo2;
  t.Lo2 := MR.Lo1;
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      s := 0.0;
      for j := ML.Lo1 to ML.Hi1 do
        s := s + ML[j,i1]*MR[i2,j];
      t[i1,i2] := s;
    end;
  result := t;
end;

{ DxMt = DiagonalMatrix(DL)*MR  }
function DxMt(const DL: IFArr1D; const MR:IFArr2D): IFArr2D;
var
  i1, i2: TInt;
  t: IFArr2D;
begin
  if (DL.Lo1<>MR.Lo1) or (DL.Hi1<>MR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr2D.Create(DL.Lo1,DL.Hi1,MR.Lo2,MR.Hi2);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
      t[i1,i2] := DL[i1]*MR[i1,i2];
  result := t;
end;

{ MtxD = ML*DiagonalMatrix(DR) }
function MtxD(const ML: IFArr2D; const DR: IFArr1D): IFArr2D;
var
  i1, i2: TInt;
  t: IFArr2D;
begin
  if (ML.Lo2<>DR.Lo1) or (ML.Hi2<>DR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);

  t := TFArr2D.Create(ML);
  for i2 := t.Lo2 to t.Hi2 do
    for i1 := t.Lo1 to t.Hi1 do
      t[i1,i2] := ML[i1,i2]*DR[i2];
  result := t;
end;

function MtxDxMtT(const M: IFArr2D; const D: IFArr1D): IFArr2D;
var
  t: IFArr2D;
begin
  t := MtxD(M, D);
  result := MtxMtT(t, M);
end;

function MtTxDxMt(const M: IFArr2D; const D: IFArr1D): IFArr2D;
var
  t: IFArr2D;
begin
  t := DxMt(D, M);
  result := MtTxMt(M, t);
end;

function Mt1xDxMt2T(const M1: IFArr2D; const D: IFArr1D; const M2: IFArr2D): IFArr2D;
var
  t: IFArr2D;
begin
  t := MtxD(M1, D);
  result := MtxMtT(t, M2);
end;

function Mt1TxDxMt2(const M1: IFArr2D; const D: IFArr1D; const M2: IFArr2D): IFArr2D;
var
  t: IFArr2D;
begin
  t := DxMt(D, M2);
  result := MtTxMt(M1, t);
end;

{ Implementation of Complex Algebra routines }

{ Transpose(M) }
function Transpose(const M: ICArr2D): ICArr2D;
var
  t: ICArr2D;
begin
  t := TCArr2D.Create(M.Lo2,M.Hi2, M.Lo1,M.Hi1);
  MtT(M.Re, t.Re);
  MtT(M.Im, t.Im);
  result := t;
end;

{ complex Conjugate }
function Conjugate(const A: ICArr1D): ICArr1D;
begin
  result := TCArr1D.Create(A,true);
  result.Im.Times(-1.0);
end;

{ complex Conjugate }
function Conjugate(const A: ICArr2D): ICArr2D; overload;
begin
  result := TCArr2D.Create(A,true);
  result.Im.Times(-1.0);
end;

{ complex Hermitian Conjugate = Transpose (Conjugate(A)) }
function HConjugate(const A: ICArr2D): ICArr2D;
var
  t: ICArr2D;
begin
  t := Transpose(A);
  result := Conjugate(t);
end;

{ Dot (scalar) product of two vectors }
function CVtxCVt(const VL, VR: ICArr1D): Complex;
begin
  Result.Re := VtxVt(VL.Re, VR.Re)-VtxVt(VL.Im, VR.Im);
  Result.Im := VtxVt(VL.Re, VR.Im)+VtxVt(VL.Im, VR.Re);
end;

{ vector-Vector product: CVtHxVRt = HConjugate(VL)*VR }
function CVtHxCVt(const VL, VR: ICArr1D): Complex;
begin
  Result.Re := VtxVt(VL.Re, VR.Re)+VtxVt(VL.Im, VR.Im);
  Result.Im := VtxVt(VL.Re, VR.Im)-VtxVt(VL.Im, VR.Re);
end;

{ CVtxCMt = V*M = MT*V }
function CVtxCMt(const VL: ICArr1D; const MR: ICArr2D): ICArr1D;
var
  t: ICArr1D;
  j, k: TInt;
  rsum, isum: TFloat;
begin
  if (VL.Lo1<>MR.Lo1) or (VL.Re.Hi1<>MR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TCArr1D.Create(MR.Lo2,MR.Hi2);
  for j := t.Lo1 to t.Hi1 do
    begin
      rsum := 0.0;
      isum := 0.0;
      for k := VL.Lo1 to VL.Hi1 do
      begin
        rsum := rsum + VL.Re[k]*MR.Re[k,j]-VL.Im[k]*MR.Im[k,j];
        isum := isum + VL.Re[k]*MR.Im[k,j]+VL.Im[k]*MR.Re[k,j];
      end;
      t.Re[j] := rsum;
      t.Im[j] := isum;
    end;
  result := t;
end;

{ CMtxCVt = M*V = V*MT }
function CMtxCVt(const ML: ICArr2D; const VR: ICArr1D): ICArr1D;
var
  t: ICArr1D;
  i, k: TInt;
  rsum, isum: TFloat;
begin
  if (ML.Lo2<>VR.Lo1) or (ML.Hi2<>VR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TCArr1D.Create(ML.Lo1,ML.Hi1);
  for i := t.Re.Lo1 to t.Re.Hi1 do
  begin
    rsum := 0.0;
    isum := 0.0;
    for k := ML.Re.Lo2 to ML.Re.Hi2 do
    begin
      rsum := rsum + ML.Re[i,k]*VR.Re[k]-ML.Im[i,k]*VR.Im[k];
      isum := isum + ML.Re[i,k]*VR.Im[k]+ML.Im[i,k]*VR.Re[k];
    end;
    t.Re[i] := rsum;
    t.Im[i] := isum;
  end;
  result := t;
end;

{ CMtxCMt = ML*MR  }
function CMtxCMt(const ML, MR: ICArr2D): ICArr2D;
var
  t: ICArr2D;
  i, j, k: TInt;
  rsum, isum: TFloat;
begin
  if (ML.Lo2<>MR.Lo1) or (ML.Hi2<>MR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TCArr2D.Create(ML.Lo1,ML.Hi1,MR.Lo2,MR.Hi2);
  for i := t.Lo1 to t.Hi1 do
    for j := t.Lo2 to t.Hi2 do
    begin
      rsum := 0.0;
      isum := 0.0;
      for k := ML.Lo2 to ML.Hi2 do
      begin
        rsum := rsum + ML.Re[i,k]*MR.Re[k,j]-ML.Im[i,k]*MR.Im[k,j];
        isum := isum + ML.Re[i,k]*MR.Im[k,j]+ML.Im[i,k]*MR.Re[k,j];
      end;
      t.Re[i,j] := rsum;
      t.Im[i,j] := isum;
    end;
  result := t;
end;

{ DxCMt = DiagonalMatrix(DL)*MR }
function DxCMt(const DL: IFArr1D; const MR: ICArr2D): ICArr2D;
var
  i1, i2: TInt;
  t: ICArr2D;
begin
  if (DL.Lo1<>MR.Lo1) or (DL.Hi1<>MR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);

  t := TCArr2D.Create(DL.Lo1,DL.Hi1,MR.Lo2,MR.Hi2);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      t.Re[i1,i2] := DL[i1]*MR.Re[i1,i2];
      t.Im[i1,i2] := DL[i1]*MR.Im[i1,i2];
    end;
  result := t;
end;

{ CMtxD = ML*DiagonalMatrix(DR) }
function CMtxD(const ML: ICArr2D; const DR: IFArr1D): ICArr2D;
var
  t: ICArr2D;
  i1, i2: TInt;
begin
  if (ML.Lo2<>DR.Lo1) or (ML.Hi2<>DR.Hi1) then
    Raise ERangeError.Create(RS_LimMismatch);
  t := TCArr2D.Create(ML.Lo1,ML.Hi1,ML.Lo2,ML.Hi2);
  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      t.Re[i1,i2] := ML.Re[i1,i2]*DR[i2];
      t.Im[i1,i2] := ML.Im[i1,i2]*DR[i2];
    end;
  result := t;
end;

function DiagMt(const a: IFArr2D; var code: TInt;
  sort: boolean = true; vectors: boolean = true; Ascending: boolean = true;
  Epsilon : TFloat = SqrtMinFloat; IterMax: TInt = 999): IEigenSys;
var
  e: IFArr1D;
  ES: IEigenSys;
begin
  code := -1;
  ES := TEigenSys.Create(a); // check for a is square
  code := 0;
  if a.Lo1 = a.Hi1 then
  begin
    ES.Values[a.Lo1] := a[a.Lo1,a.Lo1];
    ES.Vectors[a.Lo1,a.Lo1] := 1.0;
  end else
  begin
    e := TFArr1D.Create(a.Lo1,a.Hi1);
    ES.Vectors.Assign(a);
    tred2(ES.Vectors, ES.Values, e, vectors);
    tqli(ES.Values, e, ES.Vectors, code, vectors);
    if sort then
      SortVal(ES.Values, ES.Vectors, vectors, Ascending);
  end;
  SetStringArr(ES.Names, ' ');
  result := ES;
end;

function EigenSysToMt(const ES: IEigenSys): IFArr2D;
begin
  result := MtxDxMtT(ES.Vectors,ES.Values);
end;

function SVDMt(const a: IFArr2D; fullmat: boolean = false;
               eps : TFloat = cMachEps; tol : TFloat = MinFloat): ISVDSys;
var
  SVDS: ISVDSys;
begin

  SVDS := TSVDSys.Create(a, fullmat);
  if a.Dim1 < a.Dim2 then
  begin
    xsvd(eps, tol, Transpose(a), SVDS.Values, SVDS.Vvectors, SVDS.Uvectors,
         true, true, fullmat);
  end else
    xsvd(eps, tol, a, SVDS.Values, SVDS.Uvectors, SVDS.Vvectors,
         true, true, fullmat);
  SortSVD(SVDS.Values, SVDS.Uvectors, SVDS.Vvectors, true, true);
  result := SVDS;
end;

function SVDSysToMt(const SVDS: ISVDSys): IFArr2D;
var
  A: IFArr2D;
  L1, H1, L2, H2, i1, i2, k, nm: TInt;
  s: TFloat;

begin
  L1 := SVDS.Uvectors.Lo1;
  H1 := SVDS.Uvectors.Hi1;
  L2 := SVDS.Vvectors.Lo1;
  H2 := SVDS.Vvectors.Hi1;
  A := TFArr2D.Create(L1,H1,L2,H2);
  nm := min(A.Dim2,A.Dim1);
  for i1 := L1 to H1 do
    for i2 := L2 to H2 do
    begin
      s := 0.0;
      for k := 1 to nm do
        s := s + SVDS.Uvectors[i1,k]*SVDS.Values[k]*SVDS.Vvectors[i2,k];
      A[i1,i2] := s;
    end;
 result := A;
end;

{ author: Nikolai V. Shokhirev 2002 Diagonalization of Hermitian matrix }
function DiagHMt(const a: ICArr2D; var code: TInt; sort: boolean = true;
                vectors: boolean = true; Ascending: boolean = true): IHEigenSys;
var
  e, e2: IFArr1D;
  tau: IFArr2D;
  HES: IHEigenSys;
  n, i, m : TInt;
  R1, R2 : IRestore;
begin
  HES := THEigenSys.Create(a.Re);
  n := a.Dim1;
  if vectors then m := n else m := 0;
  e := TFArr1D.Create(n);
  e2 := TFArr1D.Create(n);
  tau := TFArr2D.Create(2,n);

  R1 := TRestore1D.Create([HES.Values]);
  R2 := TRestore2D.Create([a.Re, a.Im, HES.Vectors]);
  for i := 1 to n do
    HES.Vectors.Re[i,i] := 1.0;
  htridi(a.Re, a.Im, tau, HES.Values, e, e2);
  tqli(HES.Values, e, HES.Vectors.Re, code);
  if code <> 0 then
    raise Exception.Create('tqli: code <> 0');
  if vectors then
    htribk(a.Re, a.Im, tau, m, HES.Vectors.Re, HES.Vectors.Im);
  if vectors and sort then
    SortValC(HES.Values, HES.Vectors.Re, HES.Vectors.Im, m, true, ascending);
  result := HES;
end;

function HEigenSysToMt(const HES: IHEigenSys): ICArr2D;
var
  A: ICArr2D;
begin
  A := CMtxD(HES.Vectors,HES.Values);
  result := CMtxCMt(A,HConjugate(HES.Vectors));
end;

function PseudoinverseMt(const a: IFArr2D; k: TInt = _; e : TFloat = MinFloat;
                                                    s : TFloat = 0.0): IFArr2D;
var
  SVDS: ISVDSys;
  i1, i2, j, jmax, kmax: TInt;
  val: TFloat;
  t: IFArr2D;
begin
  SVDS := SVDMt(a);
  t := TFArr2D.Create(a.Lo2,a.Hi2,a.Lo1,a.Hi1);
  if k = _ then
    kmax := SVDS.Values.Hi1
  else
    kmax := max(1, k);

  for j := 1 to kmax do
  begin
    jmax := j;
    val := SVDS.Values[j];
    if val > e then
      SVDS.Values[j] := 1.0/(val + s)
    else
      break;
  end;

  for i1 := t.Lo1 to t.Hi1 do
    for i2 := t.Lo2 to t.Hi2 do
    begin
      val := 0.0;
      for j := 1 to jmax do
        val := val + SVDS.Vvectors[i1,j]*SVDS.Values[j]*SVDS.Uvectors[i2,j];
      t[i1,i2] := val;
    end;
  result := t;
end;

end.
