{*
 Rectangular table data object
 Nikolai V. Shokhirev, 2006-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2006.10.09 - created
 @version 2007.08.04: added NullValues, switched to Lines: TStrings
 @version 2009.07.12: added btDate, btTime, btFloat0. Improved data type inference.
 }
unit uPanelData;

interface

uses
    SysUtils
  , uMatTypes
  , uDynArrays
  , Classes
  ;

type
  /// TBasicType = (btUnknown, btString, btInteger, btFloat, btBoolean, btDateTime, btNULL)
  TBasicType = (btUnknown, btString, btInteger, btFloat, btBoolean, btDateTime,
                btDate, btTime, btFloat0, btNULL);
  /// Num = [.0-9]
  TSymbol = (syComma, syTab, sySemicolon, sySpace, syNum, syLetter);

  TSymbolCountArray = array[TSymbol] of integer;

  /// TDataType = set of TBasicType
  TDataTypeSet = set of TBasicType;

  TIntDateFormat = (dtfUnknown, dtfDateInt6, dtfDateInt8);

  /// Empty or no data
  ENoInputData = class(Exception);
  /// Different number of fields in input rows
  EFieldNumberError = class(Exception);
  /// Cannot determine type
  EDataTypeError = class(Exception);

/// Interface for PanelData
  IPanelData = interface(IComment)
  ['{EB973BEB-CB7F-41F6-A8FC-BA856CD8AE39}']
    function GetValue(r, c: TInt): string;  /// Getter
    procedure SetValue(r, c: TInt; const aValue: string);  /// Setter
    function GetInt(r, c: TInt): TInt;  /// Getter
    procedure SetInt(r, c: TInt; const aValue: TInt);  /// Setter
    function GetFloat(r, c: TInt): TFloat;  /// Getter
    procedure SetFloat(r, c: TInt; const aValue: TFloat);  /// Setter
    function GetDateTime(r, c: TInt): TDateTime;  /// Getter
    procedure SetDateTime(r, c: TInt; const Value: TDateTime);  /// Setter
    function GetTitle(c: TInt): string;  /// Getter
    procedure SetTitle(c: TInt; const aValue: string);  /// Setter
    function GetType(c: TInt): TBasicType;  /// Getter
    procedure SetType(c: TInt; const Value: TBasicType);  /// Setter
    function GetLines: TStrings;  /// Getter
    function GetNCol: TInt;  /// Getter
    function GetNRow: TInt;  /// Getter
    function GetNLine: TInt;  /// Getter
    procedure SetLengths(aNRow, aNCol: TInt);  /// Set New Lengths
    {*
     @param FileName: string;
     @param delimiter: char = ',';
     @param @param CommentSymbol: char = '#'      }
    procedure SaveToFile(const FileName: string; delimiter: char = ',';
                                                   CommentSymbol: char = '#');
    {*
     @param FileName: string;
     @param Error: integer;
     @param ErrMsg: string;
     @param delimiter: char = ',';
     @param @param CommentSymbol: char = '#'     }
    procedure LoadFromFile(const FileName: string; var Error: integer; var ErrMsg: string;
                           delimiter: char = ','; CommentSymbol: char = '#');
    property NCol: TInt read GetNCol; /// Number of Columns
    property NRow: TInt read GetNRow; /// Number of Rows
    property NLines: TInt read GetNLine; /// Number of Comment Lines
    property Value[r, c: TInt]: string read GetValue write SetValue; default;  /// property
    property AsInt[r, c: TInt]: TInt read GetInt write SetInt;  /// property
    property AsFloat[r, c: TInt]: TFloat read GetFloat write SetFloat;  /// property
    property AsDateTime[r, c: TInt]: TDateTime read GetDateTime write SetDateTime;  /// property
    property Titles[c: TInt]: string read GetTitle write SetTitle; /// Field Name
    property Types[c: TInt]: TBasicType read GetType write SetType;  /// property
    property Lines: TStrings read GetLines;  /// property
    function GetFloatArr(c: TInt): IFArr1D;  /// 1 - based array for the c-th column
    procedure CopyFloatArr(c: TInt; const A: IFArr1D);  /// copy 1 - based array to the c-th column
    procedure PriceToReturn(c: TInt);  /// C[r] := C[r]/C[r-1] r = NRow..2
    procedure ReturnToPrice(c: TInt);  /// C[r] := C[r]*C[r-1] r = 2..NRow
    procedure IntToDateTime(c: TInt; df: TIntDateFormat);  /// YYMMDD (n=6) of YYYYMMDD (n=8) to DateTime
    procedure DateTimeToInt(c: TInt; df: TIntDateFormat);  /// DateTime to YYMMDD (n=6) of YYYYMMDD (n=8)
  end;
{
          +-----------------------------------------------+
          |# Comment Line 0                               |  Optional
          +-----------------------------------------------+
          |# Comment Line 1                               |  Optional
          +-----------------------------------------------+
          |# Comment Line .....                           |  . . .
          +-----------+-----------+-----------+-----------+
          | Title 1   | Title 2   |  . . . .  | Tite NCol |  Optional
          +-----------+-----------+-----------+-----------+
          | Type 1    | Type 2    |  . . . .  | Type NCol |  Optional
          +-----------+-----------+-----------+-----------+
 Row  1   |    1,   1 |    1,   2 |  . . . .  |    1,NCol |
          +-----------+-----------+-----------+-----------+
 Row  2   |    2,   1 |    2,   2 |  . . . .  |    2,NCol |
          +-----------+-----------+-----------+-----------+
          |  . . . .  |  . . . .  |  . . . .  |  . . . .  |
          +-----------+-----------+-----------+-----------+
 Row NRow | NRow,   1 | NRow,   2 |  . . . .  | NRow,NCol |
          +-----------+-----------+-----------+-----------+
             Col  1      Col  2                  Col NCol
}
/// Object for PanelData
  TPanelData = class(TComment, IPanelData)
  private
    fValues: array of array of string;
    fTitles: array of string;
    fTypes: array of TBasicType;
    fLines: TStrings;
    fNRow, fNCol: TInt;
    function GetDateTime(r, c: TInt): TDateTime;  /// Getter
    procedure SetDateTime(r, c: TInt; const Value: TDateTime);  /// Setter
    function GetLines: TStrings;  /// Getter
    function GetType(c: TInt): TBasicType;  /// Getter
    procedure SetType(c: TInt; const Value: TBasicType);  /// Setter
  protected
    function GetValue(r, c: TInt): string;  /// Getter
    procedure SetValue(r, c: TInt; const aValue: string);  /// Setter
    function GetInt(r, c: TInt): TInt;  /// Getter
    procedure SetInt(r, c: TInt; const aValue: TInt);  /// Setter
    function GetFloat(r, c: TInt): TFloat;  /// Getter
    procedure SetFloat(r, c: TInt; const aValue: TFloat);  /// Setter
    function GetTitle(c: TInt): string;  /// Getter
    procedure SetTitle(c: TInt; const aValue: string);  /// Setter
    function GetNCol: TInt;  /// Getter
    function GetNRow: TInt;  /// Getter
    function GetNLine: TInt;  /// Getter
  public
    constructor Create(aNRow, aNCol: TInt); overload;  /// constructor
    constructor Create; overload;  /// constructor
    destructor Destroy; override;  /// destructor
    procedure SetLengths(aNRow, aNCol: TInt);  /// Setter
    procedure SaveToFile(const FileName: string; delimiter: char = ',';
                                                   CommentSymbol: char = '#');
    procedure LoadFromFile(const FileName: string; var Error: integer; var ErrMsg: string;
                           delimiter: char = ','; CommentSymbol: char = '#');
    property NCol: TInt read GetNCol; /// Number of Columns
    property NRow: TInt read GetNRow; /// Number of Rows
    property NLines: TInt read GetNLine; /// Number of Comment Lines
    property Value[r, c: TInt]: string read GetValue write SetValue; default;  /// 1 - based array
    property AsInt[r, c: TInt]: TInt read GetInt write SetInt;  /// 1 - based array
    property AsFloat[r, c: TInt]: TFloat read GetFloat write SetFloat;  /// 1 - based array
    property AsDateTime[r, c: TInt]: TDateTime read GetDateTime write SetDateTime;  /// // 1 - based array
    property Titles[c: TInt]: string read GetTitle write SetTitle;  /// 1 - based array
    property Types[c: TInt]: TBasicType read GetType write SetType;  /// 1 - based array
    property Lines: TStrings read GetLines;  /// property
    function GetFloatArr(c: TInt): IFArr1D;  /// 1 - based array for the c-th column
    procedure CopyFloatArr(c: TInt; const A: IFArr1D);  /// copy 1 - based array to the c-th column
    procedure PriceToReturn(c: TInt);  /// C[r] := C[r]/C[r-1] r = NRow..2
    procedure ReturnToPrice(c: TInt);  /// C[r] := C[r]*C[r-1] r = 2..NRow
    procedure IntToDateTime(c: TInt; df: TIntDateFormat);  /// YYMMDD (n=6) of YYYYMMDD (n=8) to DateTime
    procedure DateTimeToInt(c: TInt; df: TIntDateFormat);  /// DateTime to YYMMDD (n=6) of YYYYMMDD (n=8)
end;

///  BasicType To String
function BasicTypeToStr(bt: TBasicType): string;

/// String To TBasicType
function StrToBasicType(sbt: string): TBasicType;

implementation

uses
  math
  , uStrFun
  , TypInfo
  , uSGUtils
  , DateUtils
  , Dialogs
  ;

ResourceString
  /// RS_IndexError = '%s index = %d : out of range [%d, %d]';
  RS_IndexError = '%s index = %d : out of range [%d, %d]';
  /// RS_IndexOutOfRange = 'Index out of range';
  RS_IndexOutOfRange = 'Index out of range';
  /// RS_FieldNumberError = ' @ line %d';
  RS_FieldNumberError = ' @ line %d';
  /// RS_DataTypeError = ' @ Col %d';
  RS_DataTypeError = ' @ Col %d';

function BasicTypeToStr(bt: TBasicType): string;
begin
  result := Copy(GetEnumName(TypeInfo(TBasicType), ord(bt)),3,64) ;
end;

function StrToBasicType(sbt: string): TBasicType;
var
  bt: TBasicType;
  s: string;
begin
  result := btUnknown;
  s := 'bt'+sbt;
  for bt := Low(TBasicType) to High(TBasicType) do
    if s = GetEnumName(TypeInfo(TBasicType), ord(bt)) then
      result := bt;
end;

/// check s for being one of known types
function GetDataType(const s: string): TBasicType;
begin
  if IsInt(s) then
    result := btInteger
  else if IsFloat(s) then
    result := btFloat
  else if IsFloat0(s) then
    result := btFloat0
  else if IsDate(s) then
    result := btDate
  else if IsTime(s) then
    result := btTime
  else if IsDateTime(s) then
    result := btDateTime
  else if IsBool(s) then
    result := btBoolean
  else if IsNULL(s) then
    result := btNULL
  else
    result := btString;
end;

///  TDataTypeSet -> TBasicType
///  it is assumed that DataTypeSet has only one element
function SetDataType(dts: TDataTypeSet): TBasicType;
begin
  if btFloat in dts then
    result := btFloat
  else if btFloat0 in dts then
    result := btFloat0
  else if btInteger in dts then
    result := btInteger
  else if btDateTime in dts then
    result := btDateTime
  else if btDate in dts then
    result := btDate
  else if btTime in dts then
    result := btTime
  else if btBoolean in dts then
    result := btBoolean
  else if btNULL in dts then
    result := btNULL
  else
    result := btString;
end;

function GetSymbolCounts(s: string): TSymbolCountArray;
var
  sy: TSymbol;
  c: char;
  i: integer;
begin
  for sy := Low(TSymbol) to High(TSymbol) do
    result[sy] := 0;

  for i := 1 to Length(s) do
  begin
    c := s[i];
    case c of
    '.','0'..'9': result[syNum]       := result[syNum] + 1;
    ','         : result[syComma]     := result[syComma] + 1;
    #09         : result[syTab]       := result[syTab] + 1;
    ';'         : result[sySemicolon] := result[sySemicolon] + 1;
    ' '         : result[sySpace]     := result[sySpace] + 1;
    else
        result[syLetter] := result[syLetter] + 1;
    end;
  end;
end;

function GetNonDataRecNum(const SL: TStrings): integer;
var
  s: string;
  counts: TSymbolCountArray;
  t: integer;
begin
  result := 0;
  s := RemoveDuplicateSpaces(Trim(SL[result]));
  t := Length(s) div 2;
  counts := GetSymbolCounts(s);
  while (counts[syNum] < t) or (counts[syNum] = 0) do
  begin
    inc(result);
    s := RemoveDuplicateSpaces(Trim(SL[result]));
    t := Length(s) div 2;
    counts := GetSymbolCounts(s);
  end;
end;

function GetSeparator(const SL: TStrings; StartIndex: integer = 1): char;
var
  sy, syMax: TSymbol;
  counts: TSymbolCountArray;
  i: integer;
begin
  i := StartIndex;
  counts := GetSymbolCounts(RemoveDuplicateSpaces(Trim(SL[i])));
  syMax := syLetter;

  for sy := syComma to sySpace do
    if counts[sy] > counts[syMax] then
    begin
      syMax := sy;
    end;

  case syMax of
  syComma     : result := ',';
  syTab       : result := #09;
  sySemicolon : result := ';';
  sySpace     : result := ' ';
  else
    result := #00;
  end;
end;

{ TPanelData }

constructor TPanelData.Create;
begin
  inherited Create;
  fNRow := 0;
  fNCol := 0;
  fLines := TStringList.Create;
end;

procedure TPanelData.SetLengths(aNRow, aNCol: TInt);
//var
//  r: integer;
begin
    SetLength(fValues, aNRow, aNCol);
    SetLength(fTitles, aNCol);
    SetLength(fTypes, aNCol);
{
  if (fNRow <> aNRow) then
  begin
    SetLength(fValues, aNRow);
    if aNRow > fNRow then
      for r := fNRow to aNRow-1 do
        SetLength(fValues[r], fNCol);
  end;

  if fNCol <> aNCol then
  begin
    SetLength(fTitles, aNCol);
    SetLength(fTypes, aNCol);
    for r := 0 to aNRow-1 do
      SetLength(fValues[r], aNCol);
  end;
}
  fNRow := aNRow;
  fNCol := aNCol;
end;

constructor TPanelData.Create(aNRow, aNCol: TInt);
begin
  Create;
  SetLengths(aNRow, aNCol);
end;

destructor TPanelData.Destroy;
begin
  if fNCol <> 0 then
  begin
    SetLength(fValues, 0, 0);
    SetLength(fTitles, 0);
    SetLength(fTypes, 0);
  end;
  fLines.Free;
  fValues := nil;
  fTitles := nil;
  fTypes := nil;
  fLines := nil;
  inherited;
end;

procedure TPanelData.SaveToFile(const FileName: string; delimiter: char = ',';
                                                     CommentSymbol: char = '#');
var
  i, r, c : TInt;
  sl, row: TStringlist;
  IsEmpty: boolean;
begin
  sl := TStringlist.Create;
  row := TStringlist.Create;
  row.Delimiter := delimiter;
  try
    for i := 0 to fLines.Count-1 do
      sl.Append(CommentSymbol+fLines[i]);

    // Check if the Tiles are emply
    IsEmpty := true;
    for c := 1 to NCol do
      if Trim(Titles[c]) <> '' then
        IsEmpty := false;

    // Record Field Names (Column Titles)
    if not IsEmpty then
    begin
      row.Clear;
      for c := 1 to NCol do
        row.Append(Titles[c]);
      sl.Append(row.DelimitedText);
    end;

    // Record Data
    for r := 1 to NRow do
    begin
      row.Clear;
      for c := 1 to NCol do
        row.Append(Value[r, c]);
      sl.Append(row.DelimitedText);
    end;

    sl.SaveToFile(FileName);

  finally
    sl.Free;
    row.Free;
  end;
end;

///  Loads Data
procedure TPanelData.LoadFromFile(const FileName: string; 
                                  var Error: integer; var ErrMsg: string;
                                  delimiter: char = ','; CommentSymbol: char = '#');
var
  CommentCount, LineCount, ColCount, NRow, NCol: TInt;
  r, i, n, c, NonDataRecNum: TInt;
  sl, row: TStringlist;
  dts: array of TDataTypeSet;
  ch: char;

  /// TryToGetDataTypes - in one line
  function TryToGetDataTypes(idx: integer): boolean;
  var
    c, ColCount: TInt;
  begin
    row.DelimitedText := sl[idx];
    ColCount := row.Count;
    if idx = (LineCount-1) then
    begin
      NCol := ColCount;
      SetLength(dts, NCol);
      for c := 0 to NCol-1 do
        dts[c] := [];
    end;
    if ColCount <> NCol then
        raise EFieldNumberError.CreateFmt(RS_FieldNumberError,[idx]);

    result := true; // No btNULL
    for c := 0 to NCol-1 do
    begin
      Include(dts[c], GetDataType(row[c]));
      if dts[c] = [btNULL] then
        raise EDataTypeError.CreateFmt(RS_DataTypeError,[idx]);
    end;
  end;

  /// try untill found of end of data
  procedure GetDataTypes;
  var
    c, idx: TInt;
    done: boolean;
  begin
    idx := LineCount-1;
    repeat
      done := TryToGetDataTypes(idx) or (idx = n);
      dec(idx);
    until done;

    // check types are set
    for c := 0 to NCol-1 do
    begin
      Exclude(dts[c], btNULL);
      if dts[c] = [] then
        EDataTypeError.CreateFmt(RS_DataTypeError, [c]);
    end;
    // set types
    SetLength(fTypes, NCol);
    SetLength(fTitles, NCol);
    for c := 0 to NCol-1 do
      fTypes[c] := SetDataType(dts[c]);
  end;

  /// Check if it is a Title line
  function IsTitleLine(idx: integer): boolean;
  var
    c, ColCount: TInt;
  begin
    if delimiter <> ' ' then
      row.DelimitedText := QuotedCSV(sl[idx])
    else
      row.DelimitedText := sl[idx];
    ColCount := row.Count;

    if ColCount <> NCol then
      raise EFieldNumberError.CreateFmt(RS_FieldNumberError,[idx]);

    result := false; // No btNULL
    for c := 0 to NCol-1 do
    begin
      if (GetDataType(row[c]) <> fTypes[c]) and (GetDataType(row[c]) <> btNULL) then
        result := true;
    end;
  end;

begin
  Error := 0;
  ErrMsg := '';
  sl := TStringlist.Create;
  row := TStringlist.Create;
  row.Delimiter := delimiter;
  try
    sl.LoadFromFile(FileName);
    LineCount := sl.Count;
    if (LineCount = 0) then
      Raise ENoInputData.Create('Empty file');

    // Skip comments for now
    n := 0;
    while (Pos(CommentSymbol, sl[n]) > 0) and (n < LineCount) do
      inc(n);

    CommentCount := n;
    if (n = LineCount) then
      Raise ENoInputData.Create('No data after comments');

    NonDataRecNum := GetNonDataRecNum(sl);

    if NonDataRecNum < CommentCount then
    begin
      Error := -1;
      ErrMsg := 'NonDataRecNum < CommentCount';
      exit;
    end;

    ch := GetSeparator(sl,NonDataRecNum);
    if ch = #00 then
      ch := delimiter;
      
    if ch <> delimiter then
    begin
      case ch of
      ','         : Error := 1;
      #09         : Error := 2;
      ';'         : Error := 3;
      ' '         : Error := 4;
      else
          Error := 5;
      end;
      if Error = 5 then
        ErrMsg := 'Unknown delimiter:'''+ch+''''
      else
        ErrMsg := 'old delimiter: '''+delimiter+''' found delimiter: '''+ch+'''';
      exit;
    end;

    GetDataTypes;

    if NonDataRecNum > 0 then
      if IsTitleLine(n) then
      begin
        for c := 0 to NCol-1 do
          fTitles[c] := row[c];  // copy to Titles
        inc(n)                   // advance for one line
      end;

    // Set correct dimensions and copy Data
    NRow := sl.Count - n;
    SetLengths(NRow, NCol);
    r := 0;
    for i := n to sl.Count-1 do
    begin
      row.DelimitedText := sl[i];
      ColCount := row.Count;
      if ColCount <> NCol then
        raise EFieldNumberError.CreateFmt(RS_FieldNumberError,[i]);
      for c := 0 to NCol-1 do
        fValues[r, c] := row[c];
      inc(r);
    end;

    // copy comments if all previous is OK
    fLines.Clear;
    for i := 0 to CommentCount-1 do
      fLines.Append(Copy(sl[i],2,MaxInt));

  finally
    sl.Free;
    row.Free;
    SetLength(dts, 0);
  end;
end;

function TPanelData.GetDateTime(r, c: TInt): TDateTime;
begin
  if (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  result := StrToDatetime(Value[r, c]);
end;

function TPanelData.GetFloat(r, c: TInt): TFloat;
var
  Code: Integer;
  x: TFloat;
begin
//  Val(Value[r, c], x, Code);
//  if (Code = 0)
  if TryStrToFloat(Value[r, c], x)

    then result := x
    else result := NaN;
end;

function TPanelData.GetInt(r, c: TInt): TInt;
var
  I, Code: Integer;
begin
  Val(fValues[r, c], I, Code);
  if Code = 0
    then result := I
    else result := iNaN;
end;

function TPanelData.GetLines: TStrings;
begin
  result := fLines;
end;

function TPanelData.GetNCol: TInt;
begin
  if (fValues[0] = nil)
    then result := 0
    else result := High(fValues[0])+1;
end;

function TPanelData.GetNLine: TInt;
begin
  result := fLines.Count;
end;

function TPanelData.GetNRow: TInt;
begin
  if (fValues = nil)
    then result := 0
    else result := High(fValues)+1;
end;

function TPanelData.GetTitle(c: TInt): string;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  result := fTitles[c-1];
end;

function TPanelData.GetValue(r, c: TInt): string;
begin
  if (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  result := fValues[r-1,c-1];
end;

procedure TPanelData.SetDateTime(r, c: TInt; const Value: TDateTime);
begin
  if (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fValues[r-1,c-1] := DateTimeToStr(Value);
end;

procedure TPanelData.SetFloat(r, c: TInt; const aValue: TFloat);
begin
  if (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fValues[r-1,c-1] := FloatToStr(aValue);
end;

procedure TPanelData.SetInt(r, c: TInt; const aValue: TInt);
begin
  if (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fValues[r-1,c-1] := IntToStr(aValue);
end;

procedure TPanelData.SetTitle(c: TInt; const aValue: string);
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fTitles[c-1] := aValue;
end;

procedure TPanelData.SetValue(r, c: TInt; const aValue: string);
begin
  If (r < 1) or (r > fNRow) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Row',r, 1, fNRow]);
  If (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fValues[r-1,c-1] := aValue;
end;
{
function TPanelData.GetType(c: TInt): string;
begin
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  result := fTypes[c-1];
end;

procedure TPanelData.SetType(c: TInt; const Value: string);
begin
begin
end;
}
function TPanelData.GetType(c: TInt): TBasicType;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  result := fTypes[c-1];
end;

procedure TPanelData.SetType(c: TInt; const Value: TBasicType);
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  fTypes[c-1] := Value;
end;

function TPanelData.GetFloatArr(c: TInt): IFArr1D;
var
  r: integer;
begin
  result := nil;
  if (Types[c] <> btFloat) and (Types[c] <> btInteger) then
    exit;
  result := TFArr1D.Create(NRow);
  for r := 1 to NRow do
    result[r] := self.AsFloat[r, c];
end;

procedure TPanelData.CopyFloatArr(c: TInt; const A: IFArr1D);
var
  r: integer;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  if (A.Dim1 <> fNRow) then
     Raise ERangeError.Create('Dim mismatch');
  A.Lo1 := 1; //just in case
  for r := 1 to fNRow do
    self.AsFloat[r, c] := A[r];
end;

procedure TPanelData.PriceToReturn(c: TInt);
var
  r: integer;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  for r := fNRow downto 2 do
    self.AsFloat[r, c] := self.AsFloat[r, c]/self.AsFloat[r-1, c];
end;

/// C[r] := C[r]*C[r-1] r = 2..NRow
procedure TPanelData.ReturnToPrice(c: TInt);
var
  r: integer;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  for r := 2 to fNRow do
    self.AsFloat[r, c] := self.AsFloat[r, c]*self.AsFloat[r-1, c];
end;

procedure TPanelData.DateTimeToInt(c: TInt; df: TIntDateFormat);
var
  s, sy: string;
  dt: TDateTime;
  AYear, AMonth, ADay: Word;
  r: TInt;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  if Self.Types[c] <> btDateTime then
     Raise ERangeError.Create('Non DateTime type');
  for r := 1 to fNRow do
  begin
    dt := self.AsDateTime[r, c];
    DecodeDate(dt, AYear, AMonth, ADay);
    sy := IntToStr(AYear);
    case df of
    dtfDateInt6:
    begin
      sy := CopyFromTo(sy,3,4);
    end;
    dtfDateInt8:
    begin
      // do nothing;
    end;
    else
      assert(false, 'Unknown type');
    end;

    s := IntToStr(AMonth);
    if AMonth < 10 then s := '0'+s;
    sy := sy + s;
    s := IntToStr(ADay);
    if ADay < 10 then s := '0'+s;
    sy := sy + s;
    self[r, c] := sy;
  end;
  Self.Types[c] := btInteger;
end;

procedure TPanelData.IntToDateTime(c: TInt; df: TIntDateFormat);
var
  s, sy,sm, sd: string;
  dt: TDateTime;
  AYear, AMonth, ADay: Word;
  r, n: TInt;
begin
  if (c < 1) or (c > fNCol) then
     Raise ERangeError.CreateFmt(RS_IndexError,['Col',c, 1, fNCol]);
  if Self.Types[c] <> btInteger then
     Raise ERangeError.Create('Non Int type');
  for r := 1 to fNRow do
  begin
    s := self[r, c];
    case df of
    dtfDateInt6:
     begin
       n := 6;
       assert(Length(s)=n, 'wrong Format at row '+IntToStr(r));
       sy := '19'+CopyFromTo(s,1,2);
       sm := CopyFromTo(s,3,4);
       sd := CopyFromTo(s,5,6);
     end;
    dtfDateInt8:
     begin
       n := 8;
       assert(Length(s)=n, 'wrong Format at row '+IntToStr(r));
       sy := CopyFromTo(s,1,4);
       sm := CopyFromTo(s,5,6);
       sd := CopyFromTo(s,7,8);
     end;
    else
      assert(false, 'Unsupported Format');
    end;
    AYear := StrToInt(sy);
    AMonth := StrToInt(sm);
    ADay := StrToInt(sd);
    dt := EncodeDate(AYear, AMonth, ADay);
    self.AsDateTime[r, c] := dt;
  end;
  Self.Types[c] := btDateTime;
end;

end.
