{*
 Pearson Hash (http://en.wikipedia.org/wiki/Pearson_hashing) -pascal implementation
 Nikolai V. Shokhirev, 2001-2010
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2010.10.31 - created
 @version 2010.11.11 - Refactoring
}
unit uPearsonHashing;

interface

uses
  Classes
  ;

type

  TIndex = 0..255;

  TIndexSet = Set of TIndex;

  TStringArray = array of string;

  { Use:
        Variant 1 (not perfect hash)
          h := TPearsonHashing.Create;
          b := h.Hash(s); // for any s
        Variant 2 (not perfect hash)
          h := TPearsonHashing.Create;
          // Generate Perfect Hash
          h.Perfect(Keys); // Keys is array/string list of unique strings
          b := h.Hash(s); // b is unique if s is from Keys
          i := h.GetIndex(s); // returns index of s in Kys. i = 0, 1, 2, ..
  }
  // 8-bit hash function
  TPearsonHashing = class(TObject)
  private
    { Private declarations }
    fIndex: array[0..255] of integer;  // map T to Index:
    T: array[0..255] of byte;          // lookup table
    procedure GenerateTable;           // lookup table generation
    procedure GeneratePerfectHash(Keys: TStrings);
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;
//    procedure GenerateTable;
    function Hash(s: AnsiString): byte;
    // Keys is array of unique strings: [s1, s2, s3, ..]
    procedure Perfect(Keys: TStringArray); overload;
    procedure Perfect(Keys: TStrings); overload;
    function GetIndex(Key: AnsiString): byte;
  end;

implementation

uses
  SysUtils
  ;

{ TPearsonHashing }

constructor TPearsonHashing.Create;
begin
  inherited;
  GenerateTable;
end;

destructor TPearsonHashing.Destroy;
begin

  inherited;
end;

procedure TPearsonHashing.GenerateTable;
var
  i, j, b: byte;
begin
  for i := 0 to 255 do
    T[i] := i;

  randomize;
  // Shuffle
  for i := 0 to 255 do
  begin
    j := random(256);
    b := T[i];
    T[i] := T[j];
    T[j] := b;
  end;
end;

function TPearsonHashing.Hash(s: AnsiString): byte;
var
  h, index, c: byte;
  i: integer;
begin
  h := 0;
  for i := 1 to Length(s) do
  begin
    c := ord(s[i]);
    index := h xor c;
    h := T[index];
  end;
  result := h;
end;

procedure TPearsonHashing.GeneratePerfectHash(Keys: TStrings);
var
  IndexSet : TIndexSet;
  h: byte;
  i: integer;
  s: AnsiString;
  IsUnique : boolean;
begin
  repeat
    GenerateTable;

    for h := 0 to 255 do
      fIndex[h] := -1;

    IndexSet := [];
    IsUnique := true;
    for i := 0 to Keys.Count-1 do
    begin
      s := Keys[i];
      h := Hash(s);
//      if h in IndexSet then IsUnique := false;
      if h in IndexSet then
      begin
         IsUnique := false;
         break;
      end;
      IndexSet := IndexSet + [h];
      fIndex[h] := i;
    end;
  until IsUnique;
end;

procedure TPearsonHashing.Perfect(Keys: TStrings);
var
  i, n: integer;
  s: AnsiString;
  sl: TStringList;
begin
  n := Keys.Count;
  if n > 256 then
    raise Exception.Create('Key count = ' + IntToStr(n) + ' Exceeds 256');
  sl := TStringList.Create;
  try
    for i := 0 to n-1 do
    begin
      s := UpperCase(Keys[i]);
      if sl.IndexOf(s) = -1 then
        sl.Add(s)
      else
        raise Exception.Create('Duplicate key: ' + s);
    end;

    GeneratePerfectHash(sl);
    {
    repeat
      GenerateTable;

      for h := 0 to 255 do
        fIndex[h] := -1;

      IndexSet := [];
      IsUnique := true;
      for i := 0 to sl.Count-1 do
      begin
        s := sl[i];
        h := Hash(s);
        if h in IndexSet then IsUnique := false;
        IndexSet := IndexSet + [h];
        fIndex[h] := i;
      end;
    until IsUnique;
    }
  finally
     sl.Free;
  end;
end;


procedure TPearsonHashing.Perfect(Keys: TStringArray);
var
  i, n: integer;
  s: AnsiString;
  sl: TStringList;
begin
  n := Length(Keys);
  if n > 256 then
    raise Exception.Create('Key count = ' + IntToStr(n) + ' Exceeds 256');
  sl := TStringList.Create;
  try
    for i := 0 to Length(Keys)-1 do
    begin
      s := UpperCase(Keys[i]);
      if sl.IndexOf(s) = -1 then
        sl.Add(s)
      else
        raise Exception.Create('Duplicate key: ' + s);
    end;
    GeneratePerfectHash(sl);
  finally
     sl.Free;
  end;
end;

function TPearsonHashing.GetIndex(Key: AnsiString): byte;
var
  h: byte;
  i: integer;
begin
  h := Hash(UpperCase(Key));
  i := fIndex[h];
  if i > -1 then
    result := byte(i)
  else
     raise Exception.Create('Invalid key: ' + Key);;
end;

end.
