{*
 Simple wrappers over dynamic arrays
 Nikolai V. Shokhirev, 2001-2007
 @author Nikolai V. Shokhirev <nikolai@shokhirev.com> http://www.shokhirev.com/nikolai.html
 @version 2007.09.03 - created
 @version 2007.09.05 - added TSimpleFloatList
 @version 2007.09.06 - Added TSimpleStrList, Save/Load
 @version 2008.08.23 - Added IndexOfSortedList
 @version 2008.09.09 - Added TBin, TBinList, THistorgam
 @version 2009.03.03 - Added ICount, IIntList, IFloatList, IStringList, TCount
 @version 2009.03.09 - TSimpleIntList, TSimpleFloatList, TSimpleStrList implement interfaces
 }
unit uTypedLists;

interface

uses
  uMatTypes
  , SysUtils
  , Contnrs
  ;

type

  ICount = interface
  ['{BA7AEBF5-EA05-4D56-B4C9-AF802992260E}']
    function GetCount: integer;
    procedure Clear;
    property Count: integer read GetCount;
  end;

  IIntList = interface(ICount)
    procedure Grow;
    function GetValue(i: integer): TInt;
    procedure SetValue(i: integer; const Value: TInt);
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
    procedure Sort;
    procedure Append(Value: TInt);
    property Value[i: integer]: TInt read GetValue write SetValue; default;
  end;

  IFloatList = interface(ICount)
    procedure Grow;
    function GetValue(i: integer): TFloat;
    procedure SetValue(i: integer; const Value: TFloat);
    procedure SaveToFile(const FileName: string;
                         Format: TFloatFormat = ffGeneral;
                         Precision: integer = 12; Digits: integer = 2);
    procedure LoadFromFile(const FileName: string);
    procedure Sort; // Ascending
    procedure Append(Value: TFloat);
    property Value[i: integer]: TFloat read GetValue write SetValue; default;
  end;

  IStringList = interface(ICount)
    procedure Grow;
    function GetValue(i: integer): string;
    procedure SetValue(i: integer; const Value: string);
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
    procedure Append(Value: string);
    property Value[i: integer]: string read GetValue write SetValue; default;
  end;

  TCount = class(TInterfacedObject)
  private
    fCount: Integer;
    function GetCount: integer;
  public
    constructor Create;
    procedure Clear; virtual;
    property Count: integer read GetCount;
  end;

  TSimpleIntList = class(TCount, IIntList)
  private
//    fCount: Integer;
    fCapacity: Integer;
    fList: array of TInt;
    procedure Grow;
    function GetValue(i: integer): TInt;
    procedure SetValue(i: integer; const Value: TInt);
  public
    constructor Create;
    destructor Destroy; override;
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
    procedure Clear; override;
    procedure Sort;
    procedure Append(Value: TInt);
//    property Count: integer read fCount write fCount;
    property Value[i: integer]: TInt read GetValue write SetValue; default;
  end;

  TSimpleFloatList = class(TCount, IFloatList)
  private
//    fCount: Integer;
    fCapacity: Integer;
//    Precision, Digits: integer;
    fList: array of TFloat;
//    ascSorted: boolean;
    procedure Grow;
    function GetValue(i: integer): TFloat;
    procedure SetValue(i: integer; const Value: TFloat);
  public
    constructor Create;
    destructor Destroy; override;
    procedure SaveToFile(const FileName: string;
                         Format: TFloatFormat = ffGeneral;
                         Precision: integer = 12; Digits: integer = 2);
    procedure LoadFromFile(const FileName: string);
    procedure Clear; override;
    procedure Sort; // Ascending
    procedure Append(Value: TFloat);
//    property Count: integer read fCount write fCount;
    property Value[i: integer]: TFloat read GetValue write SetValue; default;
  end;

  TSimpleStrList = class(TCount, IStringList)
  private
//    fCount: Integer;
    fCapacity: Integer;
    fList: array of string;
    procedure Grow;
    function GetValue(i: integer): string;
    procedure SetValue(i: integer; const Value: string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure SaveToFile(const FileName: string);
    procedure LoadFromFile(const FileName: string);
    procedure Clear; override;
    procedure Append(Value: string);
//    property Count: integer read fCount write fCount;
    property Value[i: integer]: string read GetValue write SetValue; default;
  end;

  // Last Mod 8008.10.01
  TBin = class
    BinNumber: integer;
    BinCount: integer;
    constructor Create (aBinNumber: integer; aBinCount: integer = 1);
  end;

  TBinList = class
  private
    fList: TObjectList;
    function GetBin(Idx: integer): TBin;
    function GetCount: integer;
    function GetEmptyCount: integer;
    function GetMaxBinNumber: integer;
    function GetMinBinNumber: integer;
  public
   constructor Create;
   destructor Destroy; override;
   procedure Clear;
   function IndexOfBinNumber(const BinNumber: integer): integer;
   procedure IncBinCount(const BinNumber: integer);
   property Count: integer read GetCount;
   property EmptyCount: integer read GetEmptyCount;
   property MinBinNumber: integer read GetMinBinNumber;
   property MaxBinNumber: integer read GetMaxBinNumber;
   property Bin[Idx: integer]: TBin read GetBin; default;
  end;

  THistorgam = class (TBinList)
  private
    fReferenceValue: TFloat;
    fStep: TFloat;
    fHalfStep: TFloat;
    function GetMaxValue: TFloat;
    function GetMinValue: TFloat;
    function GetBinCount(BinNumber: integer): integer;
    function NumberFromValue(Value: TFloat): integer;
  public
    constructor Create(ReferenceValue, Step: TFloat); overload;
    constructor Create; overload;
    procedure Init(ReferenceValue, Step: TFloat);
    procedure AddValue(Value: TFloat);
    function ValueFromNumber(BinNumber: integer): TFloat;
    function ValueFromIndex(Idx: integer): TFloat;
    property MinValue: TFloat read GetMinValue;
    property MaxValue: TFloat read GetMaxValue;
    property BinCount[BinNumber: integer]: integer read GetBinCount;
    procedure GetStat(var AvgBinNumber: TFloat; var BinNumberOfMaxCount: integer;
                      var MaxCountRatio: TFloat; BinNumberInterval: integer);
  end;

  // Returnrs idx so that
  // List[idx] <= value < List[idx+1]
  // if value < MinValue then idx = -1;
  // if value >= MaxValue then idx = Count-1
  function IndexOfSortedList(value: TFloat; List: TSimpleFloatList): integer; overload;

  function IndexOfSortedList(value: TFloat; List: IFloatList): integer; overload;

implementation

uses
  Math
  , Types
  ;

ResourceString
  /// 'idex = %d : out of range [%d, %d]';
  RS_IndexError = 'index = %d : out of range [%d, %d]';

{ unit functions }

// Returnrs idx so that
// List[idx] <= value < List[idx+1]
// if value < MinValue then idx = -1;
// if value >= MaxValue then idx = Count-1
function IndexOfSortedList(value: TFloat; List: TSimpleFloatList): integer;
var
  i, j, k: integer;
begin
//  if not ascSorted then
//    Sort;
  if value < List[0] then
    result := -1
  else if value >= List[List.Count-1] then
    result := List.Count-1
  else
  begin
    i := 0;
    j := List.Count;
    repeat // binary search
      k := (i + j) div 2;
      if (value < List[k]) then j := k
                           else i := k;
    until (j <= i + 1);
    result := i;
  end;
end;

// Returnrs idx so that
// List[idx] <= value < List[idx+1]
// if value < MinValue then idx = -1;
// if value >= MaxValue then idx = Count-1
function IndexOfSortedList(value: TFloat; List: IFloatList): integer;
var
  i, j, k: integer;
begin
//  if not ascSorted then
//    Sort;
  if value < List[0] then
    result := -1
  else if value >= List[List.Count-1] then
    result := List.Count-1
  else
  begin
    i := 0;
    j := List.Count;
    repeat // binary search
      k := (i + j) div 2;
      if (value < List[k]) then j := k
                           else i := k;
    until (j <= i + 1);
    result := i;
  end;
end;

function CompareNumbers(Item1, Item2: Pointer): Integer;
begin
  result := CompareValue(TBin(Item1).BinNumber, TBin(Item2).BinNumber);
end;


{ TCount }

procedure TCount.Clear;
begin
  fCount := 0;
end;

constructor TCount.Create;
begin
  inherited;
  fCount := 0;
end;

function TCount.GetCount: integer;
begin
  result := fCount;
end;

{ TSimpleIntList }

procedure TSimpleIntList.Clear;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
  SetLength(fList,fCapacity);
end;

destructor TSimpleIntList.Destroy;
begin
  SetLength(fList, 0);
  inherited;
end;

procedure TSimpleIntList.Append(Value: TInt);
begin
  if (fCount = fCapacity) then
    Grow;
  fList[fCount] := Value;
  inc(fCount);
end;

constructor TSimpleIntList.Create;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
end;

function TSimpleIntList.GetValue(i: integer): TInt;
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  result := fList[i];
end;

procedure TSimpleIntList.SetValue(i: integer; const Value: TInt);
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  fList[i] := Value;
end;

procedure TSimpleIntList.Grow;
var
  Delta: Integer;
begin
  if fCapacity > 64 then Delta := fCapacity div 4 else
    if fCapacity > 8 then Delta := 16 else
      Delta := 4;
  inc(fCapacity, Delta);
  SetLength(fList, fCapacity);
end;

procedure TSimpleIntList.LoadFromFile(const FileName: string);
var
  tf: TextFile;
  s: TInt;
begin
  Clear;
  AssignFile(tf, FileName);
  try
    Reset(tf);
    while not Eof(tf) do
    begin
       Readln(tf,s);
       Append(s);
    end;
  finally
    CloseFile(tf);
  end;
end;

procedure TSimpleIntList.SaveToFile(const FileName: string);
var
  tf: TextFile;
  i: integer;
begin
  AssignFile(tf, FileName);
  try
    Rewrite(tf);
    for i := 0 to fCount-1 do
      Writeln(tf,fList[i]);
  finally
    CloseFile(tf);
  end;
end;

procedure TSimpleIntList.Sort;

  procedure Swap(i, j: integer);
  var
    v: integer;
  begin
    v        := fList[j];
    fList[j] := fList[i];
    fList[i] := v;
  end;

  procedure Quicksort(l, r: integer);
  var
    i, j: TInt;
    v: TFloat;
  begin
    v := fList[(l + r) div 2]; { Wert des Teilungselements }
    i := l; j := r;
    repeat
      while fList[i] < v do
        inc(i);
      while v < fList[j] do
        dec(j);
      if i <= j then
      begin
        Swap(i,j);
        inc(i);
        dec(j); { sonst kreuzen die sich nie }
      end;
    until i > j; { bis die Zeiger sich kreuzen }
    if l < j then Quicksort(l, j);
    if i < r then Quicksort(i, r);
  end;
begin
  Quicksort(0, fCount-1);
end;

{ TSimpleFloatList }

procedure TSimpleFloatList.Clear;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
  SetLength(fList,fCapacity);
end;

destructor TSimpleFloatList.Destroy;
begin
  SetLength(fList, 0);
  inherited;
end;

procedure TSimpleFloatList.Append(Value: TFloat);
begin
  if (fCount = fCapacity) then
    Grow;
  fList[fCount] := Value;
{  //Keep it simple
  if ascSorted and (fCount > 1) then
    if (Value < fList[fCount-1]) then
      ascSorted := false;
}
  inc(fCount);
end;

constructor TSimpleFloatList.Create;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
//  Precision := 16;
//  Digits := 9;
//  ascSorted := true;
end;

function TSimpleFloatList.GetValue(i: integer): TFloat;
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  result := fList[i];
end;

procedure TSimpleFloatList.SetValue(i: integer; const Value: TFloat);
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  fList[i] := Value;
end;

procedure TSimpleFloatList.Grow;
var
  Delta: integer;
begin
  if fCapacity > 64 then Delta := fCapacity div 4 else
    if fCapacity > 8 then Delta := 16 else
      Delta := 4;
  inc(fCapacity, Delta);
  SetLength(fList, fCapacity);
end;

procedure TSimpleFloatList.LoadFromFile(const FileName: string);
var
  tf: TextFile;
  s: TFloat;
begin
  Clear;
  AssignFile(tf, FileName);
  try
    Reset(tf);
    while not Eof(tf) do
    begin
       Readln(tf,s);
       Append(s);
    end;
  finally
    CloseFile(tf);
  end;
end;

procedure TSimpleFloatList.SaveToFile(const FileName: string;
                              Format: TFloatFormat; Precision, Digits: integer);
var
  tf: TextFile;
  i: integer;
begin
  AssignFile(tf, FileName);
  try
    Rewrite(tf);
    for i := 0 to fCount-1 do
      Writeln(tf,FloatToStrF(fList[i], Format, Precision, Digits));
  finally
    CloseFile(tf);
  end;
end;

{ Ascending array sorting
  Based on: Sebastian Boung, Sorting Algorithms in Pascal
  http://privat.schlund.de/b/bossung/prog/delphi/sorting.html }
procedure TSimpleFloatList.Sort;

  procedure Swap(i, j: integer);
  var
    v: TFloat;
  begin
    v        := fList[j];
    fList[j] := fList[i];
    fList[i] := v;
  end;

  procedure Quicksort(l, r: integer);
  var
    i, j: TInt;
    v: TFloat;
  begin
    v := fList[(l + r) div 2]; { Wert des Teilungselements }
    i := l; j := r;
    repeat
      while fList[i] < v do
        inc(i);
      while v < fList[j] do
        dec(j);
      if i <= j then
      begin
        Swap(i,j);
        inc(i);
        dec(j); { sonst kreuzen die sich nie }
      end;
    until i > j; { bis die Zeiger sich kreuzen }
    if l < j then Quicksort(l, j);
    if i < r then Quicksort(i, r);
  end;

begin
  Quicksort(0, fCount-1);
//  ascSorted := true;
end;

{ TSimpleStrList }

procedure TSimpleStrList.Clear;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
  SetLength(fList,fCapacity);
end;

destructor TSimpleStrList.Destroy;
begin
  SetLength(fList, 0);
  inherited;
end;

procedure TSimpleStrList.Append(Value: string);
begin
  if (fCount = fCapacity) then
    Grow;
  fList[fCount] := Value;
  inc(fCount);
end;

constructor TSimpleStrList.Create;
begin
  inherited;
//  fCount := 0;
  fCapacity := 0;
end;

function TSimpleStrList.GetValue(i: integer): string;
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  result := fList[i];
end;

procedure TSimpleStrList.SetValue(i: integer; const Value: string);
begin
  If (i < 0) or (i >= fCount) then
     Raise ERangeError.CreateFmt(RS_IndexError,[i, 0, fCount-1]);
  fList[i] := Value;
end;

procedure TSimpleStrList.Grow;
var
  Delta: integer;
begin
  if fCapacity > 64 then Delta := fCapacity div 4 else
    if fCapacity > 8 then Delta := 16 else
      Delta := 4;
  inc(fCapacity, Delta);
  SetLength(fList, fCapacity);
end;

procedure TSimpleStrList.LoadFromFile(const FileName: string);
var
  tf: TextFile;
  s: string;
begin
  AssignFile(tf, FileName);
  try
    Reset(tf);
    while not Eof(tf) do
    begin
       Readln(tf,s);
       Append(s);
    end;
  finally
    CloseFile(tf);
  end;
end;

procedure TSimpleStrList.SaveToFile(const FileName: string);
var
  tf: TextFile;
  i: integer;
begin
  AssignFile(tf, FileName);
  try
    Rewrite(tf);
    for i := 0 to fCount-1 do
      Writeln(tf,fList[i]);
  finally
    CloseFile(tf);
  end;
end;

{ TBin } 

constructor TBin.Create(aBinNumber, aBinCount: integer);
begin 
  inherited Create;;
  BinNumber := aBinNumber; 
  BinCount := aBinCount;
end; 


{ TBinList }

function TBinList.IndexOfBinNumber (const BinNumber: integer): integer;
var 
  pivot: TBin; 
  left, mid, right: integer; 
begin 
  result := -1;;

  if (fList.Count = 0) then
    exit; 

  left := 0; 
  if (BinNumber < Bin[left].BinNumber) then 
    exit; 

  right := Count-1; 
  case CompareValue(Bin[right].BinNumber, BinNumber) of
    LessThanValue: exit;
    EqualsValue: begin
                   result := right;
                   exit;
                 end;
  else 
    while (left < right) and (result = -1) do 
    begin 
      mid := (left + right) div 2; pivot := Bin[mid]; 
      if (BinNumber < pivot.BinNumber) then 
        right := mid 
      else 
        begin 
          if (BinNumber = pivot.BinNumber) then 
            result := mid 
          else 
            left := mid + 1; 
        end; 
    end;
  end;
end;

procedure TBinList.IncBinCount(const BinNumber: integer); 
var 
  idx : integer; 
begin 
  idx := IndexOfBinNumber(BinNumber); 
  if (idx >= 0) then 
    TBin(fList[idx]).BinCount := TBin(fList[idx]).BinCount + 1 
  else 
    begin 
//      idx := flist.Add(TBin.Create(BinNumber));
      flist.Add(TBin.Create(BinNumber)); 
//      fList.Sort(@CompareNumbers);
      fList.Sort(CompareNumbers);
    end;
end;

procedure TBinList.Clear; 
begin 
  fList.Clear; 
end;

constructor TBinList.Create;
begin 
  inherited Create; 
  fList := TObjectList.Create; 
end; 

destructor TBinList.Destroy; 
begin 
  fList.Free; 
  inherited; 
end; 

function TBinList.GetBin(Idx: integer): TBin; 
begin 
  result := TBin(fList[idx]); 
end; 

function TBinList.GetCount: integer; 
begin 
  result := fList.Count; 
end; 

function TBinList.GetEmptyCount: integer;
begin 
  if (Count <= 1) then 
    result := 0 
  else 
    result := MaxBinNumber - MinBinNumber - Count + 1; 
end;

function TBinList.GetMaxBinNumber: integer; 
begin 
  result := TBin(fList.Last).BinNumber;
end; 

function TBinList.GetMinBinNumber: integer; 
begin 
  result := TBin(fList.First).BinNumber;
end; 

{ THistogram }

procedure THistorgam.AddValue(Value: TFloat); 
var 
  n: integer; 
begin 
  n := NumberFromValue(Value); IncBinCount(n); 
end; 

procedure THistorgam.GetStat(var AvgBinNumber: TFloat; var BinNumberOfMaxCount: integer;
                             var MaxCountRatio: TFloat; BinNumberInterval: integer);
var
  idx, n, MaxCount, c: integer;
  s, a, fn, fc: TFloat;
begin
  BinNumberOfMaxCount := MinBinNumber;
  MaxCount := 0;
  a := 0.0;
  s := 0.0;
  for idx := 0 to Count-1 do
  begin
    c := Bin[idx].BinCount;
    n := Bin[idx].BinNumber;
    if c > MaxCount then
    begin
      BinNumberOfMaxCount := n;
      MaxCount := c;
    end;
    fc := c;
    s := s + fc;
    fn := n;
    a := a + fn*fc;
  end;

  if s > 0.0 then AvgBinNumber := a/s
             else AvgBinNumber := 0.0;

  a := MaxCount;
  s := 0.0;
  n := (BinNumberInterval + 1)*BinNumberOfMaxCount;
  for idx := 0 to Count-1 do
  begin
     if Bin[idx].BinNumber >= n then
       break;

     s := s + Bin[idx].BinCount;
  end;

  if s > 0.0 then MaxCountRatio := a/s
             else MaxCountRatio := 0.0;
end;

constructor THistorgam.Create;
begin
  inherited Create;
end;

procedure THistorgam.Init(ReferenceValue, Step: TFloat);
begin
  fReferenceValue := ReferenceValue;
  fStep := Step;
  fHalfStep := Step/2.0;
end;

constructor THistorgam.Create(ReferenceValue, Step: TFloat);
begin 
  Create;
  Init(ReferenceValue, Step);
end;

function THistorgam.GetMaxValue: TFloat; 
begin 
  result := ValueFromNumber(MaxBinNumber); 
end;

function THistorgam.GetMinValue: TFloat; 
begin 
  Result := ValueFromNumber(MinBinNumber); 
end; 

function THistorgam.GetBinCount(BinNumber: integer): integer; 
var 
  i: integer; 
begin 
  i := IndexOfBinNumber(BinNumber); 
  if (i = -1) then 
    result := 0 
  else 
    result := Bin[i].BinCount;
end;

function THistorgam.NumberFromValue(Value: TFloat): integer; 
begin 
  result := Floor( (Value - fReferenceValue + fHalfStep)/fStep); 
end; 

function THistorgam.ValueFromNumber(BinNumber: integer): TFloat;
begin 
  result := fReferenceValue + BinNumber*fStep; 
end; 

function THistorgam.ValueFromIndex(Idx: integer): TFloat;
begin
  result := ValueFromNumber(Bin[Idx].BinNumber);
end;

end.
